/* SnappyMail Webmail (c) SnappyMail | Licensed under AGPL v3 */
(function () {
	'use strict';

	let keyScopeFake = 'all';

	const ScopeMenu = 'Menu',

		doc = document,

		$htmlCL = doc.documentElement.classList,

		elementById = id => doc.getElementById(id),

		Settings = rl.settings,
		SettingsGet = Settings.get,
		SettingsCapa = name => name && !!(SettingsGet('Capa') || {})[name],

		dropdownVisibility = ko.observable(false).extend({ rateLimit: 0 }),

		moveAction = ko.observable(false),
		leftPanelDisabled = ko.observable(false),

		createElement = (name, attr) => {
			let el = doc.createElement(name);
			attr && Object.entries(attr).forEach(([k,v]) => el.setAttribute(k,v));
			return el;
		},

		fireEvent = (name, detail, cancelable) => dispatchEvent(
			new CustomEvent(name, {detail:detail, cancelable: !!cancelable})
		),

		// keys / shortcuts
		keyScopeReal = ko.observable('all'),
		keyScope = value => {
			if (!value) {
				return keyScopeFake;
			}
			if (ScopeMenu !== value) {
				keyScopeFake = value;
				if (dropdownVisibility()) {
					value = ScopeMenu;
				}
			}
			keyScopeReal(value);
			shortcuts.setScope(value);
		};

	dropdownVisibility.subscribe(value => {
		if (value) {
			keyScope(ScopeMenu);
		} else if (ScopeMenu === shortcuts.getScope()) {
			keyScope(keyScopeFake);
		}
	});

	leftPanelDisabled.toggle = () => leftPanelDisabled(!leftPanelDisabled());
	leftPanelDisabled.subscribe(value => {
		value && moveAction() && moveAction(false);
		$htmlCL.toggle('rl-left-panel-disabled', value);
	});

	moveAction.subscribe(value => value && leftPanelDisabled() && leftPanelDisabled(false));

	/* eslint quote-props: 0 */

	const /**
	 * @enum {number}
	 */
	SaveSettingStatus = {
		Saving: -2,
		Idle: -1,
		Success: 1,
		Failed: 0
	},

	/**
	 * @enum {number}
	 */
	Notification = {
		RequestError: 1,
		RequestAborted: 2,
		RequestTimeout: 3,

		// Global
		InvalidToken: 101,
		AuthError: 102,

		// User
		ConnectionError: 104,
		DomainNotAllowed: 109,
		AccountNotAllowed: 110,

		ContactsSyncError: 140,

		CantGetMessageList: 201,
		CantGetMessage: 202,
		CantDeleteMessage: 203,
		CantMoveMessage: 204,
		CantCopyMessage: 205,

		CantSaveMessage: 301,
		CantSendMessage: 302,
		InvalidRecipients: 303,

		CantSaveFilters: 351,
		CantGetFilters: 352,
		CantActivateFiltersScript: 353,
		CantDeleteFiltersScript: 354,
	//	FiltersAreNotCorrect: 355,

		CantCreateFolder: 400,
		CantRenameFolder: 401,
		CantDeleteFolder: 402,
		CantSubscribeFolder: 403,
		CantUnsubscribeFolder: 404,
		CantDeleteNonEmptyFolder: 405,

	//	CantSaveSettings: 501,

		DomainAlreadyExists: 601,

		DemoSendMessageError: 750,
		DemoAccountError: 751,

		AccountAlreadyExists: 801,
		AccountDoesNotExist: 802,
		AccountSwitchFailed: 803,

		MailServerError: 901,
		ClientViewError: 902,
		InvalidInputArgument: 903,

		JsonFalse: 950,
		JsonParse: 952,
	//	JsonTimeout: 953,

		UnknownNotification: 998,
		UnknownError: 999,

		// Admin
		CantInstallPackage: 701,
		CantDeletePackage: 702,
		InvalidPluginPackage: 703,
		UnsupportedPluginPackage: 704,
		CantSavePluginSettings: 705
	};

	let __themeTimer = 0,
		__themeJson = null;

	const isArray = Array.isArray,
		arrayLength = array => isArray(array) && array.length,
		isFunction = v => typeof v === 'function',
		pString = value => null != value ? '' + value : '',

		forEachObjectValue = (obj, fn) => Object.values(obj).forEach(fn),

		forEachObjectEntry = (obj, fn) => Object.entries(obj).forEach(([key, value]) => fn(key, value)),

		pInt = (value, defaultValue = 0) => {
			value = parseInt(value, 10);
			return isNaN(value) || !isFinite(value) ? defaultValue : value;
		},

		convertThemeName = theme => theme
			.replace(/@custom$/, '')
			.replace(/([A-Z])/g, ' $1')
			.replace(/[^a-zA-Z0-9]+/g, ' ')
			.trim(),

		defaultOptionsAfterRender = (domItem, item) =>
			item && undefined !== item.disabled && domItem?.classList.toggle('disabled', domItem.disabled = item.disabled),

		changeTheme = (value, themeTrigger = ()=>0) => {
			const themeStyle = elementById('app-theme-style'),
				clearTimer = () => {
					__themeTimer = setTimeout(() => themeTrigger(SaveSettingStatus.Idle), 1000);
					__themeJson = null;
				},
				url = themeStyle.dataset.href.replace(/(Admin|User)\/-\/[^/]+\//, '$1/-/' + value + '/') + 'Json/';

			clearTimeout(__themeTimer);

			themeTrigger(SaveSettingStatus.Saving);

			if (__themeJson) {
				__themeJson.abort();
			}
			let init = {};
			if (window.AbortController) {
				__themeJson = new AbortController();
				init.signal = __themeJson.signal;
			}
			rl.fetchJSON(url, init)
				.then(data => {
					if (2 === arrayLength(data)) {
						themeStyle.textContent = data[1];
						themeTrigger(SaveSettingStatus.Success);
					}
				})
				.then(clearTimer, clearTimer);
		},

		getKeyByValue = (o, v) => Object.keys(o).find(key => o[key] === v);

	const HASH_PREFIX = '#/',
		SERVER_PREFIX = './?',
		VERSION = Settings.app('version'),
		adminPath = () => rl.adminArea() && !Settings.app('adminHostUse'),

		prefix = () => SERVER_PREFIX + (adminPath() ? Settings.app('adminPath') : '');

	const SUB_QUERY_PREFIX = '&q[]=',

		/**
		 * @param {string=} startupUrl
		 * @returns {string}
		 */
		root = () => HASH_PREFIX,

		/**
		 * @returns {string}
		 */
		logoutLink = () => adminPath() ? prefix() : './',

		/**
		 * @param {string} type
		 * @param {string} hash
		 * @param {string=} customSpecSuffix
		 * @returns {string}
		 */
		serverRequestRaw = (type, hash) =>
			SERVER_PREFIX + '/Raw/' + SUB_QUERY_PREFIX + '/'
			+ '0/' // AuthAccountHash ?
			+ (type
				? type + '/' + (hash ? SUB_QUERY_PREFIX + '/' + hash : '')
				: ''),

		/**
		 * @param {string} type
		 * @returns {string}
		 */
		serverRequest = type => prefix() + '/' + type + '/' + SUB_QUERY_PREFIX + '/0/',

		/**
		 * @param {string} lang
		 * @param {boolean} isAdmin
		 * @returns {string}
		 */
		langLink = (lang, isAdmin) =>
			SERVER_PREFIX + '/Lang/0/' + (isAdmin ? 'Admin' : 'App') + '/' + encodeURI(lang) + '/' + VERSION + '/',

		/**
		 * @param {string=} screenName = ''
		 * @returns {string}
		 */
		settings = (screenName = '') => HASH_PREFIX + 'settings' + (screenName ? '/' + screenName : '');

	let I18N_DATA = {};

	const
		i18nToNode = element => {
			const key = element.dataset.i18n;
			if (key) {
				if ('[' === key.slice(0, 1)) {
					switch (key.slice(0, 6)) {
						case '[html]':
							element.innerHTML = i18n(key.slice(6));
							break;
						case '[place':
							element.placeholder = i18n(key.slice(13));
							break;
						case '[title':
							element.title = i18n(key.slice(7));
							break;
						// no default
					}
				} else {
					element.textContent = i18n(key);
				}
			}
		},

		init = () => {
			if (rl.I18N) {
				I18N_DATA = rl.I18N;
				Date.defineRelativeTimeFormat(rl.relativeTime || {});
				rl.I18N = null;
				return 1;
			}
		},

		i18nKey = key => key.replace(/([a-z])([A-Z])/g, '$1_$2').toUpperCase(),

		getNotificationMessage = code => {
			let key = getKeyByValue(Notification, code);
			return key ? I18N_DATA.NOTIFICATIONS[i18nKey(key).replace('_NOTIFICATION', '_ERROR')] : '';
		};

	const trigger = ko.observable(false),

		/**
		 * @param {string} key
		 * @param {Object=} valueList
		 * @param {string=} defaulValue
		 * @returns {string}
		 */
		i18n = (key, valueList, defaulValue) => {
			let result = defaulValue || key;
			if (key) {
				let path = key.split('/');
				if (I18N_DATA[path[0]] && path[1]) {
					result = I18N_DATA[path[0]][path[1]] || result;
				}
			}
			if (valueList) {
				forEachObjectEntry(valueList, (key, value) => {
					result = result.replace('%' + key + '%', value);
				});
			}
			return result;
		},

		/**
		 * @param {Object} elements
		 * @param {boolean=} animate = false
		 */
		i18nToNodes = element =>
			setTimeout(() =>
				element.querySelectorAll('[data-i18n]').forEach(item => i18nToNode(item))
			, 1),

		timestampToString = (timeStampInUTC, formatStr) => {
			const now = Date.now(),
				time = 0 < timeStampInUTC ? Math.min(now, timeStampInUTC * 1000) : (0 === timeStampInUTC ? now : 0);

			if (31536000000 < time) {
				const m = new Date(time);
				switch (formatStr) {
					case 'FROMNOW':
						return m.fromNow();
					case 'SHORT': {
						if (4 >= (now - time) / 3600000)
							return m.fromNow();
						const mt = m.getTime(), date = new Date,
							dt = date.setHours(0,0,0,0);
						if (mt > dt)
							return i18n('MESSAGE_LIST/TODAY_AT', {TIME: m.format('LT')});
						if (mt > dt - 86400000)
							return i18n('MESSAGE_LIST/YESTERDAY_AT', {TIME: m.format('LT')});
						if (date.getFullYear() === m.getFullYear())
							return m.format('d M');
						return m.format('LL');
					}
					case 'FULL':
						return m.format('LLL');
					default:
						return m.format(formatStr);
				}
			}

			return '';
		},

		timeToNode = (element, time) => {
			try {
				if (time) {
					element.dateTime = (new Date(time * 1000)).format('Y-m-d\\TH:i:s');
				} else {
					time = Date.parse(element.dateTime) / 1000;
				}

				let key = element.dataset.momentFormat;
				if (key) {
					element.textContent = timestampToString(time, key);
				}

				if ((key = element.dataset.momentFormatTitle)) {
					element.title = timestampToString(time, key);
				}
			} catch (e) {
				// prevent knockout crashes
				console.error(e);
			}
		},

		reloadTime = () => setTimeout(() =>
				doc.querySelectorAll('time').forEach(element => timeToNode(element))
				, 1),

		/**
		 * @param {Function} startCallback
		 * @param {Function=} langCallback = null
		 */
		initOnStartOrLangChange = (startCallback, langCallback = null) => {
			startCallback?.();
			startCallback && trigger.subscribe(startCallback);
			langCallback && trigger.subscribe(langCallback);
		},

		/**
		 * @param {number} code
		 * @param {*=} message = ''
		 * @param {*=} defCode = null
		 * @returns {string}
		 */
		getNotification = (code, message = '', defCode = 0) => {
			code = parseInt(code, 10) || 0;
			if (Notification.ClientViewError === code && message) {
				return message;
			}

			return getNotificationMessage(code)
				|| getNotificationMessage(parseInt(defCode, 10))
				|| '';
		},

		/**
		 * @param {boolean} admin
		 * @param {string} language
		 */
		translatorReload = (admin, language) =>
			new Promise((resolve, reject) => {
				const script = createElement('script');
				script.onload = () => {
					// reload the data
					if (init()) {
						i18nToNodes(doc);
						admin || reloadTime();
						trigger(!trigger());
					}
					script.remove();
					resolve();
				};
				script.onerror = () => reject(new Error('Language '+language+' failed'));
				script.src = langLink(language, admin);
		//		script.async = true;
				doc.head.append(script);
			}),

		/**
		 *
		 * @param {string} language
		 * @param {boolean=} isEng = false
		 * @returns {string}
		 */
		convertLangName = (language, isEng = false) =>
			i18n(
				'LANGS_NAMES' + (true === isEng ? '_EN' : '') + '/' + language,
				null,
				language
			);

	init();

	const FormDataToObject = formData => {
		var object = {};
		formData.forEach((value, key) => {
			if (!Reflect.has(object, key)){
				object[key] = value;
			} else {
				isArray(object[key]) || (object[key] = [object[key]]);
				object[key].push(value);
			}
		});
		return object;
	};

	var bootstrap = App => {

		rl.app = App;
		rl.logoutReload = App.logoutReload;

		rl.i18n = i18n;

		rl.Enums = {
			StorageResultType: {
				Success: 0,
				Error: 1,
				Abort: 2
			}
		};

		rl.route = {
			root: () => {
				rl.route.off();
				hasher.setHash(root());
			},
			reload: () => {
				rl.route.root();
				setTimeout(() => location.reload(), 100);
			},
			off: () => hasher.active = false,
			on: () => hasher.active = true
		};

		rl.fetch = (resource, init, postData) => {
			init = Object.assign({
				mode: 'same-origin',
				cache: 'no-cache',
				redirect: 'error',
				referrerPolicy: 'no-referrer',
				credentials: 'same-origin',
				headers: {}
			}, init);
			if (postData) {
				init.method = 'POST';
				init.headers['Content-Type'] = 'application/json';
				postData = (postData instanceof FormData) ? FormDataToObject(postData) : postData;
				postData.XToken = Settings.app('token');
				init.body = JSON.stringify(postData);
			}

			return fetch(resource, init);
		};

		rl.fetchJSON = (resource, init, postData) => {
			init = Object.assign({ headers: {} }, init);
			init.headers.Accept = 'application/json';
			return rl.fetch(resource, init, postData).then(response => {
				if (!response.ok) {
					return Promise.reject('Network response error: ' + response.status);
				}
				/* TODO: use this for non-developers?
				response.clone()
				let data = response.text();
				try {
					return JSON.parse(data);
				} catch (e) {
					console.error(e);
	//				console.log(data);
					return Promise.reject(Notification.JsonParse);
					return {
						Result: false,
						ErrorCode: 952, // Notification.JsonParse
						ErrorMessage: e.message,
						ErrorMessageAdditional: data
					}
				}
				*/
				return response.json();
			});
		};

	};

	const errorTip = (element, value) => value
				? setTimeout(() => element.setAttribute('data-rainloopErrorTip', value), 100)
				: element.removeAttribute('data-rainloopErrorTip'),

		/**
		 * The value of the pureComputed observable shouldn’t vary based on the
		 * number of evaluations or other “hidden” information. Its value should be
		 * based solely on the values of other observables in the application
		 */
		koComputable = fn => ko.computed(fn, {'pure':true}),

		addObservablesTo = (target, observables) =>
			forEachObjectEntry(observables, (key, value) =>
				target[key] || (target[key] = /*isArray(value) ? ko.observableArray(value) :*/ ko.observable(value)) ),

		addComputablesTo = (target, computables) =>
			forEachObjectEntry(computables, (key, fn) => target[key] = koComputable(fn)),

		addSubscribablesTo = (target, subscribables) =>
			forEachObjectEntry(subscribables, (key, fn) => target[key].subscribe(fn)),

		dispose = disposable => isFunction(disposable?.dispose) && disposable.dispose();

	Object.assign(ko.bindingHandlers, {
		tooltipErrorTip: {
			init: (element, fValueAccessor) => {
				doc.addEventListener('click', () => {
					let value = fValueAccessor();
					ko.isObservable(value) && !ko.isComputed(value) && value('');
					errorTip(element);
				});
			},
			update: (element, fValueAccessor) => {
				let value = ko.unwrap(fValueAccessor());
				value = isFunction(value) ? value() : value;
				errorTip(element, value);
			}
		},

		onEnter: {
			init: (element, fValueAccessor, fAllBindings, viewModel) => {
				let fn = event => {
					if ('Enter' == event.key) {
						element.dispatchEvent(new Event('change'));
						fValueAccessor().call(viewModel);
					}
				};
				element.addEventListener('keydown', fn);
				ko.utils.domNodeDisposal.addDisposeCallback(element, () => element.removeEventListener('keydown', fn));
			}
		},

		onEsc: {
			init: (element, fValueAccessor, fAllBindings, viewModel) => {
				let fn = event => {
					if ('Escape' == event.key) {
						element.dispatchEvent(new Event('change'));
						fValueAccessor().call(viewModel);
					}
				};
				element.addEventListener('keyup', fn);
				ko.utils.domNodeDisposal.addDisposeCallback(element, () => element.removeEventListener('keyup', fn));
			}
		},

		onSpace: {
			init: (element, fValueAccessor, fAllBindings, viewModel) => {
				let fn = event => {
					if (' ' == event.key) {
						fValueAccessor().call(viewModel, event);
					}
				};
				element.addEventListener('keyup', fn);
				ko.utils.domNodeDisposal.addDisposeCallback(element, () => element.removeEventListener('keyup', fn));
			}
		},

		i18nUpdate: {
			update: (element, fValueAccessor) => {
				ko.unwrap(fValueAccessor());
				i18nToNodes(element);
			}
		},

		title: {
			update: (element, fValueAccessor) => element.title = ko.unwrap(fValueAccessor())
		},

		command: {
			init: (element, fValueAccessor, fAllBindings, viewModel, bindingContext) => {
				const command = fValueAccessor();

				if (!command || !command.canExecute) {
					throw new Error('Value should be a command');
				}

				ko.bindingHandlers['FORM'==element.nodeName ? 'submit' : 'click'].init(
					element,
					fValueAccessor,
					fAllBindings,
					viewModel,
					bindingContext
				);
			},
			update: (element, fValueAccessor) => {
				const cl = element.classList,
					command = fValueAccessor();

				let disabled = !command.canExecute();
				cl.toggle('disabled', disabled);

				if (element.matches('INPUT,TEXTAREA,BUTTON')) {
					element.disabled = disabled;
				}
			}
		},

		saveTrigger: {
			init: (element) => {
				let icon = element;
				if (element.matches('input,select,textarea')) {
					element.classList.add('settings-save-trigger-input');
					element.after(element.saveTriggerIcon = icon = createElement('span'));
				}
				icon.classList.add('settings-save-trigger');
			},
			update: (element, fValueAccessor) => {
				const value = parseInt(ko.unwrap(fValueAccessor()),10);
				let cl = (element.saveTriggerIcon || element).classList;
				if (element.saveTriggerIcon) {
					cl.toggle('saving', value === SaveSettingStatus.Saving);
					cl.toggle('success', value === SaveSettingStatus.Success);
					cl.toggle('error', value === SaveSettingStatus.Failed);
				}
				cl = element.classList;
				cl.toggle('success', value === SaveSettingStatus.Success);
				cl.toggle('error', value === SaveSettingStatus.Failed);
			}
		}
	});

	// extenders

	ko.extenders.toggleSubscribeProperty = (target, options) => {
		const prop = options[1];
		if (prop) {
			target.subscribe(
				prev => prev?.[prop]?.(false),
				options[0],
				'beforeChange'
			);

			target.subscribe(next => next?.[prop]?.(true), options[0]);
		}

		return target;
	};

	ko.extenders.falseTimeout = (target, option) => {
		target.subscribe((() => target(false)).debounce(parseInt(option, 10) || 0));
		return target;
	};

	// functions

	ko.observable.fn.askDeleteHelper = function() {
		return this.extend({ falseTimeout: 3000, toggleSubscribeProperty: [this, 'askDelete'] });
	};

	let iJsonErrorCount = 0;

	const getURL = (add = '') => serverRequest('Json') + add,

	checkResponseError = data => {
		const err = data ? data.ErrorCode : null;
		if (Notification.InvalidToken === err) {
			alert(getNotification(err));
			rl.logoutReload();
		} else if ([
				Notification.AuthError,
				Notification.ConnectionError,
				Notification.DomainNotAllowed,
				Notification.AccountNotAllowed,
				Notification.MailServerError,
				Notification.UnknownNotification,
				Notification.UnknownError
			].includes(err)
		) {
			if (7 < ++iJsonErrorCount) {
				rl.logoutReload();
			}
		}
	},

	oRequests = {},

	abort = (sAction, sReason, bClearOnly) => {
		if (oRequests[sAction]) {
			if (!bClearOnly && oRequests[sAction].abort) {
	//			oRequests[sAction].__aborted = true;
				oRequests[sAction].abort(sReason || 'AbortError');
			}

			oRequests[sAction] = null;
			delete oRequests[sAction];
		}
	},

	fetchJSON = (action, sGetAdd, params, timeout, jsonCallback) => {
		params = params || {};
		if (params instanceof FormData) {
			params.set('Action', action);
		} else {
			params.Action = action;
		}
	//	abort(action);
		const controller = new AbortController(),
			signal = controller.signal;
		oRequests[action] = controller;
		// Currently there is no way to combine multiple signals, so AbortSignal.timeout() not possible
		timeout && setTimeout(() => abort(action, 'TimeoutError'), timeout);
		return rl.fetchJSON(getURL(sGetAdd), {signal: signal}, sGetAdd ? null : params).then(jsonCallback).catch(err => {
			err.aborted = signal.aborted;
			err.reason = signal.reason;
			return Promise.reject(err);
		});
	};

	class FetchError extends Error
	{
		constructor(code, message) {
			super(message);
			this.code = code || Notification.JsonFalse;
		}
	}

	class AbstractFetchRemote
	{
		abort(sAction) {
			abort(sAction);
			return this;
		}

		/**
		 * Allows quicker visual responses to the user.
		 * Can be used to stream lines of json encoded data, but does not work on all servers.
		 * Apache needs 'flushpackets' like in <Proxy "fcgi://...." flushpackets=on></Proxy>
		 */
		streamPerLine(fCallback, sGetAdd, postData) {
			rl.fetch(getURL(sGetAdd), {}, postData)
			.then(response => response.body)
			.then(body => {
				let buffer = '';
				const
					// Firefox TextDecoderStream is not defined
	//				reader = body.pipeThrough(new TextDecoderStream()).getReader();
					reader = body.getReader(),
					re = /\r\n|\n|\r/gm,
					utf8decoder = new TextDecoder(),
					processText = ({ done, value }) => {
						buffer += value ? utf8decoder.decode(value, {stream: true}) : '';
						for (;;) {
							let result = re.exec(buffer);
							if (!result) {
								if (done) {
									break;
								}
								reader.read().then(processText);
								return;
							}
							fCallback(buffer.slice(0, result.index));
							buffer = buffer.slice(result.index + 1);
							re.lastIndex = 0;
						}
						if (buffer.length) {
							// last line didn't end in a newline char
							fCallback(buffer);
						}
					};
				reader.read().then(processText);
			});
		}

		/**
		 * @param {?Function} fCallback
		 * @param {string} sAction
		 * @param {Object=} oParameters
		 * @param {?number=} iTimeout
		 * @param {string=} sGetAdd = ''
		 */
		request(sAction, fCallback, params, iTimeout, sGetAdd, abortActions) {
			params = params || {};

			const start = Date.now();

			abortActions && console.error('abortActions is obsolete');

			fetchJSON(sAction, pString(sGetAdd),
				params,
				undefined === iTimeout ? 30000 : pInt(iTimeout),
				data => {
					let cached = false;
					if (data?.Time) {
						cached = pInt(data.Time) > Date.now() - start;
					}

					let iError = 0;
					if (sAction && oRequests[sAction]) {
						if (oRequests[sAction].__aborted) {
							iError = 2;
						}
						abort(sAction, 0, 1);
					}

					if (!iError && data) {
	/*
						if (sAction !== data.Action) {
							console.log(sAction + ' !== ' + data.Action);
						}
	*/
						if (data.Result) {
							iJsonErrorCount = 0;
						} else {
							checkResponseError(data);
							iError = data.ErrorCode || Notification.UnknownError;
						}
					}

					fCallback && fCallback(
						iError,
						data,
						cached,
						sAction,
						params
					);
				}
			)
			.catch(err => {
				console.error({fetchError:err});
				fCallback && fCallback(
					'TimeoutError' == err.reason ? 3 : (err.name == 'AbortError' ? 2 : 1),
					err
				);
			});
		}

		/**
		 * @param {?Function} fCallback
		 */
		getPublicKey(fCallback) {
			this.request('GetPublicKey', fCallback);
		}

		setTrigger(trigger, value) {
			if (trigger) {
				value = !!value;
				(isArray(trigger) ? trigger : [trigger]).forEach(fTrigger => {
					fTrigger?.(value);
				});
			}
		}

		post(action, fTrigger, params, timeOut) {
			this.setTrigger(fTrigger, true);
			return fetchJSON(action, '', params, pInt(timeOut, 30000),
				data => {
					abort(action, 0, 1);

					if (!data) {
						return Promise.reject(new FetchError(Notification.JsonParse));
					}
	/*
					let isCached = false, type = '';
					if (data?.Time) {
						isCached = pInt(data.Time) > microtime() - start;
					}
					// backward capability
					switch (true) {
						case 'success' === textStatus && data?.Result && action === data.Action:
							type = AbstractFetchRemote.SUCCESS;
							break;
						case 'abort' === textStatus && (!data || !data.__aborted__):
							type = AbstractFetchRemote.ABORT;
							break;
						default:
							type = AbstractFetchRemote.ERROR;
							break;
					}
	*/
					this.setTrigger(fTrigger, false);

					if (!data.Result || action !== data.Action) {
						checkResponseError(data);
						return Promise.reject(new FetchError(
							data ? data.ErrorCode : 0,
							data ? (data.ErrorMessageAdditional || data.ErrorMessage) : ''
						));
					}

					return data;
				}
			);
		}
	}

	Object.assign(AbstractFetchRemote.prototype, {
		SUCCESS : 0,
		ERROR : 1,
		ABORT : 2
	});

	class RemoteAdminFetch extends AbstractFetchRemote {

		/**
		 * @param {string} key
		 * @param {?scalar} value
		 * @param {?Function} fCallback
		 */
		saveSetting(key, value, fCallback) {
			this.request('AdminSettingsUpdate', fCallback, {[key]: value});
		}

	}

	var Remote = new RemoteAdminFetch();

	class AbstractScreen {
		constructor(screenName, viewModels = []) {
			this.__cross = null;
			this.screenName = screenName;
			this.viewModels = isArray(viewModels) ? viewModels : [];
		}

		/**
		 * @returns {?Array)}
		 */
		routes() {
			return null;
		}

	/*
		onBuild(viewModelDom) {}
		onShow() {}
		onHide() {}
		__started
		__builded
	*/

		/**
		 * @returns {void}
		 */
		onStart() {
			const routes = this.routes();
			if (arrayLength(routes)) {
				let route = new Crossroads(),
					fMatcher = (this.onRoute || (()=>0)).bind(this);

				routes.forEach(item => item && (route.addRoute(item[0], fMatcher).rules = item[1]));

				this.__cross = route;
			}
		}
	}

	const VIEW_MODELS = [];

	class AbstractSettingsScreen extends AbstractScreen {
		/**
		 * @param {Array} viewModels
		 */
		constructor(viewModels) {
			super('settings', viewModels);

			this.menu = ko.observableArray();

			this.oCurrentSubScreen = null;
		}

		onRoute(subName) {
			let settingsScreen = null,
				viewModelDom = null,
				RoutedSettingsViewModel = VIEW_MODELS.find(
					SettingsViewModel => subName === SettingsViewModel.__rlSettingsData.route
				);

			if (RoutedSettingsViewModel) {
				const vmPlace = elementById('rl-settings-subscreen');
				if (RoutedSettingsViewModel.__vm) {
					settingsScreen = RoutedSettingsViewModel.__vm;
				} else if (vmPlace) {
					viewModelDom = createElement('div',{
						id: 'V-Settings-' + RoutedSettingsViewModel.name.replace(/(User|Admin)Settings/,''),
						hidden: ''
					});
					vmPlace.append(viewModelDom);

					settingsScreen = new RoutedSettingsViewModel();
					settingsScreen.viewModelDom = viewModelDom;

					RoutedSettingsViewModel.__dom = viewModelDom;
					RoutedSettingsViewModel.__vm = settingsScreen;

					ko.applyBindingAccessorsToNode(
						viewModelDom,
						{
							template: () => ({ name: RoutedSettingsViewModel.__rlSettingsData.template })
						},
						settingsScreen
					);

					settingsScreen.onBuild?.(viewModelDom);
				} else {
					console.log('Cannot find sub settings view model position: SettingsSubScreen');
				}

				if (settingsScreen) {
					setTimeout(() => {
						// hide
						this.onHide();
						// --

						this.oCurrentSubScreen = settingsScreen;

						// show
						settingsScreen.beforeShow?.();
						i18nToNodes(settingsScreen.viewModelDom);
						settingsScreen.viewModelDom.hidden = false;
						settingsScreen.onShow?.();

						this.menu.forEach(item => {
							item.selected(
								item.route === RoutedSettingsViewModel.__rlSettingsData.route
							);
						});

						(vmPlace || {}).scrollTop = 0;
						// --
					}, 1);
				}
			} else {
				hasher.replaceHash(settings());
			}
		}

		onHide() {
			let subScreen = this.oCurrentSubScreen;
			if (subScreen) {
				subScreen.onHide?.();
				subScreen.viewModelDom.hidden = true;
			}
		}

		onBuild() {
			VIEW_MODELS.forEach(SettingsViewModel => this.menu.push(SettingsViewModel.__rlSettingsData));
		}

		routes() {
			const DefaultViewModel = VIEW_MODELS.find(
					SettingsViewModel => SettingsViewModel.__rlSettingsData.isDefault
				),
				defaultRoute =
					DefaultViewModel ? DefaultViewModel.__rlSettingsData.route : 'general',
				rules = {
					subname: /^(.*)$/,
					normalize_: (rquest, vals) => {
						vals.subname = null == vals.subname ? defaultRoute : pString(vals.subname);
						return [vals.subname];
					}
				};

			return [
				['{subname}/', rules],
				['{subname}', rules],
				['', rules]
			];
		}
	}

	/**
	 * @param {Function} SettingsViewModelClass
	 * @param {string} template
	 * @param {string} labelName
	 * @param {string} route
	 * @param {boolean=} isDefault = false
	 * @returns {void}
	 */
	function settingsAddViewModel(SettingsViewModelClass, template, labelName, route, isDefault = false) {
		let name = SettingsViewModelClass.name.replace(/(User|Admin)Settings/, '');
		SettingsViewModelClass.__rlSettingsData = {
			label: labelName || 'SETTINGS_LABELS/LABEL_' + name.toUpperCase() + '_NAME',
			route: route || name.toLowerCase(),
			selected: ko.observable(false),
			template: template || SettingsViewModelClass.name,
			isDefault: !!isDefault
		};

		VIEW_MODELS.push(SettingsViewModelClass);
	}

	let
		SCREENS = {},
		currentScreen = null,
		defaultScreenName = '';

	const
		autofocus = dom => dom.querySelector('[autofocus]')?.focus(),

		visiblePopups = new Set,

		/**
		 * @param {string} screenName
		 * @returns {?Object}
		 */
		screen = screenName => (screenName && SCREENS[screenName]) || null,

		/**
		 * @param {Function} ViewModelClass
		 * @param {Object=} vmScreen
		 * @returns {*}
		 */
		buildViewModel = (ViewModelClass, vmScreen) => {
			if (ViewModelClass && !ViewModelClass.__builded) {
				let vmDom = null;
				const
					vm = new ViewModelClass(vmScreen),
					id = vm.viewModelTemplateID,
					position = vm.viewType || '',
					dialog = ViewTypePopup === position,
					vmPlace = position ? doc.getElementById('rl-' + position.toLowerCase()) : null;

				ViewModelClass.__builded = true;
				ViewModelClass.__vm = vm;

				if (vmPlace) {
					vmDom = Element.fromHTML(dialog
						? '<dialog id="V-'+ id + '"></dialog>'
						: '<div id="V-'+ id + '" hidden=""></div>');
					vmPlace.append(vmDom);

					vm.viewModelDom = vmDom;
					ViewModelClass.__dom = vmDom;

					if (ViewTypePopup === position) {
						vm.close = () => hideScreenPopup(ViewModelClass);

						// Firefox < 98 / Safari < 15.4 HTMLDialogElement not defined
						if (!vmDom.showModal) {
							vmDom.showModal = () => {
								vmDom.backdrop ||
									vmDom.before(vmDom.backdrop = Element.fromHTML('<div class="dialog-backdrop"></div>'));
								vmDom.setAttribute('open','');
								vmDom.open = true;
								vmDom.returnValue = null;
								vmDom.backdrop.hidden = false;
							};
							vmDom.close = v => {
	//							if (vmDom.dispatchEvent(new CustomEvent('cancel', {cancelable:true}))) {
									vmDom.backdrop.hidden = true;
									vmDom.returnValue = v;
									vmDom.removeAttribute('open', null);
									vmDom.open = false;
	//								vmDom.dispatchEvent(new CustomEvent('close'));
	//							}
							};
						}
						// https://developer.mozilla.org/en-US/docs/Web/API/HTMLDialogElement/cancel_event
	//					vmDom.addEventListener('cancel', event => (false === vm.onClose() && event.preventDefault()));
	//					vmDom.addEventListener('close', () => vm.modalVisible(false));

						// show/hide popup/modal
						const endShowHide = e => {
							if (e.target === vmDom) {
								if (vmDom.classList.contains('animate')) {
									autofocus(vmDom);
									vm.afterShow?.();
								} else {
									vmDom.close();
									vm.afterHide?.();
								}
							}
						};

						vm.modalVisible.subscribe(value => {
							if (value) {
								i18nToNodes(vmDom);
								visiblePopups.add(vm);
								vmDom.style.zIndex = 3001 + (visiblePopups.size * 2);
								vmDom.showModal();
								if (vmDom.backdrop) {
									vmDom.backdrop.style.zIndex = 3000 + (visiblePopups.size * 2);
								}
								vm.keyScope.set();
								requestAnimationFrame(() => { // wait just before the next paint
									vmDom.offsetHeight; // force a reflow
									vmDom.classList.add('animate'); // trigger the transitions
								});
							} else {
								visiblePopups.delete(vm);
								vm.onHide?.();
								vm.keyScope.unset();
								vmDom.classList.remove('animate'); // trigger the transitions
							}
							arePopupsVisible(0 < visiblePopups.size);
						});
						vmDom.addEventListener('transitionend', endShowHide);
					}

					ko.applyBindingAccessorsToNode(
						vmDom,
						{
							template: () => ({ name: id })
						},
						vm
					);

					vm.onBuild?.(vmDom);

					fireEvent('rl-view-model', vm);
				} else {
					console.log('Cannot find view model position: ' + position);
				}
			}

			return ViewModelClass && ViewModelClass.__vm;
		},

		forEachViewModel = (screen, fn) => {
			screen.viewModels.forEach(ViewModelClass => {
				if (
					ViewModelClass.__vm &&
					ViewModelClass.__dom &&
					ViewTypePopup !== ViewModelClass.__vm.viewType
				) {
					fn(ViewModelClass.__vm, ViewModelClass.__dom);
				}
			});
		},

		hideScreen = (screenToHide, destroy) => {
			screenToHide.onHide?.();
			forEachViewModel(screenToHide, (vm, dom) => {
				dom.hidden = true;
				vm.onHide?.();
				destroy && vm.viewModelDom.remove();
			});
		},

		/**
		 * @param {Function} ViewModelClassToHide
		 * @returns {void}
		 */
		hideScreenPopup = ViewModelClassToHide => {
			if (ViewModelClassToHide?.__vm && ViewModelClassToHide?.__dom) {
				ViewModelClassToHide.__vm.modalVisible(false);
			}
		},

		/**
		 * @param {string} screenName
		 * @param {string} subPart
		 * @returns {void}
		 */
		screenOnRoute = (screenName, subPart) => {
			let vmScreen = null,
				isSameScreen = false;

			if (null == screenName || '' == screenName) {
				screenName = defaultScreenName;
			}

			if (fireEvent('sm-show-screen', screenName, 1)) {

				// Close all popups
				for (let vm of visiblePopups) {
					(false === vm.onClose()) || vm.close();
				}

				if (screenName) {
					vmScreen = screen(screenName);
					if (!vmScreen) {
						vmScreen = screen(defaultScreenName);
						if (vmScreen) {
							subPart = screenName + '/' + subPart;
							screenName = defaultScreenName;
						}
					}

					if (vmScreen?.__started) {
						isSameScreen = currentScreen && vmScreen === currentScreen;

						if (!vmScreen.__builded) {
							vmScreen.__builded = true;

							vmScreen.viewModels.forEach(ViewModelClass =>
								buildViewModel(ViewModelClass, vmScreen)
							);

							vmScreen.onBuild?.();
						}

						setTimeout(() => {
							// hide screen
							if (currentScreen && !isSameScreen) {
								hideScreen(currentScreen);
							}
							// --

							currentScreen = vmScreen;

							// show screen
							if (!isSameScreen) {
								vmScreen.onShow?.();

								forEachViewModel(vmScreen, (vm, dom) => {
									vm.beforeShow?.();
									i18nToNodes(dom);
									dom.hidden = false;
									vm.onShow?.();
									autofocus(dom);
								});
							}
							// --

							vmScreen.__cross?.parse(subPart);
						}, 1);
					}
				}
			}
		};


	const
		ViewTypePopup = 'Popups',

		/**
		 * @param {Function} ViewModelClassToShow
		 * @param {Array=} params
		 * @returns {void}
		 */
		showScreenPopup = (ViewModelClassToShow, params = []) => {
			const vm = buildViewModel(ViewModelClassToShow) && ViewModelClassToShow.__dom && ViewModelClassToShow.__vm;

			if (vm) {
				params = params || [];

				vm.beforeShow?.(...params);

				vm.modalVisible(true);

				vm.onShow?.(...params);
			}
		},

		arePopupsVisible = ko.observable(false),

		/**
		 * @param {Array} screensClasses
		 * @returns {void}
		 */
		startScreens = screensClasses => {
			hasher.clear();
			forEachObjectValue(SCREENS, screen => hideScreen(screen, 1));
			SCREENS = {};
			currentScreen = null,
			defaultScreenName = '';

			screensClasses.forEach(CScreen => {
				if (CScreen) {
					const vmScreen = new CScreen(),
						screenName = vmScreen.screenName;
					defaultScreenName || (defaultScreenName = screenName);
					SCREENS[screenName] = vmScreen;
				}
			});

			forEachObjectValue(SCREENS, vmScreen => {
				if (!vmScreen.__started) {
					vmScreen.onStart();
					vmScreen.__started = true;
				}
			});

			const cross = new Crossroads();
			cross.addRoute(/^([a-zA-Z0-9-]*)\/?(.*)$/, screenOnRoute);

			hasher.add(cross.parse.bind(cross));
			hasher.init();

			setTimeout(() => $htmlCL.remove('rl-started-trigger'), 100);

			const c = elementById('rl-content'), l = elementById('rl-loading');
			c && (c.hidden = false);
			l?.remove();
		},

		/**
		 * Used by ko.bindingHandlers.command (template data-bind="command: ")
		 * to enable/disable click/submit action.
		 */
		decorateKoCommands = (thisArg, commands) =>
			forEachObjectEntry(commands, (key, canExecute) => {
				let command = thisArg[key],
					fn = (...args) => fn.canExecute() && command.apply(thisArg, args);

				fn.canExecute = koComputable(() => canExecute.call(thisArg, thisArg));

				thisArg[key] = fn;
			});

	ko.decorateCommands = decorateKoCommands;

	class AbstractView {
		constructor(templateID, type)
		{
	//		Object.defineProperty(this, 'viewModelTemplateID', { value: templateID });
			this.viewModelTemplateID = templateID || this.constructor.name.replace('UserView', '');
			this.viewType = type;
			this.viewModelDom = null;

			this.keyScope = {
				scope: 'none',
				previous: 'none',
				set: function() {
					this.previous = keyScope();
					keyScope(this.scope);
				},
				unset: function() {
					keyScope(this.previous);
				}
			};
		}

	/*
		onBuild() {}
		beforeShow() {} // Happens before: hidden = false
		onShow() {}       // Happens after: hidden = false
		onHide() {}
	*/

		querySelector(selectors) {
			return this.viewModelDom.querySelector(selectors);
		}

		addObservables(observables) {
			addObservablesTo(this, observables);
		}

		addComputables(computables) {
			addComputablesTo(this, computables);
		}

		addSubscribables(subscribables) {
			addSubscribablesTo(this, subscribables);
		}

	}

	class AbstractViewPopup extends AbstractView
	{
		constructor(name)
		{
			super('Popups' + name, ViewTypePopup);
			this.keyScope.scope = name;
			this.modalVisible = ko.observable(false).extend({ rateLimit: 0 });
			shortcuts.add('escape,close', '', name, () => {
				if (this.modalVisible() && false !== this.onClose()) {
					this.close();
				}
				return false;
	//			return true; Issue with supported modal close
			});
		}

		// Happens when user hits Escape or Close key
		// return false to prevent closing
		onClose() {}

	/*
		beforeShow() {} // Happens before showModal()
		onShow() {}     // Happens after  showModal()
		afterShow() {}  // Happens after  showModal() animation transitionend
		onHide() {}     // Happens before animation transitionend
		afterHide() {}  // Happens after  animation transitionend

		close() {}
	*/
	}

	AbstractViewPopup.showModal = function(params = []) {
		showScreenPopup(this, params);
	};

	AbstractViewPopup.hidden = function() {
		return !this.__vm || !this.__vm.modalVisible();
	};

	class AbstractViewLeft extends AbstractView
	{
		constructor(templateID)
		{
			super(templateID, 'Left');
			this.leftPanelDisabled = leftPanelDisabled;
		}
	}

	class AbstractViewRight extends AbstractView
	{
		constructor(templateID)
		{
			super(templateID, 'Right');
		}
	}

	class AbstractViewSettings
	{
	/*
		onBuild(viewModelDom) {}
		beforeShow() {}
		onShow() {}
		onHide() {}
		viewModelDom
	*/
		addSetting(name, valueCb)
		{
			let prop = name[0].toLowerCase() + name.slice(1),
				trigger = prop + 'Trigger';
			addObservablesTo(this, {
				[prop]: SettingsGet(name),
				[trigger]: SaveSettingStatus.Idle,
			});
			addSubscribablesTo(this, {
				[prop]: (value => {
					this[trigger](SaveSettingStatus.Saving);
					valueCb?.(value);
					rl.app.Remote.saveSetting(name, value,
						iError => {
							this[trigger](iError ? SaveSettingStatus.Failed : SaveSettingStatus.Success);
							setTimeout(() => this[trigger](SaveSettingStatus.Idle), 1000);
						}
					);
				}).debounce(999),
			});
		}

		addSettings(names)
		{
			names.forEach(name => {
				let prop = name[0].toLowerCase() + name.slice(1);
				this[prop] || (this[prop] = ko.observable(SettingsGet(name)));
				this[prop].subscribe(value => rl.app.Remote.saveSetting(name, value));
			});
		}
	}

	class AbstractViewLogin extends AbstractView {
		constructor(templateID) {
			super(templateID, 'Content');
			this.formError = ko.observable(false).extend({ falseTimeout: 500 });
		}

		onBuild(dom) {
			dom.classList.add('LoginView');
		}

		onShow() {
			elementById('rl-left').hidden = true;
			elementById('rl-right').hidden = true;
			rl.route.off();
		}

		onHide() {
			elementById('rl-left').hidden = false;
			elementById('rl-right').hidden = false;
		}

		submitForm() {
	//		return false;
		}
	}

	const USER_VIEW_MODELS_HOOKS = [],
		ADMIN_VIEW_MODELS_HOOKS = [];

	/**
	 * @param {Function} callback
	 * @param {string} action
	 * @param {Object=} parameters
	 * @param {?number=} timeout
	 */
	rl.pluginRemoteRequest = (callback, action, parameters, timeout) => {
		rl.app.Remote.request('Plugin' + action, callback, parameters, timeout);
	};

	/**
	 * @param {Function} SettingsViewModelClass
	 * @param {string} labelName
	 * @param {string} template
	 * @param {string} route
	 */
	rl.addSettingsViewModel = (SettingsViewModelClass, template, labelName, route) => {
		USER_VIEW_MODELS_HOOKS.push([SettingsViewModelClass, template, labelName, route]);
	};

	/**
	 * @param {Function} SettingsViewModelClass
	 * @param {string} labelName
	 * @param {string} template
	 * @param {string} route
	 */
	rl.addSettingsViewModelForAdmin = (SettingsViewModelClass, template, labelName, route) => {
		ADMIN_VIEW_MODELS_HOOKS.push([SettingsViewModelClass, template, labelName, route]);
	};

	/**
	 * @param {boolean} admin
	 */
	function runSettingsViewModelHooks(admin) {
		(admin ? ADMIN_VIEW_MODELS_HOOKS : USER_VIEW_MODELS_HOOKS).forEach(view => {
			settingsAddViewModel(view[0], view[1], view[2], view[3]);
		});
	}

	/**
	 * @param {string} pluginSection
	 * @param {string} name
	 * @returns {?}
	 */
	rl.pluginSettingsGet = (pluginSection, name) =>
		SettingsGet('Plugins')?.[pluginSection]?.[name];

	rl.pluginPopupView = AbstractViewPopup;

	const ThemeStore = {
		theme: ko.observable(''),
		themes: ko.observableArray(),
		userBackgroundName: ko.observable(''),
		userBackgroundHash: ko.observable(''),
		isMobile: ko.observable($htmlCL.contains('rl-mobile')),

		populate: () => {
			const themes = Settings.app('themes');

			ThemeStore.themes(isArray(themes) ? themes : []);
			ThemeStore.theme(SettingsGet('Theme'));
			if (!ThemeStore.isMobile()) {
				ThemeStore.userBackgroundName(SettingsGet('UserBackgroundName'));
				ThemeStore.userBackgroundHash(SettingsGet('UserBackgroundHash'));
			}

			leftPanelDisabled(ThemeStore.isMobile());
		}
	};

	ThemeStore.isMobile.subscribe(value => $htmlCL.toggle('rl-mobile', value));

	ThemeStore.userBackgroundHash.subscribe(value => {
		if (value) {
			$htmlCL.add('UserBackground');
			doc.body.style.backgroundImage = "url("+serverRequestRaw('UserBackground', value)+")";
		} else {
			$htmlCL.remove('UserBackground');
			doc.body.removeAttribute('style');
		}
	});

	const LanguageStore = {
		language: ko.observable(''),
		languages: ko.observableArray(),
		userLanguage: ko.observable(''),

		populate: function() {
			const aLanguages = Settings.app('languages');
			this.languages(isArray(aLanguages) ? aLanguages : []);
			this.language(SettingsGet('Language'));
			this.userLanguage(SettingsGet('UserLanguage'));
		}
	};

	class LanguagesPopupView extends AbstractViewPopup {
		constructor() {
			super('Languages');

			this.fLang = null;
			this.userLanguage = ko.observable('');

			this.langs = ko.observableArray();

			this.languages = koComputable(() => {
				const userLanguage = this.userLanguage();
				return this.langs.map(language => ({
					key: language,
					user: language === userLanguage,
					selected: ko.observable(false),
					fullName: convertLangName(language)
				}));
			});

			this.langs.subscribe(() => this.setLanguageSelection());
		}

		languageTooltipName(language) {
			return convertLangName(language, true);
		}

		setLanguageSelection() {
			const currentLang = this.fLang ? ko.unwrap(this.fLang) : '';
			this.languages().forEach(item => item.selected(item.key === currentLang));
		}

		beforeShow() {
			this.fLang = null;
			this.userLanguage('');

			this.langs([]);
		}

		onShow(fLanguage, langs, userLanguage) {
			this.fLang = fLanguage;
			this.userLanguage(userLanguage || '');

			this.langs(langs);
		}

		changeLanguage(lang) {
			this.fLang?.(lang);
			this.close();
		}
	}

	class AdminSettingsGeneral extends AbstractViewSettings {
		constructor() {
			super();

			this.language = LanguageStore.language;
			this.languages = LanguageStore.languages;

			const aLanguagesAdmin = Settings.app('languagesAdmin');
			this.languagesAdmin = ko.observableArray(isArray(aLanguagesAdmin) ? aLanguagesAdmin : []);
			this.languageAdmin = ko.observable(SettingsGet('LanguageAdmin'));

			this.theme = ThemeStore.theme;
			this.themes = ThemeStore.themes;

			this.addSettings(['AllowLanguagesOnSettings']);

			addObservablesTo(this, {
				attachmentLimitTrigger: SaveSettingStatus.Idle,
				themeTrigger: SaveSettingStatus.Idle,
				capaThemes: SettingsCapa('Themes'),
				capaUserBackground: SettingsCapa('UserBackground'),
				capaAdditionalAccounts: SettingsCapa('AdditionalAccounts'),
				capaIdentities: SettingsCapa('Identities'),
				capaAttachmentThumbnails: SettingsCapa('AttachmentThumbnails'),
				dataFolderAccess: false
			});

			this.weakPassword = rl.app.weakPassword;

			/** https://github.com/RainLoop/rainloop-webmail/issues/1924
			if (this.weakPassword) {
				fetch('./data/VERSION?' + Math.random()).then(response => this.dataFolderAccess(response.ok));
			}
			*/

			this.attachmentLimit = ko
				.observable(SettingsGet('AttachmentLimit') / (1024 * 1024))
				.extend({ debounce: 500 });

			this.addSetting('Language');
			this.addSetting('AttachmentLimit');
			this.addSetting('Theme', value => changeTheme(value, this.themeTrigger));

			this.uploadData = SettingsGet('PhpUploadSizes');
			this.uploadDataDesc =
				(this.uploadData?.upload_max_filesize || this.uploadData?.post_max_size)
					? [
							this.uploadData.upload_max_filesize
								? 'upload_max_filesize = ' + this.uploadData.upload_max_filesize + '; '
								: '',
							this.uploadData.post_max_size ? 'post_max_size = ' + this.uploadData.post_max_size : ''
					  ].join('')
					: '';

			addComputablesTo(this, {
				themesOptions: () => this.themes.map(theme => ({ optValue: theme, optText: convertThemeName(theme) })),

				languageFullName: () => convertLangName(this.language()),
				languageAdminFullName: () => convertLangName(this.languageAdmin())
			});

			this.languageAdminTrigger = ko.observable(SaveSettingStatus.Idle).extend({ debounce: 100 });

			const fReloadLanguageHelper = (saveSettingsStep) => () => {
					this.languageAdminTrigger(saveSettingsStep);
					setTimeout(() => this.languageAdminTrigger(SaveSettingStatus.Idle), 1000);
				},
				fSaveHelper = key => value => Remote.saveSetting(key, value);

			addSubscribablesTo(this, {
				languageAdmin: value => {
					this.languageAdminTrigger(SaveSettingStatus.Saving);
					translatorReload(true, value)
						.then(fReloadLanguageHelper(SaveSettingStatus.Success), fReloadLanguageHelper(SaveSettingStatus.Failed))
						.then(() => Remote.saveSetting('LanguageAdmin', value));
				},

				capaAdditionalAccounts: fSaveHelper('CapaAdditionalAccounts'),

				capaIdentities: fSaveHelper('CapaIdentities'),

				capaAttachmentThumbnails: fSaveHelper('CapaAttachmentThumbnails'),

				capaThemes: fSaveHelper('CapaThemes'),

				capaUserBackground: fSaveHelper('CapaUserBackground')
			});
		}

		selectLanguage() {
			showScreenPopup(LanguagesPopupView, [this.language, this.languages(), LanguageStore.userLanguage()]);
		}

		selectLanguageAdmin() {
			showScreenPopup(LanguagesPopupView, [
				this.languageAdmin,
				this.languagesAdmin(),
				SettingsGet('UserLanguageAdmin')
			]);
		}
	}

	const DomainAdminStore = ko.observableArray();

	DomainAdminStore.loading = ko.observable(false);

	DomainAdminStore.fetch = () => {
		DomainAdminStore.loading(true);
		Remote.request('AdminDomainList',
			(iError, data) => {
				DomainAdminStore.loading(false);
				if (!iError) {
					DomainAdminStore(
						data.Result.map(item => {
							item.disabled = ko.observable(item.disabled);
							item.askDelete = ko.observable(false);
							return item;
						})
					);
				}
			}, {
				IncludeAliases: 1
			});
	};

	class AskPopupView extends AbstractViewPopup {
		constructor() {
			super('Ask');

			this.addObservables({
				askDesc: '',
				yesButton: '',
				noButton: '',
				username: '',
				askUsername: false,
				passphrase: '',
				askPass: false
			});

			this.fYesAction = null;
			this.fNoAction = null;

			this.focusOnShow = true;
		}

		yesClick() {
			this.close();

			isFunction(this.fYesAction) && this.fYesAction();
		}

		noClick() {
			this.close();

			isFunction(this.fNoAction) && this.fNoAction();
		}

		/**
		 * @param {string} sAskDesc
		 * @param {Function=} fYesFunc
		 * @param {Function=} fNoFunc
		 * @param {boolean=} focusOnShow = true
		 * @returns {void}
		 */
		onShow(sAskDesc, fYesFunc = null, fNoFunc = null, focusOnShow = true, ask = 0, btnText = '') {
			this.askDesc(sAskDesc || '');
			this.askUsername(ask & 2);
			this.askPass(ask & 1);
			this.username('');
			this.passphrase('');
			this.yesButton(i18n(btnText || 'GLOBAL/YES'));
			this.noButton(i18n(ask ? 'GLOBAL/CANCEL' : 'GLOBAL/NO'));
			this.fYesAction = fYesFunc;
			this.fNoAction = fNoFunc;
			this.focusOnShow = focusOnShow
				? (ask ? 'input[type="'+(ask&2?'text':'password')+'"]' : '.buttonYes')
				: '';
		}

		afterShow() {
			this.focusOnShow && this.querySelector(this.focusOnShow).focus();
		}

		onClose() {
			this.noClick();
			return false;
		}

		onBuild() {
	//		shortcuts.add('tab', 'shift', 'Ask', () => {
			shortcuts.add('tab,arrowright,arrowleft', '', 'Ask', () => {
				let yes = this.querySelector('.buttonYes'),
					no = this.querySelector('.buttonNo');
				if (yes.matches(':focus')) {
					no.focus();
					return false;
				} else if (no.matches(':focus')) {
					yes.focus();
					return false;
				}
			});
		}
	}

	AskPopupView.password = function(sAskDesc, btnText) {
		return new Promise(resolve => {
			this.showModal([
				sAskDesc,
				() => resolve(this.__vm.passphrase()),
				() => resolve(null),
				true,
				1,
				btnText
			]);
		});
	};

	AskPopupView.credentials = function(sAskDesc, btnText) {
		return new Promise(resolve => {
			this.showModal([
				sAskDesc,
				() => resolve({username:this.__vm.username(), password:this.__vm.passphrase()}),
				() => resolve(null),
				true,
				3,
				btnText
			]);
		});
	};

	const domainToParams = oDomain => ({
				Name: oDomain.name(),

				IncHost: oDomain.imapHost(),
				IncPort: oDomain.imapPort(),
				IncSecure: oDomain.imapSecure(),

				UseSieve: oDomain.useSieve() ? 1 : 0,
				SieveHost: oDomain.sieveHost(),
				SievePort: oDomain.sievePort(),
				SieveSecure: oDomain.sieveSecure(),

				OutHost: oDomain.smtpHost(),
				OutPort: oDomain.smtpPort(),
				OutSecure: oDomain.smtpSecure(),
				OutAuth: oDomain.smtpAuth() ? 1 : 0,
				OutUsePhpMail: oDomain.smtpPhpMail() ? 1 : 0
			});

	class DomainPopupView extends AbstractViewPopup {
		constructor() {
			super('Domain');

			this.addObservables(this.getDefaults());
			this.addObservables({
				edit: false,

				saving: false,

				testing: false,
				testingDone: false,
				testingImapError: false,
				testingSieveError: false,
				testingSmtpError: false,
				testingImapErrorDesc: '',
				testingSieveErrorDesc: '',
				testingSmtpErrorDesc: '',

				imapHostFocus: false,
				sieveHostFocus: false,
				smtpHostFocus: false,
			});

			this.addComputables({
				headerText: () => {
					const name = this.name(),
						aliasName = this.aliasName();

					let result = '';

					if (this.edit()) {
						result = i18n('POPUPS_DOMAIN/TITLE_EDIT_DOMAIN', { NAME: name });
						if (aliasName) {
							result += ' ⫘ ' + aliasName;
						}
					} else {
						result = name
								? i18n('POPUPS_DOMAIN/TITLE_ADD_DOMAIN_WITH_NAME', { NAME: name })
								: i18n('POPUPS_DOMAIN/TITLE_ADD_DOMAIN');
					}

					return result;
				},

				domainDesc: () => {
					const name = this.name();
					return !this.edit() && name ? i18n('POPUPS_DOMAIN/NEW_DOMAIN_DESC', { NAME: '*@' + name }) : '';
				},

				domainIsComputed: () => {
					const usePhpMail = this.smtpPhpMail(),
						useSieve = this.useSieve();

					return (
						this.name() &&
						this.imapHost() &&
						this.imapPort() &&
						(useSieve ? this.sieveHost() && this.sievePort() : true) &&
						((this.smtpHost() && this.smtpPort()) || usePhpMail)
					);
				},

				canBeTested: () => !this.testing() && this.domainIsComputed(),
				canBeSaved: () => !this.saving() && this.domainIsComputed()
			});

			this.addSubscribables({
				testingImapError: value => value || this.testingImapErrorDesc(''),
				testingSieveError: value => value || this.testingSieveErrorDesc(''),
				testingSmtpError: value => value || this.testingSmtpErrorDesc(''),

				// smart form improvements
				imapHostFocus: value =>
					value && this.name() && !this.imapHost() && this.imapHost(this.name().replace(/[.]?[*][.]?/g, '')),

				sieveHostFocus: value =>
					value && this.imapHost() && !this.sieveHost() && this.sieveHost(this.imapHost()),

				smtpHostFocus: value => value && this.imapHost() && !this.smtpHost()
					&& this.smtpHost(this.imapHost().replace(/imap/gi, 'smtp')),

				imapSecure: value => {
					if (this.enableSmartPorts()) {
						const port = pInt(this.imapPort());
						switch (pInt(value)) {
							case 0:
							case 2:
								if (993 === port) {
									this.imapPort(143);
								}
								break;
							case 1:
								if (143 === port) {
									this.imapPort(993);
								}
								break;
							// no default
						}
					}
				},

				smtpSecure: value => {
					if (this.enableSmartPorts()) {
						const port = pInt(this.smtpPort());
						switch (pInt(value)) {
							case 0:
								if (465 === port || 587 === port) {
									this.smtpPort(25);
								}
								break;
							case 1:
								if (25 === port || 587 === port) {
									this.smtpPort(465);
								}
								break;
							case 2:
								if (25 === port || 465 === port) {
									this.smtpPort(587);
								}
								break;
							// no default
						}
					}
				}
			});

			decorateKoCommands(this, {
				createOrAddCommand: self => self.canBeSaved(),
				testConnectionCommand: self => self.canBeTested()
			});
		}

		createOrAddCommand() {
			this.saving(true);
			Remote.request('AdminDomainSave',
				this.onDomainCreateOrSaveResponse.bind(this),
				Object.assign(domainToParams(this), {
					Create: this.edit() ? 0 : 1,

					IncShortLogin: this.imapShortLogin() ? 1 : 0,

					OutShortLogin: this.smtpShortLogin() ? 1 : 0,
					OutSetSender: this.smtpSetSender() ? 1 : 0,

					WhiteList: this.whiteList()
				})
			);
		}

		testConnectionCommand() {
			this.clearTesting(false);
			// https://github.com/the-djmaze/snappymail/issues/477
			AskPopupView.credentials('IMAP', 'GLOBAL/TEST').then(credentials => {
				if (credentials) {
					this.testing(true);
					const params = domainToParams(this);
					params.username = credentials.username;
					params.password = credentials.password;
					Remote.request('AdminDomainTest',
						(iError, oData) => {
							this.testing(false);
							if (iError) {
								this.testingImapError(true);
								this.testingSieveError(true);
								this.testingSmtpError(true);
							} else {
								this.testingDone(true);
								this.testingImapError(true !== oData.Result.Imap);
								this.testingSieveError(true !== oData.Result.Sieve);
								this.testingSmtpError(true !== oData.Result.Smtp);

								if (this.testingImapError() && oData.Result.Imap) {
									this.testingImapErrorDesc('');
									this.testingImapErrorDesc(oData.Result.Imap);
								}

								if (this.testingSieveError() && oData.Result.Sieve) {
									this.testingSieveErrorDesc('');
									this.testingSieveErrorDesc(oData.Result.Sieve);
								}

								if (this.testingSmtpError() && oData.Result.Smtp) {
									this.testingSmtpErrorDesc('');
									this.testingSmtpErrorDesc(oData.Result.Smtp);
								}
							}
						},
						params
					);
				}
			});
		}

		onDomainCreateOrSaveResponse(iError) {
			this.saving(false);
			if (iError) {
				this.savingError(getNotification(iError));
			} else {
				DomainAdminStore.fetch();
				this.close();
			}
		}

		clearTesting() {
			this.testing(false);
			this.testingDone(false);
			this.testingImapError(false);
			this.testingSieveError(false);
			this.testingSmtpError(false);
		}

		onShow(oDomain) {
			this.saving(false);
			this.clearTesting();
			this.clearForm();
			if (oDomain) {
				this.enableSmartPorts(false);
				this.edit(true);
				forEachObjectEntry(oDomain, (key, value) => this[key]?.(value));
				this.enableSmartPorts(true);
			}
		}

		getDefaults() {
			return {
				enableSmartPorts: false,

				savingError: '',

				name: '',

				imapHost: '',
				imapPort: 143,
				imapSecure: 0,
				imapShortLogin: false,

				useSieve: false,
				sieveHost: '',
				sievePort: 4190,
				sieveSecure: 0,

				smtpHost: '',
				smtpPort: 25,
				smtpSecure: 0,
				smtpShortLogin: false,
				smtpAuth: true,
				smtpSetSender: false,
				smtpPhpMail: false,

				whiteList: '',
				aliasName: ''
			};
		}

		clearForm() {
			this.edit(false);
			forEachObjectEntry(this.getDefaults(), (key, value) => this[key](value));
			this.enableSmartPorts(true);
		}
	}

	class DomainAliasPopupView extends AbstractViewPopup {
		constructor() {
			super('DomainAlias');

			this.addObservables({
				saving: false,
				savingError: '',

				name: '',

				alias: ''
			});

			this.addComputables({
				domains: () => DomainAdminStore.filter(item => item && !item.alias),

				domainsOptions: () => this.domains().map(item => ({ optValue: item.name, optText: item.name })),

				canBeSaved: () => !this.saving() && this.name() && this.alias()
			});

			decorateKoCommands(this, {
				createCommand: self => self.canBeSaved()
			});
		}

		createCommand() {
			this.saving(true);
			Remote.request('AdminDomainAliasSave',
				iError => {
					this.saving(false);
					if (iError) {
						this.savingError(getNotification(iError));
					} else {
						DomainAdminStore.fetch();
						this.close();
					}
				}, {
					Name: this.name(),
					Alias: this.alias()
				});
		}

		onShow() {
			this.clearForm();
		}

		clearForm() {
			this.saving(false);
			this.savingError('');

			this.name('');

			this.alias('');
		}
	}

	class AdminSettingsDomains /*extends AbstractViewSettings*/ {
		constructor() {
			this.domains = DomainAdminStore;

			this.domainForDeletion = ko.observable(null).askDeleteHelper();
		}

		createDomain() {
			showScreenPopup(DomainPopupView);
		}

		createDomainAlias() {
			showScreenPopup(DomainAliasPopupView);
		}

		deleteDomain(domain) {
			DomainAdminStore.remove(domain);
			Remote.request('AdminDomainDelete', DomainAdminStore.fetch, {
				Name: domain.name
			});
		}

		disableDomain(domain) {
			domain.disabled(!domain.disabled());
			Remote.request('AdminDomainDisable', DomainAdminStore.fetch, {
				Name: domain.name,
				Disabled: domain.disabled() ? 1 : 0
			});
		}

		onBuild(oDom) {
			oDom.addEventListener('click', event => {
				let el = event.target.closestWithin('.b-admin-domains-list-table .e-action', oDom);
				el && ko.dataFor(el) && Remote.request('AdminDomainLoad',
					(iError, oData) => iError || showScreenPopup(DomainPopupView, [oData.Result]),
					{
						Name: ko.dataFor(el).name
					}
				);

			});

			DomainAdminStore.fetch();
		}
	}

	class AdminSettingsLogin extends AbstractViewSettings {
		constructor() {
			super();
			this.addSetting('LoginDefaultDomain');
			this.addSettings(['DetermineUserLanguage','DetermineUserDomain','AllowLanguagesOnLogin']);
		}
	}

	class AdminSettingsContacts extends AbstractViewSettings {
		constructor() {
			super();
			this.defaultOptionsAfterRender = defaultOptionsAfterRender;

			this.addSetting('ContactsPdoDsn');
			this.addSetting('ContactsPdoUser');
			this.addSetting('ContactsPdoPassword');
			this.addSetting('ContactsPdoType', () => {
				this.testContactsSuccess(false);
				this.testContactsError(false);
				this.testContactsErrorMessage('');
			});

			this.addSettings(['ContactsEnable','ContactsSync']);

			addObservablesTo(this, {
				testing: false,
				testContactsSuccess: false,
				testContactsError: false,
				testContactsErrorMessage: ''
			});

			const supportedTypes = SettingsGet('supportedPdoDrivers') || [],
				types = [{
					id:'sqlite',
					name:'SQLite'
				},{
					id:'mysql',
					name:'MySQL'
				},{
					id:'pgsql',
					name:'PostgreSQL'
				}].filter(type => supportedTypes.includes(type.id));

			this.contactsSupported = 0 < types.length;

			this.contactsTypesOptions = types;

			this.mainContactsType = ko
				.computed({
					read: this.contactsPdoType,
					write: value => {
						if (value !== this.contactsPdoType()) {
							if (supportedTypes.includes(value)) {
								this.contactsPdoType(value);
							} else if (types.length) {
								this.contactsPdoType('');
							}
						} else {
							this.contactsPdoType.valueHasMutated();
						}
					}
				})
				.extend({ notify: 'always' });

			decorateKoCommands(this, {
				testContactsCommand: self => self.contactsPdoDsn() && self.contactsPdoUser()
			});
		}

		testContactsCommand() {
			this.testContactsSuccess(false);
			this.testContactsError(false);
			this.testContactsErrorMessage('');
			this.testing(true);

			Remote.request('AdminContactsTest',
				(iError, data) => {
					this.testContactsSuccess(false);
					this.testContactsError(false);
					this.testContactsErrorMessage('');

					if (!iError && data.Result.Result) {
						this.testContactsSuccess(true);
					} else {
						this.testContactsError(true);
						if (data?.Result) {
							this.testContactsErrorMessage(data.Result.Message || '');
						} else {
							this.testContactsErrorMessage('');
						}
					}

					this.testing(false);
				}, {
					ContactsPdoType: this.contactsPdoType(),
					ContactsPdoDsn: this.contactsPdoDsn(),
					ContactsPdoUser: this.contactsPdoUser(),
					ContactsPdoPassword: this.contactsPdoPassword()
				}
			);
		}

		onShow() {
			this.testContactsSuccess(false);
			this.testContactsError(false);
			this.testContactsErrorMessage('');
		}
	}

	class AdminSettingsSecurity extends AbstractViewSettings {
		constructor() {
			super();

			this.addSettings(['UseLocalProxyForExternalImages','VerifySslCertificate','AllowSelfSigned']);

			this.weakPassword = rl.app.weakPassword;

			addObservablesTo(this, {
				adminLogin: SettingsGet('AdminLogin'),
				adminLoginError: false,
				adminPassword: '',
				adminPasswordNew: '',
				adminPasswordNew2: '',
				adminPasswordNewError: false,
				adminTOTP: '',

				saveError: false,
				saveSuccess: false,

				viewQRCode: '',

				capaOpenPGP: SettingsCapa('OpenPGP')
			});

			const reset = () => {
				this.saveError(false);
				this.saveSuccess(false);
				this.adminPasswordNewError(false);
			};

			addSubscribablesTo(this, {
				adminPassword: () => {
					this.saveError(false);
					this.saveSuccess(false);
				},

				adminLogin: () => this.adminLoginError(false),

				adminTOTP: value => {
					if (/[A-Z2-7]{16,}/.test(value) && 0 == value.length * 5 % 8) {
						Remote.request('AdminQRCode', (iError, data) => {
							if (!iError) {
								console.dir({data:data});
								this.viewQRCode(data.Result);
							}
						}, {
							'username': this.adminLogin(),
							'TOTP': this.adminTOTP()
						});
					} else {
						this.viewQRCode('');
					}
				},

				adminPasswordNew: reset,

				adminPasswordNew2: reset,

				capaOpenPGP: value => Remote.saveSetting('CapaOpenPGP', value)
			});

			this.adminTOTP(SettingsGet('AdminTOTP'));

			decorateKoCommands(this, {
				saveAdminUserCommand: self => self.adminLogin().trim() && self.adminPassword()
			});
		}

		saveAdminUserCommand() {
			if (!this.adminLogin().trim()) {
				this.adminLoginError(true);
				return false;
			}

			if (this.adminPasswordNew() !== this.adminPasswordNew2()) {
				this.adminPasswordNewError(true);
				return false;
			}

			this.saveError(false);
			this.saveSuccess(false);

			Remote.request('AdminPasswordUpdate', (iError, data) => {
				if (iError) {
					this.saveError(true);
				} else {
					this.adminPassword('');
					this.adminPasswordNew('');
					this.adminPasswordNew2('');

					this.saveSuccess(true);

					this.weakPassword(!!data.Result.Weak);
				}
			}, {
				'Login': this.adminLogin(),
				'Password': this.adminPassword(),
				'NewPassword': this.adminPasswordNew(),
				'TOTP': this.adminTOTP()
			});

			return true;
		}

		onHide() {
			this.adminPassword('');
			this.adminPasswordNew('');
			this.adminPasswordNew2('');
		}
	}

	const PackageAdminStore = ko.observableArray();

	PackageAdminStore.real = ko.observable(true);

	PackageAdminStore.loading = ko.observable(false);

	PackageAdminStore.error = ko.observable('');

	PackageAdminStore.fetch = () => {
		PackageAdminStore.loading(true);
		Remote.request('AdminPackagesList', (iError, data) => {
			PackageAdminStore.loading(false);
			if (iError) {
				PackageAdminStore.real(false);
			} else {
				PackageAdminStore.real(!!data.Result.Real);
				PackageAdminStore.error(data.Result.Error);

				const loading = {};
				PackageAdminStore.forEach(item => {
					if (item?.loading()) {
						loading[item.file] = item;
					}
				});

				let list = [];
				if (isArray(data.Result.List)) {
					list = data.Result.List.filter(v => v).map(item => {
						item.loading = ko.observable(loading[item.file] !== undefined);
						item.enabled = ko.observable(item.enabled);
						return item;
					});
				}

				PackageAdminStore(list);
			}
		});
	};

	class PluginPopupView extends AbstractViewPopup {
		constructor() {
			super('Plugin');

			this.addObservables({
				saveError: '',
				id: '',
				name: '',
				readme: ''
			});

			this.config = ko.observableArray();

			this.addComputables({
				hasReadme: () => !!this.readme(),
				hasConfiguration: () => 0 < this.config().length
			});

			this.keyScope.scope = 'all';

			decorateKoCommands(this, {
				saveCommand: self => self.hasConfiguration()
			});
		}

		saveCommand() {
			const oConfig = {
				Id: this.id(),
				Settings: {}
			},
			setItem = item => {
				let value = item.value();
				if (false === value || true === value) {
					value = value ? 1 : 0;
				}
				oConfig.Settings[item.Name] = value;
			};

			this.config.forEach(oItem => {
				if (7 == oItem.Type) {
					// Group
					oItem.config.forEach(oSubItem => setItem(oSubItem));
				} else {
					setItem(oItem);
				}
			});

			this.saveError('');
			Remote.request('AdminPluginSettingsUpdate',
				iError => iError
					? this.saveError(getNotification(iError))
					: this.close(),
				oConfig);
		}

		onShow(oPlugin) {
			this.id('');
			this.name('');
			this.readme('');
			this.config([]);

			if (oPlugin) {
				this.id(oPlugin.Id);
				this.name(oPlugin.Name);
				this.readme(oPlugin.Readme);

				const config = oPlugin.Config;
				if (arrayLength(config)) {
					this.config(
						config.map(item => {
							if (7 == item.Type) {
								// Group
								item.config.forEach(subItem => {
									subItem.value = ko.observable(subItem.value);
								});
							} else {
								item.value = ko.observable(item.value);
							}
							return item;
						})
					);
				}
			}
		}

		onClose() {
			if (AskPopupView.hidden()) {
				showScreenPopup(AskPopupView, [
					i18n('POPUPS_ASK/DESC_WANT_CLOSE_THIS_WINDOW'),
					() => this.close()
				]);
			}
			return false;
		}
	}

	class AdminSettingsPackages extends AbstractViewSettings {
		constructor() {
			super();

			this.addSettings(['EnabledPlugins']);

			addObservablesTo(this, {
				packagesError: ''
			});

			this.packages = PackageAdminStore;

			addComputablesTo(this, {
				packagesCurrent: () => PackageAdminStore().filter(item => item?.installed && !item.canBeUpdated),
				packagesUpdate: () => PackageAdminStore().filter(item => item?.installed && item.canBeUpdated),
				packagesAvailable: () => PackageAdminStore().filter(item => !item?.installed),

				visibility: () => (PackageAdminStore.loading() ? 'visible' : 'hidden')
			});
		}

		onShow() {
			this.packagesError('');
		}

		onBuild(oDom) {
			PackageAdminStore.fetch();

			oDom.addEventListener('click', event => {
				// configurePlugin
				let el = event.target.closestWithin('.package-configure', oDom),
					data = el ? ko.dataFor(el) : 0;
				data && Remote.request('AdminPluginLoad',
					(iError, data) => iError || showScreenPopup(PluginPopupView, [data.Result]),
					{
						Id: data.id
					}
				);
				// disablePlugin
				el = event.target.closestWithin('.package-active', oDom);
				data = el ? ko.dataFor(el) : 0;
				data && this.disablePlugin(data);
			});
		}

		requestHelper(packageToRequest, install) {
			return (iError, data) => {
				PackageAdminStore.forEach(item => {
					if (packageToRequest && item?.loading?.() && packageToRequest.file === item.file) {
						packageToRequest.loading(false);
						item.loading(false);
					}
				});

				if (iError) {
					this.packagesError(
						getNotification(install ? Notification.CantInstallPackage : Notification.CantDeletePackage)
						+ (data.ErrorMessage ? ':\n' + data.ErrorMessage : '')
					);
				} else if (data.Result.Reload) {
					location.reload();
				} else {
					PackageAdminStore.fetch();
				}
			};
		}

		deletePackage(packageToDelete) {
			if (packageToDelete) {
				packageToDelete.loading(true);
				Remote.request('AdminPackageDelete',
					this.requestHelper(packageToDelete, false),
					{
						Id: packageToDelete.id
					}
				);
			}
		}

		installPackage(packageToInstall) {
			if (packageToInstall) {
				packageToInstall.loading(true);
				Remote.request('AdminPackageInstall',
					this.requestHelper(packageToInstall, true),
					{
						Id: packageToInstall.id,
						Type: packageToInstall.type,
						File: packageToInstall.file
					},
					60000
				);
			}
		}

		disablePlugin(plugin) {
			let disable = plugin.enabled();
			plugin.enabled(!disable);
			Remote.request('AdminPluginDisable',
				(iError, data) => {
					if (iError) {
						plugin.enabled(disable);
						this.packagesError(
							(Notification.UnsupportedPluginPackage === iError && data?.ErrorMessage)
							? data.ErrorMessage
							: getNotification(iError)
						);
					}
	//				PackageAdminStore.fetch();
				}, {
					Id: plugin.id,
					Disabled: disable ? 1 : 0
				}
			);
		}

	}

	class AdminSettingsAbout /*extends AbstractViewSettings*/ {
		constructor() {
			this.version = Settings.app('version');
			this.phpextensions = ko.observableArray();

			addObservablesTo(this, {
				coreReal: true,
				coreUpdatable: true,
				coreWarning: false,
				coreVersion: '',
				coreVersionCompare: -2,
				errorDesc: ''
			});
			this.coreChecking = ko.observable(false).extend({ throttle: 100 });
			this.coreUpdating = ko.observable(false).extend({ throttle: 100 });

			this.coreVersionHtmlDesc = ko.computed(() => {
				trigger();
				return i18n('TAB_ABOUT/HTML_NEW_VERSION', { 'VERSION': this.coreVersion() });
			});

			this.statusType = ko.computed(() => {
				let type = '';
				const versionToCompare = this.coreVersionCompare(),
					isChecking = this.coreChecking(),
					isUpdating = this.coreUpdating(),
					isReal = this.coreReal();

				if (isChecking) {
					type = 'checking';
				} else if (isUpdating) {
					type = 'updating';
				} else if (!isReal) {
					type = 'error';
					this.errorDesc('Cannot access the repository at the moment.');
				} else if (0 === versionToCompare) {
					type = 'up-to-date';
				} else if (-1 === versionToCompare) {
					type = 'available';
				}

				return type;
			});
		}

		onBuild() {
			Remote.request('AdminPHPExtensions', (iError, data) => iError || this.phpextensions(data.Result));

	//	beforeShow() {
			this.coreChecking(true);
			Remote.request('AdminUpdateInfo', (iError, data) => {
				this.coreChecking(false);
				if (!iError && data?.Result) {
					this.coreReal(true);
					this.coreUpdatable(!!data.Result.Updatable);
					this.coreWarning(!!data.Result.Warning);
					this.coreVersion(data.Result.Version || '');
					this.coreVersionCompare(data.Result.VersionCompare);
				} else {
					this.coreReal(false);
					this.coreWarning(false);
					this.coreVersion('');
					this.coreVersionCompare(-2);
				}
			});
		}

		updateCoreData() {
			if (!this.coreUpdating()) {
				this.coreUpdating(true);
				Remote.request('AdminUpgradeCore', (iError, data) => {
					this.coreUpdating(false);
					this.coreVersion('');
					this.coreVersionCompare(-2);
					if (!iError && data?.Result) {
						this.coreReal(true);
						window.location.reload();
					} else {
						this.coreReal(false);
					}
				}, {}, 90000);
			}
		}
	}

	class AdminSettingsBranding extends AbstractViewSettings {
		constructor() {
			super();
			this.addSetting('Title');
			this.addSetting('LoadingDescription');
			this.addSetting('FaviconUrl');
		}
	}

	class AdminSettingsConfig /*extends AbstractViewSettings*/ {

		constructor() {
			this.config = ko.observableArray();
			this.saved = ko.observable(false).extend({ falseTimeout: 5000 });
		}

		beforeShow() {
			Remote.request('AdminSettingsGet', (iError, data) => {
				if (!iError) {
					const cfg = [],
						getInputType = (value, pass) => {
							switch (typeof value)
							{
							case 'boolean': return 'checkbox';
							case 'number': return 'number';
							}
							return pass ? 'password' : 'text';
						};
					forEachObjectEntry(data.Result, (key, items) => {
						const section = {
							name: key,
							items: []
						};
						forEachObjectEntry(items, (skey, item) => {
							'admin_password' === skey ||
							section.items.push({
								key: `config[${key}][${skey}]`,
								name: skey,
								value: item[0],
								type: getInputType(item[0], skey.includes('password')),
								comment: item[1]
							});
						});
						cfg.push(section);
					});
					this.config(cfg);
				}
			});
		}

		saveConfig(form) {
			const data = new FormData(form),
				config = {};
			this.config.forEach(section => {
				if (!config[section.name]) {
					config[section.name] = {};
				}
				section.items.forEach(item => {
					let value = data.get(item.key);
					switch (typeof item.value) {
						case 'boolean':
							value = 'on' == value;
							break;
						case 'number':
							value = parseInt(value, 10);
							break;
					}
					config[section.name][item.name] = value;
				});
			});
			Remote.post('AdminSettingsSet', null, {config:config}).then(result => {
				result.Result && this.saved(true);
			});
		}
	}

	class MenuSettingsAdminView extends AbstractViewLeft {
		/**
		 * @param {?} screen
		 */
		constructor(screen) {
			super('AdminMenu');

			this.menu = screen.menu;
		}

		link(route) {
			return '#/' + route;
		}
	}

	class PaneSettingsAdminView extends AbstractViewRight {
		constructor() {
			super('AdminPane');
			this.leftPanelDisabled = leftPanelDisabled;
		}

		logoutClick() {
			Remote.request('AdminLogout', () => rl.logoutReload());
		}
	}

	class SettingsAdminScreen extends AbstractSettingsScreen {
		constructor() {
			super([MenuSettingsAdminView, PaneSettingsAdminView]);

			[
				AdminSettingsGeneral,
				AdminSettingsDomains,
				AdminSettingsLogin,
				AdminSettingsBranding,
				AdminSettingsContacts,
				AdminSettingsSecurity,
				AdminSettingsPackages,
				AdminSettingsConfig,
				AdminSettingsAbout
			].forEach((item, index) =>
				settingsAddViewModel(item, 0, 0, 0, 0 === index)
			);

			runSettingsViewModelHooks(true);
		}

		onShow() {
			rl.setWindowTitle();
		}
	}

	class AdminLoginView extends AbstractViewLogin {
		constructor() {
			super('AdminLogin');

			this.addObservables({
				login: '',
				password: '',
				totp: '',

				loginError: false,
				passwordError: false,

				submitRequest: false,
				submitError: ''
			});

			this.addSubscribables({
				login: () => this.loginError(false),
				password: () => this.passwordError(false)
			});

			decorateKoCommands(this, {
				submitCommand: self => !self.submitRequest()
			});
		}

		submitCommand(self, event) {
			let form = event.target.form,
				data = new FormData(form),
				valid = form.reportValidity() && fireEvent('sm-admin-login', data, 1);

			this.loginError(!this.login());
			this.passwordError(!this.password());
			this.formError(!valid);

			if (valid) {
				this.submitRequest(true);

				Remote.request('AdminLogin',
					(iError, oData) => {
						fireEvent('sm-admin-login-response', {
							error: iError,
							data: oData
						});
						if (iError) {
							this.submitRequest(false);
							this.submitError(getNotification(iError));
						} else {
							rl.setData(oData.Result);
						}
					},
					data
				);
			}

			return valid;
		}
	}

	class LoginAdminScreen extends AbstractScreen {
		constructor() {
			super('login', [AdminLoginView]);
		}

		onShow() {
			rl.setWindowTitle();
		}
	}

	class SelectComponent {
		/**
		 * @param {Object} params
		 */
		constructor(params) {
			this.value = params.value || '';
			this.label = params.label || '';
			this.enable = null == params.enable ? true : params.enable;
			this.trigger = params.trigger?.subscribe ? params.trigger : null;
			this.placeholder = params.placeholder || '';

			this.labeled = null != params.label;

			let size = 0 < params.size ? 'span' + params.size : '';
			if (this.trigger) {
				const
					classForTrigger = ko.observable(''),
					setTriggerState = value => {
						switch (pInt(value)) {
							case SaveSettingStatus.Success:
								classForTrigger('success');
								break;
							case SaveSettingStatus.Failed:
								classForTrigger('error');
								break;
							default:
								classForTrigger('');
								break;
						}
					};

				setTriggerState(this.trigger());

				this.className = koComputable(() =>
					(size + ' settings-save-trigger-input ' + classForTrigger()).trim()
				);

				this.disposables = [
					this.trigger.subscribe(setTriggerState, this),
					this.className
				];
			} else {
				this.className = size;
				this.disposables = [];
			}

			this.options = params.options || '';

			this.optionsText = params.optionsText || null;
			this.optionsValue = params.optionsValue || null;
			this.optionsCaption = i18n(params.optionsCaption || null);

			this.defaultOptionsAfterRender = defaultOptionsAfterRender;
		}

		dispose() {
			this.disposables.forEach(dispose);
		}
	}

	class CheckboxMaterialDesignComponent {
		constructor(params = {}) {
			this.value = ko.isObservable(params.value) ? params.value
				: ko.observable(!!params.value);

			this.enable = ko.isObservable(params.enable) ? params.enable
				: ko.observable(undefined === params.enable || !!params.enable);

			this.label = params.label;
			this.inline = params.inline;

			this.labeled = null != params.label;
		}

		click() {
			this.enable() && this.value(!this.value());
		}
	}

	class AbstractApp {
		/**
		 * @param {RemoteStorage|AdminRemoteStorage} Remote
		 */
		constructor(Remote) {
			this.Remote = Remote;
		}

		logoutReload() {
			const url = logoutLink();

			if (location.href !== url) {
				setTimeout(() => location.href = url, 100);
			} else {
				rl.route.reload();
			}
		}

		bootstart() {
			const register = (key, ClassObject, templateID) => ko.components.register(key, {
					template: { element: templateID || (key + 'Component') },
					viewModel: {
						createViewModel: (params, componentInfo) => {
							params = params || {};
							i18nToNodes(componentInfo.element);
							if (params.inline) {
								componentInfo.element.style.display = 'inline-block';
							}
							return new ClassObject(params);
						}
					}
				});

			register('Select', SelectComponent);
			register('Checkbox', CheckboxMaterialDesignComponent, 'CheckboxMaterialDesignComponent');

			initOnStartOrLangChange();

			LanguageStore.populate();
			ThemeStore.populate();

			this.start();
		}
	}

	class AdminApp extends AbstractApp {
		constructor() {
			super(Remote);
			this.weakPassword = ko.observable(false);
		}

		refresh() {
			this.start();
		}

		start() {
			if (!Settings.app('adminAllowed')) {
				rl.route.root();
				setTimeout(() => location.href = '/', 1);
			} else if (SettingsGet('Auth')) {
				this.weakPassword(!!SettingsGet('WeakPassword'));
				startScreens([SettingsAdminScreen]);
			} else {
				startScreens([LoginAdminScreen]);
			}
		}
	}

	bootstrap(new AdminApp);

})();
