/* SnappyMail Webmail (c) SnappyMail | Licensed under AGPL v3 */
(function () {
	'use strict';

	let keyScopeFake = 'all';

	const
		ScopeMenu = 'Menu',

		doc = document,

		$htmlCL = doc.documentElement.classList,

		elementById = id => doc.getElementById(id),

		Settings = rl.settings,
		SettingsGet = Settings.get,
		SettingsCapa = name => name && !!(SettingsGet('Capa') || {})[name],

		dropdowns = [],
		dropdownVisibility = ko.observable(false).extend({ rateLimit: 0 }),

		moveAction = ko.observable(false),
		leftPanelDisabled = ko.observable(false),

		createElement = (name, attr) => {
			let el = doc.createElement(name);
			attr && Object.entries(attr).forEach(([k,v]) => el.setAttribute(k,v));
			return el;
		},

		fireEvent = (name, detail, cancelable) => dispatchEvent(
			new CustomEvent(name, {detail:detail, cancelable: !!cancelable})
		),

		formFieldFocused = () => doc.activeElement?.matches('input,textarea'),

		addShortcut = (...args) => shortcuts.add(...args),

		registerShortcut = (keys, modifiers, scopes, method) =>
			addShortcut(keys, modifiers, scopes, event => formFieldFocused() ? true : method(event)),

		addEventsListener = (element, events, fn, options) =>
			events.forEach(event => element.addEventListener(event, fn, options)),

		addEventsListeners = (element, events) =>
			Object.entries(events).forEach(([event, fn]) => element.addEventListener(event, fn)),

		// keys / shortcuts
		keyScopeReal = ko.observable('all'),
		keyScope = value => {
			if (!value) {
				return keyScopeFake;
			}
			if (ScopeMenu !== value) {
				keyScopeFake = value;
				if (dropdownVisibility()) {
					value = ScopeMenu;
				}
			}
			keyScopeReal(value);
			shortcuts.setScope(value);
		};

	dropdownVisibility.subscribe(value => {
		if (value) {
			keyScope(ScopeMenu);
		} else if (ScopeMenu === shortcuts.getScope()) {
			keyScope(keyScopeFake);
		}
	});

	leftPanelDisabled.toggle = () => leftPanelDisabled(!leftPanelDisabled());
	leftPanelDisabled.subscribe(value => {
		value && moveAction() && moveAction(false);
		$htmlCL.toggle('rl-left-panel-disabled', value);
	});

	moveAction.subscribe(value => value && leftPanelDisabled() && leftPanelDisabled(false));

	/* eslint quote-props: 0 */

	const

	/**
	 * @enum {string}
	 */
	Scope = {
		MessageList: 'MessageList',
		FolderList: 'FolderList',
		MessageView: 'MessageView',
		Settings: 'Settings'
	},

	/**
	 * @enum {number}
	 */
	UploadErrorCode = {
		Normal: 0,
		FileIsTooBig: 1,
		FilePartiallyUploaded: 3,
		NoFileUploaded: 4,
		MissingTempFolder: 6,
		OnSavingFile: 7,
		FileType: 98,
		Unknown: 99
	},

	/**
	 * @enum {number}
	 */
	SaveSettingStatus = {
		Saving: -2,
		Idle: -1,
		Success: 1,
		Failed: 0
	},

	/**
	 * @enum {number}
	 */
	Notification$1 = {
		RequestError: 1,
		RequestAborted: 2,
		RequestTimeout: 3,

		// Global
		InvalidToken: 101,
		AuthError: 102,

		// User
		ConnectionError: 104,
		DomainNotAllowed: 109,
		AccountNotAllowed: 110,

		ContactsSyncError: 140,

		CantGetMessageList: 201,
		CantGetMessage: 202,
		CantDeleteMessage: 203,
		CantMoveMessage: 204,
		CantCopyMessage: 205,

		CantSaveMessage: 301,
		CantSendMessage: 302,
		InvalidRecipients: 303,

		CantSaveFilters: 351,
		CantGetFilters: 352,
		CantActivateFiltersScript: 353,
		CantDeleteFiltersScript: 354,
	//	FiltersAreNotCorrect: 355,

		CantCreateFolder: 400,
		CantRenameFolder: 401,
		CantDeleteFolder: 402,
		CantSubscribeFolder: 403,
		CantUnsubscribeFolder: 404,
		CantDeleteNonEmptyFolder: 405,

	//	CantSaveSettings: 501,

		DomainAlreadyExists: 601,

		DemoSendMessageError: 750,
		DemoAccountError: 751,

		AccountAlreadyExists: 801,
		AccountDoesNotExist: 802,
		AccountSwitchFailed: 803,

		MailServerError: 901,
		ClientViewError: 902,
		InvalidInputArgument: 903,

		JsonFalse: 950,
		JsonParse: 952,
	//	JsonTimeout: 953,

		UnknownNotification: 998,
		UnknownError: 999,

		// Admin
		CantInstallPackage: 701,
		CantDeletePackage: 702,
		InvalidPluginPackage: 703,
		UnsupportedPluginPackage: 704,
		CantSavePluginSettings: 705
	};

	let __themeTimer = 0,
		__themeJson = null;

	const
		isArray = Array.isArray,
		arrayLength = array => isArray(array) && array.length,
		isFunction = v => typeof v === 'function',
		pString = value => null != value ? '' + value : '',

		forEachObjectValue = (obj, fn) => Object.values(obj).forEach(fn),

		forEachObjectEntry = (obj, fn) => Object.entries(obj).forEach(([key, value]) => fn(key, value)),

		pInt = (value, defaultValue = 0) => {
			value = parseInt(value, 10);
			return isNaN(value) || !isFinite(value) ? defaultValue : value;
		},

		convertThemeName = theme => theme
			.replace(/@custom$/, '')
			.replace(/([A-Z])/g, ' $1')
			.replace(/[^a-zA-Z0-9]+/g, ' ')
			.trim(),

		defaultOptionsAfterRender = (domItem, item) =>
			item && undefined !== item.disabled && domItem?.classList.toggle('disabled', domItem.disabled = item.disabled),

		// unescape(encodeURIComponent()) makes the UTF-16 DOMString to an UTF-8 string
		b64EncodeJSON = data => btoa(unescape(encodeURIComponent(JSON.stringify(data)))),
	/* 	// Without deprecated 'unescape':
		b64EncodeJSON = data => btoa(encodeURIComponent(JSON.stringify(data)).replace(
			/%([0-9A-F]{2})/g, (match, p1) => String.fromCharCode('0x' + p1)
		)),
	*/
		b64EncodeJSONSafe = data => b64EncodeJSON(data).replace(/\+/g, '-').replace(/\//g, '_').replace(/=+$/, ''),

		changeTheme = (value, themeTrigger = ()=>0) => {
			const themeStyle = elementById('app-theme-style'),
				clearTimer = () => {
					__themeTimer = setTimeout(() => themeTrigger(SaveSettingStatus.Idle), 1000);
					__themeJson = null;
				},
				url = themeStyle.dataset.href.replace(/(Admin|User)\/-\/[^/]+\//, '$1/-/' + value + '/') + 'Json/';

			clearTimeout(__themeTimer);

			themeTrigger(SaveSettingStatus.Saving);

			if (__themeJson) {
				__themeJson.abort();
			}
			let init = {};
			if (window.AbortController) {
				__themeJson = new AbortController();
				init.signal = __themeJson.signal;
			}
			rl.fetchJSON(url, init)
				.then(data => {
					if (2 === arrayLength(data)) {
						themeStyle.textContent = data[1];
						themeTrigger(SaveSettingStatus.Success);
					}
				})
				.then(clearTimer, clearTimer);
		},

		getKeyByValue = (o, v) => Object.keys(o).find(key => o[key] === v);

	const
		HASH_PREFIX = '#/',
		SERVER_PREFIX = './?',
		VERSION = Settings.app('version'),
		VERSION_PREFIX = 'snappymail/v/' + VERSION + '/',

		adminPath = () => rl.adminArea() && !Settings.app('adminHostUse'),

		prefix = () => SERVER_PREFIX + (adminPath() ? Settings.app('adminPath') : '');

	const SUB_QUERY_PREFIX = '&q[]=',

		/**
		 * @param {string=} startupUrl
		 * @returns {string}
		 */
		root = () => HASH_PREFIX,

		/**
		 * @returns {string}
		 */
		logoutLink = () => adminPath() ? prefix() : './',

		/**
		 * @param {string} type
		 * @param {string} hash
		 * @param {string=} customSpecSuffix
		 * @returns {string}
		 */
		serverRequestRaw = (type, hash) =>
			SERVER_PREFIX + '/Raw/' + SUB_QUERY_PREFIX + '/'
			+ '0/' // AuthAccountHash ?
			+ (type
				? type + '/' + (hash ? SUB_QUERY_PREFIX + '/' + hash : '')
				: ''),

		/**
		 * @param {string} download
		 * @param {string=} customSpecSuffix
		 * @returns {string}
		 */
		attachmentDownload = (download, customSpecSuffix) =>
			serverRequestRaw('Download', download),

		proxy = url =>
			SERVER_PREFIX + '/ProxyExternal/'
				+ btoa(url.replace(/ /g, '%20')).replace(/\+/g, '-').replace(/\//g, '_').replace(/=+$/, ''),

		/**
		 * @param {string} type
		 * @returns {string}
		 */
		serverRequest = type => prefix() + '/' + type + '/' + SUB_QUERY_PREFIX + '/0/',

		/**
		 * @param {string} lang
		 * @param {boolean} isAdmin
		 * @returns {string}
		 */
		langLink = (lang, isAdmin) =>
			SERVER_PREFIX + '/Lang/0/' + (isAdmin ? 'Admin' : 'App') + '/' + encodeURI(lang) + '/' + VERSION + '/',

		/**
		 * @param {string} path
		 * @returns {string}
		 */
		staticLink = path => VERSION_PREFIX + 'static/' + path,

		/**
		 * @param {string} theme
		 * @returns {string}
		 */
		themePreviewLink = theme => {
			let prefix = VERSION_PREFIX;
			if ('@custom' === theme.slice(-7)) {
				theme = theme.slice(0, theme.length - 7).trim();
				prefix = '';
			}

			return prefix + 'themes/' + encodeURI(theme) + '/images/preview.png';
		},

		/**
		 * @param {string} inboxFolderName = 'INBOX'
		 * @returns {string}
		 */
		mailbox = (inboxFolderName = 'INBOX') => HASH_PREFIX + 'mailbox/' + inboxFolderName,

		/**
		 * @param {string=} screenName = ''
		 * @returns {string}
		 */
		settings = (screenName = '') => HASH_PREFIX + 'settings' + (screenName ? '/' + screenName : ''),

		/**
		 * @param {string} folder
		 * @param {number=} page = 1
		 * @param {string=} search = ''
		 * @param {number=} threadUid = 0
		 * @returns {string}
		 */
		mailBox = (folder, page, search, threadUid) => {
			let result = [HASH_PREFIX + 'mailbox'];

			if (folder) {
				result.push(folder + (threadUid ? '~' + threadUid : ''));
			}

			page = pInt(page, 1);
			if (1 < page) {
				result.push('p' + page);
			}

			search = pString(search);
			if (search) {
				result.push(encodeURI(search));
			}

			return result.join('/');
		};

	let I18N_DATA = {};

	const
		i18nToNode = element => {
			const key = element.dataset.i18n;
			if (key) {
				if ('[' === key.slice(0, 1)) {
					switch (key.slice(0, 6)) {
						case '[html]':
							element.innerHTML = i18n(key.slice(6));
							break;
						case '[place':
							element.placeholder = i18n(key.slice(13));
							break;
						case '[title':
							element.title = i18n(key.slice(7));
							break;
						// no default
					}
				} else {
					element.textContent = i18n(key);
				}
			}
		},

		init = () => {
			if (rl.I18N) {
				I18N_DATA = rl.I18N;
				Date.defineRelativeTimeFormat(rl.relativeTime || {});
				rl.I18N = null;
				return 1;
			}
		},

		i18nKey = key => key.replace(/([a-z])([A-Z])/g, '$1_$2').toUpperCase(),

		getNotificationMessage = code => {
			let key = getKeyByValue(Notification$1, code);
			return key ? I18N_DATA.NOTIFICATIONS[i18nKey(key).replace('_NOTIFICATION', '_ERROR')] : '';
		};

	const
		trigger = ko.observable(false),

		/**
		 * @param {string} key
		 * @param {Object=} valueList
		 * @param {string=} defaulValue
		 * @returns {string}
		 */
		i18n = (key, valueList, defaulValue) => {
			let result = defaulValue || key;
			if (key) {
				let path = key.split('/');
				if (I18N_DATA[path[0]] && path[1]) {
					result = I18N_DATA[path[0]][path[1]] || result;
				}
			}
			if (valueList) {
				forEachObjectEntry(valueList, (key, value) => {
					result = result.replace('%' + key + '%', value);
				});
			}
			return result;
		},

		/**
		 * @param {Object} elements
		 * @param {boolean=} animate = false
		 */
		i18nToNodes = element =>
			setTimeout(() =>
				element.querySelectorAll('[data-i18n]').forEach(item => i18nToNode(item))
			, 1),

		timestampToString = (timeStampInUTC, formatStr) => {
			const now = Date.now(),
				time = 0 < timeStampInUTC ? Math.min(now, timeStampInUTC * 1000) : (0 === timeStampInUTC ? now : 0);

			if (31536000000 < time) {
				const m = new Date(time);
				switch (formatStr) {
					case 'FROMNOW':
						return m.fromNow();
					case 'SHORT': {
						if (4 >= (now - time) / 3600000)
							return m.fromNow();
						const mt = m.getTime(), date = new Date,
							dt = date.setHours(0,0,0,0);
						if (mt > dt)
							return i18n('MESSAGE_LIST/TODAY_AT', {TIME: m.format('LT')});
						if (mt > dt - 86400000)
							return i18n('MESSAGE_LIST/YESTERDAY_AT', {TIME: m.format('LT')});
						if (date.getFullYear() === m.getFullYear())
							return m.format('d M');
						return m.format('LL');
					}
					case 'FULL':
						return m.format('LLL');
					default:
						return m.format(formatStr);
				}
			}

			return '';
		},

		timeToNode = (element, time) => {
			try {
				if (time) {
					element.dateTime = (new Date(time * 1000)).format('Y-m-d\\TH:i:s');
				} else {
					time = Date.parse(element.dateTime) / 1000;
				}

				let key = element.dataset.momentFormat;
				if (key) {
					element.textContent = timestampToString(time, key);
				}

				if ((key = element.dataset.momentFormatTitle)) {
					element.title = timestampToString(time, key);
				}
			} catch (e) {
				// prevent knockout crashes
				console.error(e);
			}
		},

		reloadTime = () => setTimeout(() =>
				doc.querySelectorAll('time').forEach(element => timeToNode(element))
				, 1),

		/**
		 * @param {Function} startCallback
		 * @param {Function=} langCallback = null
		 */
		initOnStartOrLangChange = (startCallback, langCallback = null) => {
			startCallback?.();
			startCallback && trigger.subscribe(startCallback);
			langCallback && trigger.subscribe(langCallback);
		},

		/**
		 * @param {number} code
		 * @param {*=} message = ''
		 * @param {*=} defCode = null
		 * @returns {string}
		 */
		getNotification = (code, message = '', defCode = 0) => {
			code = parseInt(code, 10) || 0;
			if (Notification$1.ClientViewError === code && message) {
				return message;
			}

			return getNotificationMessage(code)
				|| getNotificationMessage(parseInt(defCode, 10))
				|| '';
		},

		/**
		 * @param {*} code
		 * @returns {string}
		 */
		getUploadErrorDescByCode = code => {
			let key = getKeyByValue(UploadErrorCode, parseInt(code, 10));
			return i18n('UPLOAD/ERROR_' + (key ? i18nKey(key) : 'UNKNOWN'));
		},

		/**
		 * @param {boolean} admin
		 * @param {string} language
		 */
		translatorReload = (admin, language) =>
			new Promise((resolve, reject) => {
				const script = createElement('script');
				script.onload = () => {
					// reload the data
					if (init()) {
						i18nToNodes(doc);
						admin || reloadTime();
						trigger(!trigger());
					}
					script.remove();
					resolve();
				};
				script.onerror = () => reject(new Error('Language '+language+' failed'));
				script.src = langLink(language, admin);
		//		script.async = true;
				doc.head.append(script);
			}),

		/**
		 *
		 * @param {string} language
		 * @param {boolean=} isEng = false
		 * @returns {string}
		 */
		convertLangName = (language, isEng = false) =>
			i18n(
				'LANGS_NAMES' + (true === isEng ? '_EN' : '') + '/' + language,
				null,
				language
			);

	init();

	const FormDataToObject = formData => {
		var object = {};
		formData.forEach((value, key) => {
			if (!Reflect.has(object, key)){
				object[key] = value;
			} else {
				isArray(object[key]) || (object[key] = [object[key]]);
				object[key].push(value);
			}
		});
		return object;
	};

	var bootstrap = App => {

		rl.app = App;
		rl.logoutReload = App.logoutReload;

		rl.i18n = i18n;

		rl.Enums = {
			StorageResultType: {
				Success: 0,
				Error: 1,
				Abort: 2
			}
		};

		rl.route = {
			root: () => {
				rl.route.off();
				hasher.setHash(root());
			},
			reload: () => {
				rl.route.root();
				setTimeout(() => location.reload(), 100);
			},
			off: () => hasher.active = false,
			on: () => hasher.active = true
		};

		rl.fetch = (resource, init, postData) => {
			init = Object.assign({
				mode: 'same-origin',
				cache: 'no-cache',
				redirect: 'error',
				referrerPolicy: 'no-referrer',
				credentials: 'same-origin',
				headers: {}
			}, init);
			if (postData) {
				init.method = 'POST';
				init.headers['Content-Type'] = 'application/json';
				postData = (postData instanceof FormData) ? FormDataToObject(postData) : postData;
				postData.XToken = Settings.app('token');
				init.body = JSON.stringify(postData);
			}

			return fetch(resource, init);
		};

		rl.fetchJSON = (resource, init, postData) => {
			init = Object.assign({ headers: {} }, init);
			init.headers.Accept = 'application/json';
			return rl.fetch(resource, init, postData).then(response => {
				if (!response.ok) {
					return Promise.reject('Network response error: ' + response.status);
				}
				/* TODO: use this for non-developers?
				response.clone()
				let data = response.text();
				try {
					return JSON.parse(data);
				} catch (e) {
					console.error(e);
	//				console.log(data);
					return Promise.reject(Notification.JsonParse);
					return {
						Result: false,
						ErrorCode: 952, // Notification.JsonParse
						ErrorMessage: e.message,
						ErrorMessageAdditional: data
					}
				}
				*/
				return response.json();
			});
		};

	};

	const
		errorTip = (element, value) => value
				? setTimeout(() => element.setAttribute('data-rainloopErrorTip', value), 100)
				: element.removeAttribute('data-rainloopErrorTip'),

		/**
		 * The value of the pureComputed observable shouldn’t vary based on the
		 * number of evaluations or other “hidden” information. Its value should be
		 * based solely on the values of other observables in the application
		 */
		koComputable = fn => ko.computed(fn, {'pure':true}),

		addObservablesTo = (target, observables) =>
			forEachObjectEntry(observables, (key, value) =>
				target[key] || (target[key] = /*isArray(value) ? ko.observableArray(value) :*/ ko.observable(value)) ),

		addComputablesTo = (target, computables) =>
			forEachObjectEntry(computables, (key, fn) => target[key] = koComputable(fn)),

		addSubscribablesTo = (target, subscribables) =>
			forEachObjectEntry(subscribables, (key, fn) => target[key].subscribe(fn)),

		dispose = disposable => isFunction(disposable?.dispose) && disposable.dispose(),

		// With this we don't need delegateRunOnDestroy
		koArrayWithDestroy = data => {
			data = ko.observableArray(data);
			data.subscribe(changes =>
				changes.forEach(item =>
					'deleted' === item.status && null == item.moved && item.value.onDestroy?.()
				)
			, data, 'arrayChange');
			return data;
		};

	Object.assign(ko.bindingHandlers, {
		tooltipErrorTip: {
			init: (element, fValueAccessor) => {
				doc.addEventListener('click', () => {
					let value = fValueAccessor();
					ko.isObservable(value) && !ko.isComputed(value) && value('');
					errorTip(element);
				});
			},
			update: (element, fValueAccessor) => {
				let value = ko.unwrap(fValueAccessor());
				value = isFunction(value) ? value() : value;
				errorTip(element, value);
			}
		},

		onEnter: {
			init: (element, fValueAccessor, fAllBindings, viewModel) => {
				let fn = event => {
					if ('Enter' == event.key) {
						element.dispatchEvent(new Event('change'));
						fValueAccessor().call(viewModel);
					}
				};
				element.addEventListener('keydown', fn);
				ko.utils.domNodeDisposal.addDisposeCallback(element, () => element.removeEventListener('keydown', fn));
			}
		},

		onEsc: {
			init: (element, fValueAccessor, fAllBindings, viewModel) => {
				let fn = event => {
					if ('Escape' == event.key) {
						element.dispatchEvent(new Event('change'));
						fValueAccessor().call(viewModel);
					}
				};
				element.addEventListener('keyup', fn);
				ko.utils.domNodeDisposal.addDisposeCallback(element, () => element.removeEventListener('keyup', fn));
			}
		},

		onSpace: {
			init: (element, fValueAccessor, fAllBindings, viewModel) => {
				let fn = event => {
					if (' ' == event.key) {
						fValueAccessor().call(viewModel, event);
					}
				};
				element.addEventListener('keyup', fn);
				ko.utils.domNodeDisposal.addDisposeCallback(element, () => element.removeEventListener('keyup', fn));
			}
		},

		i18nUpdate: {
			update: (element, fValueAccessor) => {
				ko.unwrap(fValueAccessor());
				i18nToNodes(element);
			}
		},

		title: {
			update: (element, fValueAccessor) => element.title = ko.unwrap(fValueAccessor())
		},

		command: {
			init: (element, fValueAccessor, fAllBindings, viewModel, bindingContext) => {
				const command = fValueAccessor();

				if (!command || !command.canExecute) {
					throw new Error('Value should be a command');
				}

				ko.bindingHandlers['FORM'==element.nodeName ? 'submit' : 'click'].init(
					element,
					fValueAccessor,
					fAllBindings,
					viewModel,
					bindingContext
				);
			},
			update: (element, fValueAccessor) => {
				const cl = element.classList,
					command = fValueAccessor();

				let disabled = !command.canExecute();
				cl.toggle('disabled', disabled);

				if (element.matches('INPUT,TEXTAREA,BUTTON')) {
					element.disabled = disabled;
				}
			}
		},

		saveTrigger: {
			init: (element) => {
				let icon = element;
				if (element.matches('input,select,textarea')) {
					element.classList.add('settings-save-trigger-input');
					element.after(element.saveTriggerIcon = icon = createElement('span'));
				}
				icon.classList.add('settings-save-trigger');
			},
			update: (element, fValueAccessor) => {
				const value = parseInt(ko.unwrap(fValueAccessor()),10);
				let cl = (element.saveTriggerIcon || element).classList;
				if (element.saveTriggerIcon) {
					cl.toggle('saving', value === SaveSettingStatus.Saving);
					cl.toggle('success', value === SaveSettingStatus.Success);
					cl.toggle('error', value === SaveSettingStatus.Failed);
				}
				cl = element.classList;
				cl.toggle('success', value === SaveSettingStatus.Success);
				cl.toggle('error', value === SaveSettingStatus.Failed);
			}
		}
	});

	// extenders

	ko.extenders.toggleSubscribeProperty = (target, options) => {
		const prop = options[1];
		if (prop) {
			target.subscribe(
				prev => prev?.[prop]?.(false),
				options[0],
				'beforeChange'
			);

			target.subscribe(next => next?.[prop]?.(true), options[0]);
		}

		return target;
	};

	ko.extenders.falseTimeout = (target, option) => {
		target.subscribe((() => target(false)).debounce(parseInt(option, 10) || 0));
		return target;
	};

	// functions

	ko.observable.fn.askDeleteHelper = function() {
		return this.extend({ falseTimeout: 3000, toggleSubscribeProperty: [this, 'askDelete'] });
	};

	/* eslint quote-props: 0 */

	/**
	 * @enum {number}
	 */
	const FolderType = {
		User: 0,
		Inbox: 1,
		Sent: 2,
		Drafts: 3,
		Spam: 4, // JUNK
		Trash: 5,
		Archive: 6,
		NotSpam: 80
	};

	/**
	 * @enum {string}
	 */
	const FolderMetadataKeys = {
		// RFC 5464
		Comment: '/private/comment',
		CommentShared: '/shared/comment',
		// RFC 6154
		SpecialUse: '/private/specialuse',
		// Kolab
		KolabFolderType: '/private/vendor/kolab/folder-type',
		KolabFolderTypeShared: '/shared/vendor/kolab/folder-type'
	};

	/**
	 * @enum {string}
	 */
	const ComposeType = {
		Empty: 0,
		Reply: 1,
		ReplyAll: 2,
		Forward: 3,
		ForwardAsAttachment: 4,
		Draft: 5,
		EditAsNew: 6
	};

	/**
	 * @enum {number}
	 */
	const SetSystemFoldersNotification = {
		None: 0,
		Sent: 1,
		Draft: 2,
		Spam: 3,
		Trash: 4,
		Archive: 5
	};

	/**
	 * @enum {number}
	 */
	const
		ClientSideKeyNameExpandedFolders = 3,
		ClientSideKeyNameFolderListSize = 4,
		ClientSideKeyNameMessageListSize = 5,
		ClientSideKeyNameLastSignMe = 7,
		ClientSideKeyNameMessageHeaderFullInfo = 9,
		ClientSideKeyNameMessageAttachmentControls = 10;

	/**
	 * @enum {number}
	 */
	const MessageSetAction = {
		SetSeen: 0,
		UnsetSeen: 1,
		SetFlag: 2,
		UnsetFlag: 3
	};

	/**
	 * @enum {number}
	 */
	const MessagePriority = {
		Low: 5,
		Normal: 3,
		High: 1
	};

	/**
	 * @enum {string}
	 */
	const EditorDefaultType = {
		Html: 'Html',
		Plain: 'Plain'
	};

	/**
	 * @enum {number}
	 */
	const Layout = {
		NoPreview: 0,
		SidePreview: 1,
		BottomPreview: 2
	};

	const ThemeStore = {
		theme: ko.observable(''),
		themes: ko.observableArray(),
		userBackgroundName: ko.observable(''),
		userBackgroundHash: ko.observable(''),
		isMobile: ko.observable($htmlCL.contains('rl-mobile')),

		populate: () => {
			const themes = Settings.app('themes');

			ThemeStore.themes(isArray(themes) ? themes : []);
			ThemeStore.theme(SettingsGet('Theme'));
			if (!ThemeStore.isMobile()) {
				ThemeStore.userBackgroundName(SettingsGet('UserBackgroundName'));
				ThemeStore.userBackgroundHash(SettingsGet('UserBackgroundHash'));
			}

			leftPanelDisabled(ThemeStore.isMobile());
		}
	};

	ThemeStore.isMobile.subscribe(value => $htmlCL.toggle('rl-mobile', value));

	ThemeStore.userBackgroundHash.subscribe(value => {
		if (value) {
			$htmlCL.add('UserBackground');
			doc.body.style.backgroundImage = "url("+serverRequestRaw('UserBackground', value)+")";
		} else {
			$htmlCL.remove('UserBackground');
			doc.body.removeAttribute('style');
		}
	});

	const SettingsUserStore = new class {
		constructor() {
			const self = this;

			self.messagesPerPage = ko.observable(25).extend({ debounce: 999 });

			self.messageReadDelay = ko.observable(5).extend({ debounce: 999 });

			addObservablesTo(self, {
				viewHTML: 1,
				showImages: 0,
				removeColors: 0,
				useCheckboxesInList: 1,
				allowDraftAutosave: 1,
				useThreads: 0,
				replySameFolder: 0,
				hideUnsubscribed: 0,
				hideDeleted: 1,
				unhideKolabFolders: 0,
				autoLogout: 0,

				requestReadReceipt: 0,
				requestDsn: 0,
				pgpSign: 0,
				pgpEncrypt: 0,

				layout: 1,
				editorDefaultType: 'Html',
				msgDefaultAction: 1
			});

			self.init();

			self.usePreviewPane = koComputable(() => Layout.NoPreview !== self.layout() && !ThemeStore.isMobile());

			const toggleLayout = () => {
				const value = ThemeStore.isMobile() ? Layout.NoPreview : self.layout();
				$htmlCL.toggle('rl-no-preview-pane', Layout.NoPreview === value);
				$htmlCL.toggle('rl-side-preview-pane', Layout.SidePreview === value);
				$htmlCL.toggle('rl-bottom-preview-pane', Layout.BottomPreview === value);
				fireEvent('rl-layout', value);
			};
			self.layout.subscribe(toggleLayout);
			ThemeStore.isMobile.subscribe(toggleLayout);
			toggleLayout();

			let iAutoLogoutTimer;
			self.delayLogout = (() => {
				clearTimeout(iAutoLogoutTimer);
				if (0 < self.autoLogout() && !SettingsGet('AccountSignMe')) {
					iAutoLogoutTimer = setTimeout(
						rl.app.logout,
						self.autoLogout() * 60000
					);
				}
			}).throttle(5000);
		}

		init() {
			const self = this;
			self.editorDefaultType(SettingsGet('EditorDefaultType'));

			self.layout(pInt(SettingsGet('Layout')));
			self.messagesPerPage(pInt(SettingsGet('MessagesPerPage')));
			self.messageReadDelay(pInt(SettingsGet('MessageReadDelay')));
			self.autoLogout(pInt(SettingsGet('AutoLogout')));
			self.msgDefaultAction(SettingsGet('MsgDefaultAction'));

			self.viewHTML(SettingsGet('ViewHTML'));
			self.showImages(SettingsGet('ShowImages'));
			self.removeColors(SettingsGet('RemoveColors'));
			self.useCheckboxesInList(SettingsGet('UseCheckboxesInList'));
			self.allowDraftAutosave(SettingsGet('AllowDraftAutosave'));
			self.useThreads(SettingsGet('UseThreads'));
			self.replySameFolder(SettingsGet('ReplySameFolder'));

			self.hideUnsubscribed(SettingsGet('HideUnsubscribed'));
			self.hideDeleted(SettingsGet('HideDeleted'));
			self.unhideKolabFolders(SettingsGet('UnhideKolabFolders'));

			self.requestReadReceipt(SettingsGet('requestReadReceipt'));
			self.requestDsn(SettingsGet('requestDsn'));
			self.pgpSign(SettingsGet('pgpSign'));
			self.pgpEncrypt(SettingsGet('pgpEncrypt'));
		}
	};

	const
		tpl$1 = createElement('template'),
		htmlre = /[&<>"']/g,
		httpre = /^(https?:)?\/\//i,
		htmlmap = {
			'&': '&amp;',
			'<': '&lt;',
			'>': '&gt;',
			'"': '&quot;',
			"'": '&#x27;'
		},

		replaceWithChildren = node => node.replaceWith(...[...node.childNodes]),

		// Strip tracking
		stripTracking = text => text
			.replace(/tracking\.(printabout\.nl[^?]+)\?.*/gsi, (...m) => m[1])
			.replace(/^.+awstrack\.me\/.+(https:%2F%2F[^/]+)/gsi, (...m) => decodeURIComponent(m[1]))
			.replace(/([?&])utm_[a-z]+=[^&?#]*/gsi, '$1') // Urchin Tracking Module
			.replace(/([?&])ec_[a-z]+=[^&?#]*/gsi, '$1')  // Sitecore
			.replace(/&&+/, '');

	const

		/**
		 * @param {string} text
		 * @returns {string}
		 */
		encodeHtml = text => (text?.toString?.() || '' + text).replace(htmlre, m => htmlmap[m]),

		/**
		 * Clears the Message Html for viewing
		 * @param {string} text
		 * @returns {string}
		 */
		cleanHtml = (html, oAttachments) => {
			const result = {
					hasExternals: false
				},

				findAttachmentByCid = cid => oAttachments.findByCid(cid),
				findLocationByCid = cid => {
					const attachment = findAttachmentByCid(cid);
					return attachment?.contentLocation ? attachment : 0;
				},

				// convert body attributes to CSS
				tasks = {
					link: value => {
						if (/^#[a-fA-Z0-9]{3,6}$/.test(value)) {
							tpl$1.content.querySelectorAll('a').forEach(node => node.style.color = value);
						}
					},
					text: (value, node) => node.style.color = value,
					topmargin: (value, node) => node.style.marginTop = pInt(value) + 'px',
					leftmargin: (value, node) => node.style.marginLeft = pInt(value) + 'px',
					bottommargin: (value, node) => node.style.marginBottom = pInt(value) + 'px',
					rightmargin: (value, node) => node.style.marginRight = pInt(value) + 'px'
				},
				allowedAttributes = [
					// defaults
					'name',
					'dir', 'lang', 'style', 'title',
					'background', 'bgcolor', 'alt', 'height', 'width', 'src', 'href',
					'border', 'bordercolor', 'charset', 'direction',
					// a
					'download', 'hreflang',
					// body
					'alink', 'bottommargin', 'leftmargin', 'link', 'rightmargin', 'text', 'topmargin', 'vlink',
					// col
					'align', 'valign',
					// font
					'color', 'face', 'size',
					// hr
					'noshade',
					// img
					'hspace', 'sizes', 'srcset', 'vspace',
					// meter
					'low', 'high', 'optimum', 'value',
					// ol
					'reversed', 'start',
					// table
					'cols', 'rows', 'frame', 'rules', 'summary', 'cellpadding', 'cellspacing',
					// th
					'abbr', 'scope',
					// td
					'colspan', 'rowspan', 'headers'
				],
				disallowedTags = [
					'HEAD','STYLE','SVG','SCRIPT','TITLE','LINK','BASE','META',
					'INPUT','OUTPUT','SELECT','BUTTON','TEXTAREA',
					'BGSOUND','KEYGEN','SOURCE','OBJECT','EMBED','APPLET','IFRAME','FRAME','FRAMESET','VIDEO','AUDIO','AREA','MAP'
				],
				nonEmptyTags = [
					'A','B','EM','I','SPAN','STRONG'
				];

			tpl$1.innerHTML = html
	//			.replace(/<pre[^>]*>[\s\S]*?<\/pre>/gi, pre => pre.replace(/\n/g, '\n<br>'))
				.replace(/<!doctype[^>]*>/gi, '')
				.replace(/<\?xml[^>]*\?>/gi, '')
				// Not supported by <template> element
				.replace(/<(\/?)body(\s[^>]*)?>/gi, '<$1div class="mail-body"$2>')
				.replace(/<\/?(html|head)[^>]*>/gi, '')
				.trim();
			html = '';

			// \MailSo\Base\HtmlUtils::ClearComments()
			// https://github.com/the-djmaze/snappymail/issues/187
			const nodeIterator = document.createNodeIterator(tpl$1.content, NodeFilter.SHOW_COMMENT);
			while (nodeIterator.nextNode()) {
				nodeIterator.referenceNode.remove();
			}

			tpl$1.content.querySelectorAll('*').forEach(oElement => {
				const name = oElement.tagName,
					oStyle = oElement.style;

				// \MailSo\Base\HtmlUtils::ClearTags()
				if (disallowedTags.includes(name)
				 || 'none' == oStyle.display
				 || 'hidden' == oStyle.visibility
	//			 || (oStyle.lineHeight && 1 > parseFloat(oStyle.lineHeight)
	//			 || (oStyle.maxHeight && 1 > parseFloat(oStyle.maxHeight)
	//			 || (oStyle.maxWidth && 1 > parseFloat(oStyle.maxWidth)
	//			 || ('0' === oStyle.opacity
				) {
					oElement.remove();
					return;
				}
	//			if (['CENTER','FORM'].includes(name)) {
				if ('FORM' === name || 'O:P' === name || (nonEmptyTags.includes(name) && ('' == oElement.textContent.trim()))) {
					('A' !== name || !oElement.querySelector('IMG')) && replaceWithChildren(oElement);
					return;
				}
	/*
				// Idea to allow CSS
				if ('STYLE' === name) {
					msgId = '#rl-msg-061eb4d647771be4185943ce91f0039d';
					oElement.textContent = oElement.textContent
						.replace(/[^{}]+{/g, m => msgId + ' ' + m.replace(',', ', '+msgId+' '))
						.replace(/(background-)color:[^};]+/g, '');
					return;
				}
	*/
				const hasAttribute = name => oElement.hasAttribute(name),
					getAttribute = name => hasAttribute(name) ? oElement.getAttribute(name).trim() : '',
					setAttribute = (name, value) => oElement.setAttribute(name, value),
					delAttribute = name => oElement.removeAttribute(name);

				if ('mail-body' === oElement.className) {
					forEachObjectEntry(tasks, (name, cb) => {
						if (hasAttribute(name)) {
							cb(getAttribute(name), oElement);
							delAttribute(name);
						}
					});
				}

				if (oElement.hasAttributes()) {
					let i = oElement.attributes.length;
					while (i--) {
						let sAttrName = oElement.attributes[i].name.toLowerCase();
						if (!allowedAttributes.includes(sAttrName)) {
							delAttribute(sAttrName);
						}
					}
				}

				let value;

	//			if ('TABLE' === name || 'TD' === name || 'TH' === name) {
				if (!oStyle.backgroundImage) {
					if (hasAttribute('width')) {
						value = getAttribute('width');
						oStyle.width = value.includes('%') ? value : value + 'px';
						delAttribute('width');
					}
					value = oStyle.width;
					if (value && !value.includes('%')) {
						oStyle.maxWidth = value;
						if ('TD' !== name && 'TH' !== name) {
							oStyle.width = '100%';
						}
					}
					if (hasAttribute('height')) {
						value = getAttribute('height');
						oStyle.height = value.includes('%') ? value : value + 'px';
						delAttribute('height');
					}
					value = oStyle.removeProperty('height');
					if (value && !value.includes('%')) {
						oStyle.maxHeight = value;
					}
				}
	//			} else
				if ('A' === name) {
					value = oElement.href;
					if (!/^([a-z]+):/i.test(value)) {
						setAttribute('data-x-broken-href', value);
						delAttribute('href');
					} else {
						oElement.href = stripTracking(value);
						setAttribute('target', '_blank');
	//					setAttribute('rel', 'external nofollow noopener noreferrer');
					}
					setAttribute('tabindex', '-1');
				}

				// SVG xlink:href
				/*
				if (hasAttribute('xlink:href')) {
					delAttribute('xlink:href');
				}
				*/

				let skipStyle = false;
				if (hasAttribute('src')) {
					value = getAttribute('src');
					delAttribute('src');

					if ('IMG' === name) {
						oElement.loading = 'lazy';
						let attachment;
						if ((('' != getAttribute('height') && 3 > pInt(getAttribute('height')))
								|| ('' != getAttribute('width') && 3 > pInt(getAttribute('width')))
								|| [
									'email.microsoftemail.com/open',
									'github.com/notifications/beacon/',
									'mandrillapp.com/track/open',
									'list-manage.com/track/open'
								].filter(uri => value.toLowerCase().includes(uri)).length
						)) {
							skipStyle = true;
							setAttribute('style', 'display:none');
							setAttribute('data-x-src-hidden', value);
						}
						else if ((attachment = findLocationByCid(value)))
						{
							if (attachment.download) {
								oElement.src = attachment.linkPreview();
								attachment.isLinked(true);
							}
						}
						else if ('cid:' === value.slice(0, 4))
						{
							value = value.slice(4);
							setAttribute('data-x-src-cid', value);
							attachment = findAttachmentByCid(value);
							if (attachment?.download) {
								oElement.src = attachment.linkPreview();
								attachment.isInline(true);
								attachment.isLinked(true);
							}
						}
						else if (httpre.test(value))
						{
							setAttribute('data-x-src', value);
							result.hasExternals = true;
						}
						else if ('data:image/' === value.slice(0, 11))
						{
							oElement.src = value;
						}
						else
						{
							setAttribute('data-x-src-broken', value);
						}
					}
					else
					{
						setAttribute('data-x-src-broken', value);
					}
				}

				if (hasAttribute('background')) {
					oStyle.backgroundImage = 'url("' + getAttribute('background') + '")';
					delAttribute('background');
				}

				if (hasAttribute('bgcolor')) {
					oStyle.backgroundColor = getAttribute('bgcolor');
					delAttribute('bgcolor');
				}

				if (hasAttribute('color')) {
					oStyle.color = getAttribute('color');
					delAttribute('color');
				}

				if (!skipStyle) {
	/*
					if ('fixed' === oStyle.position) {
						oStyle.position = 'absolute';
					}
	*/
					oStyle.removeProperty('behavior');
					oStyle.removeProperty('cursor');
					oStyle.removeProperty('min-width');

					const
						urls_remote = [], // 'data-x-style-url'
						urls_broken = []; // 'data-x-broken-style-src'
					['backgroundImage', 'listStyleImage', 'content'].forEach(property => {
						if (oStyle[property]) {
							let value = oStyle[property],
								found = value.match(/url\s*\(([^)]+)\)/i);
							if (found) {
								oStyle[property] = null;
								found = found[1].replace(/^["'\s]+|["'\s]+$/g, '');
								let lowerUrl = found.toLowerCase();
								if ('cid:' === lowerUrl.slice(0, 4)) {
									const attachment = findAttachmentByCid(found);
									if (attachment?.linkPreview && name) {
										oStyle[property] = "url('" + attachment.linkPreview() + "')";
										attachment.isInline(true);
										attachment.isLinked(true);
									}
								} else if (httpre.test(lowerUrl)) {
									result.hasExternals = true;
									urls_remote.push([property, found]);
								} else if ('data:image/' === lowerUrl.slice(0, 11)) {
									oStyle[property] = value;
								} else {
									urls_broken.push([property, found]);
								}
							}
						}
					});
	//				oStyle.removeProperty('background-image');
	//				oStyle.removeProperty('list-style-image');

					if (urls_remote.length) {
						setAttribute('data-x-style-url', JSON.stringify(urls_remote));
					}
					if (urls_broken.length) {
						setAttribute('data-x-style-broken-urls', JSON.stringify(urls_broken));
					}

					if (11 > pInt(oStyle.fontSize)) {
						oStyle.removeProperty('font-size');
					}

					// Removes background and color
					// Many e-mails incorrectly only define one, not both
					// And in dark theme mode this kills the readability
					if (SettingsUserStore.removeColors()) {
						oStyle.removeProperty('background-color');
						oStyle.removeProperty('background-image');
						oStyle.removeProperty('color');
					}

					// Drop Microsoft Office style properties
	//				oStyle.cssText = oStyle.cssText.replace(/mso-[^:;]+:[^;]+/gi, '');
				}
			});

	//		return tpl.content.firstChild;
			result.html = tpl$1.innerHTML.trim();
			return result;
		},

		/**
		 * @param {string} html
		 * @returns {string}
		 */
		htmlToPlain = html => {
			const
				hr = '⎯'.repeat(64),
				forEach = (selector, fn) => tpl$1.content.querySelectorAll(selector).forEach(fn),
				blockquotes = node => {
					let bq;
					while ((bq = node.querySelector('blockquote'))) {
						// Convert child blockquote first
						blockquotes(bq);
						// Convert blockquote
	//					bq.innerHTML = '\n' + ('\n' + bq.innerHTML.replace(/\n{3,}/gm, '\n\n').trim() + '\n').replace(/^/gm, '&gt; ');
	//					replaceWithChildren(bq);
						bq.replaceWith(
							'\n' + ('\n' + bq.textContent.replace(/\n{3,}/g, '\n\n').trim() + '\n').replace(/^/gm, '> ')
						);
					}
				};

			html = html
				.replace(/<pre[^>]*>([\s\S]*?)<\/pre>/gim, (...args) =>
					1 < args.length ? args[1].toString().replace(/\n/g, '<br>') : '')
				.replace(/\r?\n/g, '')
				.replace(/\s+/gm, ' ');

			while (/<(div|tr)[\s>]/i.test(html)) {
				html = html.replace(/\n*<(div|tr)(\s[\s\S]*?)?>\n*/gi, '\n');
			}
			while (/<\/(div|tr)[\s>]/i.test(html)) {
				html = html.replace(/\n*<\/(div|tr)(\s[\s\S]*?)?>\n*/gi, '\n');
			}

			tpl$1.innerHTML = html
				.replace(/<t[dh](\s[\s\S]*?)?>/gi, '\t')
				.replace(/<\/tr(\s[\s\S]*?)?>/gi, '\n');

			// lines
			forEach('hr', node => node.replaceWith(`\n\n${hr}\n\n`));

			// headings
			forEach('h1,h2,h3,h4,h5,h6', h => h.replaceWith(`\n\n${'#'.repeat(h.tagName[1])} ${h.textContent}\n\n`));

			// paragraphs
			forEach('p', node => {
				node.prepend('\n\n');
				if ('' == node.textContent.trim()) {
					node.remove();
				} else {
					node.after('\n\n');
				}
			});

			// proper indenting and numbering of (un)ordered lists
			forEach('ol,ul', node => {
				let prefix = '',
					parent = node,
					ordered = 'OL' == node.tagName,
					i = 0;
				while ((parent = parent?.parentNode?.closest('ol,ul'))) {
					prefix = '    ' + prefix;
				}
				node.querySelectorAll(':scope > li').forEach(li => {
					li.prepend('\n' + prefix + (ordered ? `${++i}. ` : ' * '));
				});
				node.prepend('\n\n');
				node.after('\n\n');
			});

			// Convert anchors
			forEach('a', a => {
				let txt = a.textContent, href = a.href;
				return a.replaceWith(
					txt.trim() == href || href.includes('mailto:') ? txt : txt + ' ' + href + ' '
				);
			});

			// Bold
			forEach('b,strong', b => b.replaceWith(`**${b.textContent}**`));
			// Italic
			forEach('i,em', i => i.replaceWith(`*${i.textContent}*`));

			// Convert line-breaks
			tpl$1.innerHTML = tpl$1.innerHTML
				.replace(/\n{3,}/gm, '\n\n')
				.replace(/\n<br[^>]*>/g, '\n')
				.replace(/<br[^>]*>\n/g, '\n');
			forEach('br', br => br.replaceWith('\n'));

			// Blockquotes must be last
			blockquotes(tpl$1.content);

			return (tpl$1.content.textContent || '').trim();
		},

		/**
		 * @param {string} plain
		 * @param {boolean} findEmailAndLinksInText = false
		 * @returns {string}
		 */
		plainToHtml = plain => {
			plain = stripTracking(plain)
				.toString()
				.replace(/\r/g, '')
				.replace(/^>[> ]>+/gm, ([match]) => (match ? match.replace(/[ ]+/g, '') : match));

			let bIn = false,
				bDo = true,
				bStart = true,
				aNextText = [],
				aText = plain.split('\n');

			do {
				bDo = false;
				aNextText = [];
				aText.forEach(sLine => {
					bStart = '>' === sLine.slice(0, 1);
					if (bStart && !bIn) {
						bDo = true;
						bIn = true;
						aNextText.push('~~~blockquote~~~');
						aNextText.push(sLine.slice(1));
					} else if (!bStart && bIn) {
						if (sLine) {
							bIn = false;
							aNextText.push('~~~/blockquote~~~');
							aNextText.push(sLine);
						} else {
							aNextText.push(sLine);
						}
					} else if (bStart && bIn) {
						aNextText.push(sLine.slice(1));
					} else {
						aNextText.push(sLine);
					}
				});

				if (bIn) {
					bIn = false;
					aNextText.push('~~~/blockquote~~~');
				}

				aText = aNextText;
			} while (bDo);

			return aText.join('\n')
				// .replace(/~~~\/blockquote~~~\n~~~blockquote~~~/g, '\n')
				.replace(/&/g, '&amp;')
				.replace(/>/g, '&gt;')
				.replace(/</g, '&lt;')
				.replace(/~~~blockquote~~~\s*/g, '<blockquote>')
				.replace(/\s*~~~\/blockquote~~~/g, '</blockquote>')
				.replace(/\n/g, '<br>');
		};

	class HtmlEditor {
		/**
		 * @param {Object} element
		 * @param {Function=} onBlur
		 * @param {Function=} onReady
		 * @param {Function=} onModeChange
		 */
		constructor(element, onBlur = null, onReady = null, onModeChange = null) {
			this.blurTimer = 0;

			this.onBlur = onBlur;
			this.onModeChange = onModeChange;

			if (element) {
				let editor;

				onReady = onReady ? [onReady] : [];
				this.onReady = fn => onReady.push(fn);
				const readyCallback = () => {
					this.editor = editor;
					this.onReady = fn => fn();
					onReady.forEach(fn => fn());
				};

				if (rl.createWYSIWYG) {
					editor = rl.createWYSIWYG(element, readyCallback);
				}
				if (!editor) {
					editor = new SquireUI(element);
					setTimeout(readyCallback, 1);
				}

				editor.on('blur', () => this.blurTrigger());
				editor.on('focus', () => this.blurTimer && clearTimeout(this.blurTimer));
				editor.on('mode', () => {
					this.blurTrigger();
					this.onModeChange?.(!this.isPlain());
				});
			}
		}

		blurTrigger() {
			if (this.onBlur) {
				clearTimeout(this.blurTimer);
				this.blurTimer = setTimeout(() => this.onBlur?.(), 200);
			}
		}

		/**
		 * @returns {boolean}
		 */
		isHtml() {
			return this.editor ? !this.isPlain() : false;
		}

		/**
		 * @returns {boolean}
		 */
		isPlain() {
			return this.editor ? 'plain' === this.editor.mode : false;
		}

		/**
		 * @returns {void}
		 */
		clearCachedSignature() {
			this.onReady(() => this.editor.execCommand('insertSignature', {
				clearCache: true
			}));
		}

		/**
		 * @param {string} signature
		 * @param {bool} html
		 * @param {bool} insertBefore
		 * @returns {void}
		 */
		setSignature(signature, html, insertBefore = false) {
			this.onReady(() => this.editor.execCommand('insertSignature', {
				isHtml: html,
				insertBefore: insertBefore,
				signature: signature
			}));
		}

		/**
		 * @param {boolean=} wrapIsHtml = false
		 * @returns {string}
		 */
		getData() {
			let result = '';
			if (this.editor) {
				try {
					if (this.isPlain() && this.editor.plugins.plain && this.editor.__plain) {
						result = this.editor.__plain.getRawData();
					} else {
						result = this.editor.getData();
					}
				} catch (e) {} // eslint-disable-line no-empty
			}
			return result;
		}

		/**
		 * @returns {string}
		 */
		getDataWithHtmlMark() {
			return (this.isHtml() ? ':HTML:' : '') + this.getData();
		}

		modeWysiwyg() {
			this.onReady(() => this.editor.setMode('wysiwyg'));
		}
		modePlain() {
			this.onReady(() => this.editor.setMode('plain'));
		}

		setHtmlOrPlain(text) {
			if (':HTML:' === text.slice(0, 6)) {
				this.setHtml(text.slice(6));
			} else {
				this.setPlain(text);
			}
		}

		setData(mode, data) {
			this.onReady(() => {
				const editor = this.editor;
				this.clearCachedSignature();
				try {
					editor.setMode(mode);
					if (this.isPlain() && editor.plugins.plain && editor.__plain) {
						editor.__plain.setRawData(data);
					} else {
						editor.setData(data);
					}
				} catch (e) { console.error(e); }
			});
		}

		setHtml(html) {
			this.setData('wysiwyg', html/*.replace(/<p[^>]*><\/p>/gi, '')*/);
		}

		setPlain(txt) {
			this.setData('plain', txt);
		}

		focus() {
			this.onReady(() => this.editor.focus());
		}

		hasFocus() {
			try {
				return !!this.editor?.focusManager.hasFocus;
			} catch (e) {
				return false;
			}
		}

		blur() {
			this.onReady(() => this.editor.focusManager.blur(true));
		}

		clear() {
			this.onReady(() => this.isPlain() ? this.setPlain('') : this.setHtml(''));
		}
	}

	rl.Utils = {
		htmlToPlain: htmlToPlain,
		plainToHtml: plainToHtml
	};

	let FOLDERS_CACHE = {},
		FOLDERS_NAME_CACHE = {},
		MESSAGE_FLAGS_CACHE = {},
		inboxFolderName = 'INBOX';

	const /**
		 * @returns {string}
		 */
		getFolderInboxName = () => inboxFolderName,

		/**
		 * @returns {string}
		 */
		setFolderInboxName = name => inboxFolderName = name,

		/**
		 * @param {string} folderHash
		 * @returns {string}
		 */
		getFolderFullName = folderHash => (folderHash && FOLDERS_NAME_CACHE[folderHash]) || '',

		/**
		 * @param {string} folderHash
		 * @param {string} folderFullName
		 * @param {?FolderModel} folder
		 */
		setFolder = folder => {
			folder.hash = '';
			FOLDERS_CACHE[folder.fullName] = folder;
			FOLDERS_NAME_CACHE[folder.fullNameHash] = folder.fullName;
		},

		/**
		 * @param {string} folderFullName
		 * @param {string} folderHash
		 */
		setFolderHash = (folderFullName, folderHash) =>
			FOLDERS_CACHE[folderFullName] && (FOLDERS_CACHE[folderFullName].hash = folderHash),

		/**
		 * @param {string} folderFullName
		 * @returns {?FolderModel}
		 */
		getFolderFromCacheList = folderFullName =>
			FOLDERS_CACHE[folderFullName] ? FOLDERS_CACHE[folderFullName] : null,

		/**
		 * @param {string} folderFullName
		 */
		removeFolderFromCacheList = folderFullName => delete FOLDERS_CACHE[folderFullName];

	class MessageFlagsCache
	{
		/**
		 * @param {string} folderFullName
		 * @param {string} uid
		 * @param {string} flag
		 * @returns {bool}
		 */
		static hasFlag(folderFullName, uid, flag) {
			return MESSAGE_FLAGS_CACHE[folderFullName]?.[uid]?.includes(flag);
		}

		/**
		 * @param {string} folderFullName
		 * @param {string} uid
		 * @returns {?Array}
		 */
		static getFor(folderFullName, uid) {
			return MESSAGE_FLAGS_CACHE[folderFullName]?.[uid];
		}

		/**
		 * @param {string} folderFullName
		 * @param {string} uid
		 * @param {Array} flagsCache
		 */
		static setFor(folderFullName, uid, flags) {
			if (isArray(flags)) {
				if (!MESSAGE_FLAGS_CACHE[folderFullName]) {
					MESSAGE_FLAGS_CACHE[folderFullName] = {};
				}
				MESSAGE_FLAGS_CACHE[folderFullName][uid] = flags;
			}
		}

		/**
		 * @param {string} folderFullName
		 */
		static clearFolder(folderFullName) {
			MESSAGE_FLAGS_CACHE[folderFullName] = {};
		}

		/**
		 * @param {(MessageModel|null)} message
		 */
		static initMessage(message) {
			if (message) {
				const uid = message.uid,
					flags = this.getFor(message.folder, uid),
					thread = message.threads();

				if (isArray(flags)) {
					message.flags(flags);
				}

				if (thread.length) {
					message.hasUnseenSubMessage(!!thread.find(iSubUid =>
						(uid !== iSubUid) && !this.hasFlag(message.folder, iSubUid, '\\seen')
					));
					message.hasFlaggedSubMessage(!!thread.find(iSubUid =>
						(uid !== iSubUid) && this.hasFlag(message.folder, iSubUid, '\\flagged')
					));
				}
			}
		}

		/**
		 * @param {(MessageModel|null)} message
		 */
		static store(message) {
			message && this.setFor(message.folder, message.uid, message.flags());
		}

		/**
		 * @param {string} folder
		 * @param {string} uid
		 * @param {number} setAction
		 */
		static storeBySetAction(folder, uid, setAction) {
			let flags = this.getFor(folder, uid) || [];
			const
				unread = flags.includes('\\seen') ? 0 : 1,
				add = item => flags.includes(item) || flags.push(item),
				remove = item => flags = flags.filter(flag => flag != item);

			switch (setAction) {
				case MessageSetAction.SetSeen:
					add('\\seen');
					break;
				case MessageSetAction.UnsetSeen:
					remove('\\seen');
					break;
				case MessageSetAction.SetFlag:
					add('\\flagged');
					break;
				case MessageSetAction.UnsetFlag:
					remove('\\flagged');
					break;
				// no default
			}

			this.setFor(folder, uid, flags);

			return unread;
		}

	}

	function typeCast(curValue, newValue) {
		if (null != curValue) {
			switch (typeof curValue)
			{
			case 'boolean': return 0 != newValue && !!newValue;
			case 'number': return isFinite(newValue) ? parseFloat(newValue) : 0;
			case 'string': return null != newValue ? '' + newValue : '';
			case 'object':
				if (curValue.constructor.reviveFromJson) {
					return curValue.constructor.reviveFromJson(newValue);
				}
				if (isArray(curValue) && !isArray(newValue))
					return [];
			}
		}
		return newValue;
	}

	class AbstractModel {
		constructor() {
	/*
			if (new.target === AbstractModel) {
				throw new Error("Can't instantiate AbstractModel!");
			}
	*/
			this.disposables = [];
		}

		addObservables(observables) {
			addObservablesTo(this, observables);
		}

		addComputables(computables) {
			addComputablesTo(this, computables);
		}

		addSubscribables(subscribables) {
	//		addSubscribablesTo(this, subscribables);
			forEachObjectEntry(subscribables, (key, fn) => this.disposables.push( this[key].subscribe(fn) ) );
		}

		/** Called by delegateRunOnDestroy */
		onDestroy() {
			/** dispose ko subscribables */
			this.disposables.forEach(dispose);
			/** clear object entries */
	//		forEachObjectEntry(this, (key, value) => {
			forEachObjectValue(this, value => {
				/** clear CollectionModel */
				(ko.isObservableArray(value) ? value() : value)?.onDestroy?.();
				/** destroy ko.observable/ko.computed? */
	//			dispose(value);
				/** clear object value */
	//			this[key] = null; // TODO: issue with Contacts view
			});
	//		this.disposables = [];
		}

		/**
		 * @static
		 * @param {FetchJson} json
		 * @returns {boolean}
		 */
		static validJson(json) {
			return !!(json && ('Object/'+this.name.replace('Model', '') === json['@Object']));
		}

		/**
		 * @static
		 * @param {FetchJson} json
		 * @returns {*Model}
		 */
		static reviveFromJson(json) {
			let obj = this.validJson(json) ? new this() : null;
			obj?.revivePropertiesFromJson(json);
			return obj;
		}

		revivePropertiesFromJson(json) {
			let model = this.constructor;
			if (!model.validJson(json)) {
				return false;
			}
			forEachObjectEntry(json, (key, value) => {
				if ('@' !== key[0]) try {
					key = key[0].toLowerCase() + key.slice(1);
					switch (typeof this[key])
					{
					case 'function':
						if (ko.isObservable(this[key])) {
							this[key](typeCast(this[key](), value));
	//						console.log('Observable ' + (typeof this[key]()) + ' ' + (model.name) + '.' + key + ' revived');
						}
	//					else console.log(model.name + '.' + key + ' is a function');
						break;
					case 'boolean':
					case 'number':
					case 'object':
					case 'string':
						this[key] = typeCast(this[key], value);
						break;
						// fall through
					case 'undefined':
					default:
	//					console.log((typeof this[key])+' '+(model.name)+'.'+key+' not revived');
					}
				} catch (e) {
					console.log(model.name + '.' + key);
					console.error(e);
				}
			});
			return true;
		}

	}

	/**
	 * Parses structured e-mail addresses from an address field
	 *
	 * Example:
	 *
	 *    "Name <address@domain>"
	 *
	 * will be converted to
	 *
	 *     [{name: "Name", address: "address@domain"}]
	 *
	 * @param {String} str Address field
	 * @return {Array} An array of address objects
	 */
	function addressparser(str) {
		var tokenizer = new Tokenizer(str);
		var tokens = tokenizer.tokenize();
		var addresses = [];
		var address = [];
		var parsedAddresses = [];

		tokens.forEach(token => {
			if (token.type === 'operator' && (token.value === ',' || token.value === ';')) {
				if (address.length) {
					addresses.push(address);
				}
				address = [];
			} else {
				address.push(token);
			}
		});

		if (address.length) {
			addresses.push(address);
		}

		addresses.forEach(address => {
			address = _handleAddress(address);
			if (address.length) {
				parsedAddresses = parsedAddresses.concat(address);
			}
		});

		return parsedAddresses;
	}

	/**
	 * Converts tokens for a single address into an address object
	 *
	 * @param {Array} tokens Tokens object
	 * @return {Object} Address object
	 */
	function _handleAddress(tokens) {
		var isGroup = false;
		var state = 'text';
		var address = void 0;
		var addresses = [];
		var data = {
			address: [],
			comment: [],
			group: [],
			text: []
		};

		// Filter out <addresses>, (comments) and regular text
		tokens.forEach(token => {
			if (token.type === 'operator') {
				switch (token.value) {
					case '<':
						state = 'address';
						break;
					case '(':
						state = 'comment';
						break;
					case ':':
						state = 'group';
						isGroup = true;
						break;
					default:
						state = 'text';
				}
			} else if (token.value) {
				data[state].push(token.value);
			}
		});

		// If there is no text but a comment, replace the two
		if (!data.text.length && data.comment.length) {
			data.text = data.comment;
			data.comment = [];
		}

		if (isGroup) {
			// http://tools.ietf.org/html/rfc2822#appendix-A.1.3
			data.text = data.text.join(' ');
			addresses.push({
				name: data.text || address && address.name,
				group: data.group.length ? addressparser(data.group.join(',')) : []
			});
		} else {
			// If no address was found, try to detect one from regular text
			if (!data.address.length && data.text.length) {
				var i = data.text.length;
				while (i--) {
					if (data.text[i].match(/^[^@\s]+@[^@\s]+$/)) {
						data.address = data.text.splice(i, 1);
						break;
					}
				}

				// still no address
				if (!data.address.length) {
					i = data.text.length;
					while (i--) {
						data.text[i] = data.text[i].replace(/\s*\b[^@\s]+@[^@\s]+\b\s*/, address => {
							if (!data.address.length) {
								data.address = [address.trim()];
								return '';
							}
							return address.trim();
						});
						if (data.address.length) {
							break;
						}
					}
				}
			}

			// If there's still is no text but a comment exixts, replace the two
			if (!data.text.length && data.comment.length) {
				data.text = data.comment;
				data.comment = [];
			}

			// Keep only the first address occurence, push others to regular text
			if (data.address.length > 1) {
				data.text = data.text.concat(data.address.splice(1));
			}

			// Join values with spaces
			data.text = data.text.join(' ');
			data.address = data.address.join(' ');

			if (!data.address && isGroup) {
				return [];
			}
			address = {
				address: data.address || data.text || '',
				name: data.text || data.address || ''
			};

			if (address.address === address.name) {
				if ((address.address || '').match(/@/)) {
					address.name = '';
				} else {
					address.address = '';
				}
			}

			addresses.push(address);
		}

		return addresses;
	}

	/*
	 * Operator tokens and which tokens are expected to end the sequence
	 */
	var OPERATORS = {
	  '"': '"',
	  '(': ')',
	  '<': '>',
	  ',': '',
	  // Groups are ended by semicolons
	  ':': ';',
	  // Semicolons are not a legal delimiter per the RFC2822 grammar other
	  // than for terminating a group, but they are also not valid for any
	  // other use in this context.  Given that some mail clients have
	  // historically allowed the semicolon as a delimiter equivalent to the
	  // comma in their UI, it makes sense to treat them the same as a comma
	  // when used outside of a group.
	  ';': ''
	};

	class Tokenizer
	{
		constructor(str) {
			this.str = (str || '').toString();
			this.operatorCurrent = '';
			this.operatorExpecting = '';
			this.node = null;
			this.escaped = false;
			this.list = [];
		}

		tokenize() {
			var list = [];
			[...this.str].forEach(c => this.checkChar(c));

			this.list.forEach(node => {
				node.value = (node.value || '').toString().trim();
				if (node.value) {
					list.push(node);
				}
			});

			return list;
		}

		checkChar(chr) {
			if ((chr in OPERATORS || chr === '\\') && this.escaped) {
				this.escaped = false;
			} else if (this.operatorExpecting && chr === this.operatorExpecting) {
				this.node = {
					type: 'operator',
					value: chr
				};
				this.list.push(this.node);
				this.node = null;
				this.operatorExpecting = '';
				this.escaped = false;
				return;
			} else if (!this.operatorExpecting && chr in OPERATORS) {
				this.node = {
					type: 'operator',
					value: chr
				};
				this.list.push(this.node);
				this.node = null;
				this.operatorExpecting = OPERATORS[chr];
				this.escaped = false;
				return;
			}

			if (!this.escaped && chr === '\\') {
				this.escaped = true;
				return;
			}

			if (!this.node) {
				this.node = {
					type: 'text',
					value: ''
				};
				this.list.push(this.node);
			}

			if (this.escaped && chr !== '\\') {
				this.node.value += '\\';
			}

			this.node.value += chr;
			this.escaped = false;
		}
	}

	class EmailModel extends AbstractModel {
		/**
		 * @param {string=} email = ''
		 * @param {string=} name = ''
		 * @param {string=} dkimStatus = 'none'
		 * @param {string=} dkimValue = ''
		 */
		constructor(email = '', name = '', dkimStatus = 'none', dkimValue = '') {
			super();
			this.email = email;
			this.name = name;
			this.dkimStatus = dkimStatus;
			this.dkimValue = dkimValue;

			this.clearDuplicateName();
		}

		/**
		 * @static
		 * @param {FetchJsonEmail} json
		 * @returns {?EmailModel}
		 */
		static reviveFromJson(json) {
			const email = super.reviveFromJson(json);
			email?.clearDuplicateName();
			return email;
		}

		/**
		 * @returns {void}
		 */
		clear() {
			this.email = '';
			this.name = '';

			this.dkimStatus = 'none';
			this.dkimValue = '';
		}

		/**
		 * @returns {boolean}
		 */
		validate() {
			return this.name || this.email;
		}

		/**
		 * @param {boolean} withoutName = false
		 * @returns {string}
		 */
		hash(withoutName = false) {
			return '#' + (withoutName ? '' : this.name) + '#' + this.email + '#';
		}

		/**
		 * @returns {void}
		 */
		clearDuplicateName() {
			if (this.name === this.email) {
				this.name = '';
			}
		}

		/**
		 * @param {string} query
		 * @returns {boolean}
		 */
		search(query) {
			return (this.name + ' ' + this.email).toLowerCase().includes(query.toLowerCase());
		}

		/**
		 * @param {boolean} friendlyView = false
		 * @param {boolean} wrapWithLink = false
		 * @returns {string}
		 */
		toLine(friendlyView, wrapWithLink) {
			let result = this.email,
				name = this.name,
				toLink = text =>
					'<a href="mailto:'
					+ encodeHtml(result) + (name ? '?to=' + encodeURIComponent('"' + name + '" <' + result + '>') : '')
					+ '" target="_blank" tabindex="-1">'
					+ encodeHtml(text || result)
					+ '</a>';
			if (result) {
				if (name) {
					result = friendlyView
						? (wrapWithLink ? toLink(name) : name)
						: (wrapWithLink
							? encodeHtml('"' + name + '" <') + toLink() + encodeHtml('>')
							: '"' + name + '" <' + result + '>'
						);
				} else if (wrapWithLink) {
					result = toLink();
				}
			}
			return result;
		}

		static splitEmailLine(line) {
			const parsedResult = addressparser(line);
			if (parsedResult.length) {
				const result = [];
				let exists = false;
				parsedResult.forEach((item) => {
					const address = item.address
						? new EmailModel(item.address.replace(/^[<]+(.*)[>]+$/g, '$1'), item.name || '')
						: null;

					if (address?.email) {
						exists = true;
					}

					result.push(address ? address.toLine() : item.name);
				});

				return exists ? result : null;
			}

			return null;
		}

		static parseEmailLine(line) {
			const parsedResult = addressparser(line);
			if (parsedResult.length) {
				return parsedResult.map(item =>
					item.address ? new EmailModel(item.address.replace(/^[<]+(.*)[>]+$/g, '$1'), item.name || '') : null
				).filter(v => v);
			}

			return [];
		}

		/**
		 * @param {string} emailAddress
		 * @returns {boolean}
		 */
		parse(emailAddress) {
			emailAddress = emailAddress.trim();
			if (!emailAddress) {
				return false;
			}

			const result = addressparser(emailAddress);
			if (result.length) {
				this.name = result[0].name || '';
				this.email = result[0].address || '';
				this.clearDuplicateName();

				return true;
			}

			return false;
		}
	}

	const UNUSED_OPTION_VALUE = '__UNUSE__';

	//import Remote from 'Remote/User/Fetch'; // Circular dependency

	const FolderUserStore = new class {
		constructor() {
			const self = this;
			addObservablesTo(self, {
				/**
				 * To use "checkable" option in /#/settings/folders
				 * When true, getNextFolderNames only lists system and "checkable" folders
				 * and affects the update of unseen count
				 * Auto set to true when amount of folders > folderSpecLimit to prevent requests overload,
				 * see application.ini [labs] folders_spec_limit
				 */
				displaySpecSetting: false,

	//			sortMode: '',

				quotaLimit: 0,
				quotaUsage: 0,

				sentFolder: '',
				draftsFolder: '',
				spamFolder: '',
				trashFolder: '',
				archiveFolder: '',

				folderListOptimized: false,
				folderListError: '',

				foldersLoading: false,
				foldersCreating: false,
				foldersDeleting: false,
				foldersRenaming: false,

				foldersInboxUnreadCount: 0
			});

			self.sortMode = ko.observable('');

			self.namespace = '';

			self.folderList = ko.observableArray(/*new FolderCollectionModel*/);

			self.capabilities = ko.observableArray();

			self.currentFolder = ko.observable(null).extend({ toggleSubscribeProperty: [self, 'selected'] });

			addComputablesTo(self, {

				draftsFolderNotEnabled: () => !self.draftsFolder() || UNUSED_OPTION_VALUE === self.draftsFolder(),

				currentFolderFullName: () => (self.currentFolder() ? self.currentFolder().fullName : ''),
				currentFolderFullNameHash: () => (self.currentFolder() ? self.currentFolder().fullNameHash : ''),

				foldersChanging: () =>
					self.foldersLoading() | self.foldersCreating() | self.foldersDeleting() | self.foldersRenaming(),

				systemFoldersNames: () => {
					const list = [getFolderInboxName()],
					others = [self.sentFolder(), self.draftsFolder(), self.spamFolder(), self.trashFolder(), self.archiveFolder()];

					self.folderList().length &&
						others.forEach(name => name && UNUSED_OPTION_VALUE !== name && list.push(name));

					return list;
				},

				systemFolders: () =>
					self.systemFoldersNames().map(name => getFolderFromCacheList(name)).filter(v => v)
			});

			const
				subscribeRemoveSystemFolder = observable => {
					observable.subscribe(() => getFolderFromCacheList(observable())?.type(FolderType.User), self, 'beforeChange');
				},
				fSetSystemFolderType = type => value => getFolderFromCacheList(value)?.type(type);

			subscribeRemoveSystemFolder(self.sentFolder);
			subscribeRemoveSystemFolder(self.draftsFolder);
			subscribeRemoveSystemFolder(self.spamFolder);
			subscribeRemoveSystemFolder(self.trashFolder);
			subscribeRemoveSystemFolder(self.archiveFolder);

			addSubscribablesTo(self, {
				sentFolder: fSetSystemFolderType(FolderType.Sent),
				draftsFolder: fSetSystemFolderType(FolderType.Drafts),
				spamFolder: fSetSystemFolderType(FolderType.Spam),
				trashFolder: fSetSystemFolderType(FolderType.Trash),
				archiveFolder: fSetSystemFolderType(FolderType.Archive)
			});

			self.quotaPercentage = koComputable(() => {
				const quota = self.quotaLimit(), usage = self.quotaUsage();
				return 0 < quota ? Math.ceil((usage / quota) * 100) : 0;
			});
		}

		/**
		 * If the IMAP server supports SORT, METADATA
		 */
		hasCapability(name) {
			return this.capabilities().includes(name);
		}

		allowKolab() {
			return FolderUserStore.hasCapability('METADATA') && SettingsCapa('Kolab');
		}

		/**
		 * @returns {Array}
		 */
		getNextFolderNames(ttl) {
			const result = [],
				limit = 10,
				utc = Date.now(),
				timeout = utc - ttl,
				timeouts = [],
				bDisplaySpecSetting = this.displaySpecSetting(),
				fSearchFunction = (list) => {
					list.forEach(folder => {
						if (
							folder?.selectable() &&
							folder.exists &&
							timeout > folder.expires &&
							(folder.isSystemFolder() || (folder.isSubscribed() && (folder.checkable() || !bDisplaySpecSetting)))
						) {
							timeouts.push([folder.expires, folder.fullName]);
						}

						if (folder?.subFolders.length) {
							fSearchFunction(folder.subFolders());
						}
					});
				};

			fSearchFunction(this.folderList());

			timeouts.sort((a, b) => (a[0] < b[0]) ? -1 : (a[0] > b[0] ? 1 : 0));

			timeouts.find(aItem => {
				const folder = getFolderFromCacheList(aItem[1]);
				if (folder) {
					folder.expires = utc;
	//				result.indexOf(aItem[1]) ||
					result.push(aItem[1]);
				}

				return limit <= result.length;
			});

			return result;
		}

		saveSystemFolders(folders) {
			folders = folders || {
				Sent: FolderUserStore.sentFolder(),
				Drafts: FolderUserStore.draftsFolder(),
				Spam: FolderUserStore.spamFolder(),
				Trash: FolderUserStore.trashFolder(),
				Archive: FolderUserStore.archiveFolder()
			};
			forEachObjectEntry(folders, (k,v)=>Settings.set(k+'Folder',v));
			rl.app.Remote.request('SystemFoldersUpdate', null, folders);
		}
	};

	/* eslint key-spacing: 0 */

	const
		cache = {},
		app$1 = 'application/',
		msOffice = app$1+'vnd.openxmlformats-officedocument.',
		openDoc = app$1+'vnd.oasis.opendocument.',
		sizes = ['B', 'KiB', 'MiB', 'GiB', 'TiB'],
		lowerCase = text => text.toLowerCase().trim(),

		exts = {
			eml: 'message/rfc822',
			mime: 'message/rfc822',
			vcard: 'text/vcard',
			vcf: 'text/vcard',
			htm: 'text/html',
			html: 'text/html',
			csv: 'text/csv',
			ics: 'text/calendar',
			xml: 'text/xml',
			json: app$1+'json',
			asc: app$1+'pgp-signature',
			p10: app$1+'pkcs10',
			p7c: app$1+'pkcs7-mime',
			p7m: app$1+'pkcs7-mime',
			p7s: app$1+'pkcs7-signature',
			torrent: app$1+'x-bittorrent',

			// scripts
			js: app$1+'javascript',
			pl: 'text/perl',
			css: 'text/css',
			asp: 'text/asp',
			php: app$1+'x-php',

			// images
			jpg: 'image/jpeg',
			ico: 'image/x-icon',
			tif: 'image/tiff',
			svg: 'image/svg+xml',
			svgz: 'image/svg+xml',

			// archives
			zip: app$1+'zip',
			'7z': app$1+'x-7z-compressed',
			rar: app$1+'x-rar-compressed',
			cab: app$1+'vnd.ms-cab-compressed',
			gz: app$1+'x-gzip',
			tgz: app$1+'x-gzip',
			bz: app$1+'x-bzip',
			bz2: app$1+'x-bzip2',
			deb: app$1+'x-debian-package',

			// audio
			mp3: 'audio/mpeg',
			wav: 'audio/x-wav',
			mp4a: 'audio/mp4',
			weba: 'audio/webm',
			m3u: 'audio/x-mpegurl',

			// video
			qt: 'video/quicktime',
			mov: 'video/quicktime',
			wmv: 'video/windows-media',
			avi: 'video/x-msvideo',
			'3gp': 'video/3gpp',
			'3g2': 'video/3gpp2',
			mp4v: 'video/mp4',
			mpg4: 'video/mp4',
			ogv: 'video/ogg',
			m4v: 'video/x-m4v',
			asf: 'video/x-ms-asf',
			asx: 'video/x-ms-asf',
			wm: 'video/x-ms-wm',
			wmx: 'video/x-ms-wmx',
			wvx: 'video/x-ms-wvx',
			movie: 'video/x-sgi-movie',

			// adobe
			pdf: app$1+'pdf',
			psd: 'image/vnd.adobe.photoshop',
			ai: app$1+'postscript',
			eps: app$1+'postscript',
			ps: app$1+'postscript',

			// ms office
			doc: app$1+'msword',
			rtf: app$1+'rtf',
			xls: app$1+'vnd.ms-excel',
			ppt: app$1+'vnd.ms-powerpoint',
			docx: msOffice+'wordprocessingml.document',
			xlsx: msOffice+'spreadsheetml.sheet',
			dotx: msOffice+'wordprocessingml.template',
			pptx: msOffice+'presentationml.presentation',

			// open office
			odt: openDoc+'text',
			ods: openDoc+'spreadsheet',
			odp: openDoc+'presentation'
		};

	const FileType = {
		Unknown: 'unknown',
		Text: 'text',
		Code: 'code',
		Eml: 'eml',
		Word: 'word',
		Pdf: 'pdf',
		Image: 'image',
		Audio: 'audio',
		Video: 'video',
		Spreadsheet: 'spreadsheet',
		Presentation: 'presentation',
		Certificate: 'certificate',
		Archive: 'archive'
	};

	const FileInfo = {
		/**
		 * @param {string} fileName
		 * @returns {string}
		 */
		getExtension: fileName => {
			fileName = lowerCase(fileName);
			const result = fileName.split('.').pop();
			return result === fileName ? '' : result;
		},

		getContentType: fileName => {
			fileName = lowerCase(fileName);
			if ('winmail.dat' === fileName) {
				return app$1 + 'ms-tnef';
			}
			let ext = fileName.split('.').pop();
			if (/^(txt|text|def|list|in|ini|log|sql|cfg|conf)$/.test(ext))
				return 'text/plain';
			if (/^(mpe?g|mpe|m1v|m2v)$/.test(ext))
				return 'video/mpeg';
			if (/^aif[cf]?$/.test(ext))
				return 'audio/aiff';
			if (/^(aac|flac|midi|ogg)$/.test(ext))
				return 'audio/'+ext;
			if (/^(h26[134]|jpgv|mp4|webm)$/.test(ext))
				return 'video/'+ext;
			if (/^(otf|sfnt|ttf|woff2?)$/.test(ext))
				return 'font/'+ext;
			if (/^(png|jpeg|gif|tiff|webp)$/.test(ext))
				return 'image/'+ext;

			return exts[ext] || app$1+'octet-stream';
		},

		/**
		 * @param {string} sExt
		 * @param {string} sMimeType
		 * @returns {string}
		 */
		getType: (ext, mimeType) => {
			ext = lowerCase(ext);
			mimeType = lowerCase(mimeType).replace('csv/plain', 'text/csv');

			let key = ext + mimeType;
			if (cache[key]) {
				return cache[key];
			}

			let result = FileType.Unknown;
			const mimeTypeParts = mimeType.split('/'),
				type = mimeTypeParts[1].replace('x-','').replace('-compressed',''),
				match = str => mimeType.includes(str),
				archive = /^(zip|7z|tar|rar|gzip|bzip|bzip2)$/;

			switch (true) {
				case 'image' == mimeTypeParts[0] || ['png', 'jpg', 'jpeg', 'gif', 'webp'].includes(ext):
					result = FileType.Image;
					break;
				case 'audio' == mimeTypeParts[0] || ['mp3', 'ogg', 'oga', 'wav'].includes(ext):
					result = FileType.Audio;
					break;
				case 'video' == mimeTypeParts[0] || 'mkv' == ext || 'avi' == ext:
					result = FileType.Video;
					break;
				case ['php', 'js', 'css', 'xml', 'html'].includes(ext) || 'text/html' == mimeType:
					result = FileType.Code;
					break;
				case 'eml' == ext || ['message/delivery-status', 'message/rfc822'].includes(mimeType):
					result = FileType.Eml;
					break;
				case 'text' == mimeTypeParts[0] || 'txt' == ext || 'log' == ext:
					result = FileType.Text;
					break;
				case archive.test(type) || archive.test(ext):
					result = FileType.Archive;
					break;
				case 'pdf' == type || 'pdf' == ext:
					result = FileType.Pdf;
					break;
				case [app$1+'pgp-signature', app$1+'pgp-keys'].includes(mimeType)
					|| ['asc', 'pem', 'ppk'].includes(ext)
					|| [app$1+'pkcs7-signature'].includes(mimeType) || 'p7s' == ext:
					result = FileType.Certificate;
					break;
				case match(msOffice+'.wordprocessingml') || match(openDoc+'.text') || match('vnd.ms-word')
					|| ['rtf', 'msword', 'vnd.msword'].includes(type):
					result = FileType.Word;
					break;
				case match(msOffice+'.spreadsheetml') || match(openDoc+'.spreadsheet') || match('ms-excel'):
					result = FileType.Spreadsheet;
					break;
				case match(msOffice+'.presentationml') || match(openDoc+'.presentation') || match('ms-powerpoint'):
					result = FileType.Presentation;
					break;
				// no default
			}

			return cache[key] = result;
		},

		/**
		 * @param {string} sFileType
		 * @returns {string}
		 */
		getTypeIconClass: fileType => {
			let result = 'icon-file';
			switch (fileType) {
				case FileType.Text:
				case FileType.Eml:
				case FileType.Pdf:
				case FileType.Word:
					return result + '-text';
				case FileType.Code:
				case FileType.Image:
				case FileType.Audio:
				case FileType.Video:
				case FileType.Archive:
				case FileType.Certificate:
				case FileType.Spreadsheet:
				case FileType.Presentation:
					return result + '-' + fileType;
			}
			return result;
		},

		getIconClass: (ext, mime) => FileInfo.getTypeIconClass(FileInfo.getType(ext, mime)),

		/**
		 * @param {string} sFileType
		 * @returns {string}
		 */
		getAttachmentsIconClass: data => {
			if (arrayLength(data)) {
				let icons = data
					.map(item => item ? FileInfo.getIconClass(FileInfo.getExtension(item.fileName), item.mimeType) : '')
					.validUnique();

				return (1 === icons?.length && 'icon-file' !== icons[0])
					 ? icons[0]
					 : 'icon-attachment';
			}

			return '';
		},

		friendlySize: bytes => {
			bytes = parseInt(bytes, 10) || 0;
			let i = Math.floor(Math.log(bytes) / Math.log(1024));
			return (bytes / Math.pow(1024, i)).toFixed(2>i ? 0 : 1) + ' ' + sizes[i];
		}

	};

	class AbstractCollectionModel extends Array
	{
		constructor() {
	/*
			if (new.target === AbstractCollectionModel) {
				throw new Error("Can't instantiate AbstractCollectionModel!");
			}
	*/
			super();
		}

		onDestroy() {
			this.forEach(item => item.onDestroy?.());
		}

		/**
		 * @static
		 * @param {FetchJson} json
		 * @returns {*CollectionModel}
		 */
		static reviveFromJson(json, itemCallback) {
			const result = new this();
			if (json) {
				if ('Collection/'+this.name.replace('Model', '') === json['@Object']) {
					forEachObjectEntry(json, (key, value) => '@' !== key[0] && (result[key] = value));
					json = json['@Collection'];
				}
				if (isArray(json)) {
					json.forEach(item => {
						item && itemCallback && (item = itemCallback(item, result));
						item && result.push(item);
					});
				}
			}
			return result;
		}

	}

	let notificator = null,
		player = null,
		canPlay = type => !!player?.canPlayType(type).replace('no', ''),

		audioCtx = window.AudioContext || window.webkitAudioContext,

		play = (url, name) => {
			if (player) {
				player.src = url;
				player.play();
				name = name.trim();
				fireEvent('audio.start', name.replace(/\.([a-z0-9]{3})$/, '') || 'audio');
			}
		},

		createNewObject = () => {
			try {
				const player = new Audio;
				if (player.canPlayType && player.pause && player.play) {
					player.preload = 'none';
					player.loop = false;
					player.autoplay = false;
					player.muted = false;
					return player;
				}
			} catch (e) {
				console.error(e);
			}
			return null;
		},

		// The AudioContext is not allowed to start.
		// It must be resumed (or created) after a user gesture on the page. https://goo.gl/7K7WLu
		// Setup listeners to attempt an unlock
		unlockEvents = [
			'click','dblclick',
			'contextmenu',
			'auxclick',
			'mousedown','mouseup',
			'pointerup',
			'touchstart','touchend',
			'keydown','keyup'
		],
		unlock = () => {
			unlockEvents.forEach(type => doc.removeEventListener(type, unlock, true));
			if (audioCtx) {
				console.log('AudioContext ' + audioCtx.state);
				audioCtx.resume();
			}
	//		setTimeout(()=>SMAudio.playNotification(0,1),1);
		};

	if (audioCtx) {
		audioCtx = audioCtx ? new audioCtx : null;
		audioCtx.onstatechange = unlock;
	}
	unlockEvents.forEach(type => doc.addEventListener(type, unlock, true));

	/**
	 * Browsers can't play without user interaction
	 */

	const SMAudio = new class {
		constructor() {
			player || (player = createNewObject());

			this.supported = !!player;
			this.supportedMp3 = canPlay('audio/mpeg;');
			this.supportedWav = canPlay('audio/wav; codecs="1"');
			this.supportedOgg = canPlay('audio/ogg; codecs="vorbis"');
			if (player) {
				const stopFn = () => this.pause();
				addEventsListener(player, ['ended','error'], stopFn);
				addEventListener('audio.api.stop', stopFn);
			}

			addObservablesTo(this, {
				notifications: false
			});
		}

		paused() {
			return !player || player.paused;
		}

		stop() {
			this.pause();
		}

		pause() {
			player?.pause();
			fireEvent('audio.stop');
		}

		playMp3(url, name) {
			this.supportedMp3 && play(url, name);
		}

		playOgg(url, name) {
			this.supportedOgg && play(url, name);
		}

		playWav(url, name) {
			this.supportedWav && play(url, name);
		}

		/**
		 * Used with SoundNotification setting
		 */
		playNotification(force, silent) {
			if (force || this.notifications()) {
				if ('running' == audioCtx.state && (this.supportedMp3 || this.supportedOgg)) {
					notificator = notificator || createNewObject();
					if (notificator) {
						notificator.src = staticLink('sounds/'
							+ SettingsGet('NotificationSound')
							+ (this.supportedMp3 ? '.mp3' : '.ogg'));
						notificator.volume = silent ? 0.01 : 1;
						notificator.play();
					}
				} else {
					console.log('No audio: ' + audioCtx.state);
				}
			}
		}
	};

	class AttachmentModel extends AbstractModel {
		constructor() {
			super();

			this.checked = ko.observable(true);

			this.mimeType = '';
			this.fileName = '';
			this.fileNameExt = '';
			this.fileType = FileType.Unknown;
			this.friendlySize = '';
			this.isThumbnail = false;
			this.cid = '';
			this.contentLocation = '';
			this.download = '';
			this.folder = '';
			this.uid = '';
			this.url = '';
			this.mimeIndex = '';
			this.framed = false;

			this.addObservables({
				isInline: false,
				isLinked: false
			});
		}

		/**
		 * @static
		 * @param {FetchJsonAttachment} json
		 * @returns {?AttachmentModel}
		 */
		static reviveFromJson(json) {
			const attachment = super.reviveFromJson(json);
			if (attachment) {
				attachment.friendlySize = FileInfo.friendlySize(json.EstimatedSize);

				attachment.fileNameExt = FileInfo.getExtension(attachment.fileName);
				attachment.fileType = FileInfo.getType(attachment.fileNameExt, attachment.mimeType);
			}
			return attachment;
		}

		contentId() {
			return this.cid.replace(/^<+|>+$/g, '');
		}

		/**
		 * @returns {boolean}
		 */
		isImage() {
			return FileType.Image === this.fileType;
		}

		/**
		 * @returns {boolean}
		 */
		isMp3() {
			return FileType.Audio === this.fileType && 'mp3' === this.fileNameExt;
		}

		/**
		 * @returns {boolean}
		 */
		isOgg() {
			return FileType.Audio === this.fileType && ('oga' === this.fileNameExt || 'ogg' === this.fileNameExt);
		}

		/**
		 * @returns {boolean}
		 */
		isWav() {
			return FileType.Audio === this.fileType && 'wav' === this.fileNameExt;
		}

		/**
		 * @returns {boolean}
		 */
		hasThumbnail() {
			return this.isThumbnail;
		}

		/**
		 * @returns {boolean}
		 */
		isText() {
			return FileType.Text === this.fileType || FileType.Eml === this.fileType;
		}

		/**
		 * @returns {boolean}
		 */
		pdfPreview() {
			return null != navigator.mimeTypes['application/pdf'] && FileType.Pdf === this.fileType;
		}

		/**
		 * @returns {boolean}
		 */
		hasPreview() {
			return this.isImage() || this.pdfPreview() || this.isText();
		}

		/**
		 * @returns {boolean}
		 */
		hasPreplay() {
			return (
				(SMAudio.supportedMp3 && this.isMp3()) ||
				(SMAudio.supportedOgg && this.isOgg()) ||
				(SMAudio.supportedWav && this.isWav())
			);
		}

		/**
		 * @returns {string}
		 */
		linkDownload() {
			return this.url || attachmentDownload(this.download);
		}

		/**
		 * @returns {string}
		 */
		linkPreview() {
			return this.url || serverRequestRaw('View', this.download);
		}

		/**
		 * @returns {string}
		 */
		linkThumbnailPreviewStyle() {
			return this.hasThumbnail() ? 'background:url(' + serverRequestRaw('ViewThumbnail', this.download) + ')' : '';
		}

		/**
		 * @returns {string}
		 */
		linkPreviewMain() {
			let result = '';
			switch (true) {
				case this.isImage():
				case this.pdfPreview():
					result = this.linkPreview();
					break;
				case this.isText():
					result = serverRequestRaw('ViewAsPlain', this.download);
					break;
				// no default
			}

			return result;
		}

		/**
		 * @param {AttachmentModel} attachment
		 * @param {*} event
		 * @returns {boolean}
		 */
		eventDragStart(attachment, event) {
			const localEvent = event.originalEvent || event;
			if (attachment && localEvent && localEvent.dataTransfer && localEvent.dataTransfer.setData) {
				let link = this.linkDownload();
				if ('http' !== link.slice(0, 4)) {
					link = location.protocol + '//' + location.host + location.pathname + link;
				}
				localEvent.dataTransfer.setData('DownloadURL', this.mimeType + ':' + this.fileName + ':' + link);
			}

			return true;
		}

		/**
		 * @returns {string}
		 */
		iconClass() {
			return FileInfo.getTypeIconClass(this.fileType);
		}
	}

	class AttachmentCollectionModel extends AbstractCollectionModel
	{
		/**
		 * @param {?Array} json
		 * @returns {AttachmentCollectionModel}
		 */
		static reviveFromJson(items) {
			return super.reviveFromJson(items, attachment => AttachmentModel.reviveFromJson(attachment));
	/*
			const attachments = super.reviveFromJson(items, attachment => AttachmentModel.reviveFromJson(attachment));
			if (attachments) {
				attachments.InlineCount = attachments.reduce((accumulator, a) => accumulator + (a.isInline ? 1 : 0), 0);
			}
			return attachments;
	*/
		}

		/**
		 * @returns {boolean}
		 */
		hasVisible() {
			return !!this.filter(item => !item.isLinked()).length;
		}

		/**
		 * @param {string} cid
		 * @returns {*}
		 */
		findByCid(cid) {
			cid = cid.replace(/^<+|>+$/g, '');
			return this.find(item => cid === item.contentId());
		}
	}

	class EmailCollectionModel extends AbstractCollectionModel
	{
		/**
		 * @param {?Array} json
		 * @returns {EmailCollectionModel}
		 */
		static reviveFromJson(items) {
			return super.reviveFromJson(items, email => EmailModel.reviveFromJson(email));
		}

		/**
		 * @param {boolean=} friendlyView = false
		 * @param {boolean=} wrapWithLink = false
		 * @returns {string}
		 */
		toString(friendlyView, wrapWithLink) {
			return this.map(email => email.toLine(friendlyView, wrapWithLink)).join(', ');
		}

		/**
		 * @returns {string}
		 */
		toStringClear() {
			const result = [];
			this.forEach(email => {
				if (email?.email && email?.name) {
					result.push(email.email);
				}
			});
			return result.join(', ');
		}
	}

	var PreviewHTML = "<html>\n<head>\n\t<meta charset=\"utf-8\">\n\t<title></title>\n\t<style>\nhtml, body {\n\tmargin: 0;\n\tpadding: 0;\n}\n\nheader {\n\tbackground: rgba(125,128,128,0.3);\n\tborder-bottom: 1px solid #888;\n}\n\nheader h1 {\n\tfont-size: 120%;\n}\n\nheader * {\n\tmargin: 5px 0;\n}\n\nheader time {\n\tfloat: right;\n}\n\nblockquote {\n\tborder-left: 2px solid rgba(125,128,128,0.5);\n\tmargin: 0;\n\tpadding: 0 0 0 10px;\n}\n\npre {\n\twhite-space: pre-wrap;\n\tword-wrap: break-word;\n\tword-break: normal;\n}\n\nbody > * {\n\tpadding: 0.5em 1em;\n}\n\t</style>\n</head>\n<body></body>\n</html>\n";

	const AppUserStore = {
		allowContacts: () => !!SettingsGet('ContactsIsAllowed')
	};

	addObservablesTo(AppUserStore, {
		focusedState: 'none',

		threadsAllowed: false
	});

	AppUserStore.focusedState.subscribe(value => {
		['FolderList','MessageList','MessageView'].forEach(name => {
			if (name === value) {
				keyScope(value);
				ThemeStore.isMobile() && leftPanelDisabled('FolderList' !== value);
			}
			let dom = elementById('V-Mail'+name);
			dom?.classList.toggle('focused', name === value);
		});
	});

	let iJsonErrorCount = 0;

	const getURL = (add = '') => serverRequest('Json') + add,

	checkResponseError = data => {
		const err = data ? data.ErrorCode : null;
		if (Notification$1.InvalidToken === err) {
			alert(getNotification(err));
			rl.logoutReload();
		} else if ([
				Notification$1.AuthError,
				Notification$1.ConnectionError,
				Notification$1.DomainNotAllowed,
				Notification$1.AccountNotAllowed,
				Notification$1.MailServerError,
				Notification$1.UnknownNotification,
				Notification$1.UnknownError
			].includes(err)
		) {
			if (7 < ++iJsonErrorCount) {
				rl.logoutReload();
			}
		}
	},

	oRequests = {},

	abort = (sAction, sReason, bClearOnly) => {
		if (oRequests[sAction]) {
			if (!bClearOnly && oRequests[sAction].abort) {
	//			oRequests[sAction].__aborted = true;
				oRequests[sAction].abort(sReason || 'AbortError');
			}

			oRequests[sAction] = null;
			delete oRequests[sAction];
		}
	},

	fetchJSON = (action, sGetAdd, params, timeout, jsonCallback) => {
		params = params || {};
		if (params instanceof FormData) {
			params.set('Action', action);
		} else {
			params.Action = action;
		}
	//	abort(action);
		const controller = new AbortController(),
			signal = controller.signal;
		oRequests[action] = controller;
		// Currently there is no way to combine multiple signals, so AbortSignal.timeout() not possible
		timeout && setTimeout(() => abort(action, 'TimeoutError'), timeout);
		return rl.fetchJSON(getURL(sGetAdd), {signal: signal}, sGetAdd ? null : params).then(jsonCallback).catch(err => {
			err.aborted = signal.aborted;
			err.reason = signal.reason;
			return Promise.reject(err);
		});
	};

	class FetchError extends Error
	{
		constructor(code, message) {
			super(message);
			this.code = code || Notification$1.JsonFalse;
		}
	}

	class AbstractFetchRemote
	{
		abort(sAction) {
			abort(sAction);
			return this;
		}

		/**
		 * Allows quicker visual responses to the user.
		 * Can be used to stream lines of json encoded data, but does not work on all servers.
		 * Apache needs 'flushpackets' like in <Proxy "fcgi://...." flushpackets=on></Proxy>
		 */
		streamPerLine(fCallback, sGetAdd, postData) {
			rl.fetch(getURL(sGetAdd), {}, postData)
			.then(response => response.body)
			.then(body => {
				let buffer = '';
				const
					// Firefox TextDecoderStream is not defined
	//				reader = body.pipeThrough(new TextDecoderStream()).getReader();
					reader = body.getReader(),
					re = /\r\n|\n|\r/gm,
					utf8decoder = new TextDecoder(),
					processText = ({ done, value }) => {
						buffer += value ? utf8decoder.decode(value, {stream: true}) : '';
						for (;;) {
							let result = re.exec(buffer);
							if (!result) {
								if (done) {
									break;
								}
								reader.read().then(processText);
								return;
							}
							fCallback(buffer.slice(0, result.index));
							buffer = buffer.slice(result.index + 1);
							re.lastIndex = 0;
						}
						if (buffer.length) {
							// last line didn't end in a newline char
							fCallback(buffer);
						}
					};
				reader.read().then(processText);
			});
		}

		/**
		 * @param {?Function} fCallback
		 * @param {string} sAction
		 * @param {Object=} oParameters
		 * @param {?number=} iTimeout
		 * @param {string=} sGetAdd = ''
		 */
		request(sAction, fCallback, params, iTimeout, sGetAdd, abortActions) {
			params = params || {};

			const start = Date.now();

			abortActions && console.error('abortActions is obsolete');

			fetchJSON(sAction, pString(sGetAdd),
				params,
				undefined === iTimeout ? 30000 : pInt(iTimeout),
				data => {
					let cached = false;
					if (data?.Time) {
						cached = pInt(data.Time) > Date.now() - start;
					}

					let iError = 0;
					if (sAction && oRequests[sAction]) {
						if (oRequests[sAction].__aborted) {
							iError = 2;
						}
						abort(sAction, 0, 1);
					}

					if (!iError && data) {
	/*
						if (sAction !== data.Action) {
							console.log(sAction + ' !== ' + data.Action);
						}
	*/
						if (data.Result) {
							iJsonErrorCount = 0;
						} else {
							checkResponseError(data);
							iError = data.ErrorCode || Notification$1.UnknownError;
						}
					}

					fCallback && fCallback(
						iError,
						data,
						cached,
						sAction,
						params
					);
				}
			)
			.catch(err => {
				console.error({fetchError:err});
				fCallback && fCallback(
					'TimeoutError' == err.reason ? 3 : (err.name == 'AbortError' ? 2 : 1),
					err
				);
			});
		}

		/**
		 * @param {?Function} fCallback
		 */
		getPublicKey(fCallback) {
			this.request('GetPublicKey', fCallback);
		}

		setTrigger(trigger, value) {
			if (trigger) {
				value = !!value;
				(isArray(trigger) ? trigger : [trigger]).forEach(fTrigger => {
					fTrigger?.(value);
				});
			}
		}

		post(action, fTrigger, params, timeOut) {
			this.setTrigger(fTrigger, true);
			return fetchJSON(action, '', params, pInt(timeOut, 30000),
				data => {
					abort(action, 0, 1);

					if (!data) {
						return Promise.reject(new FetchError(Notification$1.JsonParse));
					}
	/*
					let isCached = false, type = '';
					if (data?.Time) {
						isCached = pInt(data.Time) > microtime() - start;
					}
					// backward capability
					switch (true) {
						case 'success' === textStatus && data?.Result && action === data.Action:
							type = AbstractFetchRemote.SUCCESS;
							break;
						case 'abort' === textStatus && (!data || !data.__aborted__):
							type = AbstractFetchRemote.ABORT;
							break;
						default:
							type = AbstractFetchRemote.ERROR;
							break;
					}
	*/
					this.setTrigger(fTrigger, false);

					if (!data.Result || action !== data.Action) {
						checkResponseError(data);
						return Promise.reject(new FetchError(
							data ? data.ErrorCode : 0,
							data ? (data.ErrorMessageAdditional || data.ErrorMessage) : ''
						));
					}

					return data;
				}
			);
		}
	}

	Object.assign(AbstractFetchRemote.prototype, {
		SUCCESS : 0,
		ERROR : 1,
		ABORT : 2
	});

	class RemoteUserFetch extends AbstractFetchRemote {

		/**
		 * @param {Function} fCallback
		 * @param {object} params
		 * @param {boolean=} bSilent = false
		 */
		messageList(fCallback, params, bSilent = false) {
			const
				sFolderFullName = pString(params.Folder),
				folder = getFolderFromCacheList(sFolderFullName),
				folderHash = folder?.hash || '';

			params = Object.assign({
				Offset: 0,
				Limit: SettingsUserStore.messagesPerPage(),
				Search: '',
				UidNext: folder?.uidNext || 0, // Used to check for new messages
				Sort: FolderUserStore.sortMode(),
				Hash: folderHash + SettingsGet('AccountHash')
			}, params);
			params.Folder = sFolderFullName;
			if (AppUserStore.threadsAllowed() && SettingsUserStore.useThreads()) {
				params.UseThreads = 1;
			} else {
				params.ThreadUid = 0;
			}

			let sGetAdd = '';

			if (folderHash && (!params.Search || !params.Search.includes('is:'))) {
				sGetAdd = 'MessageList/' +
					SUB_QUERY_PREFIX +
					'/' +
					b64EncodeJSONSafe(params);
				params = {};
			}

			bSilent || this.abort('MessageList');
			this.request('MessageList',
				fCallback,
				params,
				60000, // 60 seconds before aborting
				sGetAdd
			);
		}

		/**
		 * @param {?Function} fCallback
		 * @param {string} sFolderFullName
		 * @param {number} iUid
		 * @returns {boolean}
		 */
		message(fCallback, sFolderFullName, iUid) {
			sFolderFullName = pString(sFolderFullName);
			iUid = pInt(iUid);

			if (getFolderFromCacheList(sFolderFullName) && 0 < iUid) {
				this.abort('Message').request('Message',
					fCallback,
					{},
					null,
					'Message/' +
						SUB_QUERY_PREFIX +
						'/' +
						b64EncodeJSONSafe([
							sFolderFullName,
							iUid,
							AppUserStore.threadsAllowed() && SettingsUserStore.useThreads() ? 1 : 0,
							SettingsGet('AccountHash')
						])
				);

				return true;
			}

			return false;
		}

		/**
		 * @param {?Function} fCallback
		 * @param {Object} oData
		 */
		saveSettings(fCallback, oData) {
			this.request('SettingsUpdate', fCallback, oData);
		}

		/**
		 * @param {string} key
		 * @param {?scalar} value
		 * @param {?Function} fCallback
		 */
		saveSetting(key, value, fCallback) {
			this.saveSettings(fCallback, {
				[key]: value
			});
		}

	/*
		folderMove(sPrevFolderFullName, sNewFolderFullName, bSubscribe) {
			return this.post('FolderMove', FolderUserStore.foldersRenaming, {
				Folder: sPrevFolderFullName,
				NewFolder: sNewFolderFullName,
				Subscribe: bSubscribe ? 1 : 0
			});
		}
	*/
	}

	var Remote = new RemoteUserFetch();

	const
		url = /(^|\s|\n|\/?>)(https?:\/\/[-A-Z0-9+&#/%?=()~_|!:,.;]*[-A-Z0-9+&#/%=~()_|])/gi,
		// eslint-disable-next-line max-len
		email$1 = /(^|\s|\n|\/?>)((?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|"(?:[\x21\x23-\x5b\x5d-\x7f]|\\[\x21\x23-\x5b\x5d-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9]))\.){3}(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9])|[a-z0-9-]*[a-z0-9]:(?:[\x21-\x5a\x53-\x7f]|\\[\x21\x23-\x5b\x5d-\x7f])+)\]))/gi,
		// rfc3966
		tel = /(^|\s|\n|\/?>)(tel:(\+[0-9().-]+|[0-9*#().-]+(;phone-context=\+[0-9+().-]+)?))/g,

		hcont = Element.fromHTML('<div area="hidden" style="position:absolute;left:-5000px"></div>'),
		getRealHeight = el => {
			hcont.innerHTML = el.outerHTML;
			const result = hcont.clientHeight;
			hcont.innerHTML = '';
			return result;
		},

		ignoredTags = [
			// rfc5788
			'$forwarded',
			'$mdnsent',
			'$submitpending',
			'$submitted',
			// rfc9051
			'$junk',
			'$notjunk',
			'$phishing',
			// Mailo
			'sent',
			// KMail
			'$attachment',
			'$encrypted',
			'$error',
			'$ignored',
			'$invitation',
			'$queued',
			'$replied',
			'$sent',
			'$signed',
			'$todo',
			'$watched',
			// GMail
			'$replied',
			'$attachment',
			'$notphishing',
			'junk',
			'nonjunk',
			// Others
			'$readreceipt'
		],

		toggleTag = (message, keyword) => {
			const lower = keyword.toLowerCase(),
				isSet = message.flags().includes(lower);
			Remote.request('MessageSetKeyword', iError => {
				if (!iError) {
					if (isSet) {
						message.flags.remove(lower);
					} else {
						message.flags.push(lower);
					}
	//				MessageFlagsCache.setFor(message.folder, message.uid, message.flags());
				}
			}, {
				Folder: message.folder,
				Uids: message.uid,
				Keyword: keyword,
				SetAction: isSet ? 0 : 1
			});
		},

		/**
		 * @param {EmailCollectionModel} emails
		 * @param {Object} unic
		 * @param {Map} localEmails
		 */
		replyHelper = (emails, unic, localEmails) =>
			emails.forEach(email => {
				if (!unic[email.email] && !localEmails.has(email.email)) {
					localEmails.set(email.email, email);
				}
			});

	doc.body.append(hcont);

	class MessageModel extends AbstractModel {
		constructor() {
			super();

			this._reset();

			this.addObservables({
				subject: '',
				plain: '',
				html: '',
				size: 0,
				spamScore: 0,
				spamResult: '',
				isSpam: false,
				hasVirus: null, // or boolean when scanned
				dateTimeStampInUTC: 0,
				priority: MessagePriority.Normal,

				senderEmailsString: '',
				senderClearEmailsString: '',

				deleted: false,

				// Also used by Selector
				focused: false,
				selected: false,
				checked: false,

				isHtml: false,
				hasImages: false,
				hasExternals: false,

				pgpSigned: null,
				pgpVerified: null,

				pgpEncrypted: null,
				pgpDecrypted: false,

				readReceipt: '',

				hasUnseenSubMessage: false,
				hasFlaggedSubMessage: false
			});

			this.attachments = ko.observableArray(new AttachmentCollectionModel);
			this.threads = ko.observableArray();
			this.unsubsribeLinks = ko.observableArray();
			this.flags = ko.observableArray();

			this.addComputables({
				attachmentIconClass: () => FileInfo.getAttachmentsIconClass(this.attachments()),
				threadsLen: () => this.threads().length,
				hasAttachments: () => this.attachments().hasVisible(),

				isUnseen: () => !this.flags().includes('\\seen'),
				isFlagged: () => this.flags().includes('\\flagged'),
				isReadReceipt: () => this.flags().includes('$mdnsent'),
	//			isJunk: () => this.flags().includes('$junk') && !this.flags().includes('$nonjunk'),
	//			isPhishing: () => this.flags().includes('$phishing'),

				tagsToHTML: () => this.flags().map(value =>
						('\\' == value[0] || ignoredTags.includes(value))
						? ''
						: '<span class="focused msgflag-'+value+'">' + i18n('MESSAGE_TAGS/'+value,0,value) + '</span>'
					).join(' '),

				tagOptions: () => {
					const tagOptions = [];
					FolderUserStore.currentFolder().permanentFlags.forEach(value => {
						let lower = value.toLowerCase();
						if ('\\' != value[0] && !ignoredTags.includes(lower)) {
							tagOptions.push({
								css: 'msgflag-' + lower,
								value: value,
								checked: this.flags().includes(lower),
								label: i18n('MESSAGE_TAGS/'+lower, 0, lower),
								toggle: (/*obj*/) => toggleTag(this, value)
							});
						}
					});
					return tagOptions
				}

			});
		}

		_reset() {
			this.folder = '';
			this.uid = 0;
			this.hash = '';
			this.requestHash = '';
			this.emails = [];
			this.from = new EmailCollectionModel;
			this.to = new EmailCollectionModel;
			this.cc = new EmailCollectionModel;
			this.bcc = new EmailCollectionModel;
			this.replyTo = new EmailCollectionModel;
			this.deliveredTo = new EmailCollectionModel;
			this.body = null;
			this.draftInfo = [];
			this.messageId = '';
			this.inReplyTo = '';
			this.references = '';
		}

		clear() {
			this._reset();
			this.subject('');
			this.html('');
			this.plain('');
			this.size(0);
			this.spamScore(0);
			this.spamResult('');
			this.isSpam(false);
			this.hasVirus(null);
			this.dateTimeStampInUTC(0);
			this.priority(MessagePriority.Normal);

			this.senderEmailsString('');
			this.senderClearEmailsString('');

			this.deleted(false);

			this.selected(false);
			this.checked(false);

			this.isHtml(false);
			this.hasImages(false);
			this.hasExternals(false);
			this.attachments(new AttachmentCollectionModel);

			this.pgpSigned(null);
			this.pgpVerified(null);

			this.pgpEncrypted(null);
			this.pgpDecrypted(false);

			this.priority(MessagePriority.Normal);
			this.readReceipt('');

			this.threads([]);
			this.unsubsribeLinks([]);

			this.hasUnseenSubMessage(false);
			this.hasFlaggedSubMessage(false);
		}

		spamStatus() {
			let spam = this.spamResult();
			return spam ? i18n(this.isSpam() ? 'GLOBAL/SPAM' : 'GLOBAL/NOT_SPAM') + ': ' + spam : '';
		}

		/**
		 * @param {Array} properties
		 * @returns {Array}
		 */
		getEmails(properties) {
			return properties.reduce((carry, property) => carry.concat(this[property]), []).map(
				oItem => oItem ? oItem.email : ''
			).validUnique();
		}

		/**
		 * @returns {string}
		 */
		friendlySize() {
			return FileInfo.friendlySize(this.size());
		}

		computeSenderEmail() {
			const list = [FolderUserStore.sentFolder(), FolderUserStore.draftsFolder()].includes(this.folder) ? 'to' : 'from';
			this.senderEmailsString(this[list].toString(true));
			this.senderClearEmailsString(this[list].toStringClear());
		}

		/**
		 * @param {FetchJsonMessage} json
		 * @returns {boolean}
		 */
		revivePropertiesFromJson(json) {
			if ('Priority' in json && ![MessagePriority.High, MessagePriority.Low].includes(json.Priority)) {
				json.Priority = MessagePriority.Normal;
			}
			if (super.revivePropertiesFromJson(json)) {
	//			this.foundCIDs = isArray(json.FoundCIDs) ? json.FoundCIDs : [];
	//			this.attachments(AttachmentCollectionModel.reviveFromJson(json.Attachments, this.foundCIDs));

				this.computeSenderEmail();
			}
		}

		/**
		 * @returns {boolean}
		 */
		hasUnsubsribeLinks() {
			return this.unsubsribeLinks().length;
		}

		/**
		 * @returns {string}
		 */
		getFirstUnsubsribeLink() {
			return this.unsubsribeLinks()[0] || '';
		}

		/**
		 * @param {boolean} friendlyView
		 * @param {boolean=} wrapWithLink
		 * @returns {string}
		 */
		fromToLine(friendlyView, wrapWithLink) {
			return this.from.toString(friendlyView, wrapWithLink);
		}

		/**
		 * @returns {string}
		 */
		fromDkimData() {
			let result = ['none', ''];
			if (1 === arrayLength(this.from) && this.from[0]?.dkimStatus) {
				result = [this.from[0].dkimStatus, this.from[0].dkimValue || ''];
			}

			return result;
		}

		/**
		 * @param {boolean} friendlyView
		 * @param {boolean=} wrapWithLink
		 * @returns {string}
		 */
		toToLine(friendlyView, wrapWithLink) {
			return this.to.toString(friendlyView, wrapWithLink);
		}

		/**
		 * @param {boolean} friendlyView
		 * @param {boolean=} wrapWithLink
		 * @returns {string}
		 */
		ccToLine(friendlyView, wrapWithLink) {
			return this.cc.toString(friendlyView, wrapWithLink);
		}

		/**
		 * @returns {string}
		 */
		bccToLine() {
			return this.bcc.toString();
		}

		/**
		 * @returns {string}
		 */
		replyToToLine() {
			return this.replyTo.toString();
		}

		/**
		 * @return string
		 */
		lineAsCss(flags=1) {
			let classes = [];
			forEachObjectEntry({
				deleted: this.deleted(),
				selected: this.selected(),
				checked: this.checked(),
				unseen: this.isUnseen(),
				focused: this.focused(),
				priorityHigh: this.priority() === MessagePriority.High,
				withAttachments: !!this.attachments().length,
				// hasChildrenMessage: 1 < this.threadsLen(),
				hasUnseenSubMessage: this.hasUnseenSubMessage(),
				hasFlaggedSubMessage: this.hasFlaggedSubMessage()
			}, (key, value) => value && classes.push(key));
			flags && this.flags().forEach(value => classes.push('msgflag-'+value));
			return classes.join(' ');
		}

		/**
		 * @return array
		 * https://datatracker.ietf.org/doc/html/rfc5788
		 */
		keywords() {
			return this.flags().filter(value => '\\' !== value[0]);
		}

		/**
		 * @returns {string}
		 */
		fromAsSingleEmail() {
			return (isArray(this.from) && this.from[0]?.email) || '';
		}

		/**
		 * @returns {string}
		 */
		viewLink() {
			return serverRequestRaw('ViewAsPlain', this.requestHash);
		}

		/**
		 * @returns {string}
		 */
		downloadLink() {
			return serverRequestRaw('Download', this.requestHash);
		}

		/**
		 * @param {Object} excludeEmails
		 * @returns {Array}
		 */
		replyEmails(excludeEmails) {
			const
				result = new Map(),
				unic = excludeEmails || {};
			replyHelper(this.replyTo, unic, result);
			result.size || replyHelper(this.from, unic, result);
			return result.size ? [...result.values()] : [this.to[0]];
		}

		/**
		 * @param {Object} excludeEmails
		 * @returns {Array.<Array>}
		 */
		replyAllEmails(excludeEmails) {
			const
				toResult = new Map(),
				ccResult = new Map(),
				unic = excludeEmails || {};

			replyHelper(this.replyTo, unic, toResult);
			toResult.size || replyHelper(this.from, unic, toResult);

			replyHelper(this.to, unic, toResult);

			replyHelper(this.cc, unic, ccResult);

			return [[...toResult.values()], [...ccResult.values()]];
		}

		viewHtml() {
			const body = this.body;
			if (body && this.html()) {
				let result = cleanHtml(this.html(), this.attachments());
				this.hasExternals(result.hasExternals);
				this.hasImages(body.rlHasImages = !!result.hasExternals);

				body.innerHTML = result.html;

				body.classList.toggle('html', 1);
				body.classList.toggle('plain', 0);

				if (SettingsUserStore.showImages()) {
					this.showExternalImages();
				}

				this.isHtml(true);
				this.initView();
				return true;
			}
		}

		viewPlain() {
			const body = this.body;
			if (body && this.plain()) {
				body.classList.toggle('html', 0);
				body.classList.toggle('plain', 1);
				body.innerHTML = plainToHtml(
					this.plain()
						.replace(/-----BEGIN PGP (SIGNED MESSAGE-----(\r?\n[a-z][^\r\n]+)+|SIGNATURE-----[\s\S]*)/, '')
						.trim()
				)
	//				.replace(url, '$1<a href="$2" target="_blank" rel="noreferrer noopener">$2</a>')
					.replace(url, '$1<a href="$2" target="_blank">$2</a>')
					.replace(email$1, '$1<a href="mailto:$2">$2</a>')
					.replace(tel, '$1<a href="$2">$2</a>');
				this.isHtml(false);
				this.hasImages(false);
				this.initView();
				return true;
			}
		}

		initView() {
			// init BlockquoteSwitcher
			this.body.querySelectorAll('blockquote').forEach(node => {
				if (node.textContent.trim()) {
					let h = node.clientHeight || getRealHeight(node);
					if (0 === h || 100 < h) {
						const el = Element.fromHTML('<details class="sm-bq-switcher"><summary>•••</summary></details>');
						node.replaceWith(el);
						el.append(node);
					}
				}
			});
		}

		viewPopupMessage(print) {
			const timeStampInUTC = this.dateTimeStampInUTC() || 0,
				ccLine = this.ccToLine(),
				m = 0 < timeStampInUTC ? new Date(timeStampInUTC * 1000) : null,
				win = open(''),
				sdoc = win.document;
			let subject = encodeHtml(this.subject()),
				mode = this.isHtml() ? 'div' : 'pre',
				cc = ccLine ? `<div>${encodeHtml(i18n('GLOBAL/CC'))}: ${encodeHtml(ccLine)}</div>` : '',
				style = getComputedStyle(doc.querySelector('.messageView')),
				prop = property => style.getPropertyValue(property);
			sdoc.write(PreviewHTML
				.replace('<title>', '<title>'+subject)
				// eslint-disable-next-line max-len
				.replace('<body>', `<body style="background-color:${prop('background-color')};color:${prop('color')}"><header><h1>${subject}</h1><time>${encodeHtml(m ? m.format('LLL') : '')}</time><div>${encodeHtml(this.fromToLine())}</div><div>${encodeHtml(i18n('GLOBAL/TO'))}: ${encodeHtml(this.toToLine())}</div>${cc}</header><${mode}>${this.bodyAsHTML()}</${mode}>`)
			);
			sdoc.close();

			if (print) {
				setTimeout(() => win.print(), 100);
			}
		}

		/**
		 * @param {boolean=} print = false
		 */
		popupMessage() {
			this.viewPopupMessage(false);
		}

		printMessage() {
			this.viewPopupMessage(true);
		}

		/**
		 * @returns {string}
		 */
		generateUid() {
			return this.folder + '/' + this.uid;
		}

		/**
		 * @param {MessageModel} message
		 * @returns {MessageModel}
		 */
		static fromMessageListItem(message) {
			let self = new MessageModel();

			if (message) {
				self.folder = message.folder;
				self.uid = message.uid;
				self.hash = message.hash;
				self.requestHash = message.requestHash;
				self.subject(message.subject());
				self.plain(message.plain());
				self.html(message.html());

				self.size(message.size());
				self.spamScore(message.spamScore());
				self.spamResult(message.spamResult());
				self.isSpam(message.isSpam());
				self.hasVirus(message.hasVirus());
				self.dateTimeStampInUTC(message.dateTimeStampInUTC());
				self.priority(message.priority());

				self.hasExternals(message.hasExternals());

				self.emails = message.emails;

				self.from = message.from;
				self.to = message.to;
				self.cc = message.cc;
				self.bcc = message.bcc;
				self.replyTo = message.replyTo;
				self.deliveredTo = message.deliveredTo;
				self.unsubsribeLinks(message.unsubsribeLinks);

				self.flags(message.flags());

				self.priority(message.priority());

				self.selected(message.selected());
				self.checked(message.checked());
				self.attachments(message.attachments());

				self.threads(message.threads());
			}

			self.computeSenderEmail();

			return self;
		}

		showExternalImages() {
			const body = this.body;
			if (body && this.hasImages()) {
				this.hasImages(false);
				body.rlHasImages = false;

				let attr = 'data-x-src',
					src, useProxy = !!SettingsGet('UseLocalProxyForExternalImages');
				body.querySelectorAll('img[' + attr + ']').forEach(node => {
					src = node.getAttribute(attr);
					node.src = useProxy ? proxy(src) : src;
				});

				body.querySelectorAll('[data-x-style-url]').forEach(node => {
					JSON.parse(node.dataset.xStyleUrl).forEach(data =>
						node.style[data[0]] = "url('" + (useProxy ? proxy(data[1]) : data[1]) + "')"
					);
				});
			}
		}

		/**
		 * @returns {string}
		 */
		bodyAsHTML() {
			if (this.body) {
				let clone = this.body.cloneNode(true);
				clone.querySelectorAll('.sm-bq-switcher').forEach(
					node => node.replaceWith(node.lastElementChild)
				);
				return clone.innerHTML;
			}
			let result = cleanHtml(this.html(), this.attachments());
			return result.html || plainToHtml(this.plain());
		}

	}

	// Fullscreen must be on app, else other popups fail
	const
		app = elementById('rl-app'),
		appFullscreen = () => (doc.fullscreenElement || doc.webkitFullscreenElement) === app,
		exitFullscreen = () => appFullscreen() && (doc.exitFullscreen || doc.webkitExitFullscreen).call(doc),
		isFullscreen = ko.observable(false),
		toggleFullscreen = () => isFullscreen() ? exitFullscreen() : app.requestFullscreen();

	if (app) {
		let event = 'fullscreenchange';
		if (!app.requestFullscreen && app.webkitRequestFullscreen) {
			app.requestFullscreen = app.webkitRequestFullscreen;
			event = 'webkit'+event;
		}
		if (app.requestFullscreen) {
			doc.addEventListener(event, () => {
				isFullscreen(appFullscreen());
				$htmlCL.toggle('rl-fullscreen', appFullscreen());
			});
		}
	}

	const MessageUserStore = new class {
		constructor() {
			addObservablesTo(this, {
				// message viewer
				message: null,
				error: '',
				loading: false,

				// Cache mail bodies
				bodiesDom: null,
				activeDom: null
			});

			// Subscribers

			addSubscribablesTo(this, {
				message: message => {
					clearTimeout(this.MessageSeenTimer);
					elementById('rl-right').classList.toggle('message-selected', !!message);
					if (message) {
						if (!SettingsUserStore.usePreviewPane()) {
							AppUserStore.focusedState(Scope.MessageView);
						}
					} else {
						AppUserStore.focusedState(Scope.MessageList);
						exitFullscreen();
						this.hideMessageBodies();
					}
				},
			});

			this.purgeMessageBodyCache = this.purgeMessageBodyCache.throttle(30000);
		}

		purgeMessageBodyCache() {
			const messagesDom = this.bodiesDom(),
				children = messagesDom?.children;
			if (children) {
				while (15 < children.length) {
					children[0].remove();
				}
			}
		}

		hideMessageBodies() {
			const messagesDom = this.bodiesDom();
			messagesDom && Array.from(messagesDom.children).forEach(el => el.hidden = true);
		}
	};

	class MessageCollectionModel extends AbstractCollectionModel
	{
	/*
		constructor() {
			super();
			this.Filtered
			this.Folder
			this.FolderHash
			this.FolderInfo
			this.totalEmails
			this.unreadEmails
			this.MessageResultCount
			this.UidNext
			this.ThreadUid
			this.NewMessages
			this.Offset
			this.Limit
			this.Search
		}
	*/

		/**
		 * @param {?Object} json
		 * @returns {MessageCollectionModel}
		 */
		static reviveFromJson(object, cached) {
			return super.reviveFromJson(object, message => {
				message = MessageModel.reviveFromJson(message);
				if (message) {
					message.deleted(false);
					cached ? MessageFlagsCache.initMessage(message) : MessageFlagsCache.store(message);
					return message;
				}
			});
		}
	}

	const AccountUserStore = {
		accounts: koArrayWithDestroy(),
		loading: ko.observable(false).extend({ debounce: 100 }),

		getEmailAddresses: () => AccountUserStore.accounts.map(item => item.email)
	};

	addObservablesTo(AccountUserStore, {
		email: '',
		signature: ''
	});

	/**
	 * Might not work due to the new ServiceWorkerRegistration.showNotification
	 */
	const HTML5Notification = window.Notification,
		HTML5NotificationStatus = () => HTML5Notification?.permission || 'denied',
		NotificationsDenied = () => 'denied' === HTML5NotificationStatus(),
		NotificationsGranted = () => 'granted' === HTML5NotificationStatus(),
		dispatchMessage = data => {
			focus();
			if (data.Folder && data.Uid) {
				fireEvent('mailbox.message.show', data);
			} else if (data.Url) {
				hasher.setHash(data.Url);
			}
		};

	let DesktopNotifications = false,
		WorkerNotifications = navigator.serviceWorker;

	// Are Notifications supported in the service worker?
	if (WorkerNotifications && ServiceWorkerRegistration && ServiceWorkerRegistration.prototype.showNotification) {
		/* Listen for close requests from the ServiceWorker */
		WorkerNotifications.addEventListener('message', event => {
			const obj = JSON.parse(event.data);
			'notificationclick' === obj?.action && dispatchMessage(obj.data);
		});
	} else {
		WorkerNotifications = null;
		console.log('ServiceWorker Notifications not supported');
	}

	const NotificationUserStore = new class {
		constructor() {
			addObservablesTo(this, {
				enabled: false,/*.extend({ notify: 'always' })*/
				allowed: !NotificationsDenied()
			});

			this.enabled.subscribe(value => {
				DesktopNotifications = !!value;
				if (value && HTML5Notification && !NotificationsGranted()) {
					HTML5Notification.requestPermission(() =>
						this.allowed(!NotificationsDenied())
					);
				}
			});
		}

		/**
		 * Used with DesktopNotifications setting
		 */
		display(title, text, messageData, imageSrc) {
			if (DesktopNotifications && NotificationsGranted()) {
				const options = {
					body: text,
					icon: imageSrc || staticLink('css/images/icon-message-notification.png'),
					data: messageData
				};
				if (messageData?.Uid) {
					options.tag = messageData.Uid;
				}
				if (WorkerNotifications) {
					// Service-Worker-Allowed HTTP header to allow the scope.
					WorkerNotifications.register('/serviceworker.js')
	//				WorkerNotifications.register(Links.staticLink('js/serviceworker.js'), {scope:'/'})
					.then(() =>
						WorkerNotifications.ready.then(registration =>
							/* Show the notification */
							registration
								.showNotification(title, options)
								.then(() =>
									registration.getNotifications().then((/*notifications*/) => {
										/* Send an empty message so the Worker knows who the client is */
										registration.active.postMessage('');
									})
								)
						)
					)
					.catch(e => console.error(e));
				} else {
					const notification = new HTML5Notification(title, options);
					notification.show?.();
					notification.onclick = messageData ? () => dispatchMessage(messageData) : null;
					setTimeout(() => notification.close(), 7000);
				}
			}
		}
	};

	const
		isChecked = item => item.checked(),
		replaceHash = hash => {
			rl.route.off();
			hasher.replaceHash(hash);
			rl.route.on();
		},
		disableAutoSelect = ko.observable(false).extend({ falseTimeout: 500 });

	const MessagelistUserStore = ko.observableArray().extend({ debounce: 0 });

	addObservablesTo(MessagelistUserStore, {
		count: 0,
		listSearch: '',
		threadUid: 0,
		page: 1,
		pageBeforeThread: 1,
		error: '',
	//	Folder: '',

		endHash: '',
		endThreadUid: 0,

		loading: false,
		// Happens when message(s) removed from list
		isIncomplete: false,

		selectedMessage: null,
		focusedMessage: null
	});

	// Computed Observables

	addComputablesTo(MessagelistUserStore, {
		isLoading: () => {
			const value = MessagelistUserStore.loading() | MessagelistUserStore.isIncomplete();
			$htmlCL.toggle('list-loading', value);
			return value;
		},

		isArchiveFolder: () => FolderUserStore.archiveFolder() === MessagelistUserStore().Folder,

		isDraftFolder: () => FolderUserStore.draftsFolder() === MessagelistUserStore().Folder,

		isSentFolder: () => FolderUserStore.sentFolder() === MessagelistUserStore().Folder,

		isSpamFolder: () => FolderUserStore.spamFolder() === MessagelistUserStore().Folder,

		isTrashFolder: () => FolderUserStore.trashFolder() === MessagelistUserStore().Folder,

		archiveAllowed: () => ![UNUSED_OPTION_VALUE, MessagelistUserStore().Folder].includes(FolderUserStore.archiveFolder())
			&& !MessagelistUserStore.isDraftFolder(),

		canMarkAsSpam: () => !(UNUSED_OPTION_VALUE === FolderUserStore.spamFolder()
	//		| MessagelistUserStore.isArchiveFolder()
			| MessagelistUserStore.isSentFolder()
			| MessagelistUserStore.isDraftFolder()
			| MessagelistUserStore.isSpamFolder()),

		pageCount: () => Math.max(1, Math.ceil(MessagelistUserStore.count() / SettingsUserStore.messagesPerPage())),

		mainSearch: {
			read: MessagelistUserStore.listSearch,
			write: value => hasher.setHash(
				mailBox(FolderUserStore.currentFolderFullNameHash(), 1,
					value.toString().trim(), MessagelistUserStore.threadUid())
			)
		},

		listCheckedOrSelected: () => {
			const
				selectedMessage = MessagelistUserStore.selectedMessage(),
				focusedMessage = MessagelistUserStore.focusedMessage(),
				checked = MessagelistUserStore.filter(item => isChecked(item));
			return checked.length ? checked : (selectedMessage || focusedMessage ? [selectedMessage || focusedMessage] : []);
		},

		listCheckedOrSelectedUidsWithSubMails: () => {
			let result = new Set;
			MessagelistUserStore.listCheckedOrSelected().forEach(message => {
				result.add(message.uid);
				if (1 < message.threadsLen()) {
					message.threads().forEach(result.add, result);
				}
			});
			return result;
		}
	});

	MessagelistUserStore.listChecked = koComputable(
		() => MessagelistUserStore.filter(isChecked)
	).extend({ rateLimit: 0 });

	// Also used by Selector
	MessagelistUserStore.hasChecked = koComputable(
		// Issue: not all are observed?
		() => !!MessagelistUserStore.find(isChecked)
	).extend({ rateLimit: 0 });

	MessagelistUserStore.hasCheckedOrSelected = koComputable(() =>
		!!MessagelistUserStore.selectedMessage()
		| !!MessagelistUserStore.focusedMessage()
		// Issue: not all are observed?
		| !!MessagelistUserStore.find(isChecked)
	).extend({ rateLimit: 50 });

	MessagelistUserStore.notifyNewMessages = (folder, newMessages) => {
		if (getFolderInboxName() === folder && arrayLength(newMessages)) {

			SMAudio.playNotification();

			const len = newMessages.length;
			if (3 < len) {
				NotificationUserStore.display(
					AccountUserStore.email(),
					i18n('MESSAGE_LIST/NEW_MESSAGE_NOTIFICATION', {
						COUNT: len
					}),
					{ Url: mailBox(newMessages[0].Folder) }
				);
			} else {
				newMessages.forEach(item => {
					NotificationUserStore.display(
						EmailCollectionModel.reviveFromJson(item.From).toString(),
						item.subject,
						{ Folder: item.Folder, Uid: item.Uid }
					);
				});
			}
		}
	};

	MessagelistUserStore.canAutoSelect = () =>
		!/is:unseen/.test(MessagelistUserStore.mainSearch())
		&& !disableAutoSelect()
		&& SettingsUserStore.usePreviewPane();

	/**
	 * @param {boolean=} bDropPagePosition = false
	 * @param {boolean=} bDropCurrentFolderCache = false
	 */
	MessagelistUserStore.reload = (bDropPagePosition = false, bDropCurrentFolderCache = false) => {
		let iOffset = (MessagelistUserStore.page() - 1) * SettingsUserStore.messagesPerPage();

		if (bDropCurrentFolderCache) {
			setFolderHash(FolderUserStore.currentFolderFullName(), '');
		}

		if (bDropPagePosition) {
			MessagelistUserStore.page(1);
			MessagelistUserStore.pageBeforeThread(1);
			iOffset = 0;

			replaceHash(
				mailBox(
					FolderUserStore.currentFolderFullNameHash(),
					MessagelistUserStore.page(),
					MessagelistUserStore.listSearch(),
					MessagelistUserStore.threadUid()
				)
			);
		}

		MessagelistUserStore.loading(true);
		Remote.messageList(
			(iError, oData, bCached) => {
				let error = '';
				if (iError) {
					error = getNotification(iError);
					if (Notification$1.RequestAborted !== iError) {
						MessagelistUserStore([]);
					}
				} else {
					const collection = MessageCollectionModel.reviveFromJson(oData.Result, bCached);
					if (collection) {
						error = '';
						let unreadCountChange = false;

						const
							folder = getFolderFromCacheList(collection.Folder),
							folderInfo = collection.FolderInfo;
						if (folder && !bCached) {
	//						folder.revivePropertiesFromJson(result);
							folder.expires = Date.now();
							folder.uidNext = folderInfo.UidNext;
							folder.hash = collection.FolderHash;

							if (null != folderInfo.totalEmails) {
								folder.totalEmails(folderInfo.totalEmails);
							}

							if (null != folderInfo.unreadEmails) {
								if (pInt(folder.unreadEmails()) !== pInt(folderInfo.unreadEmails)) {
									unreadCountChange = true;
									MessageFlagsCache.clearFolder(folder.fullName);
								}
								folder.unreadEmails(folderInfo.unreadEmails);
							}

							folder.flags(folderInfo.Flags);
							let flags = folderInfo.PermanentFlags;
							if (flags.includes('\\*')) {
								let i = 6;
								while (--i) {
									flags.includes('$label'+i) || flags.push('$label'+i);
								}
							}
							flags.sort((a, b) => {
								a = a.toUpperCase();
								b = b.toUpperCase();
								return (a < b) ? -1 : ((a > b) ? 1 : 0);
							});
							folder.permanentFlags(flags);

							MessagelistUserStore.notifyNewMessages(folder.fullName, collection.NewMessages);
						}

						MessagelistUserStore.count(collection.MessageResultCount);
						MessagelistUserStore.listSearch(pString(collection.Search));
						MessagelistUserStore.page(Math.ceil(collection.Offset / SettingsUserStore.messagesPerPage() + 1));
						MessagelistUserStore.threadUid(collection.ThreadUid);

						MessagelistUserStore.endHash(
							collection.Folder +
							'|' + collection.Search +
							'|' + MessagelistUserStore.threadUid() +
							'|' + MessagelistUserStore.page()
						);
						MessagelistUserStore.endThreadUid(collection.ThreadUid);
						const message = MessageUserStore.message();
						if (message && collection.Folder !== message.folder) {
							MessageUserStore.message(null);
						}

						disableAutoSelect(true);

						MessagelistUserStore(collection);
						MessagelistUserStore.isIncomplete(false);

						if (folder && (bCached || unreadCountChange || SettingsUserStore.useThreads())) {
							rl.app.folderInformation(folder.fullName, collection);
						}
					} else {
						MessagelistUserStore.count(0);
						MessagelistUserStore([]);
						error = getNotification(Notification$1.CantGetMessageList);
					}
					MessagelistUserStore.error(error);
				}
				MessagelistUserStore.loading(false);
			},
			{
				Folder: FolderUserStore.currentFolderFullName(),
				Offset: iOffset,
				Limit: SettingsUserStore.messagesPerPage(),
				Search: MessagelistUserStore.listSearch(),
				ThreadUid: MessagelistUserStore.threadUid()
			}
		);
	};

	/**
	 * @param {string} sFolderFullName
	 * @param {number} iSetAction
	 * @param {Array=} messages = null
	 */
	MessagelistUserStore.setAction = (sFolderFullName, iSetAction, messages) => {
		messages = messages || MessagelistUserStore.listChecked();

		let folder,
			alreadyUnread = 0,
			rootUids = messages.map(oMessage => oMessage?.uid).validUnique(),
			length = rootUids.length;

		if (sFolderFullName && length) {
			rootUids.forEach(sSubUid =>
				alreadyUnread += MessageFlagsCache.storeBySetAction(sFolderFullName, sSubUid, iSetAction)
			);
			switch (iSetAction) {
				case MessageSetAction.SetSeen:
					length = 0;
					// fallthrough is intentionally
				case MessageSetAction.UnsetSeen:
					folder = getFolderFromCacheList(sFolderFullName);
					if (folder) {
						folder.unreadEmails(folder.unreadEmails() - alreadyUnread + length);
					}
					Remote.request('MessageSetSeen', null, {
						Folder: sFolderFullName,
						Uids: rootUids.join(','),
						SetAction: iSetAction == MessageSetAction.SetSeen ? 1 : 0
					});
					break;

				case MessageSetAction.SetFlag:
				case MessageSetAction.UnsetFlag:
					Remote.request('MessageSetFlagged', null, {
						Folder: sFolderFullName,
						Uids: rootUids.join(','),
						SetAction: iSetAction == MessageSetAction.SetFlag ? 1 : 0
					});
					break;
				// no default
			}

			MessagelistUserStore.reloadFlagsAndCachedMessage();
		}
	};

	/**
	 * @param {string} fromFolderFullName
	 * @param {Set} oUids
	 * @param {string=} toFolderFullName = ''
	 * @param {boolean=} copy = false
	 */
	MessagelistUserStore.removeMessagesFromList = (
		fromFolderFullName, oUids, toFolderFullName = '', copy = false
	) => {
		let unseenCount = 0,
			messageList = MessagelistUserStore,
			currentMessage = MessageUserStore.message();

		const trashFolder = FolderUserStore.trashFolder(),
			spamFolder = FolderUserStore.spamFolder(),
			fromFolder = getFolderFromCacheList(fromFolderFullName),
			toFolder = toFolderFullName ? getFolderFromCacheList(toFolderFullName) : null,
			messages =
				FolderUserStore.currentFolderFullName() === fromFolderFullName
					? messageList.filter(item => item && oUids.has(item.uid))
					: [];

		messages.forEach(item => item?.isUnseen() && ++unseenCount);

		if (fromFolder) {
			fromFolder.hash = '';
			if (!copy) {
				fromFolder.totalEmails(
					0 <= fromFolder.totalEmails() - oUids.size ? fromFolder.totalEmails() - oUids.size : 0
				);

				if (0 < unseenCount) {
					fromFolder.unreadEmails(
						0 <= fromFolder.unreadEmails() - unseenCount ? fromFolder.unreadEmails() - unseenCount : 0
					);
				}
			}
		}

		if (toFolder) {
			toFolder.hash = '';

			if (trashFolder === toFolder.fullName || spamFolder === toFolder.fullName) {
				unseenCount = 0;
			}

			toFolder.totalEmails(toFolder.totalEmails() + oUids.size);
			if (0 < unseenCount) {
				toFolder.unreadEmails(toFolder.unreadEmails() + unseenCount);
			}

			toFolder.actionBlink(true);
		}

		if (messages.length) {
			if (copy) {
				messages.forEach(item => item.checked(false));
			} else {
				MessagelistUserStore.isIncomplete(true);

				messages.forEach(item => {
					if (currentMessage && currentMessage.hash === item.hash) {
						currentMessage = null;
						MessageUserStore.message(null);
					}

					item.deleted(true);
				});

				setTimeout(() => messages.forEach(item => messageList.remove(item)), 350);
			}
		}

		if (MessagelistUserStore.threadUid()) {
			if (
				messageList.length &&
				messageList.find(item => item?.deleted() && item.uid == MessagelistUserStore.threadUid())
			) {
				const message = messageList.find(item => item && !item.deleted());
				let setHash;
				if (!message) {
					if (1 < MessagelistUserStore.page()) {
						MessagelistUserStore.page(MessagelistUserStore.page() - 1);
						setHash = 1;
					} else {
						MessagelistUserStore.threadUid(0);
						replaceHash(
							mailBox(
								FolderUserStore.currentFolderFullNameHash(),
								MessagelistUserStore.pageBeforeThread(),
								MessagelistUserStore.listSearch()
							)
						);
					}
				} else if (MessagelistUserStore.threadUid() != message.uid) {
					MessagelistUserStore.threadUid(message.uid);
					setHash = 1;
				}
				if (setHash) {
					replaceHash(
						mailBox(
							FolderUserStore.currentFolderFullNameHash(),
							MessagelistUserStore.page(),
							MessagelistUserStore.listSearch(),
							MessagelistUserStore.threadUid()
						)
					);
				}
			}
		}
	},

	MessagelistUserStore.reloadFlagsAndCachedMessage = () => {
		MessagelistUserStore.forEach(message => MessageFlagsCache.initMessage(message));
		MessageFlagsCache.initMessage(MessageUserStore.message());
	};

	const
		win = window,
		CLIENT_SIDE_STORAGE_INDEX_NAME = 'rlcsc',
		sName = 'localStorage',
		getStorage = () => {
			try {
				const value = localStorage.getItem(CLIENT_SIDE_STORAGE_INDEX_NAME);
				return value ? JSON.parse(value) : null;
			} catch (e) {
				return null;
			}
		};

	// Storage
	try {
		win[sName].setItem(sName, '');
		win[sName].getItem(sName);
		win[sName].removeItem(sName);
	} catch (e) {
		console.error(e);
		// initialise if there's already data
		let data = document.cookie.match(/(^|;) ?localStorage=([^;]+)/);
		data = data ? decodeURIComponent(data[2]) : null;
		data = data ? JSON.parse(data) : {};
		win[sName] = {
			getItem: key => data[key] == null ? null : data[key],
			setItem: (key, value) => {
				data[key] = ''+value; // forces the value to a string
				document.cookie = sName+'='+encodeURIComponent(JSON.stringify(data))
					+"; expires="+((new Date(Date.now()+(365*24*60*60*1000))).toGMTString())
					+"; path=/; samesite=strict";
			}
		};
	}

	/**
	 * @param {number} key
	 * @param {*} data
	 * @returns {boolean}
	 */
	function set(key, data) {
		const storageResult = getStorage() || {};
		storageResult['p' + key] = data;

		try {
			localStorage.setItem(CLIENT_SIDE_STORAGE_INDEX_NAME, JSON.stringify(storageResult));
			return true;
		} catch (e) {
			return false;
		}
	}

	/**
	 * @param {number} key
	 * @returns {*}
	 */
	function get(key) {
		try {
			return (getStorage() || {})['p' + key];
		} catch (e) {
			return null;
		}
	}

	const

	dropdownsDetectVisibility = (() =>
		dropdownVisibility(!!dropdowns.find(item => item.classList.contains('show')))
	).debounce(50),

	/**
	 * @param {string} link
	 * @returns {boolean}
	 */
	download = (link, name = "") => {
		console.log('download: '+link);
		// Firefox 98 issue https://github.com/the-djmaze/snappymail/issues/301
		if (ThemeStore.isMobile() || /firefox/i.test(navigator.userAgent)) {
			open(link, '_blank');
			focus();
		} else {
			const oLink = createElement('a', {
				href: link,
				target: '_blank',
				download: name
			});
			doc.body.appendChild(oLink).click();
			oLink.remove();
		}
	},

	/**
	 * @returns {function}
	 */
	computedPaginatorHelper = (koCurrentPage, koPageCount) => {
		return () => {
			const currentPage = koCurrentPage(),
				pageCount = koPageCount(),
				result = [],
				fAdd = (index, push = true, customName = '') => {
					const data = {
						current: index === currentPage,
						name: customName ? customName.toString() : index.toString(),
						custom: !!customName,
						title: customName ? index.toString() : '',
						value: index.toString()
					};

					if (push) {
						result.push(data);
					} else {
						result.unshift(data);
					}
				};

			let prev = 0,
				next = 0,
				limit = 2;

			if (1 < pageCount || (0 < pageCount && pageCount < currentPage)) {
				if (pageCount < currentPage) {
					fAdd(pageCount);
					prev = pageCount;
					next = pageCount;
				} else {
					if (3 >= currentPage || pageCount - 2 <= currentPage) {
						limit += 2;
					}

					fAdd(currentPage);
					prev = currentPage;
					next = currentPage;
				}

				while (0 < limit) {
					--prev;
					++next;

					if (0 < prev) {
						fAdd(prev, false);
						--limit;
					}

					if (pageCount >= next) {
						fAdd(next, true);
						--limit;
					} else if (0 >= prev) {
						break;
					}
				}

				if (3 === prev) {
					fAdd(2, false);
				} else if (3 < prev) {
					fAdd(Math.round((prev - 1) / 2), false, '…');
				}

				if (pageCount - 2 === next) {
					fAdd(pageCount - 1, true);
				} else if (pageCount - 2 > next) {
					fAdd(Math.round((pageCount + next) / 2), true, '…');
				}

				// first and last
				if (1 < prev) {
					fAdd(1, false);
				}

				if (pageCount > next) {
					fAdd(pageCount, true);
				}
			}

			return result;
		};
	},

	/**
	 * @param {string} mailToUrl
	 * @returns {boolean}
	 */
	mailToHelper = mailToUrl => {
		if ('mailto:' === mailToUrl?.slice(0, 7).toLowerCase()) {
			mailToUrl = mailToUrl.slice(7).split('?');

			const
				email = mailToUrl[0],
				params = new URLSearchParams(mailToUrl[1]),
				toEmailModel = value => null != value ? EmailModel.parseEmailLine(value) : null;

			showMessageComposer([
				ComposeType.Empty,
				null,
				params.get('to')
					? Object.values(
							toEmailModel(email + ',' + params.get('to')).reduce((result, value) => {
								if (value) {
									if (result[value.email]) {
										if (!result[value.email].name) {
											result[value.email] = value;
										}
									} else {
										result[value.email] = value;
									}
								}
								return result;
							}, {})
						)
					: EmailModel.parseEmailLine(email),
				toEmailModel(params.get('cc')),
				toEmailModel(params.get('bcc')),
				params.get('subject'),
				plainToHtml(params.get('body') || '')
			]);

			return true;
		}

		return false;
	},

	showMessageComposer = (params = []) =>
	{
		rl.app.showMessageComposer(params);
	},

	setLayoutResizer = (source, target, sClientSideKeyName, mode) =>
	{
		if (source.layoutResizer && source.layoutResizer.mode != mode) {
			target.removeAttribute('style');
			source.removeAttribute('style');
		}
		source.observer?.disconnect();
	//	source.classList.toggle('resizable', mode);
		if (mode) {
			const length = get(sClientSideKeyName + mode) || SettingsGet('Resizer' + sClientSideKeyName + mode),
				setTargetPos = mode => {
					let value;
					if ('Width' == mode) {
						value = source.offsetWidth;
						target.style.left = value + 'px';
					} else {
						value = source.offsetHeight;
						target.style.top = (4 + source.offsetTop + value) + 'px';
					}
					return value;
				};
			if (length) {
				source.style[mode.toLowerCase()] = length + 'px';
				setTargetPos(mode);
			}
			if (!source.layoutResizer) {
				const resizer = createElement('div', {'class':'resizer'}),
					save = (data => Remote.saveSettings(0, data)).debounce(500),
					size = {},
					store = () => {
						const value = setTargetPos(resizer.mode),
							prop = resizer.key + resizer.mode;
						(value == get(prop)) || set(prop, value);
						(value == SettingsGet('Resizer' + prop)) || save({['Resizer' + prop]: value});
					},
					cssint = s => {
						let value = getComputedStyle(source, null)[s].replace('px', '');
						if (value.includes('%')) {
							value = source.parentElement['offset'+resizer.mode]
								* value.replace('%', '') / 100;
						}
						return parseFloat(value);
					};
				source.layoutResizer = resizer;
				source.append(resizer);
				resizer.addEventListener('mousedown', {
					handleEvent: function(e) {
						if ('mousedown' == e.type) {
							const lmode = resizer.mode.toLowerCase();
							e.preventDefault();
							size.pos = ('width' == lmode) ? e.pageX : e.pageY;
							size.min = cssint('min-'+lmode);
							size.max = cssint('max-'+lmode);
							size.org = cssint(lmode);
							addEventListener('mousemove', this);
							addEventListener('mouseup', this);
						} else if ('mousemove' == e.type) {
							const lmode = resizer.mode.toLowerCase(),
								length = size.org + (('width' == lmode ? e.pageX : e.pageY) - size.pos);
							if (length >= size.min && length <= size.max ) {
								source.style[lmode] = length + 'px';
								source.observer || store();
							}
						} else if ('mouseup' == e.type) {
							removeEventListener('mousemove', this);
							removeEventListener('mouseup', this);
						}
					}
				});
				source.observer = window.ResizeObserver ? new ResizeObserver(store) : null;
			}
			source.layoutResizer.mode = mode;
			source.layoutResizer.key = sClientSideKeyName;
			source.observer?.observe(source, { box: 'border-box' });
		}
	},

	populateMessageBody = (oMessage, popup) => {
		if (oMessage) {
			popup || MessageUserStore.hideMessageBodies();
			popup || MessageUserStore.loading(true);
			Remote.message((iError, oData/*, bCached*/) => {
				if (iError) {
					if (Notification$1.RequestAborted !== iError && !popup) {
						MessageUserStore.message(null);
						MessageUserStore.error(getNotification(iError));
					}
				} else {
					let json = oData?.Result;

					if (
						json &&
						MessageModel.validJson(json) &&
						oMessage.folder === json.Folder
					) {
						const threads = oMessage.threads(),
							isNew = !popup && oMessage.uid != json.Uid && threads.includes(json.Uid),
							messagesDom = MessageUserStore.bodiesDom();
						if (isNew) {
							oMessage = MessageModel.reviveFromJson(json);
							if (oMessage) {
								oMessage.threads(threads);
								MessageFlagsCache.initMessage(oMessage);

								// Set clone
								MessageUserStore.message(MessageModel.fromMessageListItem(oMessage));
								oMessage = MessageUserStore.message();
							}
						}

						if (oMessage && oMessage.uid == json.Uid) {
							popup || MessageUserStore.error('');
	/*
							if (bCached) {
								delete json.Flags;
							}
	*/
							isNew || oMessage.revivePropertiesFromJson(json);
							if (messagesDom) {
								let id = 'rl-msg-' + oMessage.hash.replace(/[^a-zA-Z0-9]/g, ''),
									body = elementById(id);
								if (body) {
									oMessage.body = body;
									oMessage.isHtml(body.classList.contains('html'));
									oMessage.hasImages(body.rlHasImages);
								} else {
									body = Element.fromHTML('<div id="' + id + '" hidden="" class="b-text-part '
										+ (oMessage.pgpSigned() ? ' openpgp-signed' : '')
										+ (oMessage.pgpEncrypted() ? ' openpgp-encrypted' : '')
										+ '">'
										+ '</div>');
									oMessage.body = body;
									if (!SettingsUserStore.viewHTML() || !oMessage.viewHtml()) {
										oMessage.viewPlain();
									}

									MessageUserStore.purgeMessageBodyCache();
								}

								messagesDom.append(body);

								if (!popup) {
									MessageUserStore.activeDom(body);
									MessageUserStore.hideMessageBodies();
									oMessage.body.hidden = false;
								}
								popup && oMessage.viewPopupMessage();
							}

							MessageFlagsCache.initMessage(oMessage);
							if (oMessage.isUnseen()) {
								MessageUserStore.MessageSeenTimer = setTimeout(
									() => MessagelistUserStore.setAction(oMessage.folder, MessageSetAction.SetSeen, [oMessage]),
									SettingsUserStore.messageReadDelay() * 1000 // seconds
								);
							}

							if (isNew) {
								let selectedMessage = MessagelistUserStore.selectedMessage();
								if (
									selectedMessage &&
									(oMessage.folder !== selectedMessage.folder || oMessage.uid != selectedMessage.uid)
								) {
									MessagelistUserStore.selectedMessage(null);
									if (1 === MessagelistUserStore.length) {
										MessagelistUserStore.focusedMessage(null);
									}
								} else if (!selectedMessage) {
									selectedMessage = MessagelistUserStore.find(
										subMessage =>
											subMessage &&
											subMessage.folder === oMessage.folder &&
											subMessage.uid == oMessage.uid
									);

									if (selectedMessage) {
										MessagelistUserStore.selectedMessage(selectedMessage);
										MessagelistUserStore.focusedMessage(selectedMessage);
									}
								}
							}
						}
					}
				}
				popup || MessageUserStore.loading(false);
			}, oMessage.folder, oMessage.uid);
		}
	};

	const contentType = 'snappymail/emailaddress',
		getAddressKey = li => li?.emailaddress?.key;

	let dragAddress, datalist;

	// mailbox-list
	class EmailAddressesComponent {

		constructor(element, options) {

			if (!datalist) {
				datalist = createElement('datalist',{id:"emailaddresses-datalist"});
				doc.body.append(datalist);
			}

			const self = this,
				input = createElement('input',{type:"text", list:datalist.id,
					autocomplete:"off", autocorrect:"off", autocapitalize:"off"}),
				// In Chrome we have no access to dataTransfer.getData unless it's the 'drop' event
				// In Chrome Mobile dataTransfer.types.includes(contentType) fails, only text/plain is set
				validDropzone = () => dragAddress?.li.parentNode !== self.ul,
				fnDrag = e => validDropzone() && e.preventDefault();

			self.element = element;

			self.options = Object.assign({

				focusCallback : null,

				// simply passing an autoComplete source (array, string or function) will instantiate autocomplete functionality
				autoCompleteSource : '',

				onChange : null
			}, options);

			self._chosenValues = [];

			self._lastEdit = '';

			// Create the elements
			self.ul = createElement('ul',{class:"emailaddresses"});

			addEventsListeners(self.ul, {
				click: e => self._focus(e),
				dblclick: e => self._editTag(e),
				dragenter: fnDrag,
				dragover: fnDrag,
				drop: e => {
					if (validDropzone() && dragAddress.value) {
						e.preventDefault();
						dragAddress.source._removeDraggedTag(dragAddress.li);
						self._parseValue(dragAddress.value);
					}
				}
			});

			self.input = input;

			addEventsListeners(input, {
				focus: () => {
					self._focusTrigger(true);
					input.value || self._resetDatalist();
				},
				blur: () => {
					// prevent autoComplete menu click from causing a false 'blur'
					self._parseInput(true);
					self._focusTrigger(false);
				},
				keydown: e => {
					if ('Backspace' === e.key || 'ArrowLeft' === e.key) {
						// if our input contains no value and backspace has been pressed, select the last tag
						var lastTag = self.inputCont.previousElementSibling;
						if (lastTag && (!input.value
							|| (('selectionStart' in input) && input.selectionStart === 0 && input.selectionEnd === 0))
						) {
							e.preventDefault();
							lastTag.querySelector('a').focus();
						}
						self._updateDatalist();
					} else if (e.key == 'Enter') {
						e.preventDefault();
						self._parseInput(true);
					}
				},
				input: () => {
					self._parseInput();
					self._updateDatalist();
				}
			});

			// define starting placeholder
			if (element.placeholder) {
				input.placeholder = element.placeholder;
			}

			self.inputCont = createElement('li',{class:"emailaddresses-input"});
			self.inputCont.append(input);
			self.ul.append(self.inputCont);

			element.replaceWith(self.ul);

			// if instantiated input already contains a value, parse that junk
			if (element.value.trim()) {
				self._parseValue(element.value);
			}

			self._updateDatalist = self.options.autoCompleteSource
				? (() => {
					let value = input.value.trim();
					if (datalist.inputValue !== value) {
						datalist.inputValue = value;
						value.length && self.options.autoCompleteSource(
							value,
							items => {
								self._resetDatalist();
								let chars = value.length;
								items?.forEach(item => {
									datalist.append(new Option(item));
									chars = Math.max(chars, item.length);
								});
								// https://github.com/the-djmaze/snappymail/issues/368 and #513
								chars *= 8;
								if (input.clientWidth < chars) {
									input.style.width = chars + 'px';
								}
							}
						);
					}
				}).throttle(500)
				: () => 0;
		}

		_focusTrigger(bValue) {
			this.ul.classList.toggle('emailaddresses-focused', bValue);
			this.options.focusCallback(bValue);
		}

		_resetDatalist() {
			datalist.textContent = '';
		}

		_parseInput(force) {
			let val = this.input.value;
			if (force || val.includes(',') || val.includes(';')) {
				this._parseValue(val) && (this.input.value = '');
			}
			this._resizeInput();
		}

		_parseValue(val) {
			if (val) {
				const self = this,
					v = val.trim(),
					hook = (v && [',', ';', '\n'].includes(v.slice(-1))) ? EmailModel.splitEmailLine(val) : null,
					values = (hook || [val]).map(value => EmailModel.parseEmailLine(value))
							.flat(Infinity)
							.map(item => (item.toLine ? [item.toLine(), item] : [item, null]));

				if (values.length) {
					values.forEach(a => {
						var v = a[0].trim(),
							exists = false,
							lastIndex = -1,
							obj = {
								key : '',
								obj : null,
								value : ''
							};

						self._chosenValues.forEach((vv, kk) => {
							if (vv.value === self._lastEdit) {
								lastIndex = kk;
							}

							vv.value === v && (exists = true);
						});

						if (v !== '' && a[1] && !exists) {

							obj.key = 'mi_' + Math.random().toString( 16 ).slice( 2, 10 );
							obj.value = v;
							obj.obj = a[1];

							if (-1 < lastIndex) {
								self._chosenValues.splice(lastIndex, 0, obj);
							} else {
								self._chosenValues.push(obj);
							}

							self._lastEdit = '';
							self._renderTags();
						}
					});

					if (values.length === 1 && values[0] === '' && self._lastEdit !== '') {
						self._lastEdit = '';
						self._renderTags();
					}

					self._setValue(self._buildValue());

					return true;
				}
			}
		}

		// the input dynamically resizes based on the length of its value
		_resizeInput() {
			let input = this.input;
			if (input.clientWidth < input.scrollWidth) {
				input.style.width = (input.scrollWidth + 20) + 'px';
			}
		}

		_editTag(ev) {
			var li = ev.target.closest('li'),
				tagKey = getAddressKey(li);

			if (!tagKey) {
				return true;
			}

			var self = this,
				tagName = '',
				oPrev = null,
				next = false
			;

			self._chosenValues.forEach(v => {
				if (v.key === tagKey) {
					tagName = v.value;
					next = true;
				} else if (next && !oPrev) {
					oPrev = v;
				}
			});

			if (oPrev)
			{
				self._lastEdit = oPrev.value;
			}

			li.after(self.inputCont);

			self.input.value = tagName;
			setTimeout(() => self.input.select(), 100);

			self._removeTag(ev, li);
			self._resizeInput(ev);
		}

		_buildValue() {
			return this._chosenValues.map(v => v.value).join(',');
		}

		_setValue(value) {
			if (this.element.value !== value) {
				this.element.value = value;
				this.options.onChange(value);
			}
		}

		_renderTags() {
			let self = this;
			[...self.ul.children].forEach(node => node !== self.inputCont && node.remove());

			self._chosenValues.forEach(v => {
				if (v.obj) {
					let li = createElement('li',{title:v.obj.toLine(),draggable:'true'}),
						el = createElement('span');
					el.append(v.obj.toLine(true));
					li.append(el);

					el = createElement('a',{href:'#', class:'ficon'});
					el.append('✖');
					addEventsListeners(el, {
						click: e => self._removeTag(e, li),
						focus: () => li.className = 'emailaddresses-selected',
						blur: () => li.className = null,
						keydown: e => {
							switch (e.key) {
								case 'Delete':
								case 'Backspace':
									self._removeTag(e, li);
									break;

								// 'e' - edit tag (removes tag and places value into visible input
								case 'e':
								case 'Enter':
									self._editTag(e);
									break;

								case 'ArrowLeft':
									// select the previous tag or input if no more tags exist
									var previous = el.closest('li').previousElementSibling;
									if (previous.matches('li')) {
										previous.querySelector('a').focus();
									} else {
										self.focus();
									}
									break;

								case 'ArrowRight':
									// select the next tag or input if no more tags exist
									var next = el.closest('li').nextElementSibling;
									if (next !== this.inputCont) {
										next.querySelector('a').focus();
									} else {
										this.focus();
									}
									break;

								case 'ArrowDown':
									self._focus(e);
									break;
							}
						}
					});
					li.append(el);

					li.emailaddress = v;

					addEventsListeners(li, {
						dragstart: e => {
							dragAddress = {
								source: self,
								li: li,
								value: li.emailaddress.obj.toLine()
							};
	//						e.dataTransfer.setData(contentType, li.emailaddress.obj.toLine());
							e.dataTransfer.setData('text/plain', contentType);
	//						e.dataTransfer.setDragImage(li, 0, 0);
							e.dataTransfer.effectAllowed = 'move';
							li.style.opacity = 0.25;
						},
						dragend: () => {
							dragAddress = null;
							li.style.cssText = '';
						}
					});

					self.inputCont.before(li);
				}
			});
		}

		_removeTag(ev, li) {
			ev.preventDefault();

			var key = getAddressKey(li),
				self = this,
				indexFound = self._chosenValues.findIndex(v => key === v.key);

			indexFound > -1 && self._chosenValues.splice(indexFound, 1);

			self._setValue(self._buildValue());

			li.remove();
			setTimeout(() => self.input.focus(), 100);
		}

		_removeDraggedTag(li) {
			var
				key = getAddressKey(li),
				self = this,
				indexFound = self._chosenValues.findIndex(v => key === v.key)
			;
			if (-1 < indexFound) {
				self._chosenValues.splice(indexFound, 1);
				self._setValue(self._buildValue());
			}

			li.remove();
		}

		focus () {
			this.input.focus();
		}

		blur() {
			this.input.blur();
		}

		_focus(ev) {
			var li = ev.target.closest('li');
			if (getAddressKey(li)) {
				li.querySelector('a').focus();
			} else {
				this.focus();
			}
		}

		set value(value) {
			var self = this;
			if (self.element.value !== value) {
	//			self.input.value = '';
	//			self._resizeInput();
				self._chosenValues = [];
				self._renderTags();
				self._parseValue(self.element.value = value);
			}
		}
	}

	const

	sortFolders = folders => {
		try {
			let collator = new Intl.Collator(undefined, {numeric: true, sensitivity: 'base'});
			folders.sort((a, b) =>
				a.isInbox() ? -1 : (b.isInbox() ? 1 : collator.compare(a.fullName, b.fullName))
			);
		} catch (e) {
			console.error(e);
		}
	},

	/**
	 * @param {?Function} fCallback
	 * @param {string} folder
	 * @param {Array=} list = []
	 */
	fetchFolderInformation = (fCallback, folder, list = []) => {
		let fetch = !arrayLength(list);
		const uids = [],
			folderFromCache = getFolderFromCacheList(folder);

		if (!fetch) {
			list.forEach(messageListItem => {
				if (!MessageFlagsCache.getFor(folder, messageListItem.uid)) {
					uids.push(messageListItem.uid);
				}

				if (messageListItem.threads.length) {
					messageListItem.threads.forEach(uid => {
						if (!MessageFlagsCache.getFor(folder, uid)) {
							uids.push(uid);
						}
					});
				}
			});
			fetch = uids.length;
		}

		if (fetch) {
			Remote.request('FolderInformation', fCallback, {
				Folder: folder,
				FlagsUids: uids,
				UidNext: folderFromCache?.uidNext || 0 // Used to check for new messages
			});
		} else if (SettingsUserStore.useThreads()) {
			MessagelistUserStore.reloadFlagsAndCachedMessage();
		}
	},

	/**
	 * @param {Array=} aDisabled
	 * @param {Array=} aHeaderLines
	 * @param {Function=} fRenameCallback
	 * @param {Function=} fDisableCallback
	 * @param {boolean=} bNoSelectSelectable Used in FolderCreatePopupView
	 * @returns {Array}
	 */
	folderListOptionsBuilder = (
		aDisabled,
		aHeaderLines,
		fRenameCallback,
		fDisableCallback,
		bNoSelectSelectable,
		aList = FolderUserStore.folderList()
	) => {
		const
			aResult = [],
			sDeepPrefix = '\u00A0\u00A0\u00A0',
			// FolderSystemPopupView should always be true
			showUnsubscribed = fRenameCallback ? !SettingsUserStore.hideUnsubscribed() : true,

			foldersWalk = folders => {
				folders.forEach(oItem => {
					if (showUnsubscribed || oItem.hasSubscriptions() || !oItem.exists) {
						aResult.push({
							id: oItem.fullName,
							name:
								sDeepPrefix.repeat(oItem.deep) +
								fRenameCallback(oItem),
							system: false,
							disabled: !bNoSelectSelectable && (
								!oItem.selectable() ||
								aDisabled.includes(oItem.fullName) ||
								fDisableCallback(oItem))
						});
					}

					if (oItem.subFolders.length) {
						foldersWalk(oItem.subFolders());
					}
				});
			};


		fDisableCallback = fDisableCallback || (() => false);
		fRenameCallback = fRenameCallback || (oItem => oItem.name());
		isArray(aDisabled) || (aDisabled = []);

		isArray(aHeaderLines) && aHeaderLines.forEach(line =>
			aResult.push({
				id: line[0],
				name: line[1],
				system: false,
				disabled: false
			})
		);

		foldersWalk(aList);

		return aResult;
	},

	// Every 5 minutes
	refreshFoldersInterval = 300000,

	/**
	 * @param {string} folder
	 * @param {Array=} list = []
	 */
	folderInformation = (folder, list) => {
		if (folder?.trim()) {
			fetchFolderInformation(
				(iError, data) => {
					if (!iError && data.Result) {
						const result = data.Result,
							folderFromCache = getFolderFromCacheList(result.Folder);
						if (folderFromCache) {
							const oldHash = folderFromCache.hash,
								unreadCountChange = (folderFromCache.unreadEmails() !== result.unreadEmails);

	//							folderFromCache.revivePropertiesFromJson(result);
							folderFromCache.expires = Date.now();
							folderFromCache.uidNext = result.UidNext;
							folderFromCache.hash = result.Hash;
							folderFromCache.totalEmails(result.totalEmails);
							folderFromCache.unreadEmails(result.unreadEmails);

							if (unreadCountChange) {
								MessageFlagsCache.clearFolder(folderFromCache.fullName);
							}

							if (result.MessagesFlags.length) {
								result.MessagesFlags.forEach(message =>
									MessageFlagsCache.setFor(folderFromCache.fullName, message.Uid.toString(), message.Flags)
								);

								MessagelistUserStore.reloadFlagsAndCachedMessage();
							}

							MessagelistUserStore.notifyNewMessages(folderFromCache.fullName, result.NewMessages);

							if (!oldHash || unreadCountChange || result.Hash !== oldHash) {
								if (folderFromCache.fullName === FolderUserStore.currentFolderFullName()) {
									MessagelistUserStore.reload();
								} else if (getFolderInboxName() === folderFromCache.fullName) {
									Remote.messageList(null, {Folder: getFolderInboxName()}, true);
								}
							}
						}
					}
				},
				folder,
				list
			);
		}
	},

	/**
	 * @param {boolean=} boot = false
	 */
	folderInformationMultiply = (boot = false) => {
		const folders = FolderUserStore.getNextFolderNames(refreshFoldersInterval);
		if (arrayLength(folders)) {
			Remote.request('FolderInformationMultiply', (iError, oData) => {
				if (!iError && arrayLength(oData.Result)) {
					const utc = Date.now();
					oData.Result.forEach(item => {
						const folder = getFolderFromCacheList(item.Folder);

						if (folder) {
							const oldHash = folder.hash,
								unreadCountChange = folder.unreadEmails() !== item.unreadEmails;

	//						folder.revivePropertiesFromJson(item);
							folder.expires = utc;
							folder.hash = item.Hash;
							folder.totalEmails(item.totalEmails);
							folder.unreadEmails(item.unreadEmails);

							if (unreadCountChange) {
								MessageFlagsCache.clearFolder(folder.fullName);
							}

							if (!oldHash || item.Hash !== oldHash) {
								if (folder.fullName === FolderUserStore.currentFolderFullName()) {
									MessagelistUserStore.reload();
								}
							} else if (unreadCountChange
							 && folder.fullName === FolderUserStore.currentFolderFullName()
							 && MessagelistUserStore.length) {
								folderInformation(folder.fullName, MessagelistUserStore());
							}
						}
					});

					if (boot) {
						setTimeout(() => folderInformationMultiply(true), 2000);
					}
				}
			}, {
				Folders: folders
			});
		}
	},

	moveOrDeleteResponseHelper = (iError, oData) => {
		if (iError) {
			setFolderHash(FolderUserStore.currentFolderFullName(), '');
			alert(getNotification(iError));
		} else if (FolderUserStore.currentFolder()) {
			if (2 === arrayLength(oData.Result)) {
				setFolderHash(oData.Result[0], oData.Result[1]);
			} else {
				setFolderHash(FolderUserStore.currentFolderFullName(), '');
			}
			MessagelistUserStore.reload(!MessagelistUserStore.length);
		}
	},

	messagesMoveHelper = (fromFolderFullName, toFolderFullName, uidsForMove) => {
		const
			sSpamFolder = FolderUserStore.spamFolder(),
			isSpam = sSpamFolder === toFolderFullName,
			isHam = !isSpam && sSpamFolder === fromFolderFullName && getFolderInboxName() === toFolderFullName;

		Remote.abort('MessageList').request('MessageMove',
			moveOrDeleteResponseHelper,
			{
				FromFolder: fromFolderFullName,
				ToFolder: toFolderFullName,
				Uids: [...uidsForMove].join(','),
				MarkAsRead: (isSpam || FolderUserStore.trashFolder() === toFolderFullName) ? 1 : 0,
				Learning: isSpam ? 'SPAM' : isHam ? 'HAM' : ''
			}
		);
	},

	messagesDeleteHelper = (sFromFolderFullName, aUidForRemove) => {
		Remote.abort('MessageList').request('MessageDelete',
			moveOrDeleteResponseHelper,
			{
				Folder: sFromFolderFullName,
				Uids: [...aUidForRemove].join(',')
			}
		);
	},

	/**
	 * @param {string} sFromFolderFullName
	 * @param {Set} oUids
	 * @param {string} sToFolderFullName
	 * @param {boolean=} bCopy = false
	 */
	moveMessagesToFolder = (sFromFolderFullName, oUids, sToFolderFullName, bCopy) => {
		if (sFromFolderFullName !== sToFolderFullName && oUids?.size) {
			const oFromFolder = getFolderFromCacheList(sFromFolderFullName),
				oToFolder = getFolderFromCacheList(sToFolderFullName);

			if (oFromFolder && oToFolder) {
				if (bCopy) {
					Remote.request('MessageCopy', null, {
						FromFolder: oFromFolder.fullName,
						ToFolder: oToFolder.fullName,
						Uids: [...oUids].join(',')
					});
				} else {
					messagesMoveHelper(oFromFolder.fullName, oToFolder.fullName, oUids);
				}

				MessagelistUserStore.removeMessagesFromList(oFromFolder.fullName, oUids, oToFolder.fullName, bCopy);
				return true;
			}
		}

		return false;
	};

	const
		isPosNumeric = value => null != value && /^[0-9]*$/.test(value.toString()),

		normalizeFolder = sFolderFullName => ('' === sFolderFullName
			|| UNUSED_OPTION_VALUE === sFolderFullName
			|| null !== getFolderFromCacheList(sFolderFullName))
				? sFolderFullName
				: '',

		SystemFolders = {
			Inbox:   0,
			Sent:    0,
			Drafts:  0,
			Spam:    0,
			Trash:   0,
			Archive: 0
		},

		kolabTypes = {
			configuration: 'CONFIGURATION',
			event: 'CALENDAR',
			contact: 'CONTACTS',
			task: 'TASKS',
			note: 'NOTES',
			file: 'FILES',
			journal: 'JOURNAL'
		},

		getKolabFolderName = type => kolabTypes[type] ? 'Kolab ' + i18n('SETTINGS_FOLDERS/TYPE_' + kolabTypes[type]) : '',

		getSystemFolderName = (type, def) => {
			switch (type) {
				case FolderType.Inbox:
				case FolderType.Sent:
				case FolderType.Drafts:
				case FolderType.Trash:
				case FolderType.Archive:
					return i18n('FOLDER_LIST/' + getKeyByValue(FolderType, type).toUpperCase() + '_NAME');
				case FolderType.Spam:
					return i18n('GLOBAL/SPAM');
				// no default
			}
			return def;
		};

	const
		/**
		 * @param {string} sFullName
		 * @param {boolean} bExpanded
		 */
		setExpandedFolder = (sFullName, bExpanded) => {
			let aExpandedList = get(ClientSideKeyNameExpandedFolders);
			if (!isArray(aExpandedList)) {
				aExpandedList = [];
			}

			if (bExpanded) {
				aExpandedList.includes(sFullName) || aExpandedList.push(sFullName);
			} else {
				aExpandedList = aExpandedList.filter(value => value !== sFullName);
			}

			set(ClientSideKeyNameExpandedFolders, aExpandedList);
		},

		/**
		 * @param {?Function} fCallback
		 */
		loadFolders = fCallback => {
	//		clearTimeout(this.foldersTimeout);
			Remote.abort('Folders')
				.post('Folders', FolderUserStore.foldersLoading)
				.then(data => {
					data = FolderCollectionModel.reviveFromJson(data.Result);
					data?.storeIt();
					fCallback?.(true);
					// Repeat every 15 minutes?
	//				this.foldersTimeout = setTimeout(loadFolders, 900000);
				})
				.catch(() => fCallback && setTimeout(fCallback, 1, false));
		};

	class FolderCollectionModel extends AbstractCollectionModel
	{
	/*
		constructor() {
			super();
			this.CountRec
			this.IsThreadsSupported
			this.Namespace;
			this.Optimized
			this.SystemFolders
			this.Capabilities
		}
	*/

		/**
		 * @param {?Object} json
		 * @returns {FolderCollectionModel}
		 */
		static reviveFromJson(object) {
			const expandedFolders = get(ClientSideKeyNameExpandedFolders);
			if (object?.SystemFolders) {
				forEachObjectEntry(SystemFolders, key =>
					SystemFolders[key] = SettingsGet(key+'Folder') || object.SystemFolders[FolderType[key]]
				);
			}

			const result = super.reviveFromJson(object, oFolder => {
				let oCacheFolder = getFolderFromCacheList(oFolder.FullName),
					type = FolderType[getKeyByValue(SystemFolders, oFolder.FullName)];

				if (oCacheFolder) {
	//				oCacheFolder.revivePropertiesFromJson(oFolder);
					if (oFolder.Hash) {
						oCacheFolder.hash = oFolder.Hash;
					}
					if (null != oFolder.totalEmails) {
						oCacheFolder.totalEmails(oFolder.totalEmails);
					}
					if (null != oFolder.unreadEmails) {
						oCacheFolder.unreadEmails(oFolder.unreadEmails);
					}
				} else {
					oCacheFolder = FolderModel.reviveFromJson(oFolder);
					if (!oCacheFolder)
						return null;

					if (1 == type) {
						oCacheFolder.type(type);
						setFolderInboxName(oFolder.FullName);
					}
					setFolder(oCacheFolder);
				}

				if (1 < type) {
					oCacheFolder.type(type);
				}

				oCacheFolder.collapsed(!expandedFolders
					|| !isArray(expandedFolders)
					|| !expandedFolders.includes(oCacheFolder.fullName));

				return oCacheFolder;
			});

			let i = result.length;
			if (i) {
				sortFolders(result);
				try {
					while (i--) {
						let folder = result[i], parent = getFolderFromCacheList(folder.parentName);
						if (parent) {
							parent.subFolders.unshift(folder);
							result.splice(i,1);
						}
					}
				} catch (e) {
					console.error(e);
				}
			}

			return result;
		}

		storeIt() {
			FolderUserStore.displaySpecSetting(Settings.app('folderSpecLimit') < this.CountRec);

			if (!(
					SettingsGet('SentFolder') +
					SettingsGet('DraftsFolder') +
					SettingsGet('SpamFolder') +
					SettingsGet('TrashFolder') +
					SettingsGet('ArchiveFolder')
				)
			) {
				FolderUserStore.saveSystemFolders(SystemFolders);
			}

			FolderUserStore.folderList(this);

			FolderUserStore.namespace = this.Namespace;

			AppUserStore.threadsAllowed(!!(Settings.app('useImapThread') && this.IsThreadsSupported));

			FolderUserStore.folderListOptimized(!!this.Optimized);
			FolderUserStore.quotaUsage(this.quotaUsage);
			FolderUserStore.quotaLimit(this.quotaLimit);
			FolderUserStore.capabilities(this.Capabilities);

			FolderUserStore.sentFolder(normalizeFolder(SystemFolders.Sent));
			FolderUserStore.draftsFolder(normalizeFolder(SystemFolders.Drafts));
			FolderUserStore.spamFolder(normalizeFolder(SystemFolders.Spam));
			FolderUserStore.trashFolder(normalizeFolder(SystemFolders.Trash));
			FolderUserStore.archiveFolder(normalizeFolder(SystemFolders.Archive));

	//		FolderUserStore.folderList.valueHasMutated();
		}

	}

	class FolderModel extends AbstractModel {
		constructor() {
			super();

			this.fullName = '';
			this.delimiter = '';
			this.deep = 0;
			this.expires = 0;
			this.metadata = {};

			this.exists = true;

			this.hash = '';
	//		this.id = null;
			this.uidNext = null;

			this.addObservables({
				name: '',
				type: FolderType.User,
				selectable: false,

				focused: false,
				selected: false,
				editing: false,
				isSubscribed: true,
				checkable: false, // Check for new messages
				askDelete: false,

				nameForEdit: '',
				errorMsg: '',

				totalEmailsValue: 0,
				unreadEmailsValue: 0,

				kolabType: null,

				collapsed: true,

				tagsAllowed: false
			});

			this.flags = ko.observableArray();
			this.permanentFlags = ko.observableArray();

			this.addSubscribables({
				kolabType: sValue => this.metadata[FolderMetadataKeys.KolabFolderType] = sValue,
				permanentFlags: aValue => this.tagsAllowed(aValue.includes('\\*'))
			});

			this.subFolders = ko.observableArray(new FolderCollectionModel);
			this.actionBlink = ko.observable(false).extend({ falseTimeout: 1000 });

			this.totalEmails = koComputable({
					read: this.totalEmailsValue,
					write: (iValue) => {
						if (isPosNumeric(iValue)) {
							this.totalEmailsValue(iValue);
						} else {
							this.totalEmailsValue.valueHasMutated();
						}
					}
				})
				.extend({ notify: 'always' });

			this.unreadEmails = koComputable({
					read: this.unreadEmailsValue,
					write: (value) => {
						if (isPosNumeric(value)) {
							this.unreadEmailsValue(value);
						} else {
							this.unreadEmailsValue.valueHasMutated();
						}
					}
				})
				.extend({ notify: 'always' });
	/*
			https://www.rfc-editor.org/rfc/rfc8621.html#section-2
			"myRights": {
				"mayAddItems": true,
				"mayRename": false,
				"maySubmit": true,
				"mayDelete": false,
				"maySetKeywords": true,
				"mayRemoveItems": true,
				"mayCreateChild": true,
				"maySetSeen": true,
				"mayReadItems": true
			},
	*/
		}

		/**
		 * For url safe '/#/mailbox/...' path
		 */
		get fullNameHash() {
			return this.fullName.replace(/[^a-z0-9._-]+/giu, b64EncodeJSONSafe);
	//		return /^[a-z0-9._-]+$/iu.test(this.fullName) ? this.fullName : b64EncodeJSONSafe(this.fullName);
		}

		/**
		 * @static
		 * @param {FetchJsonFolder} json
		 * @returns {?FolderModel}
		 */
		static reviveFromJson(json) {
			const folder = super.reviveFromJson(json);
			if (folder) {
				const path = folder.fullName.split(folder.delimiter),
					type = (folder.metadata[FolderMetadataKeys.KolabFolderType]
						|| folder.metadata[FolderMetadataKeys.KolabFolderTypeShared]
						|| ''
					).split('.')[0];

				folder.deep = path.length - 1;
				path.pop();
				folder.parentName = path.join(folder.delimiter);

				type && 'mail' != type && folder.kolabType(type);

				folder.addComputables({

					isInbox: () => FolderType.Inbox === folder.type(),

					isFlagged: () => FolderUserStore.currentFolder() === folder
						&& MessagelistUserStore.listSearch().includes('flagged'),

					hasVisibleSubfolders: () => !!folder.subFolders().find(folder => folder.visible()),

					hasSubscriptions: () => folder.isSubscribed() | !!folder.subFolders().find(
							oFolder => {
								const subscribed = oFolder.hasSubscriptions();
								return !oFolder.isSystemFolder() && subscribed;
							}
						),

					canBeEdited: () => FolderType.User === folder.type() && folder.exists/* && folder.selectable()*/,

					isSystemFolder: () => FolderType.User !== folder.type()
						| (FolderUserStore.allowKolab() && !!folder.kolabType() & !SettingsUserStore.unhideKolabFolders()),

					canBeSelected: () => folder.selectable() && !folder.isSystemFolder(),

					canBeDeleted: () => folder.canBeSelected() && folder.exists,

					canBeSubscribed: () => folder.selectable()
						&& !(folder.isSystemFolder() | !SettingsUserStore.hideUnsubscribed()),

					/**
					 * Folder is visible when:
					 * - hasVisibleSubfolders()
					 * Or when all below conditions are true:
					 * - selectable()
					 * - isSubscribed() OR hideUnsubscribed = false
					 * - FolderType.User
					 * - not kolabType()
					 */
					visible: () => {
						const selectable = folder.canBeSelected(),
							visible = (folder.isSubscribed() | !SettingsUserStore.hideUnsubscribed()) && selectable;
						return folder.hasVisibleSubfolders() | visible;
					},

					hidden: () => !folder.selectable() && (folder.isSystemFolder() | !folder.hasVisibleSubfolders()),

					printableUnreadCount: () => {
						const count = folder.totalEmails(),
							unread = folder.unreadEmails(),
							type = folder.type();

						if (count) {
							if (FolderType.Drafts === type) {
								return count;
							}
							if (
								unread &&
								FolderType.Trash !== type &&
								FolderType.Archive !== type &&
								FolderType.Sent !== type
							) {
								return unread;
							}
						}

						return null;
					},

					localName: () => {
						let name = folder.name();
						if (folder.isSystemFolder()) {
							trigger();
							name = getSystemFolderName(folder.type(), name);
						}
						return name;
					},

					manageFolderSystemName: () => {
						if (folder.isSystemFolder()) {
							trigger();
							let suffix = getSystemFolderName(folder.type(), getKolabFolderName(folder.kolabType()));
							if (folder.name() !== suffix && 'inbox' !== suffix.toLowerCase()) {
								return '(' + suffix + ')';
							}
						}
						return '';
					},

					hasUnreadMessages: () => 0 < folder.unreadEmails() && folder.printableUnreadCount(),

					hasSubscribedUnreadMessagesSubfolders: () =>
						!!folder.subFolders().find(
							folder => folder.hasUnreadMessages() | folder.hasSubscribedUnreadMessagesSubfolders()
						)

	//				,href: () => folder.canBeSelected() && mailBox(folder.fullNameHash)
				});

				folder.addSubscribables({
					editing: value => value && folder.nameForEdit(folder.name()),

					unreadEmails: unread => {
						if (FolderType.Inbox === folder.type()) {
							fireEvent('mailbox.inbox-unread-count', unread);
						}
					}
				});
			}
			return folder;
		}

		/**
		 * @returns {string}
		 */
		collapsedCss() {
			return 'e-collapsed-sign ' + (this.hasVisibleSubfolders()
				? (this.collapsed() ? 'icon-right-mini' : 'icon-down-mini')
				: 'icon-none'
			);
		}
	}

	const rlContentType = 'snappymail/action',

		// In Chrome we have no access to dataTransfer.getData unless it's the 'drop' event
		// In Chrome Mobile dataTransfer.types.includes(rlContentType) fails, only text/plain is set
		dragMessages = () => 'messages' === dragData?.action,
		dragSortable = () => 'sortable' === dragData?.action,
		setDragAction = (e, action, effect, data, img) => {
			dragData = {
				action: action,
				data: data
			};
	//		e.dataTransfer.setData(rlContentType, action);
			e.dataTransfer.setData('text/plain', rlContentType+'/'+action);
			e.dataTransfer.setDragImage(img, 0, 0);
			e.dataTransfer.effectAllowed = effect;
		},

		dragTimer = {
			id: 0,
			stop: () => clearTimeout(dragTimer.id),
			start: fn => dragTimer.id = setTimeout(fn, 500)
		},

		ttn = (element, fValueAccessor) => timeToNode(element, ko.unwrap(fValueAccessor()));

	let dragImage,
		dragData;

	Object.assign(ko.bindingHandlers, {

		editor: {
			init: (element, fValueAccessor) => {
				let editor = null;

				const fValue = fValueAccessor(),
					fUpdateEditorValue = () => fValue?.__editor?.setHtmlOrPlain(fValue()),
					fUpdateKoValue = () => fValue?.__editor && fValue(fValue.__editor.getDataWithHtmlMark()),
					fOnReady = () => {
						fValue.__editor = editor;
						fUpdateEditorValue();
					};

				if (ko.isObservable(fValue)) {
					editor = new HtmlEditor(element, fUpdateKoValue, fOnReady, fUpdateKoValue);

					fValue.__fetchEditorValue = fUpdateKoValue;

					fValue.subscribe(fUpdateEditorValue);

					// ko.utils.domNodeDisposal.addDisposeCallback(element, () => {
					// });
				}
			}
		},

		moment: {
			init: ttn,
			update: ttn
		},

		emailsTags: {
			init: (element, fValueAccessor, fAllBindings) => {
				const fValue = fValueAccessor(),
					focused = fValue.focused;

				element.addresses = new EmailAddressesComponent(element, {
					focusCallback: value => focused?.(!!value),
					autoCompleteSource: fAllBindings.get('autoCompleteSource'),
					onChange: value => fValue(value)
				});

				focused?.subscribe(value =>
					element.addresses[value ? 'focus' : 'blur']()
				);
			},
			update: (element, fValueAccessor) => {
				element.addresses.value = ko.unwrap(fValueAccessor());
			}
		},

		// Start dragging checked messages
		dragmessages: {
			init: element => {
				element.addEventListener("dragstart", e => {
					dragImage || (dragImage = elementById('messagesDragImage'));
					if (dragImage && !ThemeStore.isMobile()) {
						ko.dataFor(doc.elementFromPoint(e.clientX, e.clientY))?.checked?.(true);

						const uids = MessagelistUserStore.listCheckedOrSelectedUidsWithSubMails();
						dragImage.querySelector('.text').textContent = uids.size;

						// Make sure Chrome shows it
						dragImage.style.left = e.clientX + 'px';
						dragImage.style.top = e.clientY + 'px';
						dragImage.style.right = 'auto';

						setDragAction(e, 'messages', 'copyMove', uids, dragImage);

						// Remove the Chrome visibility
						dragImage.style.cssText = '';
					} else {
						e.preventDefault();
					}

				}, false);
				element.addEventListener("dragend", () => dragData = null);
			}
		},

		// Drop selected messages on folder
		dropmessages: {
			init: (element, fValueAccessor) => {
				const folder = fValueAccessor(),
		//			folder = ko.dataFor(element),
					fnStop = e => {
						e.preventDefault();
						element.classList.remove('droppableHover');
						dragTimer.stop();
					},
					fnHover = e => {
						if (dragMessages()) {
							fnStop(e);
							e.dataTransfer.dropEffect = e.ctrlKey ? 'copy' : 'move';
							element.classList.add('droppableHover');
							if (folder.collapsed()) {
								dragTimer.start(() => {
									folder.collapsed(false);
									setExpandedFolder(folder.fullName, true);
								}, 500);
							}
						}
					};
				folder && addEventsListeners(element, {
					dragenter: fnHover,
					dragover: fnHover,
					dragleave: fnStop,
					drop: e => {
						fnStop(e);
						if (dragMessages() && 'copyMove' == e.dataTransfer.effectAllowed) {
							moveMessagesToFolder(FolderUserStore.currentFolderFullName(), dragData.data, folder.fullName, e.ctrlKey);
						}
					}
				});
			}
		},

		sortableItem: {
			init: (element, fValueAccessor) => {
				let options = ko.unwrap(fValueAccessor()) || {},
					parent = element.parentNode,
					fnHover = e => {
						if (dragSortable()) {
							e.preventDefault();
							let node = (e.target.closest ? e.target : e.target.parentNode).closest('[draggable]');
							if (node && node !== dragData.data && parent.contains(node)) {
								let rect = node.getBoundingClientRect();
								if (rect.top + (rect.height / 2) <= e.clientY) {
									if (node.nextElementSibling !== dragData.data) {
										node.after(dragData.data);
									}
								} else if (node.previousElementSibling !== dragData.data) {
									node.before(dragData.data);
								}
							}
						}
					};
				addEventsListeners(element, {
					dragstart: e => {
						dragData = {
							action: 'sortable',
							element: element
						};
						setDragAction(e, 'sortable', 'move', element, element);
						element.style.opacity = 0.25;
					},
					dragend: () => {
						element.style.opacity = null;
						if (dragSortable()) {
							dragData.data.style.cssText = '';
							let row = parent.rows[options.list.indexOf(ko.dataFor(element))];
							if (row != dragData.data) {
								row.before(dragData.data);
							}
							dragData = null;
						}
					}
				});
				if (!parent.sortable) {
					parent.sortable = true;
					addEventsListeners(parent, {
						dragenter: fnHover,
						dragover: fnHover,
						drop: e => {
							if (dragSortable()) {
								e.preventDefault();
								let data = ko.dataFor(dragData.data),
									from = options.list.indexOf(data),
									to = [...parent.children].indexOf(dragData.data);
								if (from != to) {
									let arr = options.list();
									arr.splice(to, 0, ...arr.splice(from, 1));
									options.list(arr);
								}
								dragData = null;
								options.afterMove?.();
							}
						}
					});
				}
			}
		},

		initDom: {
			init: (element, fValueAccessor) => fValueAccessor()(element)
		},

		registerBootstrapDropdown: {
			init: element => {
				dropdowns.push(element);
				element.ddBtn = new BSN.Dropdown(element.querySelector('.dropdown-toggle'));
			}
		},

		openDropdownTrigger: {
			update: (element, fValueAccessor) => {
				if (ko.unwrap(fValueAccessor())) {
					const el = element.ddBtn;
					el.open || el.toggle();
		//			el.focus();

					dropdownsDetectVisibility();
					fValueAccessor()(false);
				}
			}
		}
	});

	const ContactUserStore = koArrayWithDestroy();

	ContactUserStore.loading = ko.observable(false).extend({ debounce: 200 });
	ContactUserStore.importing = ko.observable(false).extend({ debounce: 200 });
	ContactUserStore.syncing = ko.observable(false).extend({ debounce: 200 });

	addObservablesTo(ContactUserStore, {
		allowSync: false, // Admin setting
		syncMode: 0,
		syncUrl: '',
		syncUser: '',
		syncPass: ''
	});

	// Also used by Selector
	ContactUserStore.hasChecked = koComputable(
		// Issue: not all are observed?
		() => !!ContactUserStore.find(item => item.checked())
	);

	/**
	 * @param {Function} fResultFunc
	 * @returns {void}
	 */
	ContactUserStore.sync = fResultFunc => {
		if (ContactUserStore.syncMode()
		 && !ContactUserStore.importing()
		 && !ContactUserStore.syncing()
		) {
			ContactUserStore.syncing(true);
			Remote.streamPerLine(line => {
				try {
					line = JSON.parse(line);
					if ('ContactsSync' === line.Action) {
						ContactUserStore.syncing(false);
						fResultFunc?.(line.ErrorCode, line);
					}
				} catch (e) {
					ContactUserStore.syncing(false);
					console.error(e);
					fResultFunc?.(Notification.UnknownError);
				}
			}, 'ContactsSync');
		}
	};

	ContactUserStore.init = () => {
		let value = !!SettingsGet('ContactsSyncIsAllowed');
		ContactUserStore.allowSync(value);
		if (value) {
			ContactUserStore.syncMode(SettingsGet('ContactsSyncMode'));
			ContactUserStore.syncUrl(SettingsGet('ContactsSyncUrl'));
			ContactUserStore.syncUser(SettingsGet('ContactsSyncUser'));
			ContactUserStore.syncPass(SettingsGet('ContactsSyncPassword'));
			setTimeout(ContactUserStore.sync, 10000);
			value = pInt(SettingsGet('ContactsSyncInterval'));
			value = 5 <= value ? (320 >= value ? value : 320) : 20;
			setInterval(ContactUserStore.sync, value * 60000 + 5000);
		}
	};

	const IdentityUserStore = koArrayWithDestroy();

	IdentityUserStore.loading = ko.observable(false).extend({ debounce: 100 });

	let
		SCREENS = {},
		currentScreen = null,
		defaultScreenName = '';

	const
		autofocus = dom => dom.querySelector('[autofocus]')?.focus(),

		visiblePopups = new Set,

		/**
		 * @param {string} screenName
		 * @returns {?Object}
		 */
		screen = screenName => (screenName && SCREENS[screenName]) || null,

		/**
		 * @param {Function} ViewModelClass
		 * @param {Object=} vmScreen
		 * @returns {*}
		 */
		buildViewModel = (ViewModelClass, vmScreen) => {
			if (ViewModelClass && !ViewModelClass.__builded) {
				let vmDom = null;
				const
					vm = new ViewModelClass(vmScreen),
					id = vm.viewModelTemplateID,
					position = vm.viewType || '',
					dialog = ViewTypePopup === position,
					vmPlace = position ? doc.getElementById('rl-' + position.toLowerCase()) : null;

				ViewModelClass.__builded = true;
				ViewModelClass.__vm = vm;

				if (vmPlace) {
					vmDom = Element.fromHTML(dialog
						? '<dialog id="V-'+ id + '"></dialog>'
						: '<div id="V-'+ id + '" hidden=""></div>');
					vmPlace.append(vmDom);

					vm.viewModelDom = vmDom;
					ViewModelClass.__dom = vmDom;

					if (ViewTypePopup === position) {
						vm.close = () => hideScreenPopup(ViewModelClass);

						// Firefox < 98 / Safari < 15.4 HTMLDialogElement not defined
						if (!vmDom.showModal) {
							vmDom.showModal = () => {
								vmDom.backdrop ||
									vmDom.before(vmDom.backdrop = Element.fromHTML('<div class="dialog-backdrop"></div>'));
								vmDom.setAttribute('open','');
								vmDom.open = true;
								vmDom.returnValue = null;
								vmDom.backdrop.hidden = false;
							};
							vmDom.close = v => {
	//							if (vmDom.dispatchEvent(new CustomEvent('cancel', {cancelable:true}))) {
									vmDom.backdrop.hidden = true;
									vmDom.returnValue = v;
									vmDom.removeAttribute('open', null);
									vmDom.open = false;
	//								vmDom.dispatchEvent(new CustomEvent('close'));
	//							}
							};
						}
						// https://developer.mozilla.org/en-US/docs/Web/API/HTMLDialogElement/cancel_event
	//					vmDom.addEventListener('cancel', event => (false === vm.onClose() && event.preventDefault()));
	//					vmDom.addEventListener('close', () => vm.modalVisible(false));

						// show/hide popup/modal
						const endShowHide = e => {
							if (e.target === vmDom) {
								if (vmDom.classList.contains('animate')) {
									autofocus(vmDom);
									vm.afterShow?.();
								} else {
									vmDom.close();
									vm.afterHide?.();
								}
							}
						};

						vm.modalVisible.subscribe(value => {
							if (value) {
								i18nToNodes(vmDom);
								visiblePopups.add(vm);
								vmDom.style.zIndex = 3001 + (visiblePopups.size * 2);
								vmDom.showModal();
								if (vmDom.backdrop) {
									vmDom.backdrop.style.zIndex = 3000 + (visiblePopups.size * 2);
								}
								vm.keyScope.set();
								requestAnimationFrame(() => { // wait just before the next paint
									vmDom.offsetHeight; // force a reflow
									vmDom.classList.add('animate'); // trigger the transitions
								});
							} else {
								visiblePopups.delete(vm);
								vm.onHide?.();
								vm.keyScope.unset();
								vmDom.classList.remove('animate'); // trigger the transitions
							}
							arePopupsVisible(0 < visiblePopups.size);
						});
						vmDom.addEventListener('transitionend', endShowHide);
					}

					ko.applyBindingAccessorsToNode(
						vmDom,
						{
							template: () => ({ name: id })
						},
						vm
					);

					vm.onBuild?.(vmDom);

					fireEvent('rl-view-model', vm);
				} else {
					console.log('Cannot find view model position: ' + position);
				}
			}

			return ViewModelClass && ViewModelClass.__vm;
		},

		forEachViewModel = (screen, fn) => {
			screen.viewModels.forEach(ViewModelClass => {
				if (
					ViewModelClass.__vm &&
					ViewModelClass.__dom &&
					ViewTypePopup !== ViewModelClass.__vm.viewType
				) {
					fn(ViewModelClass.__vm, ViewModelClass.__dom);
				}
			});
		},

		hideScreen = (screenToHide, destroy) => {
			screenToHide.onHide?.();
			forEachViewModel(screenToHide, (vm, dom) => {
				dom.hidden = true;
				vm.onHide?.();
				destroy && vm.viewModelDom.remove();
			});
		},

		/**
		 * @param {Function} ViewModelClassToHide
		 * @returns {void}
		 */
		hideScreenPopup = ViewModelClassToHide => {
			if (ViewModelClassToHide?.__vm && ViewModelClassToHide?.__dom) {
				ViewModelClassToHide.__vm.modalVisible(false);
			}
		},

		/**
		 * @param {string} screenName
		 * @param {string} subPart
		 * @returns {void}
		 */
		screenOnRoute = (screenName, subPart) => {
			let vmScreen = null,
				isSameScreen = false;

			if (null == screenName || '' == screenName) {
				screenName = defaultScreenName;
			}

			if (fireEvent('sm-show-screen', screenName, 1)) {

				// Close all popups
				for (let vm of visiblePopups) {
					(false === vm.onClose()) || vm.close();
				}

				if (screenName) {
					vmScreen = screen(screenName);
					if (!vmScreen) {
						vmScreen = screen(defaultScreenName);
						if (vmScreen) {
							subPart = screenName + '/' + subPart;
							screenName = defaultScreenName;
						}
					}

					if (vmScreen?.__started) {
						isSameScreen = currentScreen && vmScreen === currentScreen;

						if (!vmScreen.__builded) {
							vmScreen.__builded = true;

							vmScreen.viewModels.forEach(ViewModelClass =>
								buildViewModel(ViewModelClass, vmScreen)
							);

							vmScreen.onBuild?.();
						}

						setTimeout(() => {
							// hide screen
							if (currentScreen && !isSameScreen) {
								hideScreen(currentScreen);
							}
							// --

							currentScreen = vmScreen;

							// show screen
							if (!isSameScreen) {
								vmScreen.onShow?.();

								forEachViewModel(vmScreen, (vm, dom) => {
									vm.beforeShow?.();
									i18nToNodes(dom);
									dom.hidden = false;
									vm.onShow?.();
									autofocus(dom);
								});
							}
							// --

							vmScreen.__cross?.parse(subPart);
						}, 1);
					}
				}
			}
		};


	const
		ViewTypePopup = 'Popups',

		/**
		 * @param {Function} ViewModelClassToShow
		 * @param {Array=} params
		 * @returns {void}
		 */
		showScreenPopup = (ViewModelClassToShow, params = []) => {
			const vm = buildViewModel(ViewModelClassToShow) && ViewModelClassToShow.__dom && ViewModelClassToShow.__vm;

			if (vm) {
				params = params || [];

				vm.beforeShow?.(...params);

				vm.modalVisible(true);

				vm.onShow?.(...params);
			}
		},

		arePopupsVisible = ko.observable(false),

		/**
		 * @param {Array} screensClasses
		 * @returns {void}
		 */
		startScreens = screensClasses => {
			hasher.clear();
			forEachObjectValue(SCREENS, screen => hideScreen(screen, 1));
			SCREENS = {};
			currentScreen = null,
			defaultScreenName = '';

			screensClasses.forEach(CScreen => {
				if (CScreen) {
					const vmScreen = new CScreen(),
						screenName = vmScreen.screenName;
					defaultScreenName || (defaultScreenName = screenName);
					SCREENS[screenName] = vmScreen;
				}
			});

			forEachObjectValue(SCREENS, vmScreen => {
				if (!vmScreen.__started) {
					vmScreen.onStart();
					vmScreen.__started = true;
				}
			});

			const cross = new Crossroads();
			cross.addRoute(/^([a-zA-Z0-9-]*)\/?(.*)$/, screenOnRoute);

			hasher.add(cross.parse.bind(cross));
			hasher.init();

			setTimeout(() => $htmlCL.remove('rl-started-trigger'), 100);

			const c = elementById('rl-content'), l = elementById('rl-loading');
			c && (c.hidden = false);
			l?.remove();
		},

		/**
		 * Used by ko.bindingHandlers.command (template data-bind="command: ")
		 * to enable/disable click/submit action.
		 */
		decorateKoCommands = (thisArg, commands) =>
			forEachObjectEntry(commands, (key, canExecute) => {
				let command = thisArg[key],
					fn = (...args) => fn.canExecute() && command.apply(thisArg, args);

				fn.canExecute = koComputable(() => canExecute.call(thisArg, thisArg));

				thisArg[key] = fn;
			});

	ko.decorateCommands = decorateKoCommands;

	class AbstractView {
		constructor(templateID, type)
		{
	//		Object.defineProperty(this, 'viewModelTemplateID', { value: templateID });
			this.viewModelTemplateID = templateID || this.constructor.name.replace('UserView', '');
			this.viewType = type;
			this.viewModelDom = null;

			this.keyScope = {
				scope: 'none',
				previous: 'none',
				set: function() {
					this.previous = keyScope();
					keyScope(this.scope);
				},
				unset: function() {
					keyScope(this.previous);
				}
			};
		}

	/*
		onBuild() {}
		beforeShow() {} // Happens before: hidden = false
		onShow() {}       // Happens after: hidden = false
		onHide() {}
	*/

		querySelector(selectors) {
			return this.viewModelDom.querySelector(selectors);
		}

		addObservables(observables) {
			addObservablesTo(this, observables);
		}

		addComputables(computables) {
			addComputablesTo(this, computables);
		}

		addSubscribables(subscribables) {
			addSubscribablesTo(this, subscribables);
		}

	}

	class AbstractViewPopup extends AbstractView
	{
		constructor(name)
		{
			super('Popups' + name, ViewTypePopup);
			this.keyScope.scope = name;
			this.modalVisible = ko.observable(false).extend({ rateLimit: 0 });
			shortcuts.add('escape,close', '', name, () => {
				if (this.modalVisible() && false !== this.onClose()) {
					this.close();
				}
				return false;
	//			return true; Issue with supported modal close
			});
		}

		// Happens when user hits Escape or Close key
		// return false to prevent closing
		onClose() {}

	/*
		beforeShow() {} // Happens before showModal()
		onShow() {}     // Happens after  showModal()
		afterShow() {}  // Happens after  showModal() animation transitionend
		onHide() {}     // Happens before animation transitionend
		afterHide() {}  // Happens after  animation transitionend

		close() {}
	*/
	}

	AbstractViewPopup.showModal = function(params = []) {
		showScreenPopup(this, params);
	};

	AbstractViewPopup.hidden = function() {
		return !this.__vm || !this.__vm.modalVisible();
	};

	class AbstractViewLeft extends AbstractView
	{
		constructor(templateID)
		{
			super(templateID, 'Left');
			this.leftPanelDisabled = leftPanelDisabled;
		}
	}

	class AbstractViewRight extends AbstractView
	{
		constructor(templateID)
		{
			super(templateID, 'Right');
		}
	}

	class AbstractViewSettings
	{
	/*
		onBuild(viewModelDom) {}
		beforeShow() {}
		onShow() {}
		onHide() {}
		viewModelDom
	*/
		addSetting(name, valueCb)
		{
			let prop = name[0].toLowerCase() + name.slice(1),
				trigger = prop + 'Trigger';
			addObservablesTo(this, {
				[prop]: SettingsGet(name),
				[trigger]: SaveSettingStatus.Idle,
			});
			addSubscribablesTo(this, {
				[prop]: (value => {
					this[trigger](SaveSettingStatus.Saving);
					valueCb?.(value);
					rl.app.Remote.saveSetting(name, value,
						iError => {
							this[trigger](iError ? SaveSettingStatus.Failed : SaveSettingStatus.Success);
							setTimeout(() => this[trigger](SaveSettingStatus.Idle), 1000);
						}
					);
				}).debounce(999),
			});
		}

		addSettings(names)
		{
			names.forEach(name => {
				let prop = name[0].toLowerCase() + name.slice(1);
				this[prop] || (this[prop] = ko.observable(SettingsGet(name)));
				this[prop].subscribe(value => rl.app.Remote.saveSetting(name, value));
			});
		}
	}

	class AbstractViewLogin extends AbstractView {
		constructor(templateID) {
			super(templateID, 'Content');
			this.formError = ko.observable(false).extend({ falseTimeout: 500 });
		}

		onBuild(dom) {
			dom.classList.add('LoginView');
		}

		onShow() {
			elementById('rl-left').hidden = true;
			elementById('rl-right').hidden = true;
			rl.route.off();
		}

		onHide() {
			elementById('rl-left').hidden = false;
			elementById('rl-right').hidden = false;
		}

		submitForm() {
	//		return false;
		}
	}

	class OpenPgpKeyPopupView extends AbstractViewPopup {
		constructor() {
			super('OpenPgpKey');

			this.addObservables({
				key: '',
				keyDom: null
			});
		}

		selectKey() {
			const el = this.keyDom();
			if (el) {
				let sel = getSelection(),
					range = doc.createRange();
				sel.removeAllRanges();
				range.selectNodeContents(el);
				sel.addRange(range);
			}
			if (navigator.clipboard) {
				navigator.clipboard.writeText(this.key()).then(
					() => console.log('Copied to clipboard'),
					err => console.error(err)
				);
			}
		}

		onShow(openPgpKey) {
			// TODO: show more info
			this.key(openPgpKey ? openPgpKey.armor : '');
	/*
			this.key = key;
			const aEmails = [];
			if (key.users) {
				key.users.forEach(user => user.userID.email && aEmails.push(user.userID.email));
			}
			this.id = key.getKeyID().toHex();
			this.fingerprint = key.getFingerprint();
			this.can_encrypt = !!key.getEncryptionKey();
			this.can_sign = !!key.getSigningKey();
			this.emails = aEmails;
			this.armor = armor;
			this.askDelete = ko.observable(false);
			this.openForDeletion = ko.observable(null).askDeleteHelper();

			key.id = key.subkeys[0].keyid;
			key.fingerprint = key.subkeys[0].fingerprint;
			key.uids.forEach(uid => uid.email && aEmails.push(uid.email));
			key.emails = aEmails;
			"disabled": false,
			"expired": false,
			"revoked": false,
			"is_secret": true,
			"can_sign": true,
			"can_decrypt": true
			"can_verify": true
			"can_encrypt": true,
			"uids": [
				{
					"name": "demo",
					"comment": "",
					"email": "demo@snappymail.eu",
					"uid": "demo <demo@snappymail.eu>",
					"revoked": false,
					"invalid": false
				}
			],
			"subkeys": [
				{
					"fingerprint": "2C223F20EA2ADB4CB68F81D95F3A5CDC09AD8AE3",
					"keyid": "5F3A5CDC09AD8AE3",
					"timestamp": 1643381672,
					"expires": 0,
					"is_secret": false,
					"invalid": false,
					"can_encrypt": false,
					"can_sign": true,
					"disabled": false,
					"expired": false,
					"revoked": false,
					"can_certify": true,
					"can_authenticate": false,
					"is_qualified": false,
					"is_de_vs": false,
					"pubkey_algo": 303,
					"length": 256,
					"keygrip": "5A1A6C7310D0508C68E8E74F15068301E83FD1AE",
					"is_cardkey": false,
					"curve": "ed25519"
				},
				{
					"fingerprint": "3CD720549D8833872C267D08F1230DCE2A561ADE",
					"keyid": "F1230DCE2A561ADE",
					"timestamp": 1643381672,
					"expires": 0,
					"is_secret": false,
					"invalid": false,
					"can_encrypt": true,
					"can_sign": false,
					"disabled": false,
					"expired": false,
					"revoked": false,
					"can_certify": false,
					"can_authenticate": false,
					"is_qualified": false,
					"is_de_vs": false,
					"pubkey_algo": 302,
					"length": 256,
					"keygrip": "886921A7E06BE56F8E8C51797BB476BB26DF21BF",
					"is_cardkey": false,
					"curve": "cv25519"
				}
			]
	*/
		}

		onBuild() {
			addShortcut('a', 'meta', 'OpenPgpKey', () => {
				this.selectKey();
				return false;
			});
		}
	}

	class AskPopupView extends AbstractViewPopup {
		constructor() {
			super('Ask');

			this.addObservables({
				askDesc: '',
				yesButton: '',
				noButton: '',
				username: '',
				askUsername: false,
				passphrase: '',
				askPass: false
			});

			this.fYesAction = null;
			this.fNoAction = null;

			this.focusOnShow = true;
		}

		yesClick() {
			this.close();

			isFunction(this.fYesAction) && this.fYesAction();
		}

		noClick() {
			this.close();

			isFunction(this.fNoAction) && this.fNoAction();
		}

		/**
		 * @param {string} sAskDesc
		 * @param {Function=} fYesFunc
		 * @param {Function=} fNoFunc
		 * @param {boolean=} focusOnShow = true
		 * @returns {void}
		 */
		onShow(sAskDesc, fYesFunc = null, fNoFunc = null, focusOnShow = true, ask = 0, btnText = '') {
			this.askDesc(sAskDesc || '');
			this.askUsername(ask & 2);
			this.askPass(ask & 1);
			this.username('');
			this.passphrase('');
			this.yesButton(i18n(btnText || 'GLOBAL/YES'));
			this.noButton(i18n(ask ? 'GLOBAL/CANCEL' : 'GLOBAL/NO'));
			this.fYesAction = fYesFunc;
			this.fNoAction = fNoFunc;
			this.focusOnShow = focusOnShow
				? (ask ? 'input[type="'+(ask&2?'text':'password')+'"]' : '.buttonYes')
				: '';
		}

		afterShow() {
			this.focusOnShow && this.querySelector(this.focusOnShow).focus();
		}

		onClose() {
			this.noClick();
			return false;
		}

		onBuild() {
	//		shortcuts.add('tab', 'shift', 'Ask', () => {
			shortcuts.add('tab,arrowright,arrowleft', '', 'Ask', () => {
				let yes = this.querySelector('.buttonYes'),
					no = this.querySelector('.buttonNo');
				if (yes.matches(':focus')) {
					no.focus();
					return false;
				} else if (no.matches(':focus')) {
					yes.focus();
					return false;
				}
			});
		}
	}

	AskPopupView.password = function(sAskDesc, btnText) {
		return new Promise(resolve => {
			this.showModal([
				sAskDesc,
				() => resolve(this.__vm.passphrase()),
				() => resolve(null),
				true,
				1,
				btnText
			]);
		});
	};

	AskPopupView.credentials = function(sAskDesc, btnText) {
		return new Promise(resolve => {
			this.showModal([
				sAskDesc,
				() => resolve({username:this.__vm.username(), password:this.__vm.passphrase()}),
				() => resolve(null),
				true,
				3,
				btnText
			]);
		});
	};

	const
		askPassphrase = async (privateKey, btnTxt = 'LABEL_SIGN') =>
			await AskPopupView.password('GnuPG key<br>' + privateKey.id + ' ' + privateKey.emails[0], 'OPENPGP/'+btnTxt),

		findGnuPGKey = (keys, query, sign) =>
			keys.find(key =>
				key[sign ? 'can_sign' : 'can_decrypt']
				&& (key.emails.includes(query) || key.subkeys.find(key => query == key.keyid || query == key.fingerprint))
			);

	const GnuPGUserStore = new class {
		constructor() {
			/**
			 * PECL gnupg / PEAR Crypt_GPG
			 * [ {email, can_encrypt, can_sign}, ... ]
			 */
			this.keyring;
			this.publicKeys = ko.observableArray();
			this.privateKeys = ko.observableArray();
		}

		loadKeyrings() {
			this.keyring = null;
			this.publicKeys([]);
			this.privateKeys([]);
			Remote.request('GnupgGetKeys',
				(iError, oData) => {
					if (oData?.Result) {
						this.keyring = oData.Result;
						const initKey = (key, isPrivate) => {
							const aEmails = [];
							key.id = key.subkeys[0].keyid;
							key.fingerprint = key.subkeys[0].fingerprint;
							key.uids.forEach(uid => uid.email && aEmails.push(uid.email));
							key.emails = aEmails;
							key.askDelete = ko.observable(false);
							key.openForDeletion = ko.observable(null).askDeleteHelper();
							key.remove = () => {
								if (key.askDelete()) {
									Remote.request('GnupgDeleteKey',
										(iError, oData) => {
											if (oData) {
												if (iError) {
													alert(oData.ErrorMessage);
												} else if (oData.Result) {
													if (isPrivate) {
														this.privateKeys.remove(key);
													} else {
														this.publicKeys.remove(key);
													}
												}
											}
										}, {
											KeyId: key.id,
											isPrivate: isPrivate
										}
									);
								}
							};
							key.view = () => {
								const fetch = pass => Remote.request('GnupgExportKey',
										(iError, oData) => {
											if (oData?.Result) {
												key.armor = oData.Result;
												showScreenPopup(OpenPgpKeyPopupView, [key]);
											}
										}, {
											KeyId: key.id,
											isPrivate: isPrivate,
											Passphrase: pass
										}
									);
								if (isPrivate) {
									askPassphrase(key, 'POPUP_VIEW_TITLE').then(passphrase => {
										(null !== passphrase) && fetch(passphrase);
									});
								} else {
									fetch('');
								}
							};
							return key;
						};
						this.publicKeys(oData.Result.public.map(key => initKey(key, 0)));
						this.privateKeys(oData.Result.private.map(key => initKey(key, 1)));
						console.log('gnupg ready');
					}
				}
			);
		}

		/**
		 * @returns {boolean}
		 */
		isSupported() {
			return SettingsCapa('GnuPG');
		}

		importKey(key, callback) {
			Remote.request('GnupgImportKey',
				(iError, oData) => {
					if (oData?.Result/* && (oData.Result.imported || oData.Result.secretimported)*/) {
						this.loadKeyrings();
					}
					callback?.(iError, oData);
				}, {
					Key: key
				}
			);
		}

		/**
			keyPair.privateKey
			keyPair.publicKey
			keyPair.revocationCertificate
			keyPair.onServer
			keyPair.inGnuPG
		 */
		storeKeyPair(keyPair, callback) {
			Remote.request('PgpStoreKeyPair',
				(iError, oData) => {
					if (oData?.Result) ;
					callback?.(iError, oData);
				}, keyPair
			);
		}

		/**
		 * Checks if verifying/encrypting a message is possible with given email addresses.
		 */
		hasPublicKeyForEmails(recipients) {
			const count = recipients.length,
				length = count ? recipients.filter(email =>
	//				(key.can_verify || key.can_encrypt) &&
					this.publicKeys.find(key => key.emails.includes(email))
				).length : 0;
			return length && length === count;
		}

		getPublicKeyFingerprints(recipients) {
			const fingerprints = [];
			recipients.forEach(email => {
				fingerprints.push(this.publicKeys.find(key => key.emails.includes(email)).fingerprint);
			});
			return fingerprints;
		}

		getPrivateKeyFor(query, sign) {
			return findGnuPGKey(this.privateKeys, query, sign);
		}

		async decrypt(message) {
			const
				pgpInfo = message.pgpEncrypted();
			if (pgpInfo) {
				let ids = [message.to[0].email].concat(pgpInfo.KeyIds),
					i = ids.length, key;
				while (i--) {
					key = findGnuPGKey(this.privateKeys, ids[i]);
					if (key) {
						break;
					}
				}
				if (key) {
					// Also check message.from[0].email
					let params = {
						Folder: message.folder,
						Uid: message.uid,
						PartId: pgpInfo.PartId,
						KeyId: key.id,
						Passphrase: await askPassphrase(key, 'BUTTON_DECRYPT'),
						Data: '' // message.plain() optional
					};
					if (null !== params.Passphrase) {
						const result = await Remote.post('GnupgDecrypt', null, params);
						if (result?.Result) {
							return result.Result;
						}
					}
				}
			}
		}

		async verify(message) {
			let data = message.pgpSigned(); // { BodyPartId: "1", SigPartId: "2", MicAlg: "pgp-sha256" }
			if (data) {
				data = { ...data }; // clone
	//			const sender = message.from[0].email;
	//			let mode = await this.hasPublicKeyForEmails([sender]);
				data.Folder = message.folder;
				data.Uid = message.uid;
				if (data.BodyPart) {
					data.BodyPart = data.BodyPart.raw;
					data.SigPart = data.SigPart.body;
				}
				let response = await Remote.post('MessagePgpVerify', null, data);
				if (response?.Result) {
					return {
						fingerprint: response.Result.fingerprint,
						success: 0 == response.Result.status // GOODSIG
					};
				}
			}
		}

		async sign(privateKey) {
			return await askPassphrase(privateKey);
		}

	};

	/**
	 * OpenPGP.js
	 */

	const
		findOpenPGPKey = (keys, query/*, sign*/) =>
			keys.find(key =>
				key.emails.includes(query) || query == key.id || query == key.fingerprint
			),

		decryptKey = async (privateKey, btnTxt = 'LABEL_SIGN') => {
			if (privateKey.key.isDecrypted()) {
				return privateKey.key;
			}
			const passphrase = await AskPopupView.password(
				'OpenPGP.js key<br>' + privateKey.id + ' ' + privateKey.emails[0],
				'OPENPGP/'+btnTxt
			);
			if (null !== passphrase) {
				return await openpgp.decryptKey({
					privateKey: privateKey.key,
					passphrase
				});
			}
		},

		/**
		 * OpenPGP.js v5 removed the localStorage (keyring)
		 * This should be compatible with the old OpenPGP.js v2
		 */
		publicKeysItem = 'openpgp-public-keys',
		privateKeysItem = 'openpgp-private-keys',
		storage = window.localStorage,
		loadOpenPgpKeys = async itemname => {
			let keys = [], key,
				armoredKeys = JSON.parse(storage.getItem(itemname)),
				i = arrayLength(armoredKeys);
			while (i--) {
				key = await openpgp.readKey({armoredKey:armoredKeys[i]});
				if (!key.err) {
					keys.push(new OpenPgpKeyModel(armoredKeys[i], key));
				}
			}
			return keys;
		},
		storeOpenPgpKeys = (keys, section) => {
			let armoredKeys = keys.map(item => item.armor);
			if (armoredKeys.length) {
				storage.setItem(section, JSON.stringify(armoredKeys));
			} else {
				storage.removeItem(section);
			}
		};

	class OpenPgpKeyModel {
		constructor(armor, key) {
			this.key = key;
			const aEmails = [];
			if (key.users) {
				key.users.forEach(user => user.userID.email && aEmails.push(user.userID.email));
			}
			this.id = key.getKeyID().toHex().toUpperCase();
			this.fingerprint = key.getFingerprint();
			this.can_encrypt = !!key.getEncryptionKey();
			this.can_sign = !!key.getSigningKey();
			this.emails = aEmails;
			this.armor = armor;
			this.askDelete = ko.observable(false);
			this.openForDeletion = ko.observable(null).askDeleteHelper();
	//		key.getUserIDs()
	//		key.getPrimaryUser()
		}

		view() {
			showScreenPopup(OpenPgpKeyPopupView, [this]);
		}

		remove() {
			if (this.askDelete()) {
				if (this.key.isPrivate()) {
					OpenPGPUserStore.privateKeys.remove(this);
					storeOpenPgpKeys(OpenPGPUserStore.privateKeys, privateKeysItem);
				} else {
					OpenPGPUserStore.publicKeys.remove(this);
					storeOpenPgpKeys(OpenPGPUserStore.publicKeys, publicKeysItem);
				}
			}
		}
	/*
		toJSON() {
			return this.armor;
		}
	*/
	}

	const OpenPGPUserStore = new class {
		constructor() {
			this.publicKeys = ko.observableArray();
			this.privateKeys = ko.observableArray();
		}

		loadKeyrings() {
			loadOpenPgpKeys(publicKeysItem).then(keys => {
				this.publicKeys(keys || []);
				console.log('openpgp.js public keys loaded');
			});
			loadOpenPgpKeys(privateKeysItem).then(keys => {
				this.privateKeys(keys || []);
				console.log('openpgp.js private keys loaded');
			});
		}

		/**
		 * @returns {boolean}
		 */
		isSupported() {
			return !!window.openpgp;
		}

		importKey(armoredKey) {
			openpgp.readKey({armoredKey:armoredKey}).then(key => {
				if (!key.err) {
					if (key.isPrivate()) {
						this.privateKeys.push(new OpenPgpKeyModel(armoredKey, key));
						storeOpenPgpKeys(this.privateKeys, privateKeysItem);
					} else {
						this.publicKeys.push(new OpenPgpKeyModel(armoredKey, key));
						storeOpenPgpKeys(this.publicKeys, publicKeysItem);
					}
				}
			});
		}

		/**
			keyPair.privateKey
			keyPair.publicKey
			keyPair.revocationCertificate
		 */
		storeKeyPair(keyPair) {
			openpgp.readKey({armoredKey:keyPair.publicKey}).then(key => {
				this.publicKeys.push(new OpenPgpKeyModel(keyPair.publicKey, key));
				storeOpenPgpKeys(this.publicKeys, publicKeysItem);
			});
			openpgp.readKey({armoredKey:keyPair.privateKey}).then(key => {
				this.privateKeys.push(new OpenPgpKeyModel(keyPair.privateKey, key));
				storeOpenPgpKeys(this.privateKeys, privateKeysItem);
			});
		}

		/**
		 * Checks if verifying/encrypting a message is possible with given email addresses.
		 */
		hasPublicKeyForEmails(recipients) {
			const count = recipients.length,
				length = count ? recipients.filter(email =>
					this.publicKeys().find(key => key.emails.includes(email))
				).length : 0;
			return length && length === count;
		}

		getPrivateKeyFor(query/*, sign*/) {
			return findOpenPGPKey(this.privateKeys, query/*, sign*/);
		}

		/**
		 * https://docs.openpgpjs.org/#encrypt-and-decrypt-string-data-with-pgp-keys
		 */
		async decrypt(armoredText, sender)
		{
			const message = await openpgp.readMessage({ armoredMessage: armoredText }),
				privateKeys = this.privateKeys(),
				msgEncryptionKeyIDs = message.getEncryptionKeyIDs().map(key => key.bytes);
			// Find private key that can decrypt message
			let i = privateKeys.length, privateKey;
			while (i--) {
				if ((await privateKeys[i].key.getDecryptionKeys()).find(
					key => msgEncryptionKeyIDs.includes(key.getKeyID().bytes)
				)) {
					privateKey = privateKeys[i];
					break;
				}
			}
			if (privateKey) try {
				const decryptedKey = await decryptKey(privateKey, 'BUTTON_DECRYPT');
				if (decryptedKey) {
					const publicKey = findOpenPGPKey(this.publicKeys, sender/*, sign*/);
					return await openpgp.decrypt({
						message,
						verificationKeys: publicKey?.key,
	//					expectSigned: true,
	//					signature: '', // Detached signature
						decryptionKeys: decryptedKey
					});
				}
			} catch (err) {
				alert(err);
				console.error(err);
			}
		}

		/**
		 * https://docs.openpgpjs.org/#sign-and-verify-cleartext-messages
		 */
		async verify(message) {
			const data = message.pgpSigned(), // { BodyPartId: "1", SigPartId: "2", MicAlg: "pgp-sha256" }
				publicKey = this.publicKeys().find(key => key.emails.includes(message.from[0].email));
			if (data && publicKey) {
				data.Folder = message.folder;
				data.Uid = message.uid;
				data.GnuPG = 0;
				let response;
				if (data.SigPartId) {
					response = await Remote.post('MessagePgpVerify', null, data);
				} else if (data.BodyPart) {
					response = { Result: { text: data.BodyPart.raw, signature: data.SigPart.body } };
				} else {
					response = { Result: { text: message.plain(), signature: null } };
				}
				if (response) {
					const signature = response.Result.signature
						? await openpgp.readSignature({ armoredSignature: response.Result.signature })
						: null;
					const signedMessage = signature
						? await openpgp.createMessage({ text: response.Result.text })
						: await openpgp.readCleartextMessage({ cleartextMessage: response.Result.text });
	//				(signature||signedMessage).getSigningKeyIDs();
					let result = await openpgp.verify({
						message: signedMessage,
						verificationKeys: publicKey.key,
	//					expectSigned: true, // !!detachedSignature
						signature: signature
					});
					return {
						fingerprint: publicKey.fingerprint,
						success: result && !!result.signatures.length
					};
				}
			}
		}

		/**
		 * https://docs.openpgpjs.org/global.html#sign
		 */
		async sign(text, privateKey, detached) {
			const signingKey = await decryptKey(privateKey);
			if (signingKey) {
				const message = detached
					? await openpgp.createMessage({ text: text })
					: await openpgp.createCleartextMessage({ text: text });
				return await openpgp.sign({
					message: message,
					signingKeys: signingKey,
					detached: !!detached
				});
			}
			throw 'Sign cancelled';
		}

		/**
		 * https://docs.openpgpjs.org/global.html#encrypt
		 */
		async encrypt(text, recipients, signPrivateKey) {
			const count = recipients.length;
			recipients = recipients.map(email => this.publicKeys().find(key => key.emails.includes(email))).filter(key => key);
			if (count === recipients.length) {
				if (signPrivateKey) {
					signPrivateKey = await decryptKey(signPrivateKey);
					if (!signPrivateKey) {
						return;
					}
				}
				return await openpgp.encrypt({
					message: await openpgp.createMessage({ text: text }),
					encryptionKeys: recipients.map(pkey => pkey.key),
					signingKeys: signPrivateKey
	//				signature
				});
			}
			throw 'Encrypt failed';
		}

	};

	const
		BEGIN_PGP_MESSAGE = '-----BEGIN PGP MESSAGE-----',
	//	BEGIN_PGP_SIGNATURE = '-----BEGIN PGP SIGNATURE-----',
	//	BEGIN_PGP_SIGNED = '-----BEGIN PGP SIGNED MESSAGE-----',

		PgpUserStore = new class {
			constructor() {
				// https://mailvelope.github.io/mailvelope/Keyring.html
				this.mailvelopeKeyring = null;
			}

			init() {
				if (SettingsCapa('OpenPGP') && window.crypto && crypto.getRandomValues) {
					rl.loadScript(SettingsGet('StaticLibsJs').replace('/libs.', '/openpgp.'))
	//				rl.loadScript(staticLink('js/min/openpgp.min.js'))
						.then(() => this.loadKeyrings())
						.catch(e => {
							this.loadKeyrings();
							console.error(e);
						});
				} else {
					this.loadKeyrings();
				}
			}

			loadKeyrings(identifier) {
				identifier = identifier || SettingsGet('Email');
				if (window.mailvelope) {
					const fn = keyring => {
							this.mailvelopeKeyring = keyring;
							console.log('mailvelope ready');
						};
					mailvelope.getKeyring().then(fn, err => {
						if (identifier) {
							// attempt to create a new keyring for this app/user
							mailvelope.createKeyring(identifier).then(fn, err => console.error(err));
						} else {
							console.error(err);
						}
					});
					addEventListener('mailvelope-disconnect', event => {
						alert('Mailvelope is updated to version ' + event.detail.version + '. Reload page');
					}, false);
				} else {
					addEventListener('mailvelope', () => this.loadKeyrings(identifier));
				}

				if (OpenPGPUserStore.isSupported()) {
					OpenPGPUserStore.loadKeyrings();
				}

				if (SettingsCapa('GnuPG')) {
					GnuPGUserStore.loadKeyrings();
				}
			}

			/**
			 * @returns {boolean}
			 */
			isSupported() {
				return !!(OpenPGPUserStore.isSupported() || GnuPGUserStore.isSupported() || window.mailvelope);
			}

			/**
			 * @returns {boolean}
			 */
			isEncrypted(text) {
				return 0 === text.trim().indexOf(BEGIN_PGP_MESSAGE);
			}

			async mailvelopeHasPublicKeyForEmails(recipients) {
				const
					keyring = this.mailvelopeKeyring,
					mailvelope = keyring && await keyring.validKeyForAddress(recipients)
						/*.then(LookupResult => Object.entries(LookupResult))*/,
					entries = mailvelope && Object.entries(mailvelope);
				return entries && entries.filter(value => value[1]).length === recipients.length;
			}

			/**
			 * Checks if verifying/encrypting a message is possible with given email addresses.
			 * Returns the first library that can.
			 */
			async hasPublicKeyForEmails(recipients) {
				const count = recipients.length;
				if (count) {
					if (GnuPGUserStore.hasPublicKeyForEmails(recipients)) {
						return 'gnupg';
					}
					if (OpenPGPUserStore.hasPublicKeyForEmails(recipients)) {
						return 'openpgp';
					}
				}
				return false;
			}

			async getMailvelopePrivateKeyFor(email/*, sign*/) {
				let keyring = this.mailvelopeKeyring;
				if (keyring && await keyring.hasPrivateKey({email:email})) {
					return ['mailvelope', email];
				}
				return false;
			}

			/**
			 * Checks if signing a message is possible with given email address.
			 * Returns the first library that can.
			 */
			async getKeyForSigning(email) {
				let key = OpenPGPUserStore.getPrivateKeyFor(email, 1);
				if (key) {
					return ['openpgp', key];
				}

				key = GnuPGUserStore.getPrivateKeyFor(email, 1);
				if (key) {
					return ['gnupg', key];
				}

		//		return await this.getMailvelopePrivateKeyFor(email, 1);
			}

			async decrypt(message) {
				const sender = message.from[0].email,
					armoredText = message.plain();

				if (!this.isEncrypted(armoredText)) {
					return;
				}

				// Try OpenPGP.js
				let result = await OpenPGPUserStore.decrypt(armoredText, sender);
				if (result) {
					return result;
				}

				// Try Mailvelope (does not support inline images)
				try {
					let key = await this.getMailvelopePrivateKeyFor(message.to[0].email);
					if (key) {
						/**
						* https://mailvelope.github.io/mailvelope/Mailvelope.html#createEncryptedFormContainer
						* Creates an iframe to display an encrypted form
						*/
		//				mailvelope.createEncryptedFormContainer('#mailvelope-form');
						/**
						* https://mailvelope.github.io/mailvelope/Mailvelope.html#createDisplayContainer
						* Creates an iframe to display the decrypted content of the encrypted mail.
						*/
						const body = message.body;
						body.textContent = '';
						result = await mailvelope.createDisplayContainer(
							'#'+body.id,
							armoredText,
							this.mailvelopeKeyring,
							{
								senderAddress: sender
							}
						);
						if (result) {
							if (result.error?.message) {
								if ('PWD_DIALOG_CANCEL' !== result.error.code) {
									alert(result.error.code + ': ' + result.error.message);
								}
							} else {
								body.classList.add('mailvelope');
								return true;
							}
						}
					}
				} catch (err) {
					console.error(err);
				}

				// Now try GnuPG
				return GnuPGUserStore.decrypt(message);
			}

			async verify(message) {
				const signed = message.pgpSigned();
				if (signed) {
					const sender = message.from[0].email,
						gnupg = GnuPGUserStore.hasPublicKeyForEmails([sender]),
						openpgp = OpenPGPUserStore.hasPublicKeyForEmails([sender]);
					// Detached signature use GnuPG first, else we must download whole message
					if (gnupg && signed.SigPartId) {
						return GnuPGUserStore.verify(message);
					}
					if (openpgp) {
						return OpenPGPUserStore.verify(message);
					}
					if (gnupg) {
						return GnuPGUserStore.verify(message);
					}
					// Mailvelope can't
					// https://github.com/mailvelope/mailvelope/issues/434
				}
			}

			/**
			 * Returns headers that should be added to an outgoing email.
			 * So far this is only the autocrypt header.
			 */
		/*
			this.mailvelopeKeyring.additionalHeadersForOutgoingEmail(headers)
			this.mailvelopeKeyring.addSyncHandler(syncHandlerObj)
			this.mailvelopeKeyring.createKeyBackupContainer(selector, options)
			this.mailvelopeKeyring.createKeyGenContainer(selector, {
		//		userIds: [],
				keySize: 4096
			})

			this.mailvelopeKeyring.exportOwnPublicKey(emailAddr).then(<AsciiArmored, Error>)
			this.mailvelopeKeyring.importPublicKey(armored)

			// https://mailvelope.github.io/mailvelope/global.html#SyncHandlerObject
			this.mailvelopeKeyring.addSyncHandler({
				uploadSync
				downloadSync
				backup
				restore
			});
		*/

		};

	const LanguageStore = {
		language: ko.observable(''),
		languages: ko.observableArray(),
		userLanguage: ko.observable(''),

		populate: function() {
			const aLanguages = Settings.app('languages');
			this.languages(isArray(aLanguages) ? aLanguages : []);
			this.language(SettingsGet('Language'));
			this.userLanguage(SettingsGet('UserLanguage'));
		}
	};

	class AccountModel extends AbstractModel {
		/**
		 * @param {string} email
		 * @param {boolean=} canBeDelete = true
		 * @param {number=} count = 0
		 */
		constructor(email/*, count = 0*/, isAdditional = true) {
			super();

			this.email = email;

			this.addObservables({
	//			count: count || 0,
				askDelete: false,
				isAdditional: isAdditional
			});
		}

	}

	class IdentityModel extends AbstractModel {
		/**
		 * @param {string} id
		 * @param {string} email
		 */
		constructor(id, email) {
			super();

			this.addObservables({
				id: id || '',
				email: email,
				name: '',

				replyTo: '',
				bcc: '',

				signature: '',
				signatureInsertBefore: false,

				askDelete: false
			});

			this.canBeDeleted = koComputable(() => !!this.id());
		}

		/**
		 * @returns {string}
		 */
		formattedName() {
			const name = this.name(),
				email = this.email();

			return name ? name + ' <' + email + '>' : email;
		}
	}

	class AbstractScreen {
		constructor(screenName, viewModels = []) {
			this.__cross = null;
			this.screenName = screenName;
			this.viewModels = isArray(viewModels) ? viewModels : [];
		}

		/**
		 * @returns {?Array)}
		 */
		routes() {
			return null;
		}

	/*
		onBuild(viewModelDom) {}
		onShow() {}
		onHide() {}
		__started
		__builded
	*/

		/**
		 * @returns {void}
		 */
		onStart() {
			const routes = this.routes();
			if (arrayLength(routes)) {
				let route = new Crossroads(),
					fMatcher = (this.onRoute || (()=>0)).bind(this);

				routes.forEach(item => item && (route.addRoute(item[0], fMatcher).rules = item[1]));

				this.__cross = route;
			}
		}
	}

	class LanguagesPopupView extends AbstractViewPopup {
		constructor() {
			super('Languages');

			this.fLang = null;
			this.userLanguage = ko.observable('');

			this.langs = ko.observableArray();

			this.languages = koComputable(() => {
				const userLanguage = this.userLanguage();
				return this.langs.map(language => ({
					key: language,
					user: language === userLanguage,
					selected: ko.observable(false),
					fullName: convertLangName(language)
				}));
			});

			this.langs.subscribe(() => this.setLanguageSelection());
		}

		languageTooltipName(language) {
			return convertLangName(language, true);
		}

		setLanguageSelection() {
			const currentLang = this.fLang ? ko.unwrap(this.fLang) : '';
			this.languages().forEach(item => item.selected(item.key === currentLang));
		}

		beforeShow() {
			this.fLang = null;
			this.userLanguage('');

			this.langs([]);
		}

		onShow(fLanguage, langs, userLanguage) {
			this.fLang = fLanguage;
			this.userLanguage(userLanguage || '');

			this.langs(langs);
		}

		changeLanguage(lang) {
			this.fLang?.(lang);
			this.close();
		}
	}

	const
		SignMeOff = 0,
		SignMeOn = 1,
		SignMeUnused = 2;


	class LoginUserView extends AbstractViewLogin {
		constructor() {
			super();

			this.addObservables({
				loadingDesc: SettingsGet('LoadingDescription'),

				email: SettingsGet('DevEmail'),
				password: SettingsGet('DevPassword'),
				signMe: false,

				emailError: false,
				passwordError: false,

				submitRequest: false,
				submitError: '',
				submitErrorAdditional: '',

				langRequest: false,

				signMeType: SignMeUnused
			});

			this.allowLanguagesOnLogin = !!SettingsGet('AllowLanguagesOnLogin');

			this.language = LanguageStore.language;
			this.languages = LanguageStore.languages;

			this.bSendLanguage = false;

			this.addComputables({

				languageFullName: () => convertLangName(this.language()),

				signMeVisibility: () => SignMeUnused !== this.signMeType()
			});

			this.addSubscribables({
				email: () => this.emailError(false),

				password: () => this.passwordError(false),

				submitError: value => value || this.submitErrorAdditional(''),

				signMeType: iValue => this.signMe(SignMeOn === iValue),

				language: value => {
					this.langRequest(true);
					translatorReload(false, value).then(
						() => {
							this.langRequest(false);
							this.bSendLanguage = true;
						},
						() => this.langRequest(false)
					);
				}
			});

			if (SettingsGet('AdditionalLoginError') && !this.submitError()) {
				this.submitError(SettingsGet('AdditionalLoginError'));
			}

			decorateKoCommands(this, {
				submitCommand: self => !self.submitRequest()
			});
		}

		submitCommand(self, event) {
			const email = this.email().trim();
			this.email(email);

			let form = event.target.form,
				data = new FormData(form),
				valid = form.reportValidity() && fireEvent('sm-user-login', data, 1);

			this.emailError(!email);
			this.passwordError(!this.password());
			this.formError(!valid);

			if (valid) {
				this.submitRequest(true);
				data.set('Language', this.bSendLanguage ? this.language() : '');
				data.set('SignMe', this.signMe() ? 1 : 0);
				Remote.request('Login',
					(iError, oData) => {
						fireEvent('sm-user-login-response', {
							error: iError,
							data: oData
						});
						if (iError) {
							this.submitRequest(false);
							if (Notification$1.InvalidInputArgument == iError) {
								iError = Notification$1.AuthError;
							}
							this.submitError(getNotification(iError, oData?.ErrorMessage,
								Notification$1.UnknownNotification));
							this.submitErrorAdditional(oData?.ErrorMessageAdditional);
						} else {
							rl.setData(oData.Result);
						}
					},
					data
				);

				set(ClientSideKeyNameLastSignMe, this.signMe() ? '-1-' : '-0-');
			}

			return valid;
		}

		onBuild(dom) {
			super.onBuild(dom);

			const signMe = (SettingsGet('SignMe') || '').toLowerCase();

			switch (signMe) {
				case 'defaultoff':
				case 'defaulton':
					this.signMeType(
						'defaulton' === signMe ? SignMeOn : SignMeOff
					);

					switch (get(ClientSideKeyNameLastSignMe)) {
						case '-1-':
							this.signMeType(SignMeOn);
							break;
						case '-0-':
							this.signMeType(SignMeOff);
							break;
						// no default
					}

					break;
				default:
					this.signMeType(SignMeUnused);
					break;
			}
		}

		selectLanguage() {
			showScreenPopup(LanguagesPopupView, [this.language, this.languages(), LanguageStore.userLanguage()]);
		}
	}

	class LoginUserScreen extends AbstractScreen {
		constructor() {
			super('login', [LoginUserView]);
		}

		onShow() {
			rl.setWindowTitle();
		}
	}

	class KeyboardShortcutsHelpPopupView extends AbstractViewPopup {
		constructor() {
			super('KeyboardShortcutsHelp');
			this.metaKey = shortcuts.getMetaKey();
		}

		onBuild(dom) {
			const tabs = dom.querySelectorAll('.tabs input'),
				last = tabs.length - 1;

	//		addShortcut('tab', 'shift',
			addShortcut('tab,arrowleft,arrowright', '',
				'KeyboardShortcutsHelp',
				event => {
					let next = 0;
					tabs.forEach((node, index) => {
						if (node.matches(':checked')) {
							if (['Tab','ArrowRight'].includes(event.key)) {
								next = index < last ? index+1 : 0;
							} else {
								next = index ? index-1 : last;
							}
						}
					});
					tabs[next].checked = true;
					return false;
				}
			);
		}
	}

	class AccountPopupView extends AbstractViewPopup {
		constructor() {
			super('Account');

			this.addObservables({
				isNew: true,

				email: '',
				password: '',

				submitRequest: false,
				submitError: '',
				submitErrorAdditional: ''
			});
		}

		submitForm(form) {
			if (!this.submitRequest() && form.reportValidity()) {
				const data = new FormData(form);
				data.set('New', this.isNew() ? 1 : 0);
				this.submitRequest(true);
				Remote.request('AccountSetup', (iError, data) => {
						this.submitRequest(false);
						if (iError) {
							this.submitError(getNotification(iError));
							this.submitErrorAdditional(data?.ErrorMessageAdditional);
						} else {
							rl.app.accountsAndIdentities();
							this.close();
						}
					}, data
				);
			}
		}

		onShow(account) {
			if (account?.isAdditional()) {
				this.isNew(false);
				this.email(account.email);
			} else {
				this.isNew(true);
				this.email('');
			}
			this.password('');

			this.submitRequest(false);
			this.submitError('');
			this.submitErrorAdditional('');
		}
	}

	/*
		oCallbacks:
			ItemSelect
			MiddleClick
			AutoSelect
			ItemGetUid
			UpOrDown
	*/

	let shiftStart;

	class Selector {
		/**
		 * @param {koProperty} koList
		 * @param {koProperty} koSelectedItem
		 * @param {koProperty} koFocusedItem
		 * @param {string} sItemSelector
		 * @param {string} sItemCheckedSelector
		 * @param {string} sItemFocusedSelector
		 */
		constructor(
			koList,
			koSelectedItem,
			koFocusedItem,
			sItemSelector,
			sItemCheckedSelector,
			sItemFocusedSelector
		) {
			koFocusedItem = (koFocusedItem || ko.observable(null)).extend({ toggleSubscribeProperty: [this, 'focused'] });
			koSelectedItem = (koSelectedItem || ko.observable(null)).extend({ toggleSubscribeProperty: [null, 'selected'] });

			this.list = koList;
			this.listChecked = koComputable(() => koList.filter(item => item.checked())).extend({ rateLimit: 0 });

			this.focusedItem = koFocusedItem;
			this.selectedItem = koSelectedItem;

			this.iSelectNextHelper = 0;
			this.iFocusedNextHelper = 0;
	//		this.oContentScrollable = null;

			this.sItemSelector = sItemSelector;
			this.sItemCheckedSelector = sItemCheckedSelector;
			this.sItemFocusedSelector = sItemFocusedSelector;

			this.sLastUid = '';
			this.oCallbacks = {};

			const itemSelectedThrottle = (item => this.itemSelected(item)).debounce(300);

			this.listChecked.subscribe(items => {
				if (items.length) {
					if (koSelectedItem()) {
						koSelectedItem(null);
					} else {
						koSelectedItem.valueHasMutated?.();
					}
				} else if (this.autoSelect()) {
					koSelectedItem(koFocusedItem());
				}
			});

			let selectedItemUseCallback = true;

			koSelectedItem.subscribe(item => {
				if (item) {
					koList.forEach(subItem => subItem.checked(false));
					if (selectedItemUseCallback) {
						itemSelectedThrottle(item);
					}
				} else if (selectedItemUseCallback) {
					this.itemSelected();
				}
			});

			koFocusedItem.subscribe(item => item && (this.sLastUid = this.getItemUid(item)));

			/**
			 * Below code is used to keep checked/focused/selected states when array is refreshed.
			 */

			let aCache = [],
				aCheckedCache = [],
				mFocused = null,
				mSelected = null;

			// Before removing old list
			koList.subscribe(
				items => {
					if (isArray(items)) {
						items.forEach(item => {
							const uid = this.getItemUid(item);
							if (uid) {
								aCache.push(uid);
								item.checked() && aCheckedCache.push(uid);
								if (!mFocused && item.focused()) {
									mFocused = uid;
								}
								if (!mSelected && item.selected()) {
									mSelected = uid;
								}
							}
						});
					}
				},
				this,
				'beforeChange'
			);

			koList.subscribe(aItems => {
				selectedItemUseCallback = false;

				koFocusedItem(null);
				koSelectedItem(null);

				if (isArray(aItems)) {
					let temp,
						isChecked;

					aItems.forEach(item => {
						const uid = this.getItemUid(item);
						if (uid) {

							if (mFocused === uid) {
								koFocusedItem(item);
								mFocused = null;
							}

							if (aCheckedCache.includes(uid)) {
								item.checked(true);
								isChecked = true;
							}

							if (!isChecked && mSelected === uid) {
								koSelectedItem(item);
								mSelected = null;
							}
						}
					});

					selectedItemUseCallback = true;

					if (
						(this.iSelectNextHelper || this.iFocusedNextHelper) &&
						aItems.length &&
						!koFocusedItem()
					) {
						temp = null;
						if (this.iFocusedNextHelper) {
							temp = aItems[-1 === this.iFocusedNextHelper ? aItems.length - 1 : 0];
						}

						if (!temp && this.iSelectNextHelper) {
							temp = aItems[-1 === this.iSelectNextHelper ? aItems.length - 1 : 0];
						}

						if (temp) {
							if (this.iSelectNextHelper) {
								koSelectedItem(temp);
							}

							koFocusedItem(temp);

							this.scrollToFocused();
							setTimeout(this.scrollToFocused, 100);
						}

						this.iSelectNextHelper = 0;
						this.iFocusedNextHelper = 0;
					}

					if (this.autoSelect() && !isChecked && !koSelectedItem()) {
						koSelectedItem(koFocusedItem());
					}
				}

				aCache = [];
				aCheckedCache = [];
				mFocused = null;
				mSelected = null;
				selectedItemUseCallback = true;
			});
		}

		itemSelected(item) {
			if (this.list.hasChecked()) {
				item || this.oCallbacks.ItemSelect?.(null);
			} else if (item) {
				this.oCallbacks.ItemSelect?.(item);
			}
		}

		unselect() {
			this.selectedItem(null);
			this.focusedItem(null);
		}

		init(contentScrollable, keyScope = 'all') {
			this.oContentScrollable = contentScrollable;

			if (contentScrollable) {
				let getItem = selector => {
					let el = event.target.closestWithin(selector, contentScrollable);
					return el ? ko.dataFor(el) : null;
				};

				addEventsListeners(contentScrollable, {
					click: event => {
						let el = event.target.closestWithin(this.sItemSelector, contentScrollable);
						el && this.actionClick(ko.dataFor(el), event);

						const item = getItem(this.sItemCheckedSelector);
						if (item) {
							if (event.shiftKey) {
								this.actionClick(item, event);
							} else {
								this.focusedItem(item);
								item.checked(!item.checked());
							}
						}
					},
					auxclick: event => {
						if (1 == event.button) {
							const item = getItem(this.sItemSelector);
							if (item) {
								this.focusedItem(item);
								this.oCallbacks.MiddleClick?.(item);
							}
						}
					}
				});

				registerShortcut('enter,open', '', keyScope, () => {
					const focused = this.focusedItem();
					if (focused && !focused.selected()) {
						this.actionClick(focused);
						return false;
					}
				});

				addShortcut('arrowup,arrowdown', 'meta', keyScope, () => false);

				addShortcut('arrowup,arrowdown', 'shift', keyScope, event => {
					this.newSelectPosition(event.key, true);
					return false;
				});
				registerShortcut('arrowup,arrowdown,home,end,pageup,pagedown,space', '', keyScope, event => {
					this.newSelectPosition(event.key, false);
					return false;
				});
			}
		}

		/**
		 * @returns {boolean}
		 */
		autoSelect() {
			return (this.oCallbacks.AutoSelect || (()=>1))() && this.focusedItem();
		}

		/**
		 * @param {Object} oItem
		 * @returns {string}
		 */
		getItemUid(item) {
			return (item && this.oCallbacks.ItemGetUid?.(item)?.toString()) || '';
		}

		/**
		 * @param {string} sEventKey
		 * @param {boolean} bShiftKey
		 * @param {boolean=} bForceSelect = false
		 */
		newSelectPosition(sEventKey, bShiftKey, bForceSelect) {
			let isArrow = 'ArrowUp' === sEventKey || 'ArrowDown' === sEventKey,
				result;

			const pageStep = 10,
				list = this.list(),
				listLen = list.length,
				focused = this.focusedItem();

			bShiftKey || (shiftStart = -1);

			if (' ' === sEventKey) {
				focused?.checked(!focused.checked());
			} else if (listLen) {
				if (focused) {
					if (isArrow) {
						let i = list.indexOf(focused),
							up = 'ArrowUp' == sEventKey;
						if (bShiftKey) {
							shiftStart = -1 < shiftStart ? shiftStart : i;
							shiftStart == i
								? focused.checked(true)
								: ((up ? shiftStart < i : shiftStart > i) && focused.checked(false));
						}
						if (up) {
							i > 0 && (result = list[--i]);
						} else if (++i < listLen) {
							result = list[i];
						}
						bShiftKey && result?.checked(true);
						result || this.oCallbacks.UpOrDown?.(up);
					} else if ('Home' === sEventKey) {
						result = list[0];
					} else if ('End' === sEventKey) {
						result = list[list.length - 1];
					} else if ('PageDown' === sEventKey) {
						let i = list.indexOf(focused);
						if (i < listLen - 1) {
							result = list[Math.min(i + pageStep, listLen - 1)];
						}
					} else if ('PageUp' === sEventKey) {
						let i = list.indexOf(focused);
						if (i > 0) {
							result = list[Math.max(0, i - pageStep)];
						}
					}
				} else if (
					'Home' == sEventKey ||
					'PageUp' == sEventKey
				) {
					result = list[0];
				} else if (
					'End' === sEventKey ||
					'PageDown' === sEventKey
				) {
					result = list[list.length - 1];
				}

				if (result) {
					this.focusedItem(result);
					if ((this.autoSelect() || bForceSelect) && !this.list.hasChecked()) {
						this.selectedItem(result);
					}
					this.scrollToFocused();
				}
			}
		}

		/**
		 * @returns {boolean}
		 */
		scrollToFocused() {
			const scrollable = this.oContentScrollable;
			if (scrollable) {
				let focused = scrollable.querySelector(this.sItemFocusedSelector);
				if (focused) {
					const fRect = focused.getBoundingClientRect(),
						sRect = scrollable.getBoundingClientRect();
					if (fRect.top < sRect.top) {
						focused.scrollIntoView(true);
					} else if (fRect.bottom > sRect.bottom) {
						focused.scrollIntoView(false);
					}
				} else {
					scrollable.scrollTop = 0;
				}
			}
		}

		/**
		 * @returns {boolean}
		 */
		scrollToTop() {
			this.oContentScrollable && (this.oContentScrollable.scrollTop = 0);
		}

		/**
		 * @param {Object} item
		 * @param {Object=} event
		 */
		actionClick(item, event = null) {
			if (item) {
				let click = true;
				if (event) {
					if (event.shiftKey && !event.ctrlKey && !event.metaKey && !event.altKey) {
						click = false;
						this.sLastUid || (this.sLastUid = this.getItemUid(item));

						item.checked(!item.checked());

						const uid = this.getItemUid(item);
						if (uid && this.sLastUid && uid !== this.sLastUid) {
							let changeRange = false,
								isInRange = false,
								checked = item.checked(),
								lineUid = '';
							this.list().forEach(listItem => {
								lineUid = this.getItemUid(listItem);
								changeRange = (lineUid === this.sLastUid || lineUid === uid);
								if (isInRange || changeRange) {
									if (changeRange) {
										isInRange = !isInRange;
									}
									listItem.checked(checked);
								}
							});
						}
						this.sLastUid = uid;

						this.focusedItem(item);
					} else if (!event.shiftKey && (event.ctrlKey || event.metaKey) && !event.altKey) {
						click = false;
						this.focusedItem(item);

						if (this.selectedItem() && item !== this.selectedItem()) {
							this.selectedItem().checked(true);
						}

						item.checked(!item.checked());
					}
				}

				click && this.selectMessageItem(item);
			}
		}

		on(eventName, callback) {
			this.oCallbacks[eventName] = callback;
		}

		selectMessageItem(messageItem) {
			this.focusedItem(messageItem);
			this.selectedItem(messageItem);
			this.scrollToFocused();
		}
	}

	/**
	 * Inspired by https://github.com/mcpar-land/vcfer
	 */

	class VCardProperty {

		/**
		 * A class describing a single vCard property.
		 * Will almost always be a member of a
		 * {@link VCard}'s [props]{@link VCard.props} map.
		 *
		 * Accepts either 2-4 arguments, or 1 argument in jCard property format.
		 * @param arg the field, or a jCard property
		 * @param value
		 * @param params
		 * @param type
		 */
		constructor(arg, value, params, type = 'text')
		{
			this.field = '';

			/**
			 * the value of the property.
			 * @example '(123) 456 7890'
			 */
			this.value = '';

			/**
			 * the type of the property value.
			 * @example 'text'
			 */
			this.type = '';

			/**
			 * https://www.rfc-editor.org/rfc/rfc6350.html#section-5
			 * An jCard parameters object.
			 * @example
			 * {
			 * 	type: ['work', 'voice', 'pref'],
			 * 	value: 'uri'
			 * }
			 */
			this.params = {};

			// Construct from arguments
			if (value !== undefined && typeof arg === 'string') {
				this.field = arg;
				this.value = value;
				this.params = params || {};
				this.type = type;
			}
			// construct from jcard
			else if (value === undefined && params === undefined && typeof arg === 'object') {
				this.parseFromJCardProperty(arg);
			}
			// invalid property
			else {
				throw new Error('invalid Property constructor');
			}
		}

		parseFromJCardProperty(jCardProp)
		{
			jCardProp = JSON.parse(JSON.stringify(jCardProp));
			this.field = jCardProp[0].toLowerCase();
			this.params = jCardProp[1];
			this.type = jCardProp[2];
			this.value = jCardProp[3];
		}

		addParam(key, value)
		{
			if (Array.isArray(this.params[key])) {
				this.params[key].push(value);
			}
			else if (this.params[key] != null) {
				this.params[key] = [this.params[key], value];
			}
			else {
				this.params[key] = value;
			}
		}

		/** Returns a copy of the Property's string value. */
		getValue()
		{
			return '' + this.value;
		}

		/**
		 * https://www.rfc-editor.org/rfc/rfc6350.html#section-5.3
		 */
		pref()
		{
			return this.params.pref || 100;
		}

		/**
		 * https://www.rfc-editor.org/rfc/rfc6350.html#section-5.6
		 */
		tags()
		{
			return this.params.type || [];
		}

		/** Returns `true` if all the following are true:
		 * - the property's value contains charactes other than `;`
		 * - the property has no parameters
		 */
		isEmpty()
		{
			return ((null == this.value || !/[^;]+/.test(this.value)) && !Object.keys(this.params).length);
		}

		notEmpty()
		{
			return !this.isEmpty();
		}

		/** Returns a readonly string copy of the property's field. */
		getField()
		{
			return '' + this.field;
		}

		/**
		 * Returns stringified JSON
		 */
		toString()
		{
			return JSON.stringify(this);
		}

		/**
		 * Returns a JSON array in a [jCard property]{@link JCardProperty} format
		 */
		toJSON()
		{
			return [
				this.field,
				this.params,
				this.type || 'text',
				this.value
			];
		}
	}

	/**
	 * https://datatracker.ietf.org/doc/html/rfc7095
	 *
	 * Inspired by https://github.com/mcpar-land/vcfer
	 */

	class JCard {

		constructor(input)
		{
			this.props = new Map();
			this.version = '4.0';
			if (input) {
				// read from jCard
				if (typeof input !== 'object') {
					throw new Error('error reading vcard')
				}
				this.parseFromJCard(input);
			}
		}

		parseFromJCard(json)
		{
			json = JSON.parse(JSON.stringify(json));
			if (!/vcard/i.test(json[0])) {
				throw new SyntaxError('Incorrect jCard format');
			}
			json[1].forEach(jprop => this.add(new VCardProperty(jprop)));
		}

		/**
		 * Retrieve an array of {@link VCardProperty} objects under the specified field.
		 * Returns [] if there are no VCardProperty objects found.
		 * Properites are always stored in an array.
		 * @param field to get.
		 * @param type If provided, only return {@link VCardProperty}s with the specified
		 * type as a param.
		 */
		get(field, type)
		{
			if (type) {
				let props = this.props.get(field);
				return props
					? props.filter(prop => {
						let types = prop.type;
						return (Array.isArray(types) ? types : [types]).includes(type);
					})
					: [];
			}
			return this.props.get(field) || [];
			// TODO with type filter-er
		}

		/**
		 * Retrieve a _single_ VCardProperty of the specified field. Attempts to pick based
		 * on the following priorities, in order:
		 * - `TYPE={type}` of the value specified in the `type` argument. Ignored
		 * if the argument isn't supplied.
		 * - `TYPE=pref` is present.
		 * - is the VCardProperty at index 0 from get(field)
		 * @param field
		 * @param type
		 */
		getOne(field, type)
		{
			return this.get(field, type || 'pref')[0] || this.get(field)[0];
		}

		/**
		 * Set the contents of a field to contain a single {@link VCardProperty}.
		 *
		 * Accepts either 2-4 arguments to construct a VCardProperty,
		 * or 1 argument of a preexisting VCardProperty object.
		 *
		 * This will always overwrite all existing properties of the given
		 * field. For just adding a new VCardProperty, see {@link VCard#add}
		 * @param arg the field, or a VCardProperty object
		 * @param value the value for the VCardProperty object
		 * @param params the parameters for the VCardProperty object
		 * @param type the type for the VCardProperty object
		 */
		set(arg, value, params, type)
		{
			if (typeof arg === 'string') {
				arg = new VCardProperty(String(arg), value, params, type);
			}
			if (!(arg instanceof VCardProperty)) {
				throw new Error('invalid argument of VCard.set(), expects string arguments or a VCardProperty');
			}
			let field = arg.getField();
			this.props.set(field, [arg]);
			return arg;
		}

		add(arg, value, params, type)
		{
			// string arguments
			if (typeof arg === 'string') {
				arg = new VCardProperty(String(arg), value, params, type);
			}
			if (!(arg instanceof VCardProperty)) {
				throw new Error('invalid argument of VCard.add(), expects string arguments or a VCardProperty');
			}
			// VCardProperty arguments
			let field = arg.getField();
			if (this.props.get(field)) this.props.get(field)?.push(arg);
			else this.props.set(field, [arg]);
			return arg;
		}

		/**
		 * Removes a {@link VCardProperty}, or all properties of the supplied field.
		 * @param arg the field, or a {@link VCardProperty} object
		 * @param paramFilter (incomplete)
		 */
		remove(arg) {
			// string arguments
			if (typeof arg === 'string') {
				// TODO filter by param
				this.props.delete(arg);
			}
			// VCardProperty argument
			else if (arg instanceof VCardProperty) {
				let propArray = this.props.get(arg.getField());
				if (!(propArray === null || propArray === void 0 ? void 0 : propArray.includes(arg)))
					throw new Error("Attempted to remove VCardProperty VCard does not have: ".concat(arg));
				propArray.splice(propArray.indexOf(arg), 1);
				if (propArray.length === 0)
					this.props.delete(arg.getField());
			}
			// incorrect arguments
			else
				throw new Error('invalid argument of VCard.remove(), expects ' +
					'string and optional param filter or a VCardProperty');
		}

		/**
		 * Returns true if the vCard has at least one @{link VCardProperty}
		 * of the given field.
		 * @param field The field to query
		 */
		has(field)
		{
			return (!!this.props.get(field) && this.props.get(field).length > 0);
		}


		/**
		 * Returns stringified JSON
		 */
		toString()
		{
			return JSON.stringify(this);
		}

		/**
		 * Returns a {@link JCard} object as a JSON array.
		 */
		toJSON()
		{
			let data = [['version', {}, 'text', '4.0']];
	/*
			this.props.forEach((props, field) =>
				(field === 'version')  || props.forEach(prop => prop.isEmpty() || data.push(prop.toJSON()))
			);
	*/
			for (const [field, props] of this.props.entries()) {
				if ('version' !== field) {
					for (const prop of props) {
						prop.isEmpty() || data.push(prop.toJSON());
					}
				}
			}

			return ['vcard', data];
		}

		/**
		 * Automatically generate the 'fn' VCardProperty from the preferred 'n' VCardProperty.
		 *
		 * #### `set` (`boolean`)
		 *
		 * - `false`: (default) return the generated full name string without
		 * modifying the VCard.
		 *
		 * - `true`: modify the VCard's `fn` VCardProperty directly, as specified
		 * by `append`
		 *
		 * #### `append` (`boolean`)
		 *
		 * (ignored if `set` is `false`)
		 *
		 * - `false`: (default) replace the existing 'fn' VCardProperty/properties with
		 * a new one.
		 *
		 * - `true`: append a new `fn` VCardProperty to the array.
		 *
		 * see: [RFC 6350 section 6.2.1](https://tools.ietf.org/html/rfc6350#section-6.2.1)
		 */
		parseFullName(options) {
			let n = this.getOne('n');
			if (n === undefined) {
				throw new Error('\'fn\' VCardProperty not present in card, cannot parse full name');
			}
			let fnString = '';
			// Position in n -> position in fn
			[3, 1, 2, 0, 4].forEach(pos => {
				let splitStr = n.value[pos];
				if (splitStr) {
					// comma separated values separated by spaces
					fnString += ' ' + splitStr.replace(',', ' ');
				}
			});
			fnString = fnString.trim();
			let fn = new VCardProperty('fn', fnString);
			if (options?.set) {
				if (options.append) {
					this.add(fn);
				} else {
					this.set(fn);
				}
			}
			return fn;
		}
	}

	//import { VCardProperty } from 'DAV/VCardProperty';

	const nProps = [
		'surName',
		'givenName',
		'middleName',
		'namePrefix',
		'nameSuffix'
	];

	/*
	const propertyMap = [
		// vCard 2.1 properties and up
		'N' => 'Text',
		'FN' => 'FlatText',
		'PHOTO' => 'Binary',
		'BDAY' => 'DateAndOrTime',
		'ADR' => 'Text',
		'TEL' => 'FlatText',
		'EMAIL' => 'FlatText',
		'GEO' => 'FlatText',
		'TITLE' => 'FlatText',
		'ROLE' => 'FlatText',
		'LOGO' => 'Binary',
		'ORG' => 'Text',
		'NOTE' => 'FlatText',
		'REV' => 'TimeStamp',
		'SOUND' => 'FlatText',
		'URL' => 'Uri',
		'UID' => 'FlatText',
		'VERSION' => 'FlatText',
		'KEY' => 'FlatText', // <uri>data:application/pgp-keys;base64,AZaz09==</uri>
		'TZ' => 'Text',

		// vCard 3.0 properties
		'CATEGORIES' => 'Text',
		'SORT-STRING' => 'FlatText',
		'PRODID' => 'FlatText',
		'NICKNAME' => 'Text',

		// rfc2739 properties
		'FBURL' => 'Uri',
		'CAPURI' => 'Uri',
		'CALURI' => 'Uri',
		'CALADRURI' => 'Uri',

		// rfc4770 properties
		'IMPP' => 'Uri',

		// vCard 4.0 properties
		'SOURCE' => 'Uri',
		'XML' => 'FlatText',
		'ANNIVERSARY' => 'DateAndOrTime',
		'CLIENTPIDMAP' => 'Text',
		'LANG' => 'LanguageTag',
		'GENDER' => 'Text',
		'KIND' => 'FlatText',
		'MEMBER' => 'Uri',
		'RELATED' => 'Uri',

		// rfc6474 properties
		'BIRTHPLACE' => 'FlatText',
		'DEATHPLACE' => 'FlatText',
		'DEATHDATE' => 'DateAndOrTime',

		// rfc6715 properties
		'EXPERTISE' => 'FlatText',
		'HOBBY' => 'FlatText',
		'INTEREST' => 'FlatText',
		'ORG-DIRECTORY' => 'FlatText
	];
	*/

	class ContactModel extends AbstractModel {
		constructor() {
			super();

			this.jCard = ['vcard',[]];

			this.addObservables({
				// Also used by Selector
				focused: false,
				selected: false,
				checked: false,

				deleted: false,
				readOnly: false,

				id: 0,
				givenName:  '', // FirstName
				surName:    '', // LastName
				middleName: '', // MiddleName
				namePrefix: '', // NamePrefix
				nameSuffix: '',  // NameSuffix
				nickname: null,
				note: null,

				// Business
				org: '',
				department: '',
				title: '',

				// Crypto
				encryptpref: '',
				signpref: ''
			});
	//		this.email = koArrayWithDestroy();
			this.email = ko.observableArray();
			this.tel   = ko.observableArray();
			this.url   = ko.observableArray();
			this.adr   = ko.observableArray();

			this.addComputables({
				hasValidName: () => !!(this.givenName() || this.surName()),

				fullName: () => [this.namePrefix(), this.givenName(), this.middleName(), this.surName()].join(' ').trim(),

				display: () => {
					let a = this.fullName(),
						b = this.email()?.[0]?.value,
						c = this.nickname();
					return a || b || c;
				}
	/*
				fullName: {
					read: () => this.givenName() + " " + this.surName(),
					write: value => {
						this.jCard.set('fn', value/*, params, group* /)
					}
				}
	*/
			});
		}

		/**
		 * @returns {Array|null}
		 */
		getNameAndEmailHelper() {
			let name = (this.givenName() + ' ' + this.surName()).trim(),
				email = this.email()[0];
	/*
	//		this.jCard.getOne('fn')?.notEmpty() ||
			this.jCard.parseFullName({set:true});
	//		let name = this.jCard.getOne('nickname'),
			let name = this.jCard.getOne('fn'),
				email = this.jCard.getOne('email');
	*/
			return email ? [email, name] : null;
		}

		/**
		 * @static
		 * @param {jCard} json
		 * @returns {?ContactModel}
		 */
		static reviveFromJson(json) {
			const contact = super.reviveFromJson(json);
			if (contact) {
				let jCard = new JCard(json.jCard),
					props = jCard.getOne('n')?.value;
				props && props.forEach((value, index) =>
					value && contact[nProps[index]](value)
				);

				['nickname', 'note', 'title'].forEach(field => {
					props = jCard.getOne(field);
					props && contact[field](props.value);
				});

				if ((props = jCard.getOne('org')?.value)) {
					contact.org(props[0]);
					contact.department(props[1] || '');
				}

				['email', 'tel', 'url'].forEach(field => {
					props = jCard.get(field);
					props && props.forEach(prop => {
						contact[field].push({
							value: ko.observable(prop.value)
	//						type: prop.params.type
						});
					});
				});

				props = jCard.get('adr');
				props && props.forEach(prop => {
					contact.adr.push({
						street: ko.observable(prop.value[2]),
						street_ext: ko.observable(prop.value[1]),
						locality: ko.observable(prop.value[3]),
						region: ko.observable(prop.value[4]),
						postcode: ko.observable(prop.value[5]),
						pobox: ko.observable(prop.value[0]),
						country: ko.observable(prop.value[6]),
						preferred: ko.observable(prop.params.pref),
						type: ko.observable(prop.params.type) // HOME | WORK
					});
				});

				props = jCard.getOne('x-crypto');
				contact.signpref(props?.params.signpref || 'Ask');
				contact.encryptpref(props?.params.encryptpref || 'Ask');
	//			contact.encryptpref(props?.params.allowed || 'PGP/INLINE,PGP/MIME,S/MIME,S/MIMEOpaque');

				contact.jCard = json.jCard;
			}
			return contact;
		}

		/**
		 * @returns {string}
		 */
		generateUid() {
			return '' + this.id;
		}

		addEmail() {
			// home, work
			this.email.push({
				value: ko.observable('')
	//			type: prop.params.type
			});
		}

		addTel() {
			// home, work, text, voice, fax, cell, video, pager, textphone, iana-token, x-name
			this.tel.push({
				value: ko.observable('')
	//			type: prop.params.type
			});
		}

		addUrl() {
			// home, work
			this.url.push({
				value: ko.observable('')
	//			type: prop.params.type
			});
		}

		addNickname() {
			// home, work
			this.nickname() || this.nickname('');
		}

		addNote() {
			this.note() || this.note('');
		}

		hasChanges()
		{
			return this.email().filter(v => v.length).length && this.toJSON().jCard != JSON.stringify(this.jCard);
		}

		toJSON()
		{
			let jCard = new JCard(this.jCard);
			jCard.set('n', [
				this.surName(),
				this.givenName(),
				this.middleName(),
				this.namePrefix(),
				this.nameSuffix()
			]/*, params, group*/);
	//		jCard.parseFullName({set:true});

			['nickname', 'note', 'title'].forEach(field =>
				this[field]() ? jCard.set(field, this[field]()/*, params, group*/) : jCard.remove(field)
			);

			if (this.org()) {
				let org = [this.org()];
				if (this.department()) {
					org.push(this.department());
				}
				let prop = jCard.getOne('org');
				prop ? prop.value = org : jCard.set('org', org);
			} else {
				jCard.remove('');
			}

			['email', 'tel', 'url'].forEach(field => {
				let values = this[field].map(item => item.value());
				jCard.get(field).forEach(prop => {
					let i = values.indexOf(prop.value);
					if (0 > i || !prop.value) {
						jCard.remove(prop);
					} else {
						values.splice(i, 1);
					}
				});
				values.forEach(value => value && jCard.add(field, value));
			});

			jCard.set('x-crypto', '', {
				allowed: 'PGP/INLINE,PGP/MIME,S/MIME,S/MIMEOpaque',
				signpref: this.signpref(),
				encryptpref: this.encryptpref()
			}, 'x-crypto');

			// Done by server
	//		jCard.set('rev', '2022-05-21T10:59:52Z')

			return {
				Uid: this.id,
				jCard: JSON.stringify(jCard)
			};
		}

		/**
		 * @return string
		 */
		lineAsCss() {
			return (this.selected() ? 'selected' : '')
				+ (this.deleted() ? ' deleted' : '')
				+ (this.checked() ? ' checked' : '')
				+ (this.focused() ? ' focused' : '');
		}
	}

	const
		CONTACTS_PER_PAGE = 50,
		ScopeContacts = 'Contacts';

	let
		bOpenCompose = false,
		sComposeRecipientsField = '';

	class ContactsPopupView extends AbstractViewPopup {
		constructor() {
			super('Contacts');

			this.addObservables({
				search: '',
				contactsCount: 0,

				selectorContact: null,

				importButton: null,

				contactsPage: 1,

				isSaving: false,

				contact: null
			});

			this.contacts = ContactUserStore;

			this.useCheckboxesInList = SettingsUserStore.useCheckboxesInList;

			this.selector = new Selector(
				ContactUserStore,
				this.selectorContact,
				null,
				'.e-contact-item .actionHandle',
				'.e-contact-item .checkboxItem',
				'.e-contact-item.focused'
			);

			this.selector.on('ItemSelect', contact => this.populateViewContact(contact));

			this.selector.on('ItemGetUid', contact => contact ? contact.generateUid() : '');

			this.addComputables({
				contactsPaginator: computedPaginatorHelper(
					this.contactsPage,
					() => Math.max(1, Math.ceil(this.contactsCount() / CONTACTS_PER_PAGE))
				),

				contactsCheckedOrSelected: () => {
					const checked = ContactUserStore.filter(item => item.checked()),
						selected = this.selectorContact();
					return checked.length ? checked : (selected ? [selected] : []);
				},

				contactsSyncEnabled: () => ContactUserStore.allowSync() && ContactUserStore.syncMode(),

				isBusy: () => ContactUserStore.syncing() | ContactUserStore.importing() | ContactUserStore.loading()
					| this.isSaving()
			});

			this.search.subscribe(() => this.reloadContactList());

			this.saveCommand = this.saveCommand.bind(this);

			decorateKoCommands(this, {
				deleteCommand: self => !self.isBusy() && 0 < self.contactsCheckedOrSelected().length,
				newMessageCommand: self => !self.isBusy() && 0 < self.contactsCheckedOrSelected().length,
				saveCommand: self => !self.isBusy(),
				syncCommand: self => !self.isBusy()
			});
		}

		newContact() {
			this.populateViewContact(new ContactModel);
			this.selectorContact(null);
		}

		deleteCommand() {
			const contacts = this.contactsCheckedOrSelected();
			if (contacts.length) {
				let selectorContact = this.selectorContact(),
					Uids = [],
					count = 0;
				contacts.forEach(contact => {
					Uids.push(contact.id());
					if (selectorContact && selectorContact.id() === contact.id()) {
						this.selectorContact(selectorContact = null);
					}
					contact.deleted(true);
					++count;
				});
				Remote.request('ContactsDelete',
					(iError, oData) => {
						if (iError) {
							alert(oData?.ErrorMessage || getNotification(iError));
						} else {
							const page = this.contactsPage();
							if (page > Math.max(1, Math.ceil((this.contactsCount() - count) / CONTACTS_PER_PAGE))) {
								this.contactsPage(page - 1);
							}
	//						contacts.forEach(contact => ContactUserStore.remove(contact));
						}
						this.reloadContactList();
					}, {
						Uids: Uids.join(',')
					}
				);
			}
		}

		newMessageCommand() {
			let aE = [],
				recipients = {to:null,cc:null,bcc:null};

			this.contactsCheckedOrSelected().forEach(oContact => {
				const data = oContact?.getNameAndEmailHelper(),
					email = data ? new EmailModel(data[0], data[1]) : null;
				email?.validate() && aE.push(email);
			});

			if (arrayLength(aE)) {
				bOpenCompose = false;
				this.close();
				recipients[sComposeRecipientsField] = aE;
				showMessageComposer([ComposeType.Empty, null, recipients.to, recipients.cc, recipients.bcc]);
			}
		}

		clearSearch() {
			this.search('');
		}

		saveCommand() {
			this.saveContact(this.contact());
		}

		saveContact(contact) {
			const data = contact.toJSON();
			if (data.jCard != JSON.stringify(contact.jCard)) {
				this.isSaving(true);
				Remote.request('ContactSave',
					(iError, oData) => {
						if (iError) {
							alert(oData?.ErrorMessage || getNotification(iError));
						} else if (oData.Result.ResultID) {
							if (contact.id()) {
								contact.id(oData.Result.ResultID);
								contact.jCard = JSON.parse(data.jCard);
							} else {
								this.reloadContactList(); // TODO: remove when e-contact-foreach is dynamic
							}
						}
						this.isSaving(false);
					}, data
				);
			}
		}

		syncCommand() {
			ContactUserStore.sync(iError => {
				iError && alert(getNotification(iError));
				this.reloadContactList(true);
			});
		}

		exportVcf() {
			download(serverRequestRaw('ContactsVcf'), 'contacts.vcf');
		}

		exportCsv() {
			download(serverRequestRaw('ContactsCsv'), 'contacts.csv');
		}

		/**
		 * @param {?ContactModel} contact
		 */
		populateViewContact(contact) {
			const oldContact = this.contact(),
				fn = () => this.contact(contact);
			if (oldContact?.hasChanges()) {
				AskPopupView.showModal([
					i18n('GLOBAL/SAVE_CHANGES'),
					() => this.saveContact(oldContact) | fn(),
					fn
				]);
			} else fn();
		}

		/**
		 * @param {boolean=} dropPagePosition = false
		 */
		reloadContactList(dropPagePosition = false) {
			let offset = (this.contactsPage() - 1) * CONTACTS_PER_PAGE;

			if (dropPagePosition) {
				this.contactsPage(1);
				offset = 0;
			}

			ContactUserStore.loading(true);
			Remote.abort('Contacts').request('Contacts',
				(iError, data) => {
					let count = 0,
						list = [];

					if (iError) {
	//					console.error(data);
						alert(data?.ErrorMessage || getNotification(iError));
					} else if (arrayLength(data.Result.List)) {
						data.Result.List.forEach(item => {
							item = ContactModel.reviveFromJson(item);
							item && list.push(item);
						});
						count = pInt(data.Result.Count);
					}

					this.contactsCount(0 < count ? count : 0);

					ContactUserStore(list);

					ContactUserStore.loading(false);
				},
				{
					Offset: offset,
					Limit: CONTACTS_PER_PAGE,
					Search: this.search()
				}
			);
		}

		onBuild(dom) {
			this.selector.init(dom.querySelector('.b-list-content'), ScopeContacts);

			registerShortcut('delete', '', ScopeContacts, () => {
				this.deleteCommand();
				return false;
			});

			registerShortcut('c,w', '', ScopeContacts, () => {
				this.newMessageCommand();
				return false;
			});

			const self = this;

			dom.addEventListener('click', event => {
				let el = event.target.closestWithin('.e-paginator a', dom);
				if (el && (el = pInt(ko.dataFor(el)?.value))) {
					self.contactsPage(el);
					self.reloadContactList();
				}
			});

			// initUploader

			if (this.importButton()) {
				const j = new Jua({
					action: serverRequest('UploadContacts'),
					limit: 1,
					disableDocumentDropPrevent: true,
					clickElement: this.importButton()
				});

				if (j) {
					j.on('onStart', () => {
						ContactUserStore.importing(true);
					}).on('onComplete', (id, result, data) => {
						ContactUserStore.importing(false);
						this.reloadContactList();
						if (!id || !result || !data || !data.Result) {
							alert(i18n('CONTACTS/ERROR_IMPORT_FILE'));
						}
					});
				}
			}
		}

		tryToClose() {
			(false === this.onClose()) || this.close();
		}

		onClose() {
			const contact = this.contact();
			if (AskPopupView.hidden() && contact?.hasChanges()) {
				AskPopupView.showModal([
					i18n('GLOBAL/SAVE_CHANGES'),
					() => this.close() | this.saveContact(contact),
					() => this.close()
				]);
				return false;
			}
		}

		onShow(bBackToCompose, sRecipientsField) {
			bOpenCompose = !!bBackToCompose;
			sComposeRecipientsField = ['to','cc','bcc'].includes(sRecipientsField) ? sRecipientsField : 'to';
			this.reloadContactList(true);
		}

		onHide() {
			this.contact(null);
			this.selectorContact(null);
			this.search('');
			this.contactsCount(0);

			ContactUserStore([]);

			bOpenCompose && showMessageComposer();
		}
	}

	//import { clearCache } from 'Common/Cache';
	//import { koComputable } from 'External/ko';

	class SystemDropDownUserView extends AbstractViewRight {
		constructor() {
			super();

			this.allowAccounts = SettingsCapa('AdditionalAccounts');

			this.accountEmail = AccountUserStore.email;

			this.accounts = AccountUserStore.accounts;
			this.accountsLoading = AccountUserStore.loading;
	/*
			this.accountsUnreadCount = : koComputable(() => 0);
			this.accountsUnreadCount = : koComputable(() => AccountUserStore.accounts().reduce((result, item) => result + item.count(), 0));
	*/

			this.addObservables({
				currentAudio: '',
				accountMenuDropdownTrigger: false
			});

			this.allowContacts = AppUserStore.allowContacts();

			addEventListener('audio.stop', () => this.currentAudio(''));
			addEventListener('audio.start', e => this.currentAudio(e.detail));
		}

		stopPlay() {
			fireEvent('audio.api.stop');
		}

		accountClick(account, event) {
			let email = account?.email;
			if (email && 0 === event.button && AccountUserStore.email() != email) {
				AccountUserStore.loading(true);
				event.preventDefault();
				event.stopPropagation();
				Remote.request('AccountSwitch',
					(iError/*, oData*/) => {
						if (iError) {
							AccountUserStore.loading(false);
							alert(getNotification(iError).replace('%EMAIL%', email));
							if (account.isAdditional()) {
								showScreenPopup(AccountPopupView, [account]);
							}
						} else {
	/*						// Not working yet
							forEachObjectEntry(oData.Result, (key, value) => rl.settings.set(key, value));
							clearCache();
	//						MessageUserStore.setMessage();
	//						MessageUserStore.purgeMessageBodyCache();
	//						MessageUserStore.hideMessageBodies();
							MessagelistUserStore([]);
	//						FolderUserStore.folderList([]);
							loadFolders(value => {
								if (value) {
	//								4. Change to INBOX = reload MessageList
	//								MessagelistUserStore.setMessageList();
								}
							});
							AccountUserStore.loading(false);
	*/
							rl.route.reload();
						}
					}, {Email:email}
				);
			}
			return true;
		}

		emailTitle() {
			return AccountUserStore.email();
		}

		settingsClick() {
			hasher.setHash(settings());
		}

		settingsHelp() {
			showScreenPopup(KeyboardShortcutsHelpPopupView);
		}

		addAccountClick() {
			this.allowAccounts && showScreenPopup(AccountPopupView);
		}

		contactsClick() {
			this.allowContacts && showScreenPopup(ContactsPopupView);
		}

		toggleLayout()
		{
			const mobile = !ThemeStore.isMobile();
			doc.cookie = 'rllayout=' + (mobile ? 'mobile' : 'desktop') + '; samesite=strict';
			ThemeStore.isMobile(mobile);
			leftPanelDisabled(mobile);
		}

		logoutClick() {
			rl.app.logout();
		}

		onBuild() {
			registerShortcut('m', '', [Scope.MessageList, Scope.MessageView, Scope.Settings], () => {
				if (!this.viewModelDom.hidden) {
	//				exitFullscreen();
					this.accountMenuDropdownTrigger(true);
					return false;
				}
			});

			// shortcuts help
			registerShortcut('?,f1,help', '', [Scope.MessageList, Scope.MessageView, Scope.Settings], () => {
				if (!this.viewModelDom.hidden) {
					showScreenPopup(KeyboardShortcutsHelpPopupView);
					return false;
				}
			});
		}
	}

	class FolderCreatePopupView extends AbstractViewPopup {
		constructor() {
			super('FolderCreate');

			this.addObservables({
				folderName: '',
				folderSubscribe: SettingsUserStore.hideUnsubscribed(),

				selectedParentValue: UNUSED_OPTION_VALUE
			});

			this.parentFolderSelectList = koComputable(() =>
				folderListOptionsBuilder(
					[],
					[['', '']],
					oItem =>
						oItem ? (oItem.isSystemFolder() ? oItem.name() + ' ' + oItem.manageFolderSystemName() : oItem.name()) : '',
					FolderUserStore.namespace
						? item => FolderUserStore.namespace !== item.fullName.slice(0, FolderUserStore.namespace.length)
						: null,
					true
				)
			);

			this.defaultOptionsAfterRender = defaultOptionsAfterRender;
		}

		submitForm(form) {
			if (form.reportValidity()) {
				const data = new FormData(form);
				data.set('Subscribe', this.folderSubscribe() ? 1 : 0);

				let parentFolderName = this.selectedParentValue();
				if (!parentFolderName && 1 < FolderUserStore.namespace.length) {
					data.set('Parent', FolderUserStore.namespace.slice(0, FolderUserStore.namespace.length - 1));
				}

				Remote.abort('Folders').post('FolderCreate', FolderUserStore.foldersCreating, data)
					.then(
						data => {
							const folder = getFolderFromCacheList(parentFolderName),
								subFolder = FolderModel.reviveFromJson(data.Result),
								folders = (folder ? folder.subFolders : FolderUserStore.folderList);
							setFolder(subFolder);
							folders.push(subFolder);
							sortFolders(folders);
	/*
							var collator = new Intl.Collator(undefined, {numeric: true, sensitivity: 'base'});
							console.log((folder ? folder.subFolders : FolderUserStore.folderList).sort(collator.compare));
	*/
						},
						error => {
							FolderUserStore.folderListError(
								getNotification(error.code, '', Notification$1.CantCreateFolder)
								+ '.\n' + error.message);
						}
					);

				this.close();
			}
		}

		onShow() {
			this.folderName('');
			this.selectedParentValue('');
		}
	}

	class ComposeAttachmentModel extends AbstractModel {
		/**
		 * @param {string} id
		 * @param {string} fileName
		 * @param {?number=} size = null
		 * @param {boolean=} isInline = false
		 * @param {boolean=} isLinked = false
		 * @param {string=} CID = ''
		 * @param {string=} contentLocation = ''
		 */
		constructor(id, fileName, size = null, isInline = false, isLinked = false, CID = '', contentLocation = '') {
			super();

			this.id = id;
			this.isInline = !!isInline;
			this.isLinked = !!isLinked;
			this.CID = CID;
			this.contentLocation = contentLocation;
			this.fromMessage = false;

			this.addObservables({
				fileName: fileName,
				size: size,
				tempName: '',

				progress: 0,
				error: '',
				waiting: true,
				uploading: false,
				enabled: true,
				complete: false
			});

			this.addComputables({
				progressText: () => {
					const p = this.progress();
					return 0 === p ? '' : '' + (98 < p ? 100 : p) + '%';
				},

				progressStyle: () => {
					const p = this.progress();
					return 0 === p ? '' : 'width:' + (98 < p ? 100 : p) + '%';
				},

				title: () => this.error() || this.fileName(),

				friendlySize: () => {
					const localSize = this.size();
					return null === localSize ? '' : FileInfo.friendlySize(localSize);
				},

				mimeType: () => FileInfo.getContentType(this.fileName()),
				fileExt: () => FileInfo.getExtension(this.fileName())
			});
		}

		/**
		 * @returns {string}
		 */
		iconClass() {
			return FileInfo.getIconClass(this.fileExt(), this.mimeType());
		}
	}

	class FolderSystemPopupView extends AbstractViewPopup {
		constructor() {
			super('FolderSystem');

			this.notification = ko.observable('');

			this.folderSelectList = koComputable(() =>
				folderListOptionsBuilder(
					FolderUserStore.systemFoldersNames(),
					[
						['', i18n('POPUPS_SYSTEM_FOLDERS/SELECT_CHOOSE_ONE')],
						[UNUSED_OPTION_VALUE, i18n('POPUPS_SYSTEM_FOLDERS/SELECT_UNUSE_NAME')]
					]
				)
			);

			this.sentFolder = FolderUserStore.sentFolder;
			this.draftsFolder = FolderUserStore.draftsFolder;
			this.spamFolder = FolderUserStore.spamFolder;
			this.trashFolder = FolderUserStore.trashFolder;
			this.archiveFolder = FolderUserStore.archiveFolder;

			const fSaveSystemFolders = (()=>FolderUserStore.saveSystemFolders()).debounce(1000);

			addSubscribablesTo(FolderUserStore, {
				sentFolder: fSaveSystemFolders,
				draftsFolder: fSaveSystemFolders,
				spamFolder: fSaveSystemFolders,
				trashFolder: fSaveSystemFolders,
				archiveFolder: fSaveSystemFolders
			});

			this.defaultOptionsAfterRender = defaultOptionsAfterRender;
		}

		/**
		 * @param {number=} notificationType = SetSystemFoldersNotification.None
		 */
		onShow(notificationType = SetSystemFoldersNotification.None) {
			let notification = '', prefix = 'POPUPS_SYSTEM_FOLDERS/NOTIFICATION_';
			switch (notificationType) {
				case SetSystemFoldersNotification.Sent:
					notification = i18n(prefix + 'SENT');
					break;
				case SetSystemFoldersNotification.Draft:
					notification = i18n(prefix + 'DRAFTS');
					break;
				case SetSystemFoldersNotification.Spam:
					notification = i18n(prefix + 'SPAM');
					break;
				case SetSystemFoldersNotification.Trash:
					notification = i18n(prefix + 'TRASH');
					break;
				case SetSystemFoldersNotification.Archive:
					notification = i18n(prefix + 'ARCHIVE');
					break;
				// no default
			}

			this.notification(notification);
		}
	}

	/*
	import { ThemeStore } from 'Stores/Theme';

	let alreadyFullscreen;
	*/
	let oLastMessage;

	const
		ScopeCompose = 'Compose',

		tpl = createElement('template'),

		base64_encode = text => btoa(unescape(encodeURIComponent(text))).match(/.{1,76}/g).join('\r\n'),

		email = new EmailModel(),
		getEmail = value => {
			email.clear();
			email.parse(value.trim());
			return email.email || false;
		},

		/**
		 * @param {Array} aList
		 * @param {boolean} bFriendly
		 * @returns {string}
		 */
		emailArrayToStringLineHelper = (aList, bFriendly) =>
			aList.map(item => item.toLine(bFriendly)).join(', '),

		reloadDraftFolder = () => {
			const draftsFolder = FolderUserStore.draftsFolder();
			if (draftsFolder && UNUSED_OPTION_VALUE !== draftsFolder) {
				setFolderHash(draftsFolder, '');
				if (FolderUserStore.currentFolderFullName() === draftsFolder) {
					MessagelistUserStore.reload(true);
				} else {
					folderInformation(draftsFolder);
				}
			}
		},

		findIdentity = addresses => {
			addresses = addresses.map(item => item.email);
			return IdentityUserStore.find(item => addresses.includes(item.email()));
		},

		/**
		 * @param {Function} fKoValue
		 * @param {Array} emails
		 */
		addEmailsTo = (fKoValue, emails) => {
			if (arrayLength(emails)) {
				const value = fKoValue().trim(),
					values = emails.map(item => item ? item.toLine() : null)
						.validUnique();

				fKoValue(value + (value ? ', ' :  '') + values.join(', ').trim());
			}
		},

		isPlainEditor = () => EditorDefaultType.Plain === SettingsUserStore.editorDefaultType(),

		/**
		 * @param {string} prefix
		 * @param {string} subject
		 * @returns {string}
		 */
		replySubjectAdd = (prefix, subject) => {
			prefix = prefix.toUpperCase().trim();
			subject = subject.replace(/\s+/g, ' ').trim();

			let drop = false,
				re = 'RE' === prefix,
				fwd = 'FWD' === prefix;

			const parts = [],
				prefixIsRe = !fwd;

			if (subject) {
				subject.split(':').forEach(part => {
					const trimmedPart = part.trim();
					if (!drop && (/^(RE|FWD)$/i.test(trimmedPart) || /^(RE|FWD)[[(][\d]+[\])]$/i.test(trimmedPart))) {
						if (!re) {
							re = !!/^RE/i.test(trimmedPart);
						}

						if (!fwd) {
							fwd = !!/^FWD/i.test(trimmedPart);
						}
					} else {
						parts.push(part);
						drop = true;
					}
				});
			}

			if (prefixIsRe) {
				re = false;
			} else {
				fwd = false;
			}

			return ((prefixIsRe ? 'Re: ' : 'Fwd: ') + (re ? 'Re: ' : '')
				+ (fwd ? 'Fwd: ' : '') + parts.join(':').trim()).trim();
		};

	ko.extenders.toggleSubscribe = (target, options) => {
		target.subscribe(options[1], options[0], 'beforeChange');
		target.subscribe(options[2], options[0]);
		return target;
	};

	class MimePart {
		constructor() {
			this.headers = {};
			this.body = '';
			this.boundary = '';
			this.children = [];
		}

		toString() {
			const hasSub = this.children.length,
				boundary = this.boundary || (this.boundary = 'part' + Jua.randomId()),
				headers = this.headers;
			if (hasSub) {
				headers['Content-Type'] += `; boundary="${boundary}"`;
			}
			let result = Object.entries(headers).map(([key, value]) => `${key}: ${value}`).join('\r\n') + '\r\n';
			if (this.body) {
				result += '\r\n' + this.body.replace(/\r?\n/g, '\r\n');
			}
			if (hasSub) {
				this.children.forEach(part => result += '\r\n--' + boundary + '\r\n' + part);
				result += '\r\n--' + boundary + '--\r\n';
			}
			return result;
		}
	}

	class ComposePopupView extends AbstractViewPopup {
		constructor() {
			super('Compose');

			const fEmailOutInHelper = (context, identity, name, isIn) => {
				const identityEmail = context && identity?.[name]();
				if (identityEmail && (isIn ? true : context[name]())) {
					let list = context[name]().trim().split(',');

					list = list.filter(email => {
						email = email.trim();
						return email && identityEmail.trim() !== email;
					});

					isIn && list.push(identityEmail);

					context[name](list.join(','));
				}
			};

			this.oEditor = null;
			this.aDraftInfo = null;
			this.sInReplyTo = '';
			this.bFromDraft = false;
			this.sReferences = '';

			this.sLastFocusedField = 'to';

			this.allowContacts = AppUserStore.allowContacts();
			this.allowIdentities = SettingsCapa('Identities');

			this.addObservables({
				identitiesDropdownTrigger: false,

				from: '',
				to: '',
				cc: '',
				bcc: '',
				replyTo: '',

				subject: '',

				isHtml: false,

				requestDsn: false,
				requestReadReceipt: false,
				markAsImportant: false,

				sendError: false,
				sendSuccessButSaveError: false,
				savedError: false,

				sendErrorDesc: '',
				savedErrorDesc: '',

				savedTime: 0,

				emptyToError: false,

				attachmentsInProcessError: false,
				attachmentsInErrorError: false,

				showCc: false,
				showBcc: false,
				showReplyTo: false,

				pgpSign: false,
				canPgpSign: false,
				pgpEncrypt: false,
				canPgpEncrypt: false,
				canMailvelope: false,

				draftsFolder: '',
				draftUid: 0,
				sending: false,
				saving: false,

				viewArea: 'body',

				attacheMultipleAllowed: false,
				addAttachmentEnabled: false,

				editorArea: null, // initDom

				currentIdentity: IdentityUserStore()[0]
			});

			// this.to.subscribe((v) => console.log(v));

			// Used by ko.bindingHandlers.emailsTags
			this.to.focused = ko.observable(false);
			this.to.focused.subscribe(value => value && (this.sLastFocusedField = 'to'));
			this.cc.focused = ko.observable(false);
			this.cc.focused.subscribe(value => value && (this.sLastFocusedField = 'cc'));
			this.bcc.focused = ko.observable(false);
			this.bcc.focused.subscribe(value => value && (this.sLastFocusedField = 'bcc'));

			this.attachments = koArrayWithDestroy();

			this.dragAndDropOver = ko.observable(false).extend({ debounce: 1 });
			this.dragAndDropVisible = ko.observable(false).extend({ debounce: 1 });

			this.currentIdentity.extend({
				toggleSubscribe: [
					this,
					(identity) => {
						fEmailOutInHelper(this, identity, 'bcc');
						fEmailOutInHelper(this, identity, 'replyTo');
					},
					(identity) => {
						fEmailOutInHelper(this, identity, 'bcc', true);
						fEmailOutInHelper(this, identity, 'replyTo', true);
					}
				]
			});

			this.tryToClose = this.tryToClose.debounce(200);

			this.iTimer = 0;

			this.addComputables({
				sendButtonSuccess: () => !this.sendError() && !this.sendSuccessButSaveError(),

				savedTimeText: () =>
					this.savedTime() ? i18n('COMPOSE/SAVED_TIME', { TIME: this.savedTime().format('LT') }) : '',

				emptyToErrorTooltip: () => (this.emptyToError() ? i18n('COMPOSE/EMPTY_TO_ERROR_DESC') : ''),

				attachmentsErrorTooltip: () => {
					let result = '';
					switch (true) {
						case this.attachmentsInProcessError():
							result = i18n('COMPOSE/ATTACHMENTS_UPLOAD_ERROR_DESC');
							break;
						case this.attachmentsInErrorError():
							result = i18n('COMPOSE/ATTACHMENTS_ERROR_DESC');
							break;
						// no default
					}
					return result;
				},

				attachmentsInProcess: () => this.attachments.filter(item => item && !item.complete()),
				attachmentsInError: () => this.attachments.filter(item => item?.error()),

				attachmentsCount: () => this.attachments.length,
				attachmentsInErrorCount: () => this.attachmentsInError.length,
				attachmentsInProcessCount: () => this.attachmentsInProcess.length,
				isDraftFolderMessage: () => this.draftsFolder() && this.draftUid(),

				identitiesOptions: () =>
					IdentityUserStore.map(item => ({
						item: item,
						optValue: item.id(),
						optText: item.formattedName()
					})),

				canBeSentOrSaved: () => !this.sending() && !this.saving()
			});

			this.addSubscribables({
				sendError: value => !value && this.sendErrorDesc(''),

				savedError: value => !value && this.savedErrorDesc(''),

				sendSuccessButSaveError: value => !value && this.savedErrorDesc(''),

				currentIdentity: value => value && this.from(value.formattedName()),

				from: value => {
					this.canPgpSign(false);
					value = getEmail(value);
					value && PgpUserStore.getKeyForSigning(value).then(result => {
						console.log({
							email: value,
							canPgpSign:result
						});
						this.canPgpSign(result);
					});
				},

				cc: value => {
					if (false === this.showCc() && value.length) {
						this.showCc(true);
					}
					this.initPgpEncrypt();
				},

				bcc: value => {
					if (false === this.showBcc() && value.length) {
						this.showBcc(true);
					}
					this.initPgpEncrypt();
				},

				replyTo: value => {
					if (false === this.showReplyTo() && value.length) {
						this.showReplyTo(true);
					}
				},

				attachmentsInErrorCount: value => {
					if (0 === value) {
						this.attachmentsInErrorError(false);
					}
				},

				to: value => {
					if (this.emptyToError() && value.length) {
						this.emptyToError(false);
					}
					this.initPgpEncrypt();
				},

				attachmentsInProcess: value => {
					if (this.attachmentsInProcessError() && arrayLength(value)) {
						this.attachmentsInProcessError(false);
					}
				}
			});

			decorateKoCommands(this, {
				sendCommand: self => self.canBeSentOrSaved(),
				saveCommand: self => self.canBeSentOrSaved(),
				deleteCommand: self => self.isDraftFolderMessage(),
				skipCommand: self => self.canBeSentOrSaved(),
				contactsCommand: self => self.allowContacts
			});

			this.from(IdentityUserStore()[0].formattedName());
		}

		sendCommand() {
			let sSentFolder = FolderUserStore.sentFolder();

			this.attachmentsInProcessError(false);
			this.attachmentsInErrorError(false);
			this.emptyToError(false);

			if (this.attachmentsInProcess().length) {
				this.attachmentsInProcessError(true);
				this.attachmentsArea();
			} else if (this.attachmentsInError().length) {
				this.attachmentsInErrorError(true);
				this.attachmentsArea();
			}

			if (!this.to().trim() && !this.cc().trim() && !this.bcc().trim()) {
				this.emptyToError(true);
			}

			if (!this.emptyToError() && !this.attachmentsInErrorError() && !this.attachmentsInProcessError()) {
				if (SettingsUserStore.replySameFolder()) {
					if (
						3 === arrayLength(this.aDraftInfo) &&
						null != this.aDraftInfo[2] &&
						this.aDraftInfo[2].length
					) {
						sSentFolder = this.aDraftInfo[2];
					}
				}

				if (!sSentFolder) {
					showScreenPopup(FolderSystemPopupView, [SetSystemFoldersNotification.Sent]);
				} else try {
					this.sendError(false);
					this.sending(true);

					if (3 === arrayLength(this.aDraftInfo)) {
						const flagsCache = MessageFlagsCache.getFor(this.aDraftInfo[2], this.aDraftInfo[1]);
						if (isArray(flagsCache)) {
							flagsCache.push(('forward' === this.aDraftInfo[0]) ? '$forwarded' : '\\answered');
							MessageFlagsCache.setFor(this.aDraftInfo[2], this.aDraftInfo[1], flagsCache);
							MessagelistUserStore.reloadFlagsAndCachedMessage();
							setFolderHash(this.aDraftInfo[2], '');
						}
					}

					sSentFolder = UNUSED_OPTION_VALUE === sSentFolder ? '' : sSentFolder;

					this.getMessageRequestParams(sSentFolder).then(params => {
						Remote.request('SendMessage',
							(iError, data) => {
								this.sending(false);
								if (iError) {
									if (Notification$1.CantSaveMessage === iError) {
										this.sendSuccessButSaveError(true);
										this.savedErrorDesc(i18n('COMPOSE/SAVED_ERROR_ON_SEND').trim());
									} else {
										this.sendError(true);
										this.sendErrorDesc(getNotification(iError, data?.ErrorMessage)
											|| getNotification(Notification$1.CantSendMessage));
									}
								} else {
									this.close();
								}
								setFolderHash(this.draftsFolder(), '');
								setFolderHash(sSentFolder, '');
								reloadDraftFolder();
							},
							params,
							30000
						);
					}).catch(e => {
						console.error(e);
						this.sendError(true);
						this.sendErrorDesc(e);
						this.sending(false);
					});
				} catch (e) {
					console.error(e);
					this.sendError(true);
					this.sendErrorDesc(e);
					this.sending(false);
				}
			}
		}

		saveCommand() {
			if (FolderUserStore.draftsFolderNotEnabled()) {
				showScreenPopup(FolderSystemPopupView, [SetSystemFoldersNotification.Draft]);
			} else {
				this.savedError(false);
				this.saving(true);
				this.autosaveStart();
				this.getMessageRequestParams(FolderUserStore.draftsFolder(), 1).then(params => {
					Remote.request('SaveMessage',
						(iError, oData) => {
							let result = false;

							this.saving(false);

							if (!iError) {
								if (oData.Result.NewFolder && oData.Result.NewUid) {
									result = true;

									if (this.bFromDraft) {
										const message = MessageUserStore.message();
										if (message && this.draftsFolder() === message.folder && this.draftUid() == message.uid) {
											MessageUserStore.message(null);
										}
									}

									this.draftsFolder(oData.Result.NewFolder);
									this.draftUid(oData.Result.NewUid);

									this.savedTime(new Date);

									if (this.bFromDraft) {
										setFolderHash(this.draftsFolder(), '');
									}
									setFolderHash(FolderUserStore.draftsFolder(), '');
								}
							}

							if (!result) {
								this.savedError(true);
								this.savedErrorDesc(getNotification(Notification$1.CantSaveMessage));
							}

							reloadDraftFolder();
						},
						params,
						200000
					);
				}).catch(e => {
					this.saving(false);
					this.savedError(true);
					this.savedErrorDesc(getNotification(Notification$1.CantSaveMessage) + ': ' + e);
				});
			}
		}

		deleteCommand() {
			AskPopupView.hidden()
			&& showScreenPopup(AskPopupView, [
				i18n('POPUPS_ASK/DESC_WANT_DELETE_MESSAGES'),
				() => {
					const
						sFromFolderFullName = this.draftsFolder(),
						oUids = new Set([this.draftUid()]);
					messagesDeleteHelper(sFromFolderFullName, oUids);
					MessagelistUserStore.removeMessagesFromList(sFromFolderFullName, oUids);
					this.close();
				}
			]);
		}

		onClose() {
			this.skipCommand();
			return false;
		}

		skipCommand() {
			ComposePopupView.inEdit(true);

			if (
				!this.saving() &&
				!this.sending() &&
				!FolderUserStore.draftsFolderNotEnabled() &&
				SettingsUserStore.allowDraftAutosave()
			) {
				this.saveCommand();
			}

			this.tryToClose();
		}

		contactsCommand() {
			if (this.allowContacts) {
				this.skipCommand();
				setTimeout(() => {
					showScreenPopup(ContactsPopupView, [true, this.sLastFocusedField]);
				}, 200);
			}
		}

		autosaveStart() {
			clearTimeout(this.iTimer);
			this.iTimer = setTimeout(()=>{
				if (this.modalVisible()
					&& !FolderUserStore.draftsFolderNotEnabled()
					&& SettingsUserStore.allowDraftAutosave()
					&& !this.isEmptyForm(false)
					&& !this.saving()
					&& !this.sending()
					&& !this.savedError()
				) {
					this.saveCommand();
				}

				this.autosaveStart();
			}, 60000);
		}

		// getAutocomplete
		emailsSource(value, fResponse) {
			Remote.abort('Suggestions').request('Suggestions',
				(iError, data) => {
					if (!iError && isArray(data.Result)) {
						fResponse(
							data.Result.map(item => (item?.[0] ? (new EmailModel(item[0], item[1])).toLine() : null))
							.filter(v => v)
						);
					} else if (Notification$1.RequestAborted !== iError) {
						fResponse([]);
					}
				},
				{
					Query: value
	//				,Page: 1
				}
			);
		}

		selectIdentity(identity) {
			identity = identity?.item;
			if (identity) {
				this.currentIdentity(identity);
				this.setSignature(identity);
			}
		}

		onHide() {
			// Stop autosave
			clearTimeout(this.iTimer);

			ComposePopupView.inEdit() || this.reset();

			this.to.focused(false);

	//		alreadyFullscreen || exitFullscreen();
		}

		dropMailvelope() {
			if (this.mailvelope) {
				elementById('mailvelope-editor').textContent = '';
				this.mailvelope = null;
			}
		}

		editor(fOnInit) {
			if (fOnInit && this.editorArea()) {
				if (this.oEditor) {
					fOnInit(this.oEditor);
				} else {
					// setTimeout(() => {
					this.oEditor = new HtmlEditor(
						this.editorArea(),
						null,
						() => fOnInit(this.oEditor),
						bHtml => this.isHtml(!!bHtml)
					);
					// }, 1000);
				}
			}
		}

		setSignature(identity, msgComposeType) {
			if (identity && ComposeType.Draft !== msgComposeType && ComposeType.EditAsNew !== msgComposeType) {
				this.editor(editor => {
					let signature = identity.signature() || '',
						isHtml = ':HTML:' === signature.slice(0, 6),
						fromLine = oLastMessage ? emailArrayToStringLineHelper(oLastMessage.from, true) : '';
					if (fromLine) {
						signature = signature.replace(/{{FROM-FULL}}/g, fromLine);
						if (!fromLine.includes(' ') && 0 < fromLine.indexOf('@')) {
							fromLine = fromLine.replace(/@\S+/, '');
						}
						signature = signature.replace(/{{FROM}}/g, fromLine);
					}
					signature = (isHtml ? signature.slice(6) : signature)
						.replace(/\r/g, '')
						.replace(/\s{1,2}?{{FROM}}/g, '')
						.replace(/\s{1,2}?{{FROM-FULL}}/g, '')
						.replace(/{{DATE}}/g, new Date().format('LLLL'))
						.replace(/{{TIME}}/g, new Date().format('LT'))
						.replace(/{{MOMENT:[^}]+}}/g, '');
					signature.length && editor.setSignature(signature, isHtml, !!identity.signatureInsertBefore());
				});
			}
		}

		/**
		 * @param {string=} type = ComposeType.Empty
		 * @param {?MessageModel|Array=} oMessageOrArray = null
		 * @param {Array=} aToEmails = null
		 * @param {Array=} aCcEmails = null
		 * @param {Array=} aBccEmails = null
		 * @param {string=} sCustomSubject = null
		 * @param {string=} sCustomPlainText = null
		 */
		onShow(type, oMessageOrArray, aToEmails, aCcEmails, aBccEmails, sCustomSubject, sCustomPlainText) {
			this.autosaveStart();

			this.viewModelDom.dataset.wysiwyg = SettingsUserStore.editorDefaultType();

			if (ComposePopupView.inEdit()) {
				type = type || ComposeType.Empty;
				if (ComposeType.Empty !== type) {
					showScreenPopup(AskPopupView, [
						i18n('COMPOSE/DISCARD_UNSAVED_DATA'),
						() => {
							this.initOnShow(type, oMessageOrArray, aToEmails, aCcEmails, aBccEmails, sCustomSubject, sCustomPlainText);
						},
						null,
						false
					]);
				} else {
					addEmailsTo(this.to, aToEmails);
					addEmailsTo(this.cc, aCcEmails);
					addEmailsTo(this.bcc, aBccEmails);

					if (sCustomSubject && !this.subject()) {
						this.subject(sCustomSubject);
					}
				}
			} else {
				this.initOnShow(type, oMessageOrArray, aToEmails, aCcEmails, aBccEmails, sCustomSubject, sCustomPlainText);
			}

			ComposePopupView.inEdit(false);
			// Chrome bug #298
	//		alreadyFullscreen = isFullscreen();
	//		alreadyFullscreen || (ThemeStore.isMobile() && toggleFullscreen());
		}

		/**
		 * @param {string=} sType = ComposeType.Empty
		 * @param {?MessageModel|Array=} oMessageOrArray = null
		 * @param {Array=} aToEmails = null
		 * @param {Array=} aCcEmails = null
		 * @param {Array=} aBccEmails = null
		 * @param {string=} sCustomSubject = null
		 * @param {string=} sCustomPlainText = null
		 */
		initOnShow(sType, oMessageOrArray, aToEmails, aCcEmails, aBccEmails, sCustomSubject, sCustomPlainText) {
			let sFrom = '',
				sTo = '',
				sCc = '',
				sDate = '',
				sSubject = '',
				sText = '',
				identity = null,
				aDraftInfo = null,
				message = 1 === arrayLength(oMessageOrArray)
					? oMessageOrArray[0]
					: (isArray(oMessageOrArray) ? null : oMessageOrArray);

			const
	//			excludeEmail = new Set(),
				excludeEmail = {},
				mEmail = AccountUserStore.email(),
				msgComposeType = sType || ComposeType.Empty;

			oLastMessage = message;

			if (mEmail) {
	//			excludeEmail.add(mEmail);
				excludeEmail[mEmail] = true;
			}

			this.reset();

			if (message) {
				switch (msgComposeType) {
					case ComposeType.Reply:
					case ComposeType.ReplyAll:
					case ComposeType.Forward:
					case ComposeType.ForwardAsAttachment:
						identity = findIdentity(message.to.concat(message.cc, message.bcc))
							/* || findIdentity(message.deliveredTo)*/;
						break;
					case ComposeType.Draft:
						identity = findIdentity(message.from.concat(message.replyTo));
						break;
					// no default
	//				case ComposeType.Empty:
				}
			}
			identity = identity || IdentityUserStore()[0];
			if (identity) {
	//			excludeEmail.add(identity.email());
				excludeEmail[identity.email()] = true;
			}

			if (arrayLength(aToEmails)) {
				this.to(emailArrayToStringLineHelper(aToEmails));
			}

			if (arrayLength(aCcEmails)) {
				this.cc(emailArrayToStringLineHelper(aCcEmails));
			}

			if (arrayLength(aBccEmails)) {
				this.bcc(emailArrayToStringLineHelper(aBccEmails));
			}

			if (msgComposeType && message) {
				sDate = timestampToString(message.dateTimeStampInUTC(), 'FULL');
				sSubject = message.subject();
				aDraftInfo = message.draftInfo;

				switch (msgComposeType) {
					case ComposeType.Reply:
					case ComposeType.ReplyAll:
						if (ComposeType.Reply === msgComposeType) {
							this.to(emailArrayToStringLineHelper(message.replyEmails(excludeEmail)));
						} else {
							let parts = message.replyAllEmails(excludeEmail);
							this.to(emailArrayToStringLineHelper(parts[0]));
							this.cc(emailArrayToStringLineHelper(parts[1]));
						}
						this.subject(replySubjectAdd('Re', sSubject));
						this.prepareMessageAttachments(message, msgComposeType);
						this.aDraftInfo = ['reply', message.uid, message.folder];
						this.sInReplyTo = message.messageId;
						this.sReferences = (message.messageId + ' ' + message.references).trim();
						break;

					case ComposeType.Forward:
					case ComposeType.ForwardAsAttachment:
						this.subject(replySubjectAdd('Fwd', sSubject));
						this.prepareMessageAttachments(message, msgComposeType);
						this.aDraftInfo = ['forward', message.uid, message.folder];
						this.sInReplyTo = message.messageId;
						this.sReferences = (message.messageId + ' ' + message.references).trim();
						break;

					case ComposeType.Draft:
						this.to(emailArrayToStringLineHelper(message.to));
						this.cc(emailArrayToStringLineHelper(message.cc));
						this.bcc(emailArrayToStringLineHelper(message.bcc));
						this.replyTo(emailArrayToStringLineHelper(message.replyTo));

						this.bFromDraft = true;

						this.draftsFolder(message.folder);
						this.draftUid(message.uid);

						this.subject(sSubject);
						this.prepareMessageAttachments(message, msgComposeType);

						this.aDraftInfo = 3 === arrayLength(aDraftInfo) ? aDraftInfo : null;
						this.sInReplyTo = message.inReplyTo;
						this.sReferences = message.references;
						break;

					case ComposeType.EditAsNew:
						this.to(emailArrayToStringLineHelper(message.to));
						this.cc(emailArrayToStringLineHelper(message.cc));
						this.bcc(emailArrayToStringLineHelper(message.bcc));
						this.replyTo(emailArrayToStringLineHelper(message.replyTo));

						this.subject(sSubject);
						this.prepareMessageAttachments(message, msgComposeType);

						this.aDraftInfo = 3 === arrayLength(aDraftInfo) ? aDraftInfo : null;
						this.sInReplyTo = message.inReplyTo;
						this.sReferences = message.references;
						break;

	//				case ComposeType.Empty:
	//					break;
					// no default
				}

				let encrypted;

				// https://github.com/the-djmaze/snappymail/issues/491
				tpl.innerHTML = message.bodyAsHTML();
				tpl.content.querySelectorAll('img').forEach(img =>
					img.dataset.xSrcCid || img.dataset.xSrc || img.replaceWith(img.alt || img.title)
				);
				sText = tpl.innerHTML.trim();

				switch (msgComposeType) {
					case ComposeType.Reply:
					case ComposeType.ReplyAll:
						sFrom = message.fromToLine(false, true);
						sText = '<br><br><p>' + i18n('COMPOSE/REPLY_MESSAGE_TITLE', { DATETIME: sDate, EMAIL: sFrom })
							+ ':</p><blockquote>'
							+ sText.trim()
							+ '</blockquote>';
						break;

					case ComposeType.Forward:
						sFrom = message.fromToLine(false, true);
						sTo = message.toToLine(false, true);
						sCc = message.ccToLine(false, true);
						sText = '<br><br><p>' + i18n('COMPOSE/FORWARD_MESSAGE_TOP_TITLE') + '</p><div>'
							+ i18n('GLOBAL/FROM') + ': ' + sFrom
							+ '<br>'
							+ i18n('GLOBAL/TO') + ': ' + sTo
							+ (sCc.length ? '<br>' + i18n('GLOBAL/CC') + ': ' + sCc : '')
							+ '<br>'
							+ i18n('COMPOSE/FORWARD_MESSAGE_TOP_SENT')
							+ ': '
							+ encodeHtml(sDate)
							+ '<br>'
							+ i18n('GLOBAL/SUBJECT')
							+ ': '
							+ encodeHtml(sSubject)
							+ '<br><br>'
							+ sText.trim()
							+ '</div>';
						break;

					case ComposeType.ForwardAsAttachment:
						sText = '';
						break;

					default:
						encrypted = PgpUserStore.isEncrypted(sText);
						if (encrypted) {
							sText = message.plain();
						}
				}

				this.editor(editor => {
					encrypted || editor.setHtml(sText);
					if (encrypted || this.isPlainEditor() || !message.isHtml()) {
						editor.modePlain();
					}
					encrypted && editor.setPlain(sText);
					this.setSignature(identity, msgComposeType);
					this.setFocusInPopup();
				});
			} else if (ComposeType.Empty === msgComposeType) {
				this.subject(null != sCustomSubject ? '' + sCustomSubject : '');
				this.editor(editor => {
					editor.setHtml(sCustomPlainText ? '' + sCustomPlainText : '');
					isPlainEditor() && editor.modePlain();
					this.setSignature(identity);
					this.setFocusInPopup();
				});
			} else if (arrayLength(oMessageOrArray)) {
				oMessageOrArray.forEach(item => this.addMessageAsAttachment(item));
				this.editor(editor => {
					isPlainEditor() ? editor.setPlain('') : editor.setHtml('');
					this.setSignature(identity, msgComposeType);
					this.setFocusInPopup();
				});
			} else {
				this.setFocusInPopup();
			}

			// item.CID item.isInline item.isLinked
			const downloads = this.attachments.filter(item => item && !item.tempName()).map(item => item.id);
			if (arrayLength(downloads)) {
				Remote.request('MessageUploadAttachments',
					(iError, oData) => {
						if (!iError) {
							forEachObjectEntry(oData.Result, (tempName, id) => {
								const attachment = this.getAttachmentById(id);
								if (attachment) {
									attachment.tempName(tempName);
									attachment
										.waiting(false)
										.uploading(false)
										.complete(true);
								}
							});
						} else {
							this.attachments.forEach(attachment => {
								if (attachment?.fromMessage) {
									attachment
										.waiting(false)
										.uploading(false)
										.complete(true)
										.error(getUploadErrorDescByCode(UploadErrorCode.NoFileUploaded));
								}
							});
						}
					},
					{
						Attachments: downloads
					},
					999000
				);
			}

			this.currentIdentity(identity);
		}

		setFocusInPopup() {
			setTimeout(() => {
				if (!this.to()) {
					this.to.focused(true);
				// TODO https://github.com/the-djmaze/snappymail/issues/501#issuecomment-1255906243
	//			} else if (!this.subject()) {
	//				this.subject.focus();
				} else {
					this.oEditor?.focus();
				}
			}, 100);
		}

		tryToClose() {
			if (AskPopupView.hidden()) {
				if (ComposePopupView.inEdit() || (this.isEmptyForm() && !this.draftUid())) {
					this.close();
				} else {
					showScreenPopup(AskPopupView, [
						i18n('POPUPS_ASK/DESC_WANT_CLOSE_THIS_WINDOW'),
						() => this.close()
					]);
				}
			}
		}

		onBuild(dom) {
			// initUploader
			const oJua = new Jua({
					action: serverRequest('Upload'),
					clickElement: dom.querySelector('#composeUploadButton'),
					dragAndDropElement: dom.querySelector('.b-attachment-place')
				}),
				attachmentSizeLimit = pInt(SettingsGet('AttachmentLimit'));

			oJua
				// .on('onLimitReached', (limit) => {
				// 	alert(limit);
				// })
				.on('onDragEnter', () => {
					this.dragAndDropOver(true);
				})
				.on('onDragLeave', () => {
					this.dragAndDropOver(false);
				})
				.on('onBodyDragEnter', () => {
					this.attachmentsArea();
					this.dragAndDropVisible(true);
				})
				.on('onBodyDragLeave', () => {
					this.dragAndDropVisible(false);
				})
				.on('onProgress', (id, loaded, total) => {
					let item = this.getAttachmentById(id);
					if (item) {
						item.progress(Math.floor((loaded / total) * 100));
					}
				})
				.on('onSelect', (sId, oData) => {
					this.dragAndDropOver(false);

					const
						size = pInt(oData.Size, null),
						attachment = new ComposeAttachmentModel(
							sId,
							oData.FileName ? oData.FileName.toString() : '',
							size
						);

					this.addAttachment(attachment, 1, oJua);

					if (0 < size && 0 < attachmentSizeLimit && attachmentSizeLimit < size) {
						attachment
							.waiting(false)
							.uploading(true)
							.complete(true)
							.error(i18n('UPLOAD/ERROR_FILE_IS_TOO_BIG'));

						return false;
					}

					return true;
				})
				.on('onStart', id => {
					let item = this.getAttachmentById(id);
					if (item) {
						item
							.waiting(false)
							.uploading(true)
							.complete(false);
					}
				})
				.on('onComplete', (id, result, data) => {
					const attachment = this.getAttachmentById(id),
						response = data?.Result || {},
						errorCode = response.ErrorCode,
						attachmentJson = result && response.Attachment;

					let error = '';
					if (null != errorCode) {
						error = getUploadErrorDescByCode(errorCode);
					} else if (!attachmentJson) {
						error = i18n('UPLOAD/ERROR_UNKNOWN');
					}

					if (attachment) {
						if (error) {
							attachment
								.waiting(false)
								.uploading(false)
								.complete(true)
								.error(error + '\n' + response.ErrorMessage);
						} else if (attachmentJson) {
							attachment
								.waiting(false)
								.uploading(false)
								.complete(true);
							attachment.fileName(attachmentJson.Name);
							attachment.size(attachmentJson.Size ? pInt(attachmentJson.Size) : 0);
							attachment.tempName(attachmentJson.TempName ? attachmentJson.TempName : '');
							attachment.isInline = false;
						}
					}
				});

			this.addAttachmentEnabled(true);

			addShortcut('q', 'meta', ScopeCompose, ()=>false);
			addShortcut('w', 'meta', ScopeCompose, ()=>false);

			addShortcut('m', 'meta', ScopeCompose, () => {
				this.identitiesDropdownTrigger(true);
				return false;
			});

			addShortcut('arrowdown', 'meta', ScopeCompose, () => {
				this.skipCommand();
				return false;
			});

			addShortcut('s', 'meta', ScopeCompose, () => {
				this.saveCommand();
				return false;
			});
			addShortcut('save', '', ScopeCompose, () => {
				this.saveCommand();
				return false;
			});

			if (Settings.app('allowCtrlEnterOnCompose')) {
				addShortcut('enter', 'meta', ScopeCompose, () => {
					this.sendCommand();
					return false;
				});
			}
			addShortcut('mailsend', '', ScopeCompose, () => {
				this.sendCommand();
				return false;
			});

			addShortcut('escape,close', 'shift', ScopeCompose, () => {
				this.tryToClose();
				return false;
			});

			this.editor(editor => editor[isPlainEditor()?'modePlain':'modeWysiwyg']());
		}

		/**
		 * @param {string} id
		 * @returns {?Object}
		 */
		getAttachmentById(id) {
			return this.attachments.find(item => item && id === item.id);
		}

		/**
		 * @returns {Object}
		 */
		prepareAttachmentsForSendOrSave() {
			const result = {};
			this.attachments.forEach(item => {
				if (item?.complete() && item?.tempName() && item?.enabled()) {
					result[item.tempName()] = [item.fileName(), item.isInline ? '1' : '0', item.CID, item.contentLocation];
				}
			});

			return result;
		}

		/**
		 * @param {MessageModel} message
		 */
		addMessageAsAttachment(message) {
			if (message) {
				let temp = message.subject();
				temp = '.eml' === temp.slice(-4).toLowerCase() ? temp : temp + '.eml';

				const attachment = new ComposeAttachmentModel(message.requestHash, temp, message.size());
				attachment.fromMessage = true;
				attachment.complete(true);
				this.addAttachment(attachment);
			}
		}

		addAttachment(attachment, view, oJua) {
			oJua || attachment.waiting(false).uploading(true);
			attachment.cancel = () => {
				this.attachments.remove(attachment);
				oJua?.cancel(attachment.id);
			};
			this.attachments.push(attachment);
			view && this.attachmentsArea();
		}

		/**
		 * @param {string} url
		 * @param {string} name
		 * @param {number} size
		 * @returns {ComposeAttachmentModel}
		 */
		addAttachmentHelper(url, name, size) {
			const attachment = new ComposeAttachmentModel(url, name, size);
			this.addAttachment(attachment, 1);
			return attachment;
		}

		/**
		 * @param {MessageModel} message
		 * @param {string} type
		 */
		prepareMessageAttachments(message, type) {
			if (message) {
				if (ComposeType.ForwardAsAttachment === type) {
					this.addMessageAsAttachment(message);
				} else if ([
					ComposeType.Reply, ComposeType.ReplyAll,
					ComposeType.Forward, ComposeType.Draft, ComposeType.EditAsNew
				].includes(type)) {
					message.attachments.forEach(item => {
						const attachment = new ComposeAttachmentModel(
							item.download,
							item.fileName,
							item.estimatedSize,
							item.isInline(),
							item.isLinked(),
							item.cid,
							item.contentLocation
						);
						attachment.fromMessage = true;
						this.addAttachment(attachment);
					});
				}
			}
		}

		/**
		 * @param {boolean=} includeAttachmentInProgress = true
		 * @returns {boolean}
		 */
		isEmptyForm(includeAttachmentInProgress = true) {
			const withoutAttachment = includeAttachmentInProgress
				? !this.attachments.length
				: !this.attachments.some(item => item?.complete());

			return (
				!this.to.length &&
				!this.cc.length &&
				!this.bcc.length &&
				!this.replyTo.length &&
				!this.subject.length &&
				withoutAttachment &&
				(!this.oEditor || !this.oEditor.getData())
			);
		}

		reset() {
			this.to('');
			this.cc('');
			this.bcc('');
			this.replyTo('');
			this.subject('');

			this.requestDsn(SettingsUserStore.requestDsn());
			this.requestReadReceipt(SettingsUserStore.requestReadReceipt());
			this.markAsImportant(false);

			this.bodyArea();

			this.aDraftInfo = null;
			this.sInReplyTo = '';
			this.bFromDraft = false;
			this.sReferences = '';

			this.sendError(false);
			this.sendSuccessButSaveError(false);
			this.savedError(false);
			this.savedTime(0);
			this.emptyToError(false);
			this.attachmentsInProcessError(false);

			this.showCc(false);
			this.showBcc(false);
			this.showReplyTo(false);

			this.pgpSign(SettingsUserStore.pgpSign());
			this.pgpEncrypt(SettingsUserStore.pgpEncrypt());

			this.attachments([]);

			this.dragAndDropOver(false);
			this.dragAndDropVisible(false);

			this.draftsFolder('');
			this.draftUid(0);

			this.sending(false);
			this.saving(false);

			this.oEditor?.clear();

			this.dropMailvelope();
		}

		mailvelopeArea() {
			if (!this.mailvelope) {
				/**
				 * Creates an iframe with an editor for a new encrypted mail.
				 * The iframe will be injected into the container identified by selector.
				 * https://mailvelope.github.io/mailvelope/Editor.html
				 */
				let text = this.oEditor.getData(),
					encrypted = PgpUserStore.isEncrypted(text),
					size = SettingsGet('PhpUploadSizes')['post_max_size'],
					quota = pInt(size);
				switch (size.slice(-1)) {
					case 'G': quota *= 1024; // fallthrough
					case 'M': quota *= 1024; // fallthrough
					case 'K': quota *= 1024;
				}
				// Issue: can't select signing key
	//			this.pgpSign(this.pgpSign() || confirm('Sign this message?'));
				mailvelope.createEditorContainer('#mailvelope-editor', PgpUserStore.mailvelopeKeyring, {
					// https://mailvelope.github.io/mailvelope/global.html#EditorContainerOptions
					quota: Math.max(2048, (quota / 1024)) - 48, // (text + attachments) limit in kilobytes
					armoredDraft: encrypted ? text : '', // Ascii Armored PGP Text Block
					predefinedText: encrypted ? '' : (this.oEditor.isHtml() ? htmlToPlain(text) : text),
	/*
					quotedMail: '', // Ascii Armored PGP Text Block mail that should be quoted
					quotedMailIndent: true, // if true the quoted mail will be indented (default: true)
					quotedMailHeader: '', // header to be added before the quoted mail
					keepAttachments: false, // add attachments of quotedMail to editor (default: false)
					// Issue: can't select signing key
					signMsg: this.pgpSign()
	*/
				}).then(editor => this.mailvelope = editor);
			}
			this.viewArea('mailvelope');
		}
		attachmentsArea() {
			this.viewArea('attachments');
		}
		bodyArea() {
			this.viewArea('body');
		}

		allRecipients() {
			return [
					// From/sender is also recipient (Sent mailbox)
	//				this.currentIdentity().email(),
					this.from(),
					this.to(),
					this.cc(),
					this.bcc()
				].join(',').split(',').map(value => getEmail(value.trim())).validUnique();
		}

		initPgpEncrypt() {
			const recipients = this.allRecipients();
			PgpUserStore.hasPublicKeyForEmails(recipients).then(result => {
				console.log({canPgpEncrypt:result});
				this.canPgpEncrypt(result);
			});
			PgpUserStore.mailvelopeHasPublicKeyForEmails(recipients).then(result => {
				console.log({canMailvelope:result});
				this.canMailvelope(result);
				if (!result) {
					'mailvelope' === this.viewArea() && this.bodyArea();
	//				this.dropMailvelope();
				}
			});
		}

		togglePgpSign() {
			this.pgpSign(!this.pgpSign()/* && this.canPgpSign()*/);
		}

		togglePgpEncrypt() {
			this.pgpEncrypt(!this.pgpEncrypt()/* && this.canPgpEncrypt()*/);
		}

		async getMessageRequestParams(sSaveFolder, draft)
		{
			const
				identity = this.currentIdentity(),
				params = {
					IdentityID: identity.id(),
					MessageFolder: this.draftsFolder(),
					MessageUid: this.draftUid(),
					SaveFolder: sSaveFolder,
					From: this.from(),
					To: this.to(),
					Cc: this.cc(),
					Bcc: this.bcc(),
					ReplyTo: this.replyTo(),
					subject: this.subject(),
					DraftInfo: this.aDraftInfo,
					InReplyTo: this.sInReplyTo,
					References: this.sReferences,
					MarkAsImportant: this.markAsImportant() ? 1 : 0,
					Attachments: this.prepareAttachmentsForSendOrSave(),
					// Only used at send, not at save:
					Dsn: this.requestDsn() ? 1 : 0,
					ReadReceiptRequest: this.requestReadReceipt() ? 1 : 0
				},
				recipients = draft ? [identity.email()] : this.allRecipients(),
				sign = !draft && this.pgpSign() && this.canPgpSign(),
				encrypt = this.pgpEncrypt() && this.canPgpEncrypt(),
				TextIsHtml = this.oEditor.isHtml();

			let Text = this.oEditor.getData();
			if (TextIsHtml) {
				let l;
				do {
					l = Text.length;
					Text = Text
						// Remove Microsoft Office styling
						.replace(/(<[^>]+[;"'])\s*mso-[a-z-]+\s*:[^;"']+/gi, '$1')
						// Remove hubspot data-hs- attributes
						.replace(/(<[^>]+)\s+data-hs-[a-z-]+=("[^"]+"|'[^']+')/gi, '$1');
				} while (l != Text.length)
				params.Html = Text;
				params.Text = htmlToPlain(Text);
			} else {
				params.Text = Text;
			}

			if (this.mailvelope && 'mailvelope' === this.viewArea()) {
				params.Encrypted = draft
					? await this.mailvelope.createDraft()
					: await this.mailvelope.encrypt(recipients);
			} else if (sign || encrypt) {
				let data = new MimePart;
				data.headers['Content-Type'] = 'text/'+(TextIsHtml?'html':'plain')+'; charset="utf-8"';
				data.headers['Content-Transfer-Encoding'] = 'base64';
				data.body = base64_encode(Text);
				if (TextIsHtml) {
					const alternative = new MimePart, plain = new MimePart;
					alternative.headers['Content-Type'] = 'multipart/alternative';
					plain.headers['Content-Type'] = 'text/plain; charset="utf-8"';
					plain.headers['Content-Transfer-Encoding'] = 'base64';
					plain.body = base64_encode(params.Text);
					// First add plain
					alternative.children.push(plain);
					// Now add HTML
					alternative.children.push(data);
					data = alternative;
				}
				if (!draft && sign?.[1]) {
					if ('openpgp' == sign[0]) {
						// Doesn't sign attachments
						params.Html = params.Text = '';
						let signed = new MimePart;
						signed.headers['Content-Type'] =
							'multipart/signed; micalg="pgp-sha256"; protocol="application/pgp-signature"';
						signed.headers['Content-Transfer-Encoding'] = '7Bit';
						signed.children.push(data);
						let signature = new MimePart;
						signature.headers['Content-Type'] = 'application/pgp-signature; name="signature.asc"';
						signature.headers['Content-Transfer-Encoding'] = '7Bit';
						signature.body = await OpenPGPUserStore.sign(data.toString(), sign[1], 1);
						signed.children.push(signature);
						params.Signed = signed.toString();
						params.Boundary = signed.boundary;
						data = signed;
					} else if ('gnupg' == sign[0]) {
						// TODO: sign in PHP fails
	//					params.SignData = data.toString();
						params.SignFingerprint = sign[1].fingerprint;
						params.SignPassphrase = await GnuPGUserStore.sign(sign[1]);
					} else {
						throw 'Signing with ' + sign[0] + ' not yet implemented';
					}
				}
				if (encrypt) {
					if ('openpgp' == encrypt) {
						// Doesn't encrypt attachments
						params.Encrypted = await OpenPGPUserStore.encrypt(data.toString(), recipients);
						params.Signed = '';
					} else if ('gnupg' == encrypt) {
						// Does encrypt attachments
						params.EncryptFingerprints = JSON.stringify(GnuPGUserStore.getPublicKeyFingerprints(recipients));
					} else {
						throw 'Encryption with ' + encrypt + ' not yet implemented';
					}
				}
			}
			return params;
		}
	}

	/**
	 * When view is closed and reopened, fill it with previous data.
	 * This, for example, happens when opening Contacts view to select recipients
	 */
	ComposePopupView.inEdit = ko.observable(false);

	class MailFolderList extends AbstractViewLeft {
		constructor() {
			super();

	//		this.oContentScrollable = null;

			this.composeInEdit = ComposePopupView.inEdit;

			this.systemFolders = FolderUserStore.systemFolders;

			this.moveAction = moveAction;

			this.foldersListWithSingleInboxRootFolder = ko.observable(false);

			this.allowContacts = AppUserStore.allowContacts();

			addComputablesTo(this, {
				folderListVisible: () => {
					let multiple = false,
						inbox, visible,
						result = FolderUserStore.folderList().filter(folder => {
							if (folder.isInbox()) {
								inbox = folder;
							}
							visible = folder.visible();
							multiple |= visible && !folder.isInbox();
							return visible;
						});
					if (inbox && !multiple) {
						inbox.collapsed(false);
					}
					this.foldersListWithSingleInboxRootFolder(!multiple);
					return result;
				}
			});
		}

		onBuild(dom) {
			const qs = s => dom.querySelector(s),
				eqs = (ev, s) => ev.target.closestWithin(s, dom);

			this.oContentScrollable = qs('.b-content');

			dom.addEventListener('click', event => {
				let el = eqs(event, '.e-collapsed-sign');
				if (el) {
					const folder = ko.dataFor(el);
					if (folder) {
						const collapsed = folder.collapsed();
						setExpandedFolder(folder.fullName, collapsed);

						folder.collapsed(!collapsed);
						event.preventDefault();
						event.stopPropagation();
						return;
					}
				}

				el = eqs(event, 'a');
				if (el?.matches('.selectable')) {
					event.preventDefault();
					const folder = ko.dataFor(el);
					if (folder) {
						if (moveAction()) {
							moveAction(false);
							moveMessagesToFolder(
								FolderUserStore.currentFolderFullName(),
								MessagelistUserStore.listCheckedOrSelectedUidsWithSubMails(),
								folder.fullName,
								event.ctrlKey
							);
						} else {
							if (!SettingsUserStore.usePreviewPane()) {
								MessageUserStore.message(null);
							}
	/*
							if (folder.fullName === FolderUserStore.currentFolderFullName()) {
								setFolderHash(folder.fullName, '');
							}
	*/
							hasher.setHash(
								mailBox(folder.fullNameHash, 1,
									(event.target.matches('.flag-icon') && !folder.isFlagged()) ? 'flagged' : ''
								)
							);
						}

						AppUserStore.focusedState(Scope.MessageList);
					}
				}
			});

			addShortcut('arrowup,arrowdown', '', Scope.FolderList, event => {
				let items = [], index = 0;
				dom.querySelectorAll('li a:not(.hidden)').forEach(node => {
					if (node.offsetHeight || node.getClientRects().length) {
						items.push(node);
						if (node.matches('.focused')) {
							node.classList.remove('focused');
							index = items.length - 1;
						}
					}
				});
				if (items.length) {
					if ('ArrowUp' === event.key) {
						index && --index;
					} else if (index < items.length - 1) {
						++index;
					}
					items[index].classList.add('focused');
					this.scrollToFocused();
				}

				return false;
			});

			addShortcut('enter,open', '', Scope.FolderList, () => {
				const item = qs('li a:not(.hidden).focused');
				if (item) {
					AppUserStore.focusedState(Scope.MessageList);
					item.click();
				}

				return false;
			});

			addShortcut('space', '', Scope.FolderList, () => {
				const item = qs('li a:not(.hidden).focused'),
					folder = item && ko.dataFor(item);
				if (folder) {
					const collapsed = folder.collapsed();
					setExpandedFolder(folder.fullName, collapsed);
					folder.collapsed(!collapsed);
				}

				return false;
			});

	//		addShortcut('tab', 'shift', Scope.FolderList, () => {
			addShortcut('escape,tab,arrowright', '', Scope.FolderList, () => {
				AppUserStore.focusedState(Scope.MessageList);
				moveAction(false);
				return false;
			});

			AppUserStore.focusedState.subscribe(value => {
				let el = qs('li a.focused');
				el?.classList.remove('focused');
				if (Scope.FolderList === value) {
					el = qs('li a.selected');
					el?.classList.add('focused');
				}
			});
		}

		scrollToFocused() {
			const scrollable = this.oContentScrollable;
			if (scrollable) {
				let block, focused = scrollable.querySelector('li a.focused');
				if (focused) {
					const fRect = focused.getBoundingClientRect(),
						sRect = scrollable.getBoundingClientRect();
					if (fRect.top < sRect.top) {
						block = 'start';
					} else if (fRect.bottom > sRect.bottom) {
						block = 'end';
					}
					block && focused.scrollIntoView(block === 'start');
				}
			}
		}

		composeClick() {
			showMessageComposer();
		}

		createFolder() {
			showScreenPopup(FolderCreatePopupView);
		}

		configureFolders() {
			hasher.setHash(settings('folders'));
		}

		contactsClick() {
			if (this.allowContacts) {
				showScreenPopup(ContactsPopupView);
			}
		}
	}

	class FolderClearPopupView extends AbstractViewPopup {
		constructor() {
			super('FolderClear');

			this.addObservables({
				folder: null,
				clearing: false
			});

			this.addComputables({
				dangerDescHtml: () => {
	//				const folder = this.folder();
	//				return i18n('POPUPS_CLEAR_FOLDER/DANGER_DESC_HTML_1', { FOLDER: folder.fullName.replace(folder.delimiter, ' / ') });
					return i18n('POPUPS_CLEAR_FOLDER/DANGER_DESC_HTML_1', { FOLDER: this.folder()?.localName() });
				}
			});

			decorateKoCommands(this, {
				clearCommand: self => !self.clearing()
			});
		}

		clearCommand() {
			const folder = this.folder();
			if (folder) {
				this.clearing(true);
				Remote.request('FolderClear', iError => {
					folder.totalEmails(0);
					folder.unreadEmails(0);
					MessageUserStore.message(null);
					MessagelistUserStore.reload(true, true);
					this.clearing(false);
					iError ? alert(getNotification(iError)) : this.close();
				}, {
					Folder: folder.fullName
				});
			}
		}

		onShow(folder) {
			this.clearing(false);
			this.folder(folder);
		}
	}

	class AdvancedSearchPopupView extends AbstractViewPopup {
		constructor() {
			super('AdvancedSearch');

			this.addObservables({
				from: '',
				to: '',
				subject: '',
				text: '',
				repliedValue: -1,
				selectedDateValue: -1,
				selectedTreeValue: '',

				hasAttachment: false,
				starred: false,
				unseen: false
			});

			this.showMultisearch = koComputable(() => FolderUserStore.hasCapability('MULTISEARCH'));

			this.repliedOptions = koComputable(() => {
				trigger();
				return [
					{ id: -1, name: '' },
					{ id: 1, name: i18n('GLOBAL/YES') },
					{ id: 0, name: i18n('GLOBAL/NO') }
				];
			});

			this.selectedDates = koComputable(() => {
				trigger();
				let prefix = 'SEARCH/LABEL_DATE_';
				return [
					{ id: -1, name: i18n(prefix + 'ALL') },
					{ id: 3, name: i18n(prefix + '3_DAYS') },
					{ id: 7, name: i18n(prefix + '7_DAYS') },
					{ id: 30, name: i18n(prefix + 'MONTH') },
					{ id: 90, name: i18n(prefix + '3_MONTHS') },
					{ id: 180, name: i18n(prefix + '6_MONTHS') },
					{ id: 365, name: i18n(prefix + 'YEAR') }
				];
			});

			this.selectedTree = koComputable(() => {
				trigger();
				let prefix = 'SEARCH/LABEL_SUBFOLDERS_';
				return [
					{ id: '', name: i18n(prefix + 'NONE') },
					{ id: 'subtree-one', name: i18n(prefix + 'SUBTREE_ONE') },
					{ id: 'subtree', name: i18n(prefix + 'SUBTREE') }
				];
			});
		}

		submitForm() {
			const search = this.buildSearchString();
			if (search) {
				MessagelistUserStore.mainSearch(search);
			}

			this.close();
		}

		buildSearchString() {
			const
				self = this,
				data = new FormData(),
				append = (key, value) => value.length && data.append(key, value);

			append('from', self.from().trim());
			append('to', self.to().trim());
			append('subject', self.subject().trim());
			append('text', self.text().trim());
			append('in', self.selectedTreeValue());
			if (-1 < self.selectedDateValue()) {
				let d = new Date();
				d.setDate(d.getDate() - self.selectedDateValue());
				append('since', d.toISOString().split('T')[0]);
			}

			let result = new URLSearchParams(data).toString();

			if (self.hasAttachment()) {
				result += '&attachment';
			}
			if (self.unseen()) {
				result += '&unseen';
			}
			if (self.starred()) {
				result += '&flagged';
			}
			if (1 == self.repliedValue()) {
				result += '&answered';
			}
			if (0 == self.repliedValue()) {
				result += '&unanswered';
			}

			return result.replace(/^&+/, '');
		}

		onShow(search) {
			const self = this,
				params = new URLSearchParams('?'+search);
			self.from(pString(params.get('from')));
			self.to(pString(params.get('to')));
			self.subject(pString(params.get('subject')));
			self.text(pString(params.get('text')));
			self.selectedTreeValue(pString(params.get('in')));
	//		self.selectedDateValue(params.get('since'));
			self.selectedDateValue(-1);
			self.hasAttachment(params.has('attachment'));
			self.starred(params.has('flagged'));
			self.unseen(params.has('unseen'));
			if (params.has('answered')) {
				self.repliedValue(1);
			} else if (params.has('unanswered')) {
				self.repliedValue(0);
			}
		}
	}

	const
		canBeMovedHelper = () => MessagelistUserStore.hasCheckedOrSelected(),

		/**
		 * @param {string} sFolderFullName
		 * @param {number} iSetAction
		 * @param {Array=} aMessages = null
		 * @returns {void}
		 */
		listAction = (...args) => MessagelistUserStore.setAction(...args),

		moveMessagesToFolderType = (toFolderType, bDelete) =>
			rl.app.moveMessagesToFolderType(
				toFolderType,
				FolderUserStore.currentFolderFullName(),
				MessagelistUserStore.listCheckedOrSelectedUidsWithSubMails(),
				bDelete
			);

	let
		iGoToUpOrDownTimeout = 0,
		sLastSearchValue = '';

	class MailMessageList extends AbstractViewRight {
		constructor() {
			super();

			this.allowDangerousActions = SettingsCapa('DangerousActions');

			this.messageList = MessagelistUserStore;
			this.archiveAllowed = MessagelistUserStore.archiveAllowed;
			this.canMarkAsSpam = MessagelistUserStore.canMarkAsSpam;
			this.isSpamFolder = MessagelistUserStore.isSpamFolder;

			this.composeInEdit = ComposePopupView.inEdit;

			this.isMobile = ThemeStore.isMobile;
			this.leftPanelDisabled = leftPanelDisabled;

			this.popupVisibility = arePopupsVisible;

			this.useCheckboxesInList = SettingsUserStore.useCheckboxesInList;

			this.userUsageProc = FolderUserStore.quotaPercentage;

			this.addObservables({
				moreDropdownTrigger: false,
				sortDropdownTrigger: false,

				focusSearch: false
			});

			// append drag and drop
			this.dragOver = ko.observable(false).extend({ throttle: 1 });
			this.dragOverEnter = ko.observable(false).extend({ throttle: 1 });

			this.addComputables({

				sortSupported: () =>
					FolderUserStore.hasCapability('SORT') | FolderUserStore.hasCapability('ESORT'),

				messageListSearchDesc: () => {
					const value = MessagelistUserStore().Search;
					return value ? i18n('MESSAGE_LIST/SEARCH_RESULT_FOR', { SEARCH: value }) : ''
				},

				messageListPaginator: computedPaginatorHelper(MessagelistUserStore.page,
					MessagelistUserStore.pageCount),

				checkAll: {
					read: () => MessagelistUserStore.hasChecked(),
					write: (value) => {
						value = !!value;
						MessagelistUserStore.forEach(message => message.checked(value));
					}
				},

				inputSearch: {
					read: MessagelistUserStore.mainSearch,
					write: value => sLastSearchValue = value
				},

				isIncompleteChecked: () => {
					const c = MessagelistUserStore.listChecked().length;
					return c && MessagelistUserStore().length > c;
				},

				mobileCheckedStateShow: () => ThemeStore.isMobile() ? MessagelistUserStore.hasChecked() : 1,

				mobileCheckedStateHide: () => ThemeStore.isMobile() ? !MessagelistUserStore.hasChecked() : 1,

				sortText: () => {
					let mode = FolderUserStore.sortMode(),
						desc = '' === mode || mode.includes('REVERSE');
					mode = mode.split(/\s+/);
					if (mode.includes('FROM')) {
						 return '@' + (desc ? '⬆' : '⬇');
					}
					if (mode.includes('SUBJECT')) {
						 return '𝐒' + (desc ? '⬆' : '⬇');
					}
					return (mode.includes('SIZE') ? '✉' : '📅') + (desc ? '⬇' : '⬆');
				}
			});

			this.selector = new Selector(
				MessagelistUserStore,
				MessagelistUserStore.selectedMessage,
				MessagelistUserStore.focusedMessage,
				'.messageListItem .actionHandle',
				'.messageListItem .checkboxMessage',
				'.messageListItem.focused'
			);

			this.selector.on('ItemSelect', message => {
				if (message) {
					MessageUserStore.message(MessageModel.fromMessageListItem(message));
					populateMessageBody(MessageUserStore.message());
				} else {
					MessageUserStore.message(null);
				}
			});

			this.selector.on('MiddleClick', message => populateMessageBody(message, true));

			this.selector.on('ItemGetUid', message => (message ? message.generateUid() : ''));

			this.selector.on('AutoSelect', () => MessagelistUserStore.canAutoSelect());

			this.selector.on('UpOrDown', v => this.goToUpOrDown(v));

			addEventListener('mailbox.message-list.selector.go-down',
				e => this.selector.newSelectPosition('ArrowDown', false, e.detail)
			);

			addEventListener('mailbox.message-list.selector.go-up',
				e => this.selector.newSelectPosition('ArrowUp', false, e.detail)
			);

			addEventListener('mailbox.message.show', e => {
				const sFolder = e.detail.Folder, iUid = e.detail.Uid;

				const message = MessagelistUserStore.find(
					item => sFolder === item?.folder && iUid == item?.uid
				);

				if ('INBOX' === sFolder) {
					hasher.setHash(mailBox(sFolder));
				}

				if (message) {
					this.selector.selectMessageItem(message);
				} else {
					if ('INBOX' !== sFolder) {
						hasher.setHash(mailBox(sFolder));
					}
					if (sFolder && iUid) {
						MessageUserStore.message(MessageModel.fromMessageListItem(null));
						MessageUserStore.message().folder = sFolder;
						MessageUserStore.message().uid = iUid;

						populateMessageBody(MessageUserStore.message());
					} else {
						MessageUserStore.message(null);
					}
				}
			});

			MessagelistUserStore.endHash.subscribe((() =>
				this.selector.scrollToFocused()
			).throttle(50));

			decorateKoCommands(this, {
				forwardCommand: canBeMovedHelper,
				deleteWithoutMoveCommand: canBeMovedHelper,
				deleteCommand: canBeMovedHelper,
				archiveCommand: canBeMovedHelper,
				spamCommand: canBeMovedHelper,
				notSpamCommand: canBeMovedHelper,
				moveCommand: canBeMovedHelper,
			});
		}

		changeSort(self, event) {
			FolderUserStore.sortMode(event.target.closest('li').dataset.sort);
			this.reload();
		}

		clear() {
			if (SettingsCapa('DangerousActions')) {
				showScreenPopup(FolderClearPopupView, [FolderUserStore.currentFolder()]);
			}
		}

		reload() {
			if (!MessagelistUserStore.isLoading()) {
				MessagelistUserStore.reload(false, true);
			}
		}

		forwardCommand() {
			showMessageComposer([
				ComposeType.ForwardAsAttachment,
				MessagelistUserStore.listCheckedOrSelected()
			]);
		}

		deleteWithoutMoveCommand() {
			if (SettingsCapa('DangerousActions')) {
				moveMessagesToFolderType(FolderType.Trash, true);
			}
		}

		deleteCommand() {
			moveMessagesToFolderType(FolderType.Trash);
		}

		archiveCommand() {
			moveMessagesToFolderType(FolderType.Archive);
		}

		spamCommand() {
			moveMessagesToFolderType(FolderType.Spam);
		}

		notSpamCommand() {
			moveMessagesToFolderType(FolderType.NotSpam);
		}

		moveCommand(vm, event) {
			if (this.mobileCheckedStateShow()) {
				if (vm && event?.preventDefault) {
					event.preventDefault();
					event.stopPropagation();
				}

				let b = moveAction();
				AppUserStore.focusedState(b ? Scope.MessageList : Scope.FolderList);
				moveAction(!b);
			}
		}

		composeClick() {
			showMessageComposer();
		}

		goToUpOrDown(up) {
			if (MessagelistUserStore.hasChecked()) {
				return false;
			}

			clearTimeout(iGoToUpOrDownTimeout);
			iGoToUpOrDownTimeout = setTimeout(() => {
				let prev, next, temp, current;

				this.messageListPaginator().find(item => {
					if (item) {
						if (current) {
							next = item;
						}

						if (item.current) {
							current = item;
							prev = temp;
						}

						if (next) {
							return true;
						}

						temp = item;
					}

					return false;
				});

				if (up ? prev : next) {
					if (SettingsUserStore.usePreviewPane() || MessageUserStore.message()) {
						this.selector.iSelectNextHelper = up ? -1 : 1;
					} else {
						this.selector.iFocusedNextHelper = up ? -1 : 1;
					}
					this.selector.unselect();
					this.gotoPage(up ? prev : next);
				}
			}, 350);

			return true;
		}

		cancelSearch() {
			MessagelistUserStore.mainSearch('');
			this.focusSearch(false);
		}

		cancelThreadUid() {
			// history.go(-1) better?
			hasher.setHash(
				mailBox(
					FolderUserStore.currentFolderFullNameHash(),
					MessagelistUserStore.pageBeforeThread(),
					MessagelistUserStore.listSearch()
				)
			);
		}

		/**
		 * @param {string} sToFolderFullName
		 * @param {boolean} bCopy
		 * @returns {boolean}
		 */
		moveSelectedMessagesToFolder(sToFolderFullName, bCopy) {
			if (MessagelistUserStore.hasCheckedOrSelected()) {
				moveMessagesToFolder(
					FolderUserStore.currentFolderFullName(),
					MessagelistUserStore.listCheckedOrSelectedUidsWithSubMails(),
					sToFolderFullName,
					bCopy
				);
			}

			return false;
		}

		listSetSeen() {
			listAction(
				FolderUserStore.currentFolderFullName(),
				MessageSetAction.SetSeen,
				MessagelistUserStore.listCheckedOrSelected()
			);
		}

		listSetAllSeen() {
			let sFolderFullName = FolderUserStore.currentFolderFullName(),
				iThreadUid = MessagelistUserStore.endThreadUid();
			if (sFolderFullName) {
				let cnt = 0;
				const uids = [];

				let folder = getFolderFromCacheList(sFolderFullName);
				if (folder) {
					MessagelistUserStore.forEach(message => {
						if (message.isUnseen()) {
							++cnt;
						}

						message.flags.push('\\seen');
	//					message.flags.valueHasMutated();
						iThreadUid && uids.push(message.uid);
					});

					if (iThreadUid) {
						folder.unreadEmails(Math.max(0, folder.unreadEmails() - cnt));
					} else {
						folder.unreadEmails(0);
					}

					MessageFlagsCache.clearFolder(sFolderFullName);

					Remote.request('MessageSetSeenToAll', null, {
						Folder: sFolderFullName,
						SetAction: 1,
						ThreadUids: uids.join(',')
					});

					MessagelistUserStore.reloadFlagsAndCachedMessage();
				}
			}
		}

		listUnsetSeen() {
			listAction(
				FolderUserStore.currentFolderFullName(),
				MessageSetAction.UnsetSeen,
				MessagelistUserStore.listCheckedOrSelected()
			);
		}

		listSetFlags() {
			listAction(
				FolderUserStore.currentFolderFullName(),
				MessageSetAction.SetFlag,
				MessagelistUserStore.listCheckedOrSelected()
			);
		}

		listUnsetFlags() {
			listAction(
				FolderUserStore.currentFolderFullName(),
				MessageSetAction.UnsetFlag,
				MessagelistUserStore.listCheckedOrSelected()
			);
		}

		flagMessages(currentMessage) {
			const checked = MessagelistUserStore.listCheckedOrSelected();
			if (currentMessage) {
				const checkedUids = checked.map(message => message.uid);
				if (checkedUids.includes(currentMessage.uid)) {
					listAction(
						currentMessage.folder,
						currentMessage.isFlagged() ? MessageSetAction.UnsetFlag : MessageSetAction.SetFlag,
						checked
					);
				} else {
					listAction(
						currentMessage.folder,
						currentMessage.isFlagged() ? MessageSetAction.UnsetFlag : MessageSetAction.SetFlag,
						[currentMessage]
					);
				}
			}
		}

		flagMessagesFast(bFlag) {
			const checked = MessagelistUserStore.listCheckedOrSelected();
			if (checked.length) {
				if (undefined === bFlag) {
					const flagged = checked.filter(message => message.isFlagged());
					listAction(
						checked[0].folder,
						checked.length === flagged.length ? MessageSetAction.UnsetFlag : MessageSetAction.SetFlag,
						checked
					);
				} else {
					listAction(
						checked[0].folder,
						!bFlag ? MessageSetAction.UnsetFlag : MessageSetAction.SetFlag,
						checked
					);
				}
			}
		}

		seenMessagesFast(seen) {
			const checked = MessagelistUserStore.listCheckedOrSelected();
			if (checked.length && null != seen) {
				listAction(
					checked[0].folder,
					seen ? MessageSetAction.SetSeen : MessageSetAction.UnsetSeen,
					checked
				);
			}
		}

		gotoPage(page) {
			page && hasher.setHash(
				mailBox(
					FolderUserStore.currentFolderFullNameHash(),
					page.value,
					MessagelistUserStore.listSearch(),
					MessagelistUserStore.threadUid()
				)
			);
		}

		gotoThread(message) {
			if (0 < message?.threadsLen()) {
				MessagelistUserStore.pageBeforeThread(MessagelistUserStore.page());

				hasher.setHash(
					mailBox(FolderUserStore.currentFolderFullNameHash(), 1, MessagelistUserStore.listSearch(), message.uid)
				);
			}
		}

		listEmptyMessage() {
			if (!this.dragOver()
			 && !MessagelistUserStore().length
			 && !MessagelistUserStore.isLoading()
			 && !MessagelistUserStore.error()) {
				 return i18n('MESSAGE_LIST/EMPTY_' + (MessagelistUserStore.listSearch() ? 'SEARCH_' : '') + 'LIST');
			}
			return '';
		}

		clearListIsVisible() {
			return (
				!this.messageListSearchDesc() &&
				!MessagelistUserStore.error() &&
				!MessagelistUserStore.endThreadUid() &&
				MessagelistUserStore().length &&
				(MessagelistUserStore.isSpamFolder() || MessagelistUserStore.isTrashFolder())
			);
		}

		onBuild(dom) {
			const b_content = dom.querySelector('.b-content'),
				eqs = (ev, s) => ev.target.closestWithin(s, dom);

			this.selector.init(b_content, Scope.MessageList);

			addEventsListeners(dom, {
				click: event => {
					ThemeStore.isMobile() && !eqs(event, '.toggleLeft') && leftPanelDisabled(true);

					if (eqs(event, '.messageList') && Scope.MessageView === AppUserStore.focusedState()) {
						AppUserStore.focusedState(Scope.MessageList);
					}

					let el = eqs(event, '.e-paginator a');
					el && this.gotoPage(ko.dataFor(el));

					eqs(event, '.checkboxCheckAll') && this.checkAll(!this.checkAll());

					el = eqs(event, '.flagParent');
					el && this.flagMessages(ko.dataFor(el));

					el = eqs(event, '.threads-len');
					el && this.gotoThread(ko.dataFor(el));
				},
				dblclick: event => {
					let el = eqs(event, '.actionHandle');
					el && this.gotoThread(ko.dataFor(el));
				}
			});

			// initUploaderForAppend

			if (Settings.app('allowAppendMessage')) {
				const oJua = new Jua({
					action: serverRequest('Append'),
					name: 'AppendFile',
					limit: 1,
					hidden: {
						Folder: () => FolderUserStore.currentFolderFullName()
					},
					dragAndDropElement: dom.querySelector('.listDragOver'),
					dragAndDropBodyElement: b_content
				});

				this.dragOver.subscribe(value => value && this.selector.scrollToTop());

				oJua
					.on('onDragEnter', () => this.dragOverEnter(true))
					.on('onDragLeave', () => this.dragOverEnter(false))
					.on('onBodyDragEnter', () => this.dragOver(true))
					.on('onBodyDragLeave', () => this.dragOver(false))
					.on('onSelect', (sUid, oData) => {
						if (sUid && oData && 'message/rfc822' === oData.Type) {
							MessagelistUserStore.loading(true);
							return true;
						}

						return false;
					})
					.on('onComplete', () => MessagelistUserStore.reload(true, true));
			}

			// initShortcuts

			addShortcut('enter,open', '', Scope.MessageList, () => {
				if (formFieldFocused()) {
					MessagelistUserStore.mainSearch(sLastSearchValue);
					return false;
				}
				if (MessageUserStore.message() && MessagelistUserStore.canAutoSelect()) {
					isFullscreen() || toggleFullscreen();
					return false;
				}
			});

			// archive (zip)
			registerShortcut('z', '', [Scope.MessageList, Scope.MessageView], () => {
				this.archiveCommand();
				return false;
			});

			// delete
			registerShortcut('delete', 'shift', Scope.MessageList, () => {
				MessagelistUserStore.listCheckedOrSelected().length && this.deleteWithoutMoveCommand();
				return false;
			});
	//		registerShortcut('3', 'shift', Scope.MessageList, () => {
			registerShortcut('delete', '', Scope.MessageList, () => {
				MessagelistUserStore.listCheckedOrSelected().length && this.deleteCommand();
				return false;
			});

			// check mail
			addShortcut('r', 'meta', [Scope.FolderList, Scope.MessageList, Scope.MessageView], () => {
				this.reload();
				return false;
			});

			// check all
			registerShortcut('a', 'meta', Scope.MessageList, () => {
				this.checkAll(!(this.checkAll() && !this.isIncompleteChecked()));
				return false;
			});

			// write/compose (open compose popup)
			registerShortcut('w,c,new', '', [Scope.MessageList, Scope.MessageView], () => {
				showMessageComposer();
				return false;
			});

			// important - star/flag messages
			registerShortcut('i', '', [Scope.MessageList, Scope.MessageView], () => {
				this.flagMessagesFast();
				return false;
			});

			registerShortcut('t', '', [Scope.MessageList], () => {
				let message = MessagelistUserStore.selectedMessage() || MessagelistUserStore.focusedMessage();
				if (0 < message?.threadsLen()) {
					this.gotoThread(message);
				}
				return false;
			});

			// move
			registerShortcut('insert', '', Scope.MessageList, () => {
				this.moveCommand();
				return false;
			});

			// read
			registerShortcut('q', '', [Scope.MessageList, Scope.MessageView], () => {
				this.seenMessagesFast(true);
				return false;
			});

			// unread
			registerShortcut('u', '', [Scope.MessageList, Scope.MessageView], () => {
				this.seenMessagesFast(false);
				return false;
			});

			addShortcut('f,mailforward', 'shift', [Scope.MessageList, Scope.MessageView], () => {
				this.forwardCommand();
				return false;
			});

			if (SettingsCapa('Search')) {
				// search input focus
				addShortcut('/', '', [Scope.MessageList, Scope.MessageView], () => {
					this.focusSearch(true);
					return false;
				});
			}

			// cancel search
			addShortcut('escape', '', Scope.MessageList, () => {
				if (this.messageListSearchDesc()) {
					this.cancelSearch();
					return false;
				} else if (MessagelistUserStore.endThreadUid()) {
					this.cancelThreadUid();
					return false;
				}
			});

			// change focused state
			addShortcut('tab', 'shift', Scope.MessageList, () => {
				AppUserStore.focusedState(Scope.FolderList);
				return false;
			});
			addShortcut('arrowleft', '', Scope.MessageList, () => {
				AppUserStore.focusedState(Scope.FolderList);
				return false;
			});
			addShortcut('tab,arrowright', '', Scope.MessageList, () => {
				if (MessageUserStore.message()){
					AppUserStore.focusedState(Scope.MessageView);
					return false;
				}
			});

			addShortcut('arrowleft', 'meta', Scope.MessageView, ()=>false);
			addShortcut('arrowright', 'meta', Scope.MessageView, ()=>false);

			addShortcut('f', 'meta', Scope.MessageList, ()=>this.advancedSearchClick());
		}

		advancedSearchClick() {
			showScreenPopup(AdvancedSearchPopupView, [MessagelistUserStore.mainSearch()]);
		}

		quotaTooltip() {
			return i18n('MESSAGE_LIST/QUOTA_SIZE', {
				SIZE: FileInfo.friendlySize(FolderUserStore.quotaUsage()),
				PROC: FolderUserStore.quotaPercentage(),
				LIMIT: FileInfo.friendlySize(FolderUserStore.quotaLimit())
			}).replace(/<[^>]+>/g, '');
		}
	}

	const
		// RFC2045
		QPDecodeIn = /=([0-9A-F]{2})/g,
		QPDecodeOut = (...args) => String.fromCharCode(parseInt(args[1], 16));

	function ParseMime(text)
	{
		class MimePart
		{
	/*
			constructor() {
				this.id = 0;
				this.start = 0;
				this.end = 0;
				this.parts = [];
				this.bodyStart = 0;
				this.bodyEnd = 0;
				this.boundary = '';
				this.bodyText = '';
				this.headers = {};
			}
	*/

			header(name) {
				return this.headers?.[name];
			}

			headerValue(name) {
				return this.header(name)?.value;
			}

			get raw() {
				return text.slice(this.start, this.end);
			}

			get bodyRaw() {
				return text.slice(this.bodyStart, this.bodyEnd);
			}

			get body() {
				let body = this.bodyRaw,
	//				charset = this.header('content-type')?.params.charset,
					encoding = this.headerValue('content-transfer-encoding');
				if ('quoted-printable' == encoding) {
					body = body.replace(/=\r?\n/g, '').replace(QPDecodeIn, QPDecodeOut);
				} else if ('base64' == encoding) {
					body = atob(body.replace(/\r?\n/g, ''));
				}
	/*
				try {
					// https://developer.mozilla.org/en-US/docs/Web/API/Encoding_API/Encodings
					return new TextDecoder(charset).decode(Uint8Array.from(body, c => c.charCodeAt(0)));
	//				return new TextDecoder(charset).decode(new TextEncoder().encode(body));
				} catch (e) {
				}
	*/
				return body;
			}

			get dataUrl() {
				let body = this.bodyRaw,
					encoding = this.headerValue('content-transfer-encoding');
				if ('base64' == encoding) {
					body = body.replace(/\r?\n/g, '');
				} else {
					if ('quoted-printable' == encoding) {
						body = body.replace(/=\r?\n/g, '').replace(QPDecodeIn, QPDecodeOut);
					}
					body = btoa(body);
				}
				return 'data:' + this.headerValue('content-type') + ';base64,' + body;
			}

			forEach(fn) {
				fn(this);
				this.parts.forEach(part => part.forEach(fn));
			}

			getByContentType(type) {
				if (type == this.headerValue('content-type')) {
					return this;
				}
				let i = 0, p = this.parts, part;
				for (i; i < p.length; ++i) {
					if ((part = p[i].getByContentType(type))) {
						return part;
					}
				}
			}
		}

		const ParsePart = (mimePart, start_pos = 0, id = '') =>
		{
			let part = new MimePart,
				head = mimePart.match(/^[\s\S]+?\r?\n\r?\n/)?.[0],
				headers = {};
			if (id) {
				part.id = id;
				part.start = start_pos;
				part.end = start_pos + mimePart.length;
			}
			part.parts = [];

			// get headers
			if (head) {
				head.replace(/\r?\n\s+/g, ' ').split(/\r?\n/).forEach(header => {
					let match = header.match(/^([^:]+):\s*([^;]+)/),
						params = {};
					if (match) {
						[...header.matchAll(/;\s*([^;=]+)=\s*"?([^;"]+)"?/g)].forEach(param =>
							params[param[1].trim().toLowerCase()] = param[2].trim()
						);
						headers[match[1].trim().toLowerCase()] = {
							value: match[2].trim(),
							params: params
						};
					}
				});

				// get body
				part.bodyStart = start_pos + head.length;
				part.bodyEnd = start_pos + mimePart.length;

				// get child parts
				let boundary = headers['content-type']?.params.boundary;
				if (boundary) {
					part.boundary = boundary;
					let regex = new RegExp('(?:^|\r?\n)--' + boundary + '(?:--)?(?:\r?\n|$)', 'g'),
						body = mimePart.slice(head.length),
						bodies = body.split(regex),
						pos = part.bodyStart;
					[...body.matchAll(regex)].forEach(([boundary], index) => {
						if (!index) {
							// Mostly something like: "This is a multi-part message in MIME format."
							part.bodyText = bodies[0];
						}
						// Not the end?
						if ('--' != boundary.trim().slice(-2)) {
							pos += bodies[index].length + boundary.length;
							part.parts.push(ParsePart(bodies[1+index], pos, ((id ? id + '.' : '') + (1+index))));
						}
					});
				}

				part.headers = headers;
			}

			return part;
		};

		return ParsePart(text);
	}

	/**
	 * @param string data
	 * @param MessageModel message
	 */
	function MimeToMessage(data, message)
	{
		let signed;
		const struct = ParseMime(data);
		if (struct.headers) {
			let html = struct.getByContentType('text/html');
			html = html ? html.body : '';

			if (struct.headers.subject) {
				message.subject(struct.headers.subject.value);
			}
			['from','to'].forEach(name => {
				if (struct.headers[name] && !message[name].length) {
					let mail = new EmailModel;
					mail.parse(struct.headers[name].value);
					message[name].push(mail);
				}
			});

			struct.forEach(part => {
				let cd = part.header('content-disposition'),
					cid = part.header('content-id'),
					type = part.header('content-type');
				if (cid || cd) {
					// if (cd && 'attachment' === cd.value) {
					let attachment = new AttachmentModel;
					attachment.mimeType = type.value;
					attachment.fileName = type.name || (cd && cd.params.filename) || '';
					attachment.fileNameExt = attachment.fileName.replace(/^.+(\.[a-z]+)$/, '$1');
					attachment.fileType = FileInfo.getType('', type.value);
					attachment.url = part.dataUrl;
					attachment.friendlySize = FileInfo.friendlySize(part.body.length);
	/*
					attachment.isThumbnail = false;
					attachment.contentLocation = '';
					attachment.download = '';
					attachment.folder = '';
					attachment.uid = '';
					attachment.mimeIndex = part.id;
					attachment.framed = false;
	*/
					attachment.cid = cid ? cid.value : '';
					if (cid && html) {
						let cid = 'cid:' + attachment.contentId(),
							found = html.includes(cid);
						attachment.isInline(found);
						attachment.isLinked(found);
						found && (html = html
							.replace('src="' + cid + '"', 'src="' + attachment.url + '"')
							.replace("src='" + cid + "'", "src='" + attachment.url + "'")
						);
					} else {
						message.attachments.push(attachment);
					}
				} else if ('multipart/signed' === type.value && 'application/pgp-signature' === type.params.protocol) {
					signed = {
						MicAlg: type.micalg,
						BodyPart: part.parts[0],
						SigPart: part.parts[1]
					};
				}
			});

			const text = struct.getByContentType('text/plain');
			message.plain(text ? text.body : '');
			message.html(html);
		} else {
			message.plain(data);
		}

		if (!signed && message.plain().includes(BEGIN_PGP_MESSAGE)) {
			signed = true;
		}
		message.pgpSigned(signed);

		// TODO: Verify instantly?
	}

	const
		oMessageScrollerDom = () => elementById('messageItem') || {},

		currentMessage = MessageUserStore.message;

	class MailMessageView extends AbstractViewRight {
		constructor() {
			super();

			const
				/**
				 * @param {Function} fExecute
				 * @param {Function} fCanExecute = true
				 * @returns {Function}
				 */
				createCommand = (fExecute, fCanExecute) => {
					let fResult = () => {
							fCanExecute() && fExecute.call(null);
							return false;
						};
					fResult.canExecute = fCanExecute;
					return fResult;
				},

				createCommandReplyHelper = type =>
					createCommand(() => this.replyOrforward(type), this.canBeRepliedOrForwarded),

				createCommandActionHelper = (folderType, bDelete) =>
					createCommand(() => {
						const message = currentMessage();
						if (message) {
							currentMessage(null);
							rl.app.moveMessagesToFolderType(folderType, message.folder, [message.uid], bDelete);
						}
					}, this.messageVisibility);

			this.msgDefaultAction = SettingsUserStore.msgDefaultAction;

			this.addObservables({
				showAttachmentControls: !!get(ClientSideKeyNameMessageAttachmentControls),
				downloadAsZipLoading: false,
				showFullInfo: '1' === get(ClientSideKeyNameMessageHeaderFullInfo),
				moreDropdownTrigger: false,

				// viewer
				viewFromShort: '',
				viewFromDkimData: ['none', ''],
				viewToShort: ''
			});

			this.moveAction = moveAction;

			this.allowMessageActions = SettingsCapa('MessageActions');

			const attachmentsActions = Settings.app('attachmentsActions');
			this.attachmentsActions = ko.observableArray(arrayLength(attachmentsActions) ? attachmentsActions : []);

			this.hasCheckedMessages = MessagelistUserStore.hasChecked;
			this.archiveAllowed = MessagelistUserStore.archiveAllowed;
			this.canMarkAsSpam = MessagelistUserStore.canMarkAsSpam;
			this.isDraftFolder = MessagelistUserStore.isDraftFolder;
			this.isSpamFolder = MessagelistUserStore.isSpamFolder;

			this.message = MessageUserStore.message;
			this.messageLoadingThrottle = MessageUserStore.loading;
			this.messagesBodiesDom = MessageUserStore.bodiesDom;
			this.messageError = MessageUserStore.error;

			this.fullScreenMode = isFullscreen;
			this.toggleFullScreen = toggleFullscreen;

			this.downloadAsZipError = ko.observable(false).extend({ falseTimeout: 7000 });

			this.messageDomFocused = ko.observable(false).extend({ rateLimit: 0 });

			// viewer
			this.viewHash = '';

			this.addComputables({
				allowAttachmentControls: () => arrayLength(attachmentsActions) && SettingsCapa('AttachmentsActions'),

				downloadAsZipAllowed: () => this.attachmentsActions.includes('zip')
					&& (currentMessage()?.attachments || [])
						.filter(item => item?.download && !item?.isLinked() && item?.checked())
						.length,

				tagsAllowed: () => FolderUserStore.currentFolder() ? FolderUserStore.currentFolder().tagsAllowed() : false,

				messageVisibility: () => !MessageUserStore.loading() && !!currentMessage(),

				canBeRepliedOrForwarded: () => !MessagelistUserStore.isDraftFolder() && this.messageVisibility(),

				viewFromDkimVisibility: () => 'none' !== this.viewFromDkimData()[0],

				viewFromDkimStatusIconClass:() => {
					switch (this.viewFromDkimData()[0]) {
						case 'none':
							return '';
						case 'pass':
							return 'icon-ok iconcolor-green';
						default:
							return 'icon-warning-alt iconcolor-red';
					}
				},

				viewFromDkimStatusTitle:() => {
					const status = this.viewFromDkimData();
					if (arrayLength(status) && status[0]) {
						return status[1] || 'DKIM: ' + status[0];
					}

					return '';
				},

				pgpSupported: () => currentMessage() && PgpUserStore.isSupported(),

				messageListOrViewLoading:
					() => MessagelistUserStore.isLoading() | MessageUserStore.loading()
			});

			this.addSubscribables({
				message: message => {
					MessageUserStore.activeDom(null);

					if (message) {
						if (this.viewHash !== message.hash) {
							this.scrollMessageToTop();
						}
						this.viewHash = message.hash;
						this.viewFromShort(message.fromToLine(true, true));
						this.viewFromDkimData(message.fromDkimData());
						this.viewToShort(message.toToLine(true, true));
					} else {
						MessagelistUserStore.selectedMessage(null);

						this.viewHash = '';

						this.scrollMessageToTop();
					}
				},

				showFullInfo: value => set(ClientSideKeyNameMessageHeaderFullInfo, value ? '1' : '0')
			});

			// commands
			this.replyCommand = createCommandReplyHelper(ComposeType.Reply);
			this.replyAllCommand = createCommandReplyHelper(ComposeType.ReplyAll);
			this.forwardCommand = createCommandReplyHelper(ComposeType.Forward);
			this.forwardAsAttachmentCommand = createCommandReplyHelper(ComposeType.ForwardAsAttachment);
			this.editAsNewCommand = createCommandReplyHelper(ComposeType.EditAsNew);

			this.deleteCommand = createCommandActionHelper(FolderType.Trash);
			this.deleteWithoutMoveCommand = createCommandActionHelper(FolderType.Trash, true);
			this.archiveCommand = createCommandActionHelper(FolderType.Archive);
			this.spamCommand = createCommandActionHelper(FolderType.Spam);
			this.notSpamCommand = createCommandActionHelper(FolderType.NotSpam);

			decorateKoCommands(this, {
				messageEditCommand: self => self.messageVisibility(),
				goUpCommand: self => !self.messageListOrViewLoading(),
				goDownCommand: self => !self.messageListOrViewLoading()
			});
		}

		closeMessage() {
			currentMessage(null);
		}

		messageEditCommand() {
			currentMessage() && showMessageComposer([ComposeType.Draft, currentMessage()]);
		}

		goUpCommand() {
			fireEvent('mailbox.message-list.selector.go-up',
				!!currentMessage() // bForceSelect
			);
		}

		goDownCommand() {
			fireEvent('mailbox.message-list.selector.go-down',
				!!currentMessage() // bForceSelect
			);
		}

		/**
		 * @param {string} sType
		 * @returns {void}
		 */
		replyOrforward(sType) {
			showMessageComposer([sType, currentMessage()]);
		}

		onBuild(dom) {
			const eqs = (ev, s) => ev.target.closestWithin(s, dom);
			dom.addEventListener('click', event => {
				ThemeStore.isMobile() && leftPanelDisabled(true);

				let el = eqs(event, 'a');
				if (el && 0 === event.button && mailToHelper(el.href)) {
					event.preventDefault();
					event.stopPropagation();
					return;
				}

				if (eqs(event, '.attachmentsPlace .attachmentIconParent')) {
					event.stopPropagation();
				}

				el = eqs(event, '.attachmentsPlace .showPreplay');
				if (el) {
					event.stopPropagation();
					const attachment = ko.dataFor(el);
					if (attachment && SMAudio.supported) {
						switch (true) {
							case SMAudio.supportedMp3 && attachment.isMp3():
								SMAudio.playMp3(attachment.linkDownload(), attachment.fileName);
								break;
							case SMAudio.supportedOgg && attachment.isOgg():
								SMAudio.playOgg(attachment.linkDownload(), attachment.fileName);
								break;
							case SMAudio.supportedWav && attachment.isWav():
								SMAudio.playWav(attachment.linkDownload(), attachment.fileName);
								break;
							// no default
						}
					}
				}

				el = eqs(event, '.attachmentsPlace .attachmentItem .attachmentNameParent');
				if (el) {
					const attachment = ko.dataFor(el);
					if (attachment?.linkDownload()) {
						if ('message/rfc822' == attachment.mimeType) {
							// TODO
							rl.fetch(attachment.linkDownload()).then(response => {
								if (response.ok) {
									response.text().then(text => {
										const oMessage = new MessageModel();
										MimeToMessage(text, oMessage);
										// cleanHTML
										oMessage.viewPopupMessage();
									});
								}
							});
						} else {
							download(attachment.linkDownload(), attachment.fileName);
						}
					}
				}

				if (eqs(event, '.messageItemHeader .subjectParent .flagParent')) {
					const message = currentMessage();
					message && MessagelistUserStore.setAction(
						message.folder,
						message.isFlagged() ? MessageSetAction.UnsetFlag : MessageSetAction.SetFlag,
						[message]
					);
				}
			});

			AppUserStore.focusedState.subscribe(value => {
				if (Scope.MessageView !== value) {
					this.scrollMessageToTop();
					this.scrollMessageToLeft();
				}
			});

			keyScopeReal.subscribe(value => this.messageDomFocused(Scope.MessageView === value));

			// initShortcuts

			// exit fullscreen, back
			addShortcut('escape', '', Scope.MessageView, () => {
				if (!this.viewModelDom.hidden && currentMessage()) {
					const preview = SettingsUserStore.usePreviewPane();
					if (isFullscreen()) {
						exitFullscreen();
						if (preview) {
							AppUserStore.focusedState(Scope.MessageList);
						}
					} else if (!preview) {
						currentMessage(null);
					} else {
						AppUserStore.focusedState(Scope.MessageList);
					}

					return false;
				}
			});

			// fullscreen
			addShortcut('enter,open', '', Scope.MessageView, () => {
				isFullscreen() || toggleFullscreen();
				return false;
			});

			// reply
			registerShortcut('r,mailreply', '', [Scope.MessageList, Scope.MessageView], () => {
				if (currentMessage()) {
					this.replyCommand();
					return false;
				}
				return true;
			});

			// replyAll
			registerShortcut('a', '', [Scope.MessageList, Scope.MessageView], () => {
				if (currentMessage()) {
					this.replyAllCommand();
					return false;
				}
			});
			registerShortcut('mailreply', 'shift', [Scope.MessageList, Scope.MessageView], () => {
				if (currentMessage()) {
					this.replyAllCommand();
					return false;
				}
			});

			// forward
			registerShortcut('f,mailforward', '', [Scope.MessageList, Scope.MessageView], () => {
				if (currentMessage()) {
					this.forwardCommand();
					return false;
				}
			});

			// message information
			registerShortcut('i', 'meta', [Scope.MessageList, Scope.MessageView], () => {
				if (currentMessage()) {
					this.showFullInfo(!this.showFullInfo());
				}
				return false;
			});

			// toggle message blockquotes
			registerShortcut('b', '', [Scope.MessageList, Scope.MessageView], () => {
				const message = currentMessage();
				if (message?.body) {
					message.body.querySelectorAll('.sm-bq-switcher > summary').forEach(node => node.click());
					return false;
				}
			});

			addShortcut('arrowup,arrowleft', 'meta', [Scope.MessageList, Scope.MessageView], () => {
				this.goUpCommand();
				return false;
			});

			addShortcut('arrowdown,arrowright', 'meta', [Scope.MessageList, Scope.MessageView], () => {
				this.goDownCommand();
				return false;
			});

			// print
			addShortcut('p,printscreen', 'meta', [Scope.MessageView, Scope.MessageList], () => {
				currentMessage()?.printMessage();
				return false;
			});

			// delete
			addShortcut('delete', '', Scope.MessageView, () => {
				this.deleteCommand();
				return false;
			});
			addShortcut('delete', 'shift', Scope.MessageView, () => {
				this.deleteWithoutMoveCommand();
				return false;
			});

			// change focused state
			addShortcut('arrowleft', '', Scope.MessageView, () => {
				if (!isFullscreen() && currentMessage() && SettingsUserStore.usePreviewPane()
				 && !oMessageScrollerDom().scrollLeft) {
					AppUserStore.focusedState(Scope.MessageList);
					return false;
				}
			});
			addShortcut('tab', 'shift', Scope.MessageView, () => {
				if (!isFullscreen() && currentMessage() && SettingsUserStore.usePreviewPane()) {
					AppUserStore.focusedState(Scope.MessageList);
				}
				return false;
			});
		}

		/**
		 * @returns {boolean}
		 */
		isDraftOrSentFolder() {
			return MessagelistUserStore.isDraftFolder() || MessagelistUserStore.isSentFolder();
		}

		composeClick() {
			showMessageComposer();
		}

		scrollMessageToTop() {
			oMessageScrollerDom().scrollTop = (50 < oMessageScrollerDom().scrollTop) ? 50 : 0;
		}

		scrollMessageToLeft() {
			oMessageScrollerDom().scrollLeft = 0;
		}

		toggleAttachmentControls() {
			const b = !this.showAttachmentControls();
			this.showAttachmentControls(b);
			set(ClientSideKeyNameMessageAttachmentControls, b);
		}

		downloadAsZip() {
			const hashes = (currentMessage() ? currentMessage().attachments : [])
				.map(item => item?.checked() && !item?.isLinked() ? item.download : '')
				.filter(v => v);
			if (hashes.length) {
				Remote.post('AttachmentsActions', this.downloadAsZipLoading, {
					Do: 'Zip',
					Hashes: hashes
				})
				.then(result => {
					let hash = result?.Result?.FileHash;
					if (hash) {
						download(attachmentDownload(hash), hash+'.zip');
					} else {
						this.downloadAsZipError(true);
					}
				})
				.catch(() => this.downloadAsZipError(true));
			}
		}

		/**
		 * @param {MessageModel} oMessage
		 * @returns {void}
		 */
		showImages() {
			currentMessage().showExternalImages();
		}

		/**
		 * @returns {string}
		 */
		printableCheckedMessageCount() {
			const cnt = MessagelistUserStore.listCheckedOrSelectedUidsWithSubMails().size;
			return 0 < cnt ? (100 > cnt ? cnt : '99+') : '';
		}

		/**
		 * @param {MessageModel} oMessage
		 * @returns {void}
		 */
		readReceipt() {
			let oMessage = currentMessage();
			if (oMessage.readReceipt()) {
				Remote.request('SendReadReceiptMessage', iError => {
					if (!iError) {
						oMessage.flags.push('$mdnsent');
	//					oMessage.flags.valueHasMutated();
						MessageFlagsCache.store(oMessage);
						MessagelistUserStore.reloadFlagsAndCachedMessage();
					}
				}, {
					MessageFolder: oMessage.folder,
					MessageUid: oMessage.uid,
					ReadReceipt: oMessage.readReceipt(),
					subject: i18n('READ_RECEIPT/SUBJECT', { SUBJECT: oMessage.subject() }),
					Text: i18n('READ_RECEIPT/BODY', { 'READ-RECEIPT': AccountUserStore.email() })
				});
			}
		}

		pgpDecrypt() {
			const oMessage = currentMessage();
			PgpUserStore.decrypt(oMessage).then(result => {
				if (result) {
					oMessage.pgpDecrypted(true);
					if (result.data) {
						MimeToMessage(result.data, oMessage);
						oMessage.html() ? oMessage.viewHtml() : oMessage.viewPlain();
						if (result.signatures?.length) {
							oMessage.pgpSigned(true);
							oMessage.pgpVerified({
								signatures: result.signatures,
								success: !!result.signatures.length
							});
						}
					}
				}
			});
		}

		pgpVerify(/*self, event*/) {
			const oMessage = currentMessage()/*, ctrl = event.target.closest('.openpgp-control')*/;
			PgpUserStore.verify(oMessage).then(result => {
				if (result) {
					oMessage.pgpVerified(result);
				}
	/*
				if (result?.success) {
					i18n('OPENPGP/GOOD_SIGNATURE', {
						USER: validKey.user + ' (' + validKey.id + ')'
					});
					message.getText()
				} else {
					const keyIds = arrayLength(signingKeyIds) ? signingKeyIds : null,
						additional = keyIds
							? keyIds.map(item => item?.toHex?.()).filter(v => v).join(', ')
							: '';

					i18n('OPENPGP/ERROR', {
						ERROR: 'message'
					}) + (additional ? ' (' + additional + ')' : '');
				}
	*/
			});
		}

	}

	class MailBoxUserScreen extends AbstractScreen {
		constructor() {
			var styleSheet = createElement('style');
			doc.head.appendChild(styleSheet);
			initOnStartOrLangChange(() =>
				styleSheet.innerText = '.subjectParent:empty::after,.subjectParent .subject:empty::after'
				+'{content:"'+i18n('MESSAGE/EMPTY_SUBJECT_TEXT')+'"}'
			);
			super('mailbox', [
				SystemDropDownUserView,
				MailFolderList,
				MailMessageList,
				MailMessageView
			]);
		}

		/**
		 * @returns {void}
		 */
		updateWindowTitle() {
			const count = Settings.app('listPermanentFiltered') ? 0 : FolderUserStore.foldersInboxUnreadCount(),
				email = AccountUserStore.email();

			rl.setWindowTitle(
				(email
					? '' + (0 < count ? '(' + count + ') ' : ' ') + email + ' - '
					: ''
				) + i18n('TITLES/MAILBOX')
			);
		}

		/**
		 * @returns {void}
		 */
		onShow() {
			this.updateWindowTitle();

			AppUserStore.focusedState('none');
			AppUserStore.focusedState(Scope.MessageList);

			ThemeStore.isMobile() && leftPanelDisabled(true);
		}

		/**
		 * @param {string} folderHash
		 * @param {number} page
		 * @param {string} search
		 * @returns {void}
		 */
		onRoute(folderHash, page, search) {
			const folder = getFolderFromCacheList(getFolderFullName(folderHash.replace(/~([\d]+)$/, '')));
			if (folder) {
				let threadUid = folderHash.replace(/^.+~(\d+)$/, '$1');

				FolderUserStore.currentFolder(folder);

				MessagelistUserStore.page(1 > page ? 1 : page);
				MessagelistUserStore.listSearch(search);
				MessagelistUserStore.threadUid((folderHash === threadUid) ? 0 : pInt(threadUid));

				MessagelistUserStore.reload();
			}
		}

		/**
		 * @returns {void}
		 */
		onStart() {
			super.onStart();

			addEventListener('mailbox.inbox-unread-count', e => {
				FolderUserStore.foldersInboxUnreadCount(e.detail);
	/*			// Disabled in SystemDropDown.html
				const email = AccountUserStore.email();
				AccountUserStore.accounts.forEach(item =>
					email === item?.email && item?.count(e.detail)
				);
	*/
				this.updateWindowTitle();
			});
		}

		/**
		 * @returns {void}
		 */
		onBuild() {
			setTimeout(() => {
				// initMailboxLayoutResizer
				const top = elementById('V-MailMessageList'),
					bottom = elementById('V-MailMessageView'),
					fToggle = () => {
						let layout = SettingsUserStore.layout();
						setLayoutResizer(top, bottom, ClientSideKeyNameMessageListSize,
							(ThemeStore.isMobile() || Layout.NoPreview === layout)
								? 0
								: (Layout.SidePreview === layout ? 'Width' : 'Height')
						);
					};
				if (top && bottom) {
					fToggle();
					addEventListener('rl-layout', fToggle);
				}
			}, 1);

			doc.addEventListener('click', event =>
				event.target.closest('#rl-right') && moveAction(false)
			);
		}

		/**
		 * Parse link as generated by mailBox()
		 * @returns {Array}
		 */
		routes() {
			const
				folder = (request, vals) => request ? pString(vals[0]) : getFolderInboxName(),
				fNormS = (request, vals) => [folder(request, vals), request ? pInt(vals[1]) : 1, decodeURI(pString(vals[2]))];

			return [
				// Folder: INBOX | INBOX.sub | Sent | fullNameHash
				[/^([^/]*)$/, { normalize_: fNormS }],
				// Search: {folder}/{string}
				[/^([a-zA-Z0-9.~_-]+)\/(.+)\/?$/, { normalize_: (request, vals) =>
					[folder(request, vals), 1, decodeURI(pString(vals[1]))]
				}],
				// Page: {folder}/p{int}(/{search})?
				[/^([a-zA-Z0-9.~_-]+)\/p([1-9][0-9]*)(?:\/(.+))?$/, { normalize_: fNormS }]
			];
		}
	}

	const VIEW_MODELS = [];

	class AbstractSettingsScreen extends AbstractScreen {
		/**
		 * @param {Array} viewModels
		 */
		constructor(viewModels) {
			super('settings', viewModels);

			this.menu = ko.observableArray();

			this.oCurrentSubScreen = null;
		}

		onRoute(subName) {
			let settingsScreen = null,
				viewModelDom = null,
				RoutedSettingsViewModel = VIEW_MODELS.find(
					SettingsViewModel => subName === SettingsViewModel.__rlSettingsData.route
				);

			if (RoutedSettingsViewModel) {
				const vmPlace = elementById('rl-settings-subscreen');
				if (RoutedSettingsViewModel.__vm) {
					settingsScreen = RoutedSettingsViewModel.__vm;
				} else if (vmPlace) {
					viewModelDom = createElement('div',{
						id: 'V-Settings-' + RoutedSettingsViewModel.name.replace(/(User|Admin)Settings/,''),
						hidden: ''
					});
					vmPlace.append(viewModelDom);

					settingsScreen = new RoutedSettingsViewModel();
					settingsScreen.viewModelDom = viewModelDom;

					RoutedSettingsViewModel.__dom = viewModelDom;
					RoutedSettingsViewModel.__vm = settingsScreen;

					ko.applyBindingAccessorsToNode(
						viewModelDom,
						{
							template: () => ({ name: RoutedSettingsViewModel.__rlSettingsData.template })
						},
						settingsScreen
					);

					settingsScreen.onBuild?.(viewModelDom);
				} else {
					console.log('Cannot find sub settings view model position: SettingsSubScreen');
				}

				if (settingsScreen) {
					setTimeout(() => {
						// hide
						this.onHide();
						// --

						this.oCurrentSubScreen = settingsScreen;

						// show
						settingsScreen.beforeShow?.();
						i18nToNodes(settingsScreen.viewModelDom);
						settingsScreen.viewModelDom.hidden = false;
						settingsScreen.onShow?.();

						this.menu.forEach(item => {
							item.selected(
								item.route === RoutedSettingsViewModel.__rlSettingsData.route
							);
						});

						(vmPlace || {}).scrollTop = 0;
						// --
					}, 1);
				}
			} else {
				hasher.replaceHash(settings());
			}
		}

		onHide() {
			let subScreen = this.oCurrentSubScreen;
			if (subScreen) {
				subScreen.onHide?.();
				subScreen.viewModelDom.hidden = true;
			}
		}

		onBuild() {
			VIEW_MODELS.forEach(SettingsViewModel => this.menu.push(SettingsViewModel.__rlSettingsData));
		}

		routes() {
			const DefaultViewModel = VIEW_MODELS.find(
					SettingsViewModel => SettingsViewModel.__rlSettingsData.isDefault
				),
				defaultRoute =
					DefaultViewModel ? DefaultViewModel.__rlSettingsData.route : 'general',
				rules = {
					subname: /^(.*)$/,
					normalize_: (rquest, vals) => {
						vals.subname = null == vals.subname ? defaultRoute : pString(vals.subname);
						return [vals.subname];
					}
				};

			return [
				['{subname}/', rules],
				['{subname}', rules],
				['', rules]
			];
		}
	}

	/**
	 * @param {Function} SettingsViewModelClass
	 * @param {string} template
	 * @param {string} labelName
	 * @param {string} route
	 * @param {boolean=} isDefault = false
	 * @returns {void}
	 */
	function settingsAddViewModel(SettingsViewModelClass, template, labelName, route, isDefault = false) {
		let name = SettingsViewModelClass.name.replace(/(User|Admin)Settings/, '');
		SettingsViewModelClass.__rlSettingsData = {
			label: labelName || 'SETTINGS_LABELS/LABEL_' + name.toUpperCase() + '_NAME',
			route: route || name.toLowerCase(),
			selected: ko.observable(false),
			template: template || SettingsViewModelClass.name,
			isDefault: !!isDefault
		};

		VIEW_MODELS.push(SettingsViewModelClass);
	}

	const USER_VIEW_MODELS_HOOKS = [],
		ADMIN_VIEW_MODELS_HOOKS = [];

	/**
	 * @param {Function} callback
	 * @param {string} action
	 * @param {Object=} parameters
	 * @param {?number=} timeout
	 */
	rl.pluginRemoteRequest = (callback, action, parameters, timeout) => {
		rl.app.Remote.request('Plugin' + action, callback, parameters, timeout);
	};

	/**
	 * @param {Function} SettingsViewModelClass
	 * @param {string} labelName
	 * @param {string} template
	 * @param {string} route
	 */
	rl.addSettingsViewModel = (SettingsViewModelClass, template, labelName, route) => {
		USER_VIEW_MODELS_HOOKS.push([SettingsViewModelClass, template, labelName, route]);
	};

	/**
	 * @param {Function} SettingsViewModelClass
	 * @param {string} labelName
	 * @param {string} template
	 * @param {string} route
	 */
	rl.addSettingsViewModelForAdmin = (SettingsViewModelClass, template, labelName, route) => {
		ADMIN_VIEW_MODELS_HOOKS.push([SettingsViewModelClass, template, labelName, route]);
	};

	/**
	 * @param {boolean} admin
	 */
	function runSettingsViewModelHooks(admin) {
		(admin ? ADMIN_VIEW_MODELS_HOOKS : USER_VIEW_MODELS_HOOKS).forEach(view => {
			settingsAddViewModel(view[0], view[1], view[2], view[3]);
		});
	}

	/**
	 * @param {string} pluginSection
	 * @param {string} name
	 * @returns {?}
	 */
	rl.pluginSettingsGet = (pluginSection, name) =>
		SettingsGet('Plugins')?.[pluginSection]?.[name];

	rl.pluginPopupView = AbstractViewPopup;

	class IdentityPopupView extends AbstractViewPopup {
		constructor() {
			super('Identity');

			this.id = '';
			this.addObservables({
				edit: false,
				owner: false,

				email: '',
				emailFocused: false,

				name: '',

				replyTo: '',
				replyToFocused: false,

				bcc: '',
				bccFocused: false,
				bccHasError: false,

				signature: '',
				signatureInsertBefore: false,

				showBcc: false,
				showReplyTo: false,

				submitRequest: false,
				submitError: ''
			});

			this.addSubscribables({
				replyTo: value => {
					if (false === this.showReplyTo() && value.length) {
						this.showReplyTo(true);
					}
				},
				bcc: value => {
					if (false === this.showBcc() && value.length) {
						this.showBcc(true);
					}
				}
			});
	/*
			this.email.valueHasMutated();
			this.replyTo.valueHasMutated();
			this.bcc.valueHasMutated();
	*/
		}

		submitForm(form) {
			if (!this.submitRequest() && form.reportValidity()) {
				this.signature?.__fetchEditorValue?.();
				this.submitRequest(true);
				const data = new FormData(form);
				data.set('Id', this.id);
				data.set('Signature', this.signature());
				data.set('SignatureInsertBefore', this.signatureInsertBefore() ? 1 : 0);
				Remote.request('IdentityUpdate', iError => {
						this.submitRequest(false);
						if (iError) {
							this.submitError(getNotification(iError));
						} else {
							rl.app.accountsAndIdentities();
							this.close();
						}
					}, data
				);
			}
		}

		/**
		 * @param {?IdentityModel} oIdentity
		 */
		onShow(identity) {
			this.showBcc(false);
			this.showReplyTo(false);

			this.submitRequest(false);
			this.submitError('');

			if (identity) {
				this.edit(true);

				this.id = identity.id() || '';
				this.name(identity.name());
				this.email(identity.email());
				this.replyTo(identity.replyTo());
				this.bcc(identity.bcc());
				this.signature(identity.signature());
				this.signatureInsertBefore(identity.signatureInsertBefore());

				this.owner(!this.id);
			} else {
				this.edit(false);

				this.id = Jua.randomId();
				this.name('');
				this.email('');
				this.replyTo('');
				this.bcc('');
				this.signature('');
				this.signatureInsertBefore(false);

				this.owner(false);
			}
		}

		afterShow() {
			this.owner() || this.emailFocused(true);
		}
	}

	class UserSettingsGeneral extends AbstractViewSettings {
		constructor() {
			super();

			this.language = LanguageStore.language;
			this.languages = LanguageStore.languages;

			this.soundNotification = SMAudio.notifications;
			this.notificationSound = ko.observable(SettingsGet('NotificationSound'));
			this.notificationSounds = ko.observableArray(SettingsGet('NewMailSounds'));

			this.desktopNotification = NotificationUserStore.enabled;
			this.isDesktopNotificationAllowed = NotificationUserStore.allowed;

			this.threadsAllowed = AppUserStore.threadsAllowed;

			['layout', 'messageReadDelay', 'messagesPerPage',
			 'editorDefaultType', 'requestReadReceipt', 'requestDsn', 'pgpSign', 'pgpEncrypt',
			 'viewHTML', 'showImages', 'removeColors', 'hideDeleted',
			 'useCheckboxesInList', 'useThreads', 'replySameFolder', 'msgDefaultAction'
			].forEach(name => this[name] = SettingsUserStore[name]);

			this.allowLanguagesOnSettings = !!SettingsGet('AllowLanguagesOnSettings');

			this.languageTrigger = ko.observable(SaveSettingStatus.Idle);

			this.identities = IdentityUserStore;

			addComputablesTo(this, {
				languageFullName: () => convertLangName(this.language()),

				identityMain: () => {
					const list = this.identities();
					return isArray(list) ? list.find(item => item && !item.id()) : null;
				},

				identityMainDesc: () => {
					const identity = this.identityMain();
					return identity ? identity.formattedName() : '---';
				},

				editorDefaultTypes: () => {
					trigger();
					return [
						{ id: EditorDefaultType.Html, name: i18n('SETTINGS_GENERAL/LABEL_EDITOR_HTML') },
						{ id: EditorDefaultType.Plain, name: i18n('SETTINGS_GENERAL/LABEL_EDITOR_PLAIN') }
					];
				},

				msgDefaultActions: () => {
					trigger();
					return [
						{ id: 1, name: i18n('MESSAGE/BUTTON_REPLY') }, // ComposeType.Reply,
						{ id: 2, name: i18n('MESSAGE/BUTTON_REPLY_ALL') } // ComposeType.ReplyAll
					];
				},

				layoutTypes: () => {
					trigger();
					return [
						{ id: Layout.NoPreview, name: i18n('SETTINGS_GENERAL/LABEL_LAYOUT_NO_SPLIT') },
						{ id: Layout.SidePreview, name: i18n('SETTINGS_GENERAL/LABEL_LAYOUT_VERTICAL_SPLIT') },
						{ id: Layout.BottomPreview, name: i18n('SETTINGS_GENERAL/LABEL_LAYOUT_HORIZONTAL_SPLIT') }
					];
				}
			});

			this.addSetting('EditorDefaultType');
			this.addSetting('MsgDefaultAction');
			this.addSetting('MessageReadDelay');
			this.addSetting('MessagesPerPage');
			this.addSetting('Layout');

			this.addSettings(['ViewHTML', 'ShowImages', 'HideDeleted', 'UseCheckboxesInList', 'ReplySameFolder',
				'requestReadReceipt', 'requestDsn', 'pgpSign', 'pgpEncrypt',
				'DesktopNotifications', 'SoundNotification']);

			const fReloadLanguageHelper = (saveSettingsStep) => () => {
					this.languageTrigger(saveSettingsStep);
					setTimeout(() => this.languageTrigger(SaveSettingStatus.Idle), 1000);
				};

			addSubscribablesTo(this, {
				language: value => {
					this.languageTrigger(SaveSettingStatus.Saving);
					translatorReload(false, value)
						.then(fReloadLanguageHelper(SaveSettingStatus.Success), fReloadLanguageHelper(SaveSettingStatus.Failed))
						.then(() => Remote.saveSetting('Language', value));
				},

				removeColors: value => {
					let dom = MessageUserStore.bodiesDom();
					if (dom) {
						dom.innerHTML = '';
					}
					Remote.saveSetting('RemoveColors', value);
				},

				notificationSound: value => {
					Remote.saveSetting('NotificationSound', value);
					Settings.set('NotificationSound', value);
				},

				useThreads: value => {
					MessagelistUserStore([]);
					Remote.saveSetting('UseThreads', value);
				}
			});
		}

		editMainIdentity() {
			const identity = this.identityMain();
			identity && showScreenPopup(IdentityPopupView, [identity]);
		}

		testSoundNotification() {
			SMAudio.playNotification(true);
		}

		testSystemNotification() {
			NotificationUserStore.display('SnappyMail', 'Test notification');
		}

		selectLanguage() {
			showScreenPopup(LanguagesPopupView, [this.language, this.languages(), LanguageStore.userLanguage()]);
		}
	}

	class UserSettingsContacts /*extends AbstractViewSettings*/ {
		constructor() {
			this.contactsAutosave = ko.observable(!!SettingsGet('ContactsAutosave'));

			this.allowContactsSync = ContactUserStore.allowSync;
			this.syncMode = ContactUserStore.syncMode;
			this.syncUrl = ContactUserStore.syncUrl;
			this.syncUser = ContactUserStore.syncUser;
			this.syncPass = ContactUserStore.syncPass;

			this.syncModeOptions = koComputable(() => {
				trigger();
				return [
					{ id: 0, name: i18n('GLOBAL/NO') },
					{ id: 1, name: i18n('GLOBAL/YES') },
					{ id: 2, name: i18n('SETTINGS_CONTACTS/SYNC_READ') },
				];
			});

			this.saveTrigger = koComputable(() =>
					[
						ContactUserStore.syncMode(),
						ContactUserStore.syncUrl(),
						ContactUserStore.syncUser(),
						ContactUserStore.syncPass()
					].join('|')
				)
				.extend({ debounce: 500 });

			this.contactsAutosave.subscribe(value =>
				Remote.saveSettings(null, { ContactsAutosave: value })
			);

			this.saveTrigger.subscribe(() =>
				Remote.request('SaveContactsSyncData', null, {
					Mode: ContactUserStore.syncMode(),
					Url: ContactUserStore.syncUrl(),
					User: ContactUserStore.syncUser(),
					Password: ContactUserStore.syncPass()
				})
			);
		}
	}

	class UserSettingsAccounts /*extends AbstractViewSettings*/ {
		constructor() {
			this.allowAdditionalAccount = SettingsCapa('AdditionalAccounts');
			this.allowIdentities = SettingsCapa('Identities');

			this.accounts = AccountUserStore.accounts;
			this.loading = AccountUserStore.loading;
			this.identities = IdentityUserStore;
			this.mainEmail = SettingsGet('MainEmail');

			this.accountForDeletion = ko.observable(null).askDeleteHelper();
			this.identityForDeletion = ko.observable(null).askDeleteHelper();
		}

		addNewAccount() {
			showScreenPopup(AccountPopupView);
		}

		editAccount(account) {
			if (account?.isAdditional()) {
				showScreenPopup(AccountPopupView, [account]);
			}
		}

		addNewIdentity() {
			showScreenPopup(IdentityPopupView);
		}

		editIdentity(identity) {
			showScreenPopup(IdentityPopupView, [identity]);
		}

		/**
		 * @param {AccountModel} accountToRemove
		 * @returns {void}
		 */
		deleteAccount(accountToRemove) {
			if (accountToRemove?.askDelete()) {
				this.accountForDeletion(null);
				if (accountToRemove) {
					this.accounts.remove((account) => accountToRemove === account);

					Remote.request('AccountDelete', (iError, data) => {
						if (!iError && data.Reload) {
							rl.route.root();
							setTimeout(() => location.reload(), 1);
						} else {
							rl.app.accountsAndIdentities();
						}
					}, {
						EmailToDelete: accountToRemove.email
					});
				}
			}
		}

		/**
		 * @param {IdentityModel} identityToRemove
		 * @returns {void}
		 */
		deleteIdentity(identityToRemove) {
			if (identityToRemove?.askDelete()) {
				this.identityForDeletion(null);

				if (identityToRemove) {
					IdentityUserStore.remove(oIdentity => identityToRemove === oIdentity);
					Remote.request('IdentityDelete', () => rl.app.accountsAndIdentities(), {
						IdToDelete: identityToRemove.id()
					});
				}
			}
		}

		accountsAndIdentitiesAfterMove() {
			Remote.request('AccountsAndIdentitiesSortOrder', null, {
				Accounts: AccountUserStore.getEmailAddresses().filter(v => v != SettingsGet('MainEmail')),
				Identities: IdentityUserStore.map(item => (item ? item.id() : ""))
			});
		}

		onBuild(oDom) {
			oDom.addEventListener('click', event => {
				let el = event.target.closestWithin('.accounts-list .e-action', oDom);
				el && ko.dataFor(el) && this.editAccount(ko.dataFor(el));

				el = event.target.closestWithin('.identities-list .e-action', oDom);
				el && ko.dataFor(el) && this.editIdentity(ko.dataFor(el));
			});
		}
	}

	//export class UserSettingsFilters /*extends AbstractViewSettings*/ {
	class UserSettingsFilters /*extends AbstractViewSettings*/ {
		constructor() {
			this.scripts = ko.observableArray();
			this.loading = ko.observable(true).extend({ debounce: 200 });
			addObservablesTo(this, {
				serverError: false,
				serverErrorDesc: ''
			});

			rl.loadScript(SettingsGet('StaticLibsJs').replace('/libs.', '/sieve.')).then(() => {
				const Sieve = window.Sieve;
				Sieve.folderList = FolderUserStore.folderList;
				Sieve.serverError.subscribe(value => this.serverError(value));
				Sieve.serverErrorDesc.subscribe(value => this.serverErrorDesc(value));
				Sieve.loading.subscribe(value => this.loading(value));
				Sieve.scripts.subscribe(value => this.scripts(value));
				Sieve.updateList();
			}).catch(e => console.error(e));

			this.scriptForDeletion = ko.observable(null).askDeleteHelper();
		}

		addScript() {
			this.editScript();
		}

		editScript(script) {
			window.Sieve.ScriptView.showModal(script ? [script] : null);
		}

		deleteScript(script) {
			window.Sieve.deleteScript(script);
		}

		toggleScript(script) {
			window.Sieve.toggleScript(script);
		}

		onBuild(oDom) {
			oDom.addEventListener('click', event => {
				const el = event.target.closestWithin('.script-item .script-name', oDom),
					script = el && ko.dataFor(el);
				script && this.editScript(script);
			});
		}

		onShow() {
			window.Sieve?.updateList();
		}
	}

	class OpenPgpImportPopupView extends AbstractViewPopup {
		constructor() {
			super('OpenPgpImport');

			this.addObservables({
				key: '',
				keyError: false,
				keyErrorMessage: '',

				saveGnuPG: true,
				saveServer: false
			});

			this.canGnuPG = GnuPGUserStore.isSupported();

			this.key.subscribe(() => {
				this.keyError(false);
				this.keyErrorMessage('');
			});
		}

		submitForm() {
			let keyTrimmed = this.key().trim();

			if (/\n/.test(keyTrimmed)) {
				keyTrimmed = keyTrimmed.replace(/\r+/g, '').replace(/\n{2,}/g, '\n\n');
			}

			this.keyError(!keyTrimmed);
			this.keyErrorMessage('');

			if (!keyTrimmed) {
				return;
			}

			let match = null,
				count = 30,
				done = false;
			// eslint-disable-next-line max-len
			const reg = /[-]{3,6}BEGIN[\s]PGP[\s](PRIVATE|PUBLIC)[\s]KEY[\s]BLOCK[-]{3,6}[\s\S]+?[-]{3,6}END[\s]PGP[\s](PRIVATE|PUBLIC)[\s]KEY[\s]BLOCK[-]{3,6}/gi;

			do {
				match = reg.exec(keyTrimmed);
				if (match && 0 < count) {
					if (match[0] && match[1] && match[2] && match[1] === match[2]) {
						this.saveGnuPG() && GnuPGUserStore.isSupported() && GnuPGUserStore.importKey(this.key(), (iError, oData) => {
							iError && alert(oData.ErrorMessage);
						});
						OpenPGPUserStore.isSupported() && OpenPGPUserStore.importKey(this.key());
					}

					--count;
					done = false;
				} else {
					done = true;
				}
			} while (!done);

			if (this.keyError()) {
				return;
			}

			this.close();
		}

		onShow() {
			this.key('');
			this.keyError(false);
			this.keyErrorMessage('');
		}
	}

	//import { pInt } from 'Common/Utils';

	class OpenPgpGeneratePopupView extends AbstractViewPopup {
		constructor() {
			super('OpenPgpGenerate');

			this.identities = IdentityUserStore;

			this.addObservables({
				email: '',
				emailError: false,

				name: '',
				password: '',
				keyType: 'ECC',

				submitRequest: false,
				submitError: '',

				backupPublicKey: true,
				backupPrivateKey: false,

				saveGnuPGPublic: true,
				saveGnuPGPrivate: false
			});

			this.canGnuPG = SettingsCapa('GnuPG');

			this.email.subscribe(() => this.emailError(false));
		}

		submitForm() {
			const type = this.keyType().toLowerCase(),
				userId = {
					name: this.name(),
					email: this.email()
				},
				cfg = {
					type: type,
					userIDs: [userId],
					passphrase: this.password().trim()
	//				format: 'armored' // output key format, defaults to 'armored' (other options: 'binary' or 'object')
				};
	/*
			if ('ecc' === type) {
				cfg.curve = 'curve25519';
			} else {
				cfg.rsaBits = pInt(this.keyBitLength());
			}
	*/
			this.emailError(!this.email().trim());
			if (this.emailError()) {
				return;
			}

			this.submitRequest(true);
			this.submitError('');

			openpgp.generateKey(cfg).then(keyPair => {
				if (keyPair) {
					const fn = () => {
						this.submitRequest(false);
						this.close();
					};

					OpenPGPUserStore.storeKeyPair(keyPair);

					keyPair.onServer = (this.backupPublicKey() ? 1 : 0) + (this.backupPrivateKey() ? 2 : 0);
					keyPair.inGnuPG = (this.saveGnuPGPublic() ? 1 : 0) + (this.saveGnuPGPrivate() ? 2 : 0);
					if (keyPair.onServer || keyPair.inGnuPG) {
						if (!this.backupPrivateKey() && !this.saveGnuPGPrivate()) {
							delete keyPair.privateKey;
						}
						GnuPGUserStore.storeKeyPair(keyPair, fn);
					} else {
						fn();
					}
				}
			})
			.catch((e) => {
				this.submitRequest(false);
				this.showError(e);
			});
		}

		showError(e) {
			console.log(e);
			if (e?.message) {
				this.submitError(e.message);
			}
		}

		onShow() {
			this.name(''/*IdentityUserStore()[0].name()*/);
			this.password('');
			this.email(''/*IdentityUserStore()[0].email()*/);
			this.emailError(false);
			this.submitError('');
		}
	}

	class UserSettingsSecurity extends AbstractViewSettings {
		constructor() {
			super();

			this.capaAutoLogout = SettingsCapa('AutoLogout');

			this.autoLogout = SettingsUserStore.autoLogout;

			let i18nLogout = (key, params) => i18n('SETTINGS_SECURITY/AUTOLOGIN_' + key, params);
			this.autoLogoutOptions = koComputable(() => {
				trigger();
				return [
					{ id: 0, name: i18nLogout('NEVER_OPTION_NAME') },
					{ id: 5, name: i18nLogout('MINUTES_OPTION_NAME', { MINUTES: 5 }) },
					{ id: 10, name: i18nLogout('MINUTES_OPTION_NAME', { MINUTES: 10 }) },
					{ id: 30, name: i18nLogout('MINUTES_OPTION_NAME', { MINUTES: 30 }) },
					{ id: 60, name: i18nLogout('MINUTES_OPTION_NAME', { MINUTES: 60 }) },
					{ id: 60 * 2, name: i18nLogout('HOURS_OPTION_NAME', { HOURS: 2 }) },
					{ id: 60 * 5, name: i18nLogout('HOURS_OPTION_NAME', { HOURS: 5 }) },
					{ id: 60 * 10, name: i18nLogout('HOURS_OPTION_NAME', { HOURS: 10 }) }
				];
			});

			if (this.capaAutoLogout) {
				this.addSetting('AutoLogout');
			}

			this.gnupgPublicKeys = GnuPGUserStore.publicKeys;
			this.gnupgPrivateKeys = GnuPGUserStore.privateKeys;

			this.openpgpkeysPublic = OpenPGPUserStore.publicKeys;
			this.openpgpkeysPrivate = OpenPGPUserStore.privateKeys;

			this.canOpenPGP = SettingsCapa('OpenPGP');
			this.canGnuPG = GnuPGUserStore.isSupported();
			this.canMailvelope = !!window.mailvelope;

			this.allowDraftAutosave = SettingsUserStore.allowDraftAutosave;

			this.allowDraftAutosave.subscribe(value => Remote.saveSetting('AllowDraftAutosave', value));
		}

		addOpenPgpKey() {
			showScreenPopup(OpenPgpImportPopupView);
		}

		generateOpenPgpKey() {
			showScreenPopup(OpenPgpGeneratePopupView);
		}

		onBuild() {
			/**
			 * Create an iframe to display the Mailvelope keyring settings.
			 * The iframe will be injected into the container identified by selector.
			 */
			window.mailvelope && mailvelope.createSettingsContainer('#mailvelope-settings'/*[, keyring], options*/);
		}
	}

	const folderForDeletion = ko.observable(null).askDeleteHelper();

	class UserSettingsFolders /*extends AbstractViewSettings*/ {
		constructor() {
			this.showKolab = FolderUserStore.allowKolab();
			this.defaultOptionsAfterRender = defaultOptionsAfterRender;
			this.kolabTypeOptions = ko.observableArray();
			let i18nFilter = key => i18n('SETTINGS_FOLDERS/TYPE_' + key);
			initOnStartOrLangChange(()=>{
				this.kolabTypeOptions([
					{ id: '', name: '' },
					{ id: 'event', name: i18nFilter('CALENDAR') },
					{ id: 'contact', name: i18nFilter('CONTACTS') },
					{ id: 'task', name: i18nFilter('TASKS') },
					{ id: 'note', name: i18nFilter('NOTES') },
					{ id: 'file', name: i18nFilter('FILES') },
					{ id: 'journal', name: i18nFilter('JOURNAL') },
					{ id: 'configuration', name: i18nFilter('CONFIGURATION') }
				]);
			});

			this.displaySpecSetting = FolderUserStore.displaySpecSetting;
			this.folderList = FolderUserStore.folderList;
			this.folderListOptimized = FolderUserStore.folderListOptimized;
			this.folderListError = FolderUserStore.folderListError;
			this.hideUnsubscribed = SettingsUserStore.hideUnsubscribed;
			this.unhideKolabFolders = SettingsUserStore.unhideKolabFolders;

			this.loading = FolderUserStore.foldersChanging;

			this.folderForDeletion = folderForDeletion;

			this.folderForEdit = ko.observable(null).extend({ toggleSubscribeProperty: [this, 'editing'] });

			SettingsUserStore.hideUnsubscribed.subscribe(value => Remote.saveSetting('HideUnsubscribed', value));
			SettingsUserStore.unhideKolabFolders.subscribe(value => Remote.saveSetting('UnhideKolabFolders', value));
		}

		folderEditOnEnter(folder) {
			const nameToEdit = folder?.nameForEdit().trim();
			if (nameToEdit && folder.name() !== nameToEdit) {
				Remote.abort('Folders').post('FolderRename', FolderUserStore.foldersRenaming, {
						Folder: folder.fullName,
						NewFolderName: nameToEdit,
						Subscribe: folder.isSubscribed() ? 1 : 0
					})
					.then(data => {
						folder.name(nameToEdit/*data.Name*/);
						if (folder.subFolders.length) {
							Remote.setTrigger(FolderUserStore.foldersLoading, true);
	//						clearTimeout(Remote.foldersTimeout);
	//						Remote.foldersTimeout = setTimeout(loadFolders, 500);
							setTimeout(loadFolders, 500);
							// TODO: rename all subfolders with folder.delimiter to prevent reload?
						} else {
							removeFolderFromCacheList(folder.fullName);
							folder.fullName = data.Result.FullName;
							setFolder(folder);
							const parent = getFolderFromCacheList(folder.parentName);
							sortFolders(parent ? parent.subFolders : FolderUserStore.folderList);
						}
					})
					.catch(error => {
						FolderUserStore.folderListError(
							getNotification(error.code, '', Notification$1.CantRenameFolder)
							+ '.\n' + error.message);
					});
			}

	//		this.folderForEdit(null);
			folder.editing(false);
		}

		folderEditOnEsc(folder) {
	//		this.folderForEdit(null);
			folder?.editing(false);
		}

		onShow() {
			FolderUserStore.folderListError('');
		}
	/*
		onBuild(oDom) {
		}
	*/
		createFolder() {
			showScreenPopup(FolderCreatePopupView);
		}

		systemFolder() {
			showScreenPopup(FolderSystemPopupView);
		}

		deleteFolder(folderToRemove) {
			if (folderToRemove
			 && folderToRemove.canBeDeleted()
			 && folderToRemove.askDelete()
			) {
				if (0 < folderToRemove.totalEmails()) {
	//				FolderUserStore.folderListError(getNotification(Notification.CantDeleteNonEmptyFolder));
					folderToRemove.errorMsg(getNotification(Notification$1.CantDeleteNonEmptyFolder));
				} else {
					folderForDeletion(null);

					if (folderToRemove) {
						Remote.abort('Folders').post('FolderDelete', FolderUserStore.foldersDeleting, {
								Folder: folderToRemove.fullName
							}).then(
								() => {
	//								folderToRemove.flags.push('\\nonexistent');
									folderToRemove.selectable(false);
	//								folderToRemove.isSubscribed(false);
	//								folderToRemove.checkable(false);
									if (!folderToRemove.subFolders.length) {
										removeFolderFromCacheList(folderToRemove.fullName);
										const folder = getFolderFromCacheList(folderToRemove.parentName);
										(folder ? folder.subFolders : FolderUserStore.folderList).remove(folderToRemove);
									}
								},
								error => {
									FolderUserStore.folderListError(
										getNotification(error.code, '', Notification$1.CantDeleteFolder)
										+ '.\n' + error.message
									);
								}
							);
					}
				}
			}
		}

		toggleFolderKolabType(folder, event) {
			let type = event.target.value;
			// TODO: append '.default' ?
			Remote.request('FolderSetMetadata', null, {
				Folder: folder.fullName,
				Key: FolderMetadataKeys.KolabFolderType,
				Value: type
			});
			folder.kolabType(type);
		}

		toggleFolderSubscription(folder) {
			let subscribe = !folder.isSubscribed();
			Remote.request('FolderSubscribe', null, {
				Folder: folder.fullName,
				Subscribe: subscribe ? 1 : 0
			});
			folder.isSubscribed(subscribe);
		}

		toggleFolderCheckable(folder) {
			let checkable = !folder.checkable();
			Remote.request('FolderCheckable', null, {
				Folder: folder.fullName,
				Checkable: checkable ? 1 : 0
			});
			folder.checkable(checkable);
		}
	}

	const themeBackground = {
		name: ThemeStore.userBackgroundName,
		hash: ThemeStore.userBackgroundHash
	};
	addObservablesTo(themeBackground, {
		uploaderButton: null,
		loading: false,
		error: ''
	});

	class UserSettingsThemes /*extends AbstractViewSettings*/ {
		constructor() {
			this.theme = ThemeStore.theme;
			this.themes = ThemeStore.themes;
			this.themesObjects = ko.observableArray();

			themeBackground.enabled = SettingsCapa('UserBackground');
			this.background = themeBackground;

			this.themeTrigger = ko.observable(SaveSettingStatus.Idle).extend({ debounce: 100 });

			ThemeStore.theme.subscribe(value => {
				this.themesObjects.forEach(theme => {
					theme.selected(value === theme.name);
				});

				changeTheme(value, this.themeTrigger);

				Remote.saveSettings(null, {
					Theme: value
				});
			});
		}

		onBuild() {
			const currentTheme = ThemeStore.theme();

			this.themesObjects(
				ThemeStore.themes.map(theme => ({
					name: theme,
					nameDisplay: convertThemeName(theme),
					selected: ko.observable(theme === currentTheme),
					themePreviewSrc: themePreviewLink(theme)
				}))
			);

			// initUploader

			if (themeBackground.uploaderButton() && themeBackground.enabled) {
				const oJua = new Jua({
					action: serverRequest('UploadBackground'),
					limit: 1,
					clickElement: themeBackground.uploaderButton()
				});

				oJua
					.on('onStart', () => {
						themeBackground.loading(true);
						themeBackground.error('');
						return true;
					})
					.on('onComplete', (id, result, data) => {
						themeBackground.loading(false);

						if (result && id && data?.Result?.Name && data?.Result?.Hash) {
							themeBackground.name(data.Result.Name);
							themeBackground.hash(data.Result.Hash);
						} else {
							themeBackground.name('');
							themeBackground.hash('');

							let errorMsg = '';
							if (data.ErrorCode) {
								switch (data.ErrorCode) {
									case UploadErrorCode.FileIsTooBig:
										errorMsg = i18n('SETTINGS_THEMES/ERROR_FILE_IS_TOO_BIG');
										break;
									case UploadErrorCode.FileType:
										errorMsg = i18n('SETTINGS_THEMES/ERROR_FILE_TYPE_ERROR');
										break;
									// no default
								}
							}

							themeBackground.error(errorMsg || data.ErrorMessage || i18n('SETTINGS_THEMES/ERROR_UNKNOWN'));
						}

						return true;
					});
			}
		}

		onShow() {
			themeBackground.error('');
		}

		clearBackground() {
			if (themeBackground.enabled) {
				Remote.request('ClearUserBackground', () => {
					themeBackground.name('');
					themeBackground.hash('');
				});
			}
		}
	}

	class SettingsMenuUserView extends AbstractViewLeft {
		/**
		 * @param {Object} screen
		 */
		constructor(screen) {
			super();

			this.menu = screen.menu;
		}

		link(route) {
			return settings(route);
		}
	}

	class SettingsPaneUserView extends AbstractViewRight {
		constructor() {
			super();

			this.isMobile = ThemeStore.isMobile;
			this.leftPanelDisabled = leftPanelDisabled;
		}

		onShow() {
			MessageUserStore.message(null);
		}

		onBuild(dom) {
			dom.addEventListener('click', () =>
				ThemeStore.isMobile() && !event.target.closestWithin('.toggleLeft', dom) && leftPanelDisabled(true)
			);
		}

		backToInbox() {
			hasher.setHash(mailbox(getFolderInboxName()));
		}
	}

	class SettingsUserScreen extends AbstractSettingsScreen {
		constructor() {
			super([SystemDropDownUserView, SettingsMenuUserView, SettingsPaneUserView]);

			const views = [
				UserSettingsGeneral
			];

			if (AppUserStore.allowContacts()) {
				views.push(UserSettingsContacts);
			}

			if (SettingsCapa('AdditionalAccounts') || SettingsCapa('Identities')) {
				views.push(UserSettingsAccounts);
			}

			if (SettingsCapa('Sieve')) {
				views.push(UserSettingsFilters);
			}

			if (SettingsCapa('AutoLogout') || SettingsCapa('OpenPGP') || SettingsCapa('GnuPG')) {
				views.push(UserSettingsSecurity);
			}

			views.push(UserSettingsFolders);

			if (SettingsCapa('Themes')) {
				views.push(UserSettingsThemes);
			}

			views.forEach((item, index) =>
				settingsAddViewModel(item, item.name.replace('User', ''),
					(item === UserSettingsAccounts && !SettingsCapa('AdditionalAccounts'))
						? 'SETTINGS_ACCOUNTS/LEGEND_IDENTITIES' : 0,
					0, 0 === index)
			);

			runSettingsViewModelHooks(false);

			initOnStartOrLangChange(
				() => this.sSettingsTitle = i18n('TITLES/SETTINGS'),
				() => this.setSettingsTitle()
			);
		}

		onShow() {
			this.setSettingsTitle();
			keyScope(Scope.Settings);
			ThemeStore.isMobile() && leftPanelDisabled(true);
		}

		setSettingsTitle() {
			const sEmail = AccountUserStore.email();
			rl.setWindowTitle((sEmail ? sEmail + ' - ' :  '') + this.sSettingsTitle);
		}
	}

	class SelectComponent {
		/**
		 * @param {Object} params
		 */
		constructor(params) {
			this.value = params.value || '';
			this.label = params.label || '';
			this.enable = null == params.enable ? true : params.enable;
			this.trigger = params.trigger?.subscribe ? params.trigger : null;
			this.placeholder = params.placeholder || '';

			this.labeled = null != params.label;

			let size = 0 < params.size ? 'span' + params.size : '';
			if (this.trigger) {
				const
					classForTrigger = ko.observable(''),
					setTriggerState = value => {
						switch (pInt(value)) {
							case SaveSettingStatus.Success:
								classForTrigger('success');
								break;
							case SaveSettingStatus.Failed:
								classForTrigger('error');
								break;
							default:
								classForTrigger('');
								break;
						}
					};

				setTriggerState(this.trigger());

				this.className = koComputable(() =>
					(size + ' settings-save-trigger-input ' + classForTrigger()).trim()
				);

				this.disposables = [
					this.trigger.subscribe(setTriggerState, this),
					this.className
				];
			} else {
				this.className = size;
				this.disposables = [];
			}

			this.options = params.options || '';

			this.optionsText = params.optionsText || null;
			this.optionsValue = params.optionsValue || null;
			this.optionsCaption = i18n(params.optionsCaption || null);

			this.defaultOptionsAfterRender = defaultOptionsAfterRender;
		}

		dispose() {
			this.disposables.forEach(dispose);
		}
	}

	class CheckboxMaterialDesignComponent {
		constructor(params = {}) {
			this.value = ko.isObservable(params.value) ? params.value
				: ko.observable(!!params.value);

			this.enable = ko.isObservable(params.enable) ? params.enable
				: ko.observable(undefined === params.enable || !!params.enable);

			this.label = params.label;
			this.inline = params.inline;

			this.labeled = null != params.label;
		}

		click() {
			this.enable() && this.value(!this.value());
		}
	}

	class AbstractApp {
		/**
		 * @param {RemoteStorage|AdminRemoteStorage} Remote
		 */
		constructor(Remote) {
			this.Remote = Remote;
		}

		logoutReload() {
			const url = logoutLink();

			if (location.href !== url) {
				setTimeout(() => location.href = url, 100);
			} else {
				rl.route.reload();
			}
		}

		bootstart() {
			const register = (key, ClassObject, templateID) => ko.components.register(key, {
					template: { element: templateID || (key + 'Component') },
					viewModel: {
						createViewModel: (params, componentInfo) => {
							params = params || {};
							i18nToNodes(componentInfo.element);
							if (params.inline) {
								componentInfo.element.style.display = 'inline-block';
							}
							return new ClassObject(params);
						}
					}
				});

			register('Select', SelectComponent);
			register('Checkbox', CheckboxMaterialDesignComponent, 'CheckboxMaterialDesignComponent');

			initOnStartOrLangChange();

			LanguageStore.populate();
			ThemeStore.populate();

			this.start();
		}
	}

	class AppUser extends AbstractApp {
		constructor() {
			super(Remote);

			// wakeUp
			const interval = 3600000; // 60m
			let lastTime = Date.now();
			setInterval(() => {
				const currentTime = Date.now();
				if (currentTime > (lastTime + interval + 1000)) {
					Remote.request('Version',
						iError => (100 < iError) && location.reload(),
						{ Version: Settings.app('version') }
					);
				}
				lastTime = currentTime;
			}, interval);

			addEventsListener(doc, ['keydown','keyup'], (ev=>$htmlCL.toggle('rl-ctrl-key-pressed', ev.ctrlKey)).debounce(500));

			addShortcut('escape,enter', '', dropdownsDetectVisibility);
			addEventListener('click', dropdownsDetectVisibility);

			this.folderList = FolderUserStore.folderList;
		}

		refresh() {
			LanguageStore.language(SettingsGet('Language'));
			ThemeStore.populate();
			changeTheme(SettingsGet('Theme'));
			this.start();
		}

		/**
		 * @param {number} iFolderType
		 * @param {string} sFromFolderFullName
		 * @param {Set} oUids
		 * @param {boolean=} bDelete = false
		 */
		moveMessagesToFolderType(iFolderType, sFromFolderFullName, oUids, bDelete) {
			let oMoveFolder = null,
				nSetSystemFoldersNotification = null;

			switch (iFolderType) {
				case FolderType.Spam:
					oMoveFolder = getFolderFromCacheList(FolderUserStore.spamFolder());
					nSetSystemFoldersNotification = SetSystemFoldersNotification.Spam;
					bDelete = bDelete || UNUSED_OPTION_VALUE === FolderUserStore.spamFolder();
					break;
				case FolderType.NotSpam:
					oMoveFolder = getFolderFromCacheList(getFolderInboxName());
					break;
				case FolderType.Trash:
					oMoveFolder = getFolderFromCacheList(FolderUserStore.trashFolder());
					nSetSystemFoldersNotification = SetSystemFoldersNotification.Trash;
					bDelete = bDelete || UNUSED_OPTION_VALUE === FolderUserStore.trashFolder()
						|| sFromFolderFullName === FolderUserStore.spamFolder()
						|| sFromFolderFullName === FolderUserStore.trashFolder();
					break;
				case FolderType.Archive:
					oMoveFolder = getFolderFromCacheList(FolderUserStore.archiveFolder());
					nSetSystemFoldersNotification = SetSystemFoldersNotification.Archive;
					bDelete = bDelete || UNUSED_OPTION_VALUE === FolderUserStore.archiveFolder();
					break;
				// no default
			}

			if (!oMoveFolder && !bDelete) {
				showScreenPopup(FolderSystemPopupView, [nSetSystemFoldersNotification]);
			} else if (bDelete) {
				showScreenPopup(AskPopupView, [
					i18n('POPUPS_ASK/DESC_WANT_DELETE_MESSAGES'),
					() => {
						messagesDeleteHelper(sFromFolderFullName, oUids);
						MessagelistUserStore.removeMessagesFromList(sFromFolderFullName, oUids);
					}
				]);
			} else if (oMoveFolder) {
				messagesMoveHelper(sFromFolderFullName, oMoveFolder.fullName, oUids);
				MessagelistUserStore.removeMessagesFromList(sFromFolderFullName, oUids, oMoveFolder.fullName);
			}
		}

		accountsAndIdentities() {
			AccountUserStore.loading(true);
			IdentityUserStore.loading(true);

			Remote.request('AccountsAndIdentities', (iError, oData) => {
				AccountUserStore.loading(false);
				IdentityUserStore.loading(false);

				if (!iError) {
					const
	//					counts = {},
						accounts = oData.Result.Accounts,
						mainEmail = SettingsGet('MainEmail');

					if (isArray(accounts)) {
	//					AccountUserStore.accounts.forEach(oAccount => counts[oAccount.email] = oAccount.count());

						AccountUserStore.accounts(
							accounts.map(
								sValue => new AccountModel(sValue/*, counts[sValue]*/)
							)
						);
	//					accounts.length &&
						AccountUserStore.accounts.unshift(new AccountModel(mainEmail/*, counts[mainEmail]*/, false));
					}

					if (isArray(oData.Result.Identities)) {
						IdentityUserStore(
							oData.Result.Identities.map(identityData => {
								const identity = new IdentityModel(
									pString(identityData.Id),
									pString(identityData.Email)
								);

								identity.name(pString(identityData.Name));
								identity.replyTo(pString(identityData.ReplyTo));
								identity.bcc(pString(identityData.Bcc));
								identity.signature(pString(identityData.Signature));
								identity.signatureInsertBefore(!!identityData.SignatureInsertBefore);

								return identity;
							})
						);
					}
				}
			});
		}

		/**
		 * @param {string} folder
		 * @param {Array=} list = []
		 */
		folderInformation(folder, list) {
			folderInformation(folder, list);
		}

		logout() {
			Remote.request('Logout', () => {
				const customLogoutLink = Settings.app('customLogoutLink');
				if (customLogoutLink) {
					location.href = customLogoutLink;
				} else {
					rl.logoutReload();
				}
			});
		}

		bootstart() {
			super.bootstart();

			addEventListener('resize', () => leftPanelDisabled(ThemeStore.isMobile() || 1000 > innerWidth));
			addEventListener('beforeunload', event => {
				if (arePopupsVisible() || (ThemeStore.isMobile() && MessageUserStore.message())) {
					event.preventDefault();
					return event.returnValue = "Are you sure you want to exit?";
				}
			}, {capture: true});
		}

		start() {
			if (SettingsGet('Auth')) {
				rl.setWindowTitle(i18n('GLOBAL/LOADING'));

				SMAudio.notifications(!!SettingsGet('SoundNotification'));
				NotificationUserStore.enabled(!!SettingsGet('DesktopNotifications'));

				AccountUserStore.email(SettingsGet('Email'));

				SettingsUserStore.init();
				ContactUserStore.init();

				loadFolders(value => {
					try {
						if (value) {
							startScreens([
								MailBoxUserScreen,
								SettingsUserScreen
							]);

							setInterval(() => {
								const cF = FolderUserStore.currentFolderFullName(),
									iF = getFolderInboxName();
								folderInformation(iF);
								if (iF !== cF) {
									folderInformation(cF);
								}
								folderInformationMultiply();
							}, refreshFoldersInterval);

							ContactUserStore.init();

							this.accountsAndIdentities();

							setTimeout(() => {
								const cF = FolderUserStore.currentFolderFullName();
								if (getFolderInboxName() !== cF) {
									folderInformation(cF);
								}
								FolderUserStore.hasCapability('LIST-STATUS') || folderInformationMultiply(true);
							}, 1000);

							setTimeout(() => Remote.request('AppDelayStart'), 35000);

							// add pointermove ?
							addEventsListener(doc, ['touchstart','mousemove','keydown'], SettingsUserStore.delayLogout, {passive:true});
							SettingsUserStore.delayLogout();

							// initLeftSideLayoutResizer
							setTimeout(() => {
								const left = elementById('rl-left'),
									right = elementById('rl-right'),
									fToggle = () =>
										setLayoutResizer(left, right, ClientSideKeyNameFolderListSize,
											(ThemeStore.isMobile() || leftPanelDisabled()) ? 0 : 'Width');
								if (left && right) {
									fToggle();
									leftPanelDisabled.subscribe(fToggle);
								}
							}, 1);

							setInterval(reloadTime(), 60000);

							PgpUserStore.init();

							// When auto-login is active
							if (navigator.registerProtocolHandler) {
								try {
									navigator.registerProtocolHandler(
										'mailto',
										location.protocol + '//' + location.host + location.pathname + '?mailto&to=%s',
										(SettingsGet('Title') || 'SnappyMail')
									);
								} catch (e) {} // eslint-disable-line no-empty
							}
							setTimeout(() => mailToHelper(SettingsGet('MailToEmail')), 500);
						} else {
							this.logout();
						}
					} catch (e) {
						console.error(e);
					}
				});

			} else {
				startScreens([LoginUserScreen]);
			}
		}

		showMessageComposer(params = [])
		{
			showScreenPopup(ComposePopupView, params);
		}
	}

	bootstrap(new AppUser);

})();
